/***************
 *
 * AveChannels.c
 *
 * This is a test harness for the basilar membrane model module.
 * This test harness requires 1000K of memory to run.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ****************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "CRLHeaders.h"

/******************************************************************************/ 
/****************************** Constant definitions **************************/ 
/******************************************************************************/ 

#define	PARAMETERS_FILE		"AveChannels.par"	/* Name of paramters file. */

/******************************************************************************/ 
/****************************** Global variables ******************************/ 
/******************************************************************************/

char	outputFile1[MAXLINE], outputFile2[MAXLINE], stParFile[MAXLINE];
char	stModuleName[MAXLINE], bMParFile[MAXLINE], bMModuleName[MAXLINE];

double	rampInterval;

/******************************************************************************/ 
/****************************** Functions and subroutines *********************/ 
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile1);
	GetPars_ParFile(fp, "%s", outputFile2);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	fclose(fp);
	Free_ParFile();
	
}

/******************************************************************************/ 
/****************************** Main Body *************************************/ 
/******************************************************************************/ 

int main()
{
	int		i;
	double	firstPeaks[MAX_CHANNELS], centreFreq[MAX_CHANNELS], binWidth;
	clock_t	startTime;
	FILE	*fp;
	EarObjectPtr	stimulus = NULL, filter = NULL, chanAverage = NULL;
	
	printf("Starting Test Harness...\n");
	startTime = clock();
	ReadParsFromFile(PARAMETERS_FILE);
	printf("\nThis test routine uses a pure tone stimulus, supplied.\n");
	printf("to a gamma tone basilar membrane filter, then calculates\n");
	printf("the average (summary) of the output channels.\n\n");
	
	/* Initialise EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((filter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);
	chanAverage = Init_EarObject("null");

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject(stimulus, filter);
	ConnectOutSignalToIn_EarObject(filter, chanAverage);

	/* Initialise modules. */

	if (!DoFun1(ReadPars, stimulus, stParFile))
		exit(1);
	DoFun(PrintPars, stimulus);
	
	if (!DoFun1(ReadPars, filter, bMParFile))
		exit(1);
	DoFun(PrintPars, filter);
	
	/* Start main process. */

	printf("Working...\n");	
	DoProcess(GenerateSignal, stimulus);
	if (!stimulus->outSignal->rampFlag ) {
		RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval );
		printf("\tAudPerGen: Stimulus has been ramped.\n" );
	}
	DoProcess(RunModel, filter);
	AveChannels_GenAnalysis(chanAverage);
	WriteOutSignal_DataFile(outputFile1, chanAverage);
	WriteOutSignal_DataFile(outputFile2, filter);

	FreeAll_EarObject();

	printf("This program took %lu seconds to run.\n", (clock() - startTime) / 
	  CLOCKS_PER_SEC);
	printf("Finished test.\n");
	return(0);
	
}

