/*******
 *
 * SynchIndex.c
 *
 * This is a simple test harness to make sure that things are working.
 * It tests the auto-correlation routine of the Analysis module.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/

#include <stdlib.h> 
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "CRLHeaders.h"

/******************************************************************************/ 
/****************************** Constant definitions **************************/ 
/******************************************************************************/ 

#define	PARAMETERS_FILE		"SynchIndex.par" /* Name of paramters file. */
#define	NUM_CHANNELS		1				/* No. of Signal Channels. */
#define CHANNEL				0				/* Work channel. */

/******************************************************************************/ 
/****************************** Global variables ******************************/ 
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], stModuleName[MAXLINE];
double	initialDuration, finalDuration, durationIncrement, pHPeriod, pHOffset;


/******************************************************************************/ 
/****************************** Functions and subroutines *********************/ 
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%lf", &initialDuration);
	GetPars_ParFile(fp, "%lf", &finalDuration);
	GetPars_ParFile(fp, "%lf", &durationIncrement);
	GetPars_ParFile(fp, "%lf", &pHPeriod);
	GetPars_ParFile(fp, "%lf", &pHOffset);
	fclose(fp);
	Free_ParFile();
	
}

/******************************************************************************/ 
/****************************** Main Body *************************************/ 
/******************************************************************************/ 

int main()
{
	int		i;
	double	intensity[NUM_CHANNELS], duration, synInd;
	FILE	*fp;
	EarObjectPtr	stimulus = NULL, histogram = NULL;
	
	printf("Starting Test Harness...\n");
	ReadParsFromFile(PARAMETERS_FILE);
	printf("\nThis test generates a %s and the calculated ", stModuleName);
	printf("intensity of the\nsignal is generated.\n");
	printf("The synchronisation index is then calculated for signal\n");
	printf("durations ranging from %g ms to %g ms in steps of %g ms.\n",
	  MSEC(initialDuration), MSEC(finalDuration), MSEC(durationIncrement));
	printf("\n");
	
	/* Initialise EarObjects */

	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	histogram = Init_EarObject("null");

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject(stimulus, histogram);

	/* Initialise modules. */

	if (!DoFun1(ReadPars, stimulus, stParFile))
		exit(1);
	DoFun(PrintPars, stimulus);
	  
	/* Start main process and print diagonstics. */

	printf("\n");
	if ((fp = fopen(outputFile, "w")) == NULL) {
		NotifyError("SynchIndex: Cannot open file '%s'.");
		exit(1);
	}
	for ( duration = initialDuration; duration < finalDuration; duration +=
	  durationIncrement) {
		printf("Reading %g / %g ms\n", MSEC(duration), MSEC(finalDuration));
	  	DoFun1(SetDuration, stimulus, duration);
		DoProcess(GenerateSignal, stimulus);
		synInd = CalcSynchIndex_ExpAnalysis(histogram, stimulus->outSignal->dt,
		  pHPeriod, pHOffset);
		fprintf(fp, "%10.5g\t%10.5g\n", duration, synInd);
	}
	fclose(fp);
	
	PrintProcessName_EarObject("Stimulus signal generated by the '%s' "\
	  "paradigm.\n", stimulus);
	FreeAll_EarObject();

	printf("Finished test.\n");
    exit(0);
	
}
