/***************
 *
 * FilterShape.c
 *
 * This program  plots the filter shape by plotting the frequency vs. intensity
 * for set characteristic (centre) frequency.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ****************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "CRLHeaders.h"

/******************************************************************************/ 
/****************************** Constant definitions **************************/ 
/******************************************************************************/ 

#define	PARAMETERS_FILE		"FilterShape.par" /* Name of paramters file. */
#define NUM_CHANNELS		1			/* The number of filter channels */
#define CHANNEL				0			/* Channel number. */

/******************************************************************************/ 
/****************************** Global variables ******************************/ 
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], pEParFile[MAXLINE];
char	bMParFile[MAXLINE], rPParFile[MAXLINE], hCParFile[MAXLINE];
char	sGParFile[MAXLINE], dEParFile[MAXLINE], mGParFile[MAXLINE];

char	stModuleName[MAXLINE], pEModuleName[MAXLINE], bMModuleName[MAXLINE];
char	rPModuleName[MAXLINE], hCModuleName[MAXLINE], sGModuleName[MAXLINE];
char	dEModuleName[MAXLINE], mGModuleName[MAXLINE];

int		numberOfFreqs;
double	lowestFrequency, highestFrequency;
double	rampInterval;

/******************************************************************************/ 
/****************************** Functions and subroutines *********************/ 
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", pEParFile, pEModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%s %s", rPParFile, rPModuleName);
	GetPars_ParFile(fp, "%s %s", hCParFile, hCModuleName);
	GetPars_ParFile(fp, "%s %s", sGParFile, sGModuleName);
	GetPars_ParFile(fp, "%s %s", dEParFile, dEModuleName);
	GetPars_ParFile(fp, "%s %s", mGParFile, mGModuleName);
	GetPars_ParFile(fp, "%lf", &lowestFrequency);
	GetPars_ParFile(fp, "%lf", &highestFrequency);
	GetPars_ParFile(fp, "%d", &numberOfFreqs);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	fclose(fp);
	Free_ParFile();
	
}

/******************************************************************************/ 
/****************************** Main Body *************************************/ 
/******************************************************************************/ 

int main()
{
	int		i;
	double	intensity[NUM_CHANNELS];
	FILE	*fp;
	EarObjectPtr	stimulus = NULL, pEFilter = NULL, bMFilter = NULL;
	EarObjectPtr	recpPotn = NULL, hairCell = NULL, spikeGeneration = NULL;
	EarObjectPtr	dendrite = NULL, neuron = NULL;
	CFListPtr	frequencies;
	
	printf("Starting Test Harness...\n");

	ReadParsFromFile(PARAMETERS_FILE);
	printf("\nThis test program generates a series of frequency vs. intensity ");
	printf("points to show the filter shape.\n");
	printf("The pure tone stimulus used is ramped with a raise interval of "\
	  "%lg ms.\n", MSEC(rampInterval));
	printf("The model process contains the following modules:\n\n");
	printf("\tStimulus generation:\t%s\n", stModuleName);
	printf("\tOuter-/middle-ear:\t%s\n", pEModuleName);
	printf("\tBasilar membrane:\t%s\n", bMModuleName);
	printf("\tIHC receptor pot.:\t%s\n", rPModuleName);
	printf("\tInner hair cell (IHC):\t%s\n", hCModuleName);
	printf("\tAuditory nerve spike generation:\t%s\n", sGModuleName);
	printf("\tDendritic filtering:\t%s\n", dEModuleName);
	printf("\tNeural cell model:\t%s\n", mGModuleName);
	printf("\n");
	
	/* Initialise EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((pEFilter = Init_EarObject(pEModuleName)) == NULL)
		exit(1);
	if ((bMFilter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);
	if ((recpPotn = Init_EarObject(rPModuleName)) == NULL)
		exit(1);
	if ((hairCell = Init_EarObject(hCModuleName)) == NULL)
		exit(1);
	if ((spikeGeneration = Init_EarObject(sGModuleName)) == NULL)
		exit(1);
	if ((dendrite = Init_EarObject(dEModuleName)) == NULL)
		exit(1);
	if ((neuron = Init_EarObject(mGModuleName)) == NULL)
		exit(1);

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject( stimulus, pEFilter );
	ConnectOutSignalToIn_EarObject( pEFilter, bMFilter );
	ConnectOutSignalToIn_EarObject( bMFilter, recpPotn );
	ConnectOutSignalToIn_EarObject( recpPotn, hairCell );
	ConnectOutSignalToIn_EarObject( hairCell, spikeGeneration );
 	ConnectOutSignalToIn_EarObject( spikeGeneration, dendrite );
 	ConnectOutSignalToIn_EarObject( dendrite, neuron );

	/* Initialise modules. */

	if (!DoFun1( ReadPars, stimulus, stParFile))
		exit(1);
	DoFun( PrintPars, stimulus );

	if (!DoFun1( ReadPars, pEFilter, pEParFile))
		exit(1);
	DoFun( PrintPars, pEFilter );

	if (!DoFun1( ReadPars, bMFilter, bMParFile))
		exit(1);
	DoFun( PrintPars, bMFilter );

	if (!DoFun1( ReadPars, recpPotn, rPParFile))
		exit(1);
	DoFun( PrintPars, recpPotn );
	
	if (!DoFun1( ReadPars, hairCell, hCParFile))
		exit(1);
	DoFun( PrintPars, hairCell );

	if (!DoFun1( ReadPars, spikeGeneration, sGParFile))
		exit(1);
	DoFun( PrintPars, spikeGeneration );

	if (!DoFun1( ReadPars, dendrite, sGParFile))
		exit(1);
	DoFun( PrintPars, dendrite );

	if (!DoFun1( ReadPars, neuron, mGParFile))
		exit(1);
	DoFun( PrintPars, neuron );
	
	/* Start main process. */

	printf("\n");
	if ((frequencies = GenerateLog_CFList(lowestFrequency, highestFrequency,
	  numberOfFreqs)) == NULL)
		exit(1);
 	if ((fp = fopen(outputFile, "w")) == NULL) {
		NotifyError("FilterShape: Cannot open '%s',.\n", outputFile);
		exit(1);
	}
	for (i = 0; i < frequencies->numChannels; i++) {
	 	printf("Frequency %d of %d, value = %lg Hz ...\n", i + 1,
	 	  frequencies->numChannels, frequencies->channel[i]);
		DoFun1(SetFrequency, stimulus, frequencies->channel[i]);
		DoProcess(GenerateSignal, stimulus);
		if (!stimulus->outSignal->rampFlag)
			RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval);
		DoProcess(RunModel, pEFilter);
		DoProcess(RunModel, bMFilter);
		DoProcess(RunModel, recpPotn);
		DoProcess(RunModel, hairCell);
		DoProcess(RunModel, spikeGeneration);
		DoProcess(RunModel, dendrite);
		DoProcess(RunModel, neuron);
		CalcIntensities_GenAnalysis(intensity, neuron->outSignal, rampInterval);
		fprintf(fp, "%18.5lf\t%10.5lf\n", frequencies->channel[i],
		  intensity[CHANNEL]);
	}
	fclose(fp);
	printf("Output data sent to file '%s'.\n", outputFile);
	FreeAll_EarObject();

	printf("Finished test.\n");
	return(0);
	
}

