/*******
 *
 * StimulusResp.c
 *
 * This is a simple test harness to make sure that things are working.
 * This harness tests the Meddis 86 hair cell response to a pure-tone
 * signal surrounded by silence.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/

#include <stdlib.h> 
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "CRLHeaders.h"

/******************************************************************************/ 
/****************************** Constant definitions **************************/ 
/******************************************************************************/ 

#define	PARAMETERS_FILE		"StimulusResp.par"/* Name of parameters file. */
#define	NUM_CHANNELS		1			/* Number of signal channels. */
#define	CHANNEL				0			/* Channel for testing. */

/******************************************************************************/ 
/****************************** Global variables ******************************/ 
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], pEParFile[MAXLINE];
char	bMParFile[MAXLINE], rPParFile[MAXLINE], hCParFile[MAXLINE];
char	sGParFile[MAXLINE], dEParFile[MAXLINE], mGParFile[MAXLINE];

char	stModuleName[MAXLINE], pEModuleName[MAXLINE], bMModuleName[MAXLINE];
char	rPModuleName[MAXLINE], hCModuleName[MAXLINE], sGModuleName[MAXLINE];
char	dEModuleName[MAXLINE], mGModuleName[MAXLINE];

double	binWidth, rampInterval;

/******************************************************************************/ 
/****************************** Functions and subroutines *********************/ 
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", pEParFile, pEModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%s %s", rPParFile, rPModuleName);
	GetPars_ParFile(fp, "%s %s", hCParFile, hCModuleName);
	GetPars_ParFile(fp, "%s %s", sGParFile, sGModuleName);
	GetPars_ParFile(fp, "%s %s", dEParFile, dEModuleName);
	GetPars_ParFile(fp, "%s %s", mGParFile, mGModuleName);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	GetPars_ParFile(fp, "%lf", &binWidth);
	fclose(fp);
	Free_ParFile();
	
}

/******************************************************************************/ 
/****************************** Main Body *************************************/ 
/******************************************************************************/ 

int main()
{
	clock_t		startTime;
	EarObjectPtr	stimulus = NULL, pEFilter = NULL, bMFilter = NULL;
	EarObjectPtr	recpPotn = NULL, hairCell = NULL, spikeGeneration = NULL;
	EarObjectPtr	dendrite = NULL, neuron = NULL, pSTHistogram = NULL;
	
	printf("Starting Test Harness...\n");
	
	ReadParsFromFile(PARAMETERS_FILE);
	printf("This test routine calculates the %s response of\n", stModuleName);
	printf("the %s hair cell.\n", hCModuleName);
	printf("The model process contains the following modules:\n\n");
	printf("\tStimulus generation:\t%s\n", stModuleName);
	printf("\tOuter-/middle-ear:\t%s\n", pEModuleName);
	printf("\tBasilar membrane:\t%s\n", bMModuleName);
	printf("\tIHC receptor pot.:\t%s\n", rPModuleName);
	printf("\tInner hair cell (IHC):\t%s\n", hCModuleName);
	printf("\tAuditory nerve spike generation:\t%s\n", sGModuleName);
	printf("\tDendritic filtering:\t%s\n", dEModuleName);
	printf("\tNeural cell model:\t%s\n", mGModuleName);
	printf("\n");

	/* Initialise EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((pEFilter = Init_EarObject(pEModuleName)) == NULL)
		exit(1);
	if ((bMFilter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);
	if ((recpPotn = Init_EarObject(rPModuleName)) == NULL)
		exit(1);
	if ((hairCell = Init_EarObject(hCModuleName)) == NULL)
		exit(1);
	if ((spikeGeneration = Init_EarObject(sGModuleName)) == NULL)
		exit(1);
	if ((dendrite = Init_EarObject(dEModuleName)) == NULL)
		exit(1);
	if ((neuron = Init_EarObject(mGModuleName)) == NULL)
		exit(1);
	pSTHistogram = Init_EarObject("null");

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject( stimulus, pEFilter );
	ConnectOutSignalToIn_EarObject( pEFilter, bMFilter );
	ConnectOutSignalToIn_EarObject( bMFilter, recpPotn );
	ConnectOutSignalToIn_EarObject( recpPotn, hairCell );
	ConnectOutSignalToIn_EarObject( hairCell, spikeGeneration );
 	ConnectOutSignalToIn_EarObject( spikeGeneration, dendrite );
 	ConnectOutSignalToIn_EarObject( dendrite, neuron );
 	ConnectOutSignalToIn_EarObject( neuron, pSTHistogram );

	/* Initialise modules. */

	printf("Module parameters...\n\n" );

	if (!DoFun1( ReadPars, stimulus, stParFile))
		exit(1);
	DoFun( PrintPars, stimulus );

	if (!DoFun1( ReadPars, pEFilter, pEParFile))
		exit(1);
	DoFun( PrintPars, pEFilter );

	if (!DoFun1( ReadPars, bMFilter, bMParFile))
		exit(1);
	DoFun( PrintPars, bMFilter );

	if (!DoFun1( ReadPars, recpPotn, rPParFile))
		exit(1);
	DoFun( PrintPars, recpPotn );
	
	if (!DoFun1( ReadPars, hairCell, hCParFile))
		exit(1);
	DoFun( PrintPars, hairCell );

	if (!DoFun1( ReadPars, spikeGeneration, sGParFile))
		exit(1);
	DoFun( PrintPars, spikeGeneration );

	if (!DoFun1( ReadPars, dendrite, dEParFile))
		exit(1);
	DoFun( PrintPars, dendrite );

	if (!DoFun1( ReadPars, neuron, mGParFile))
		exit(1);
	DoFun( PrintPars, neuron );

    /* Main simulation process */
	
	printf("\n");
	startTime = clock();
	DoProcess(GenerateSignal, stimulus);
	if (!stimulus->outSignal->rampFlag )
		RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval );
	DoProcess(RunModel, pEFilter);
	DoProcess(RunModel, bMFilter);
	DoProcess(RunModel, recpPotn);
	DoProcess(RunModel, hairCell);
	DoProcess(RunModel, spikeGeneration);
	DoProcess(RunModel, dendrite);
	DoProcess(RunModel, neuron);
    CalcPSTH_ExpAnalysis(pSTHistogram, binWidth);
    
	printf("The process took %lu seconds to run.\n", (clock() -
	  startTime) / CLOCKS_PER_SEC);
	
	WriteOutSignal_DataFile(outputFile, pSTHistogram);
	FreeAll_EarObject();
	
	printf("Finished test.\n");
	return(0);
	
}
