/**********************
 *
 * TuningCurve.c
 *
 * This program calculates the tuning curve of a filter, with respect
 * to a specified intensity threshold.
 *
 **********************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "CRLHeaders.h"

/******************************************************************************/
/****************************** Constant definitions **************************/
/******************************************************************************/

#define	PARAMETERS_FILE		"TuningCurve.par" /* Name of paramters file. */
#define NUM_CHANNELS		1		/* No. of filter channels. */
#define CHANNEL				0		/* Work filter channel. */
#define INTENSITY_ACCURACY	0.1		/* Accuracy of intensity points calculated*/

/******************************************************************************/
/****************************** Global variables ******************************/
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], pEParFile[MAXLINE];
char	bMParFile[MAXLINE], rPParFile[MAXLINE], hCParFile[MAXLINE];
char	sGParFile[MAXLINE], dEParFile[MAXLINE], mGParFile[MAXLINE];

char	stModuleName[MAXLINE], pEModuleName[MAXLINE], bMModuleName[MAXLINE];
char	rPModuleName[MAXLINE], hCModuleName[MAXLINE], sGModuleName[MAXLINE];
char	dEModuleName[MAXLINE], mGModuleName[MAXLINE];

int		numberOfFreqs;
double	lowestCFFrequency, highestCFFrequency, minimumFreq;
double	rampInterval, targetThreshold;

EarObjectPtr	stimulus = NULL, pEFilter = NULL, bMFilter = NULL;
EarObjectPtr	recpPotn = NULL, hairCell1 = NULL, spikeGeneration = NULL;
EarObjectPtr	dendrite1 = NULL, neuron = NULL;

/******************************************************************************/
/****************************** Function prototypes ***************************/
/******************************************************************************/

EarObjectPtr	RunModel(void);

/******************************************************************************/
/****************************** Functions and subroutines *********************/
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", pEParFile, pEModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%s %s", rPParFile, rPModuleName);
	GetPars_ParFile(fp, "%s %s", hCParFile, hCModuleName);
	GetPars_ParFile(fp, "%s %s", sGParFile, sGModuleName);
	GetPars_ParFile(fp, "%s %s", dEParFile, dEModuleName);
	GetPars_ParFile(fp, "%s %s", mGParFile, mGModuleName);
	GetPars_ParFile(fp, "%lf", &lowestCFFrequency);
	GetPars_ParFile(fp, "%lf", &highestCFFrequency);
	GetPars_ParFile(fp, "%d", &numberOfFreqs);
	GetPars_ParFile(fp, "%lf", &minimumFreq);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	GetPars_ParFile(fp, "%lf", &targetThreshold);
	fclose(fp);
	Free_ParFile();
	
}

/****************************** RunModel **************************************/

/*
 * This routine calculates the model being investigated.
 * It returns the model's final output intensity.
 * It accepts as input the intensity to which the model should be set - this
 * is the only parameter changed by the FindThresholdIntensity routine.
 * It works only on a single channel.
 * It is passed to the FindThresholdIntensity function.
 */

EarObjectPtr
RunModel(void)
{
	DoProcess(GenerateSignal, stimulus);
	if (!stimulus->outSignal->rampFlag )
		RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval );
	DoProcess(RunModel, pEFilter);
	DoProcess(RunModel, bMFilter);
	DoProcess(RunModel, recpPotn);
	DoProcess(RunModel, hairCell1);
	DoProcess(RunModel, spikeGeneration);
	DoProcess(RunModel, dendrite1);
	DoProcess(RunModel, neuron);
	return(neuron);

}

/******************************************************************************/
/****************************** Main Body *************************************/
/******************************************************************************/

int main()
{
	BOOLEAN	firstPass;
	int		i;
	double	intensity, samplingFrequency;
	FILE		*fp;
	CFList		*frequencies;
	
	printf("Starting test program...\n\n");
	
	ReadParsFromFile(PARAMETERS_FILE);
	printf("This test routine calculates the tuning curve for a model.\n");
	printf("If necessary the %s stimulus is ramped with a %lg ms rise time.\n",
	  stModuleName, MSEC(rampInterval));
	printf("The model is tested over the range %lg - %lg Hz, \n",
	  lowestCFFrequency, highestCFFrequency);
	printf("using a logarithmic scale with %d different frequency values.\n",
	  numberOfFreqs);
	printf("A %s probe stimulus ramped with a rise time of %lg ms\n",
	  stModuleName, MSEC(rampInterval));
	printf("and a sampling interval = 1/ 4.0 * CF is used.\n");
	printf("The sampling interval is limited by the minimum frequency of\n");
	printf("%g Hz.\n", minimumFreq);
	printf("The target threshold used is %lg dB SPL.\n", targetThreshold);
	printf("The model process contains the following modules:\n\n");
	printf("\tStimulus generation:\t%s\n", stModuleName);
	printf("\tOuter-/middle-ear:\t%s\n", pEModuleName);
	printf("\tBasilar membrane:\t%s\n", bMModuleName);
	printf("\tIHC receptor pot.:\t%s\n", rPModuleName);
	printf("\tInner hair cell (IHC):\t%s\n", hCModuleName);
	printf("\tAuditory nerve spike generation:\t%s\n", sGModuleName);
	printf("\tDendritic filtering:\t%s\n", dEModuleName);
	printf("\tNeural cell model:\t%s\n", mGModuleName);
	printf("\n");
	
	/* Initialising EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((pEFilter = Init_EarObject(pEModuleName)) == NULL)
		exit(1);
	if ((bMFilter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);
	if ((recpPotn = Init_EarObject(rPModuleName)) == NULL)
		exit(1);
	if ((hairCell1 = Init_EarObject(hCModuleName)) == NULL)
		exit(1);
	if ((spikeGeneration = Init_EarObject(sGModuleName)) == NULL)
		exit(1);
	if ((dendrite1 = Init_EarObject(dEModuleName)) == NULL)
		exit(1);
	if ((neuron = Init_EarObject(mGModuleName)) == NULL)
		exit(1);

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject( stimulus, pEFilter );
	ConnectOutSignalToIn_EarObject( pEFilter, bMFilter );
	ConnectOutSignalToIn_EarObject( bMFilter, recpPotn );
	ConnectOutSignalToIn_EarObject( recpPotn, hairCell1 );
	ConnectOutSignalToIn_EarObject( hairCell1, spikeGeneration );
 	ConnectOutSignalToIn_EarObject( spikeGeneration, dendrite1 );
 	ConnectOutSignalToIn_EarObject( dendrite1, neuron );

	/* Initialising Modules */
	
	if (!DoFun1( ReadPars, stimulus, stParFile))
		exit(1);
	DoFun( PrintPars, stimulus );

	if (!DoFun1( ReadPars, pEFilter, pEParFile))
		exit(1);
	DoFun( PrintPars, pEFilter );

	if (!DoFun1( ReadPars, bMFilter, bMParFile))
		exit(1);
	DoFun( PrintPars, bMFilter );

	if (!DoFun1( ReadPars, recpPotn, rPParFile))
		exit(1);
	DoFun( PrintPars, recpPotn );
	
	if (!DoFun1( ReadPars, hairCell1, hCParFile))
		exit(1);
	DoFun( PrintPars, hairCell1 );

	if (!DoFun1( ReadPars, spikeGeneration, sGParFile))
		exit(1);
	DoFun( PrintPars, spikeGeneration );

	if (!DoFun1( ReadPars, dendrite1, sGParFile))
		exit(1);
	DoFun( PrintPars, dendrite1 );

	if (!DoFun1( ReadPars, neuron, mGParFile))
		exit(1);
	DoFun( PrintPars, neuron );
	
	/* Start main process. */

	/* GenerateLog is used to create a frequency list, which is not directly
	 * used for the filter itself. */
	if ((frequencies = GenerateLog_CFList(lowestCFFrequency, highestCFFrequency,
	  numberOfFreqs)) == NULL)
		exit(1);
	if ((fp = fopen(outputFile, "w")) == NULL) {
		fprintf(stderr, "TuningCurve: Cannot open file '%s'\n", outputFile);
		exit(1);
	}
	for (i = 0, firstPass = TRUE; i < frequencies->numChannels; i++) {
		printf("Frequency %2d of %d = %g Hz...\n", i + 1, 
		  frequencies->numChannels, frequencies->channel[i]);
		DoFun1(SetFrequency, stimulus, frequencies->channel[i]);
		/* Set the sampling interval to min for Freq. */
		samplingFrequency = (frequencies->channel[i] >= minimumFreq) ? 
		  frequencies->channel[i]: minimumFreq;
		DoFun1(SetSamplingInterval, stimulus, 1.0 / (4.0 * samplingFrequency));
		intensity = FindThresholdIntensity_ExpAnalysis(RunModel,
		  stimulus->module->SetIntensity, targetThreshold, INTENSITY_ACCURACY,
		  rampInterval, &firstPass);
		fprintf(fp, "%10.5lf\t%10.5lf\n", frequencies->channel[i], intensity);
		
	}
	fclose(fp);
	FreeAll_EarObject();
	
	return(0);
	
}
