/***************
 *
 * ConnectionMgr.c
 *
 * This is a test harness for the basilar membrane model module.
 * It is an example of an insecure EarObject connection.
 * It requires 500k of memory.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ****************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "CRLHeaders.h"

/******************************************************************************/ 
/****************************** Constant definitions **************************/ 
/******************************************************************************/ 

#define	PARAMETERS_FILE		"ConnectionMgr.par"	/* Name of paramters file. */
#define NUM_CHANNELS		1				/* The number of filter channels */
#define CHANNEL				0				/* Work channel. */

/******************************************************************************/ 
/***************************** Macro definitions ******************************/
/******************************************************************************/ 

/* The following macro sets the sampling interval to the minimum for the
 *frequency. */

#define _SamplingInterval(FREQ)				(1.0 / (4.0 * (FREQ)))

/******************************************************************************/ 
/****************************** Global variables ******************************/ 
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], stModuleName[MAXLINE];
char	bMParFile[MAXLINE], bMModuleName[MAXLINE];
int		numReadings;
double	initialCentreFrequency, variationFromCF, frequencyIncrement;
double	dBDownDiff;
double	rampInterval;

/******************************************************************************/ 
/****************************** Functions and subroutines *********************/ 
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%d", &numReadings);
	GetPars_ParFile(fp, "%lf", &initialCentreFrequency);
	GetPars_ParFile(fp, "%lf", &variationFromCF);
	GetPars_ParFile(fp, "%lf", &frequencyIncrement);
	GetPars_ParFile(fp, "%lf", &dBDownDiff);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	fclose(fp);
	Free_ParFile();
	
}

/******************************************************************************/ 
/****************************** Main Body *************************************/ 
/******************************************************************************/ 

int main()
{
	int		i;
	double	sum, *ptr, intensity[NUM_CHANNELS], centreFreq[NUM_CHANNELS];
	double	frequency, maxIntensity, halfWidthAtXdB, qValue, percentageAccuracy;
	FILE	*fp;
	EarObjectPtr	stimulus = NULL, filter = NULL, filter2 = NULL;
	EarObjectPtr	filter3 = NULL;
	
	printf("Starting Test Harness...\n");
	
	ReadParsFromFile(PARAMETERS_FILE);
	
	/* Initialise EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((filter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);
	filter2 = Init_EarObject("null");
	filter3 = Init_EarObject("null");

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject(stimulus, filter3);
	ConnectOutSignalToIn_EarObject(stimulus, filter2);
	ConnectOutSignalToIn_EarObject(stimulus, filter);
	Free_EarObject(&filter2);

	/* Initialise modules. */

	if (!DoFun1(ReadPars, stimulus, stParFile))
		exit(1);
	DoFun(PrintPars, stimulus);
	
	if (!DoFun1(ReadPars, filter, bMParFile))
		exit(1);
	DoFun(PrintPars, filter);
	
	/* Start main process and print diagonstics. */

	printf("\nThis module tests the connection manager by showing the\n");
	printf("Q (%lg dB down) relation to\n", dBDownDiff);
	printf("characteristic frequency of a basilar membrane filter,\n");
	printf(" with an initial centre frequency(CF) of %lg Hz which is\n",
	  initialCentreFrequency);
	printf("increased logarithmically.\n");
	printf("A pure tone probe stimulus ramped with a rise time of %lg ms\n",
	  MSEC(rampInterval));
	printf("and sampling inteval = 1/ (4.0 * CF) is used.\n");
	printf("The frequencies are probed +/- %lg of the filter centre\n",
	  variationFromCF);
	printf("frequency in increments of %lg Hz.\n", frequencyIncrement);
	printf("The output is the Q value vs. characteristic frequency.\n\n");

	if ((fp = fopen(outputFile, "w")) == NULL) {
		fprintf(stderr, "TestGammaTone: Cannot open '%s',.\n", outputFile);
		exit(1);
	}
	
	for (i = 0; i < numReadings; i++) {
		printf("Reading %d of %d ...\n", i + 1, numReadings);
		centreFreq[CHANNEL] = initialCentreFrequency * pow(2.0, (double) i);
		SetCentreFreq_BasilarM_GammaT(centreFreq, sizeof centreFreq);
		SetSamplingInterval_PureTone(_SamplingInterval(centreFreq[CHANNEL]));
					/* Set the sampling interval to min for Freq. */
		for (frequency = centreFreq[CHANNEL], maxIntensity = 0.0; frequency <
		  centreFreq[0] * (1.0 + variationFromCF);
		  frequency += frequencyIncrement) {
			SetFrequency_PureTone(frequency);
			if (!DoProcess(GenerateSignal, stimulus)) {
				NotifyError("ConnectionMgr: Could not generate signal.");
				exit(1);
			}
			RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval);
			if (!DoProcess(RunModel, filter)) {
				NotifyError("ConnectionMgr: Could not run filter model.");
				exit(1);
			}
			CalcIntensities_GenAnalysis(intensity, filter->outSignal,
			  rampInterval);
			if (maxIntensity == 0.0) /* First value at the centre frequency */
				maxIntensity = intensity[CHANNEL];
			if ((maxIntensity - intensity[CHANNEL]) >= dBDownDiff) {
			  	halfWidthAtXdB = frequency - centreFreq[CHANNEL];
				qValue =  centreFreq[CHANNEL] / (2.0 * halfWidthAtXdB);
				percentageAccuracy = frequencyIncrement / halfWidthAtXdB *
				  100.0;
				fprintf(fp, "%10.5lf \t %10.5lf \t %5.2lf\n",
				  centreFreq[CHANNEL], qValue, percentageAccuracy);
				break;
			}
		}
	}
	fclose(fp);
	PrintProcessName_EarObject("Stimulus signal generated by the '%s' "\
	  "paradigm.\n", stimulus);
	PrintProcessName_EarObject("Signal filtered using the '%s'.\n", filter);
	printf("Results output to file '%s'.\n", outputFile);	
	FreeAll_EarObject();
	
	printf("Finished test.\n");
	return(0);
	
}

