/*******
 *
 * DistortProds
 *
 * This is a simple test harness to check the Basilar Membrane module.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/

#include <stdlib.h> 
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "CRLHeaders.h"

/******************************************************************************/
/****************************** Constant definitions **************************/
/******************************************************************************/

#define	PARAMETERS_FILE		"DistortProds.par"	/* Name of paramters file.*/
#define	NUM_CHANNELS		1			/* No. of channels for the filter. */
#define CHANNEL				0			/* Channel no. for test. */
#define	INPUT_INTENSITY_REFERENCE	1	/* Output dB reference - input level. */
#define LOWEST_DB			-1000.0		/* Default lowest dB. */

/******************************************************************************/
/****************************** Global variables ******************************/
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], bMParFile[MAXLINE];
char	stModuleName[MAXLINE], bMModuleName[MAXLINE];
int		numReadings, intensityOutputMode, multi1, multi2;
double	frequencyRatio, intensityInc1, intensityInc2;
double	rampInterval;

/******************************************************************************/
/****************************** Function prototypes ***************************/
/******************************************************************************/

double	OutputdB(EarObjectPtr data, double frequency, double reference);

/******************************************************************************/
/****************************** Functions and subroutines *********************/
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%d", &numReadings);
	GetPars_ParFile(fp, "%d", &intensityOutputMode);
	GetPars_ParFile(fp, "%lf", &frequencyRatio);
	GetPars_ParFile(fp, "%lf", &intensityInc1);
	GetPars_ParFile(fp, "%lf", &intensityInc2);
	GetPars_ParFile(fp, "%d", &multi1);
	GetPars_ParFile(fp, "%d", &multi2);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	fclose(fp);
	Free_ParFile();
	
}

/****************************** OutputdB **************************************/

/*
 * This routine returns a decibel value, from the output signal of an EarObject.
 * It assumes that the EarObject has been process by the FFT routine, and thus
 * has the value dF stored in the output signal's dt field.
 */

double
OutputdB(EarObjectPtr data, double frequency, double reference)
{
	double	modulusFT;
	ChanLen	index;
	
	if ((data == NULL) || !CheckPars_SignalData(data->outSignal)) {
		NotifyError("OutputdB: Data not correctly initialised.");
		exit(0);
	}
	index = (ChanLen) (frequency / data->outSignal->dt);
	if (index >= data->outSignal->length) {
		NotifyError("OutputdB: Frequency, %g Hz is too high.", frequency);
		exit(0);
	}
	modulusFT = data->outSignal->channel[CHANNEL][index];
	if (modulusFT == 0.0) {
		NotifyWarning("OutputdB: Zero modulus! Returning %g dB.", LOWEST_DB);
		return(LOWEST_DB);
	}
	return(20.0 * log10( modulusFT / (SQRT_2 *  reference)));
	
}

/******************************************************************************/
/****************************** Main Body *************************************/
/******************************************************************************/

int main()
{
	int		i;
	double	intensity1, intensity2, frequency1, frequency2;
	double	centreFreq[NUM_CHANNELS], primaryIntensity[NUM_CHANNELS];
	double	outputdB, refRMSAmplitude;
	FILE	*fp;
	EarObjectPtr	stimulus = NULL, filter = NULL, fourierTransform = NULL;
	
	printf("Starting Test Harness...\n");
	
	ReadParsFromFile(PARAMETERS_FILE);
	printf("\nThis test generates a multiple pure tone stimulus which is\n");
	printf("presented to the '%s' model.\n", bMModuleName);
	printf("The initial frequencies and intensities are read from the\n");
	printf("stimulus parameter file, but only the first frequency (f1) will ");
	printf("be used (f2 = f1 * %g).\n", frequencyRatio);
	printf("The intensities of the tones are incremented by %g and %g dB\n",
	  intensityInc1, intensityInc2);
	printf("for each of %d readings.\n", numReadings);
	printf("The '%s' filter is set with CF = %d * f1 - %d * f2 Hz.\n",
	  bMModuleName, multi1, multi2);
	printf("A rise time of %g ms is applied to the stimulus.\n",
	  MSEC(rampInterval));
	printf("\n");
	
	/* Initialise EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((filter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);
	fourierTransform = Init_EarObject("null");

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject(stimulus, filter);
	ConnectOutSignalToIn_EarObject(filter, fourierTransform);

	/* Initialise modules. */

	if (!DoFun1(ReadPars, stimulus, stParFile))
		exit(1);
	DoFun(PrintPars, stimulus);
	frequency1 = DoFun1(GetIndividualFreq, stimulus, 0);
	frequency2 = frequency1 * frequencyRatio;
	intensity1 = DoFun1(GetIndividualIntensity, stimulus, 0);
	intensity2 = DoFun1(GetIndividualIntensity, stimulus, 1);
	DoFun2(SetIndividualFreq, stimulus, 1, frequency2);

	if (!DoFun1(ReadPars, filter, bMParFile))
		exit(1);
	centreFreq[CHANNEL] = frequency1 * multi1 - frequency2 * multi2;
	DoFun2(SetCentreFreq, filter, centreFreq, sizeof (centreFreq));
	DoFun(PrintPars, filter);
	
	/* Start main process. */
	
	printf("\n");
	if ((fp = fopen(outputFile, "w")) == NULL) {
		NotifyError("DistortProds: Could not open file '%s' for output.",
		  outputFile);
		exit(1);
	}
	fprintf(fp, "int. (dB SPL)\tf1\tf2\t%df1 - %df2\n", multi1, multi2);
	for (i = 0; i < numReadings; i++) {
		printf("Reading %d of %d...\n", i + 1, numReadings);
		DoProcess(GenerateSignal, stimulus);
		RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval);
		CalcIntensities_GenAnalysis(primaryIntensity, stimulus->outSignal,
		  rampInterval);
		if (!DoProcess(RunModel, filter))
			exit(1);
		if (!Modulus_FourierT(fourierTransform))
			exit(1);
		switch (intensityOutputMode) {
		case	INPUT_INTENSITY_REFERENCE:
			refRMSAmplitude = RMS_AMP(primaryIntensity[CHANNEL]);
			break;
		default:
			refRMSAmplitude = 1.0;
		} /* Switch */
		fprintf(fp, "%10.5g\t%10.5g\t%10.5g\t%10.5g\n", intensity2, 
		  OutputdB(fourierTransform, frequency1, refRMSAmplitude),
		  OutputdB(fourierTransform, frequency2, refRMSAmplitude),
		  OutputdB(fourierTransform, centreFreq[CHANNEL], refRMSAmplitude));
		DoFun2(SetIndividualIntensity, stimulus, 0, (intensity1 +=
		  intensityInc1));
		DoFun2(SetIndividualIntensity, stimulus, 1, (intensity2 +=
		  intensityInc2));
	}
	fclose(fp);
	PrintProcessName_EarObject("Stimulus signal generated by the '%s' "\
	  "paradigm.\n", stimulus);
	PrintProcessName_EarObject("The '%s' was used.\n", filter);
	FreeAll_EarObject();

	printf("Finished test.\n");
    exit(0);
	
}
