/*******
 *
 * DistortProdsII.c
 *
 * This is a simple test harness to check the Basilar Membrane module.
 * It has some untidy code, which will allow us to record the stimulus intensity
 * just before the compression filter.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/

#include <stdlib.h> 
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "CRLHeaders.h"

/******************************************************************************/
/****************************** Constant definitions **************************/
/******************************************************************************/

#define	PARAMETERS_FILE		"DistortProdsII.par"	/* Name of paramters file.*/
#define	NUM_CHANNELS		1			/* No. of channels for the filter. */
#define CHANNEL				0			/* Channel no. for test. */
#define	INPUT_INTENSITY_REFERENCE	1	/* Output dB reference - input level. */
#define LOWEST_DB			-1000.0		/* Default lowest dB. */

/******************************************************************************/
/****************************** Global variables ******************************/
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], bMParFile[MAXLINE];
char	bMModuleName[MAXLINE];
int		numReadings, intensityOutputMode, multi1, multi2;
double	frequencyRatio, intensityInc1, intensityInc2;
double	rampInterval;

/******************************************************************************/
/****************************** Function prototypes ***************************/
/******************************************************************************/

BOOLEAN	RunModel_PartialFilter(EarObjectPtr data);

double	OutputdB(EarObjectPtr data, double frequency, double reference);

/******************************************************************************/
/****************************** Functions and subroutines *********************/
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s", stParFile);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%d", &numReadings);
	GetPars_ParFile(fp, "%d", &intensityOutputMode);
	GetPars_ParFile(fp, "%lf", &frequencyRatio);
	GetPars_ParFile(fp, "%lf", &intensityInc1);
	GetPars_ParFile(fp, "%lf", &intensityInc2);
	GetPars_ParFile(fp, "%d", &multi1);
	GetPars_ParFile(fp, "%d", &multi2);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	fclose(fp);
	Free_ParFile();
	
}

/****************************** OutputdB **************************************/

/*
 * This routine returns a decibel value, from the output signal of an EarObject.
 * It assumes that the EarObject has been process by the FFT routine, and thus
 * has the value dF stored in the output signal's dt field.
 */

double
OutputdB(EarObjectPtr data, double frequency, double reference)
{
	double	modulusFT;
	ChanLen	index;
	
	if ((data == NULL) || !CheckPars_SignalData(data->outSignal)) {
		NotifyError("OutputdB: Data not correctly initialised.");
		exit(0);
	}
	index = (ChanLen) (frequency / data->outSignal->dt);
	if (index >= data->outSignal->length) {
		NotifyError("OutputdB: Frequency, %g Hz is too high.", frequency);
		exit(0);
	}
	modulusFT = data->outSignal->channel[CHANNEL][index];
	if (modulusFT == 0.0) {
		NotifyWarning("OutputdB: Zero modulus! Returning %g dB.", LOWEST_DB);
		return(LOWEST_DB);
	}
	return(20.0 * log10( modulusFT / (SQRT_2 *  reference)));
	
}

/**************************** Simulate_Model **********************************/

/*
 * This routine filters the input signal and puts the result into the output
 * signal.
 * It checks that all initialisation has been correctly carried out by calling
 * the appropriate checking routines.
 * The appropriate coefficients are calculated at the beginning, then the memory
 * used is released before the program returns.
 * The wide filter signal is initialised by running this routine.  Thereafter
 * memory is only re-allocated if the signal changes in dimension.  The memory
 * for the wide filter signal is free'd when the module is un-initialised.
 */

BOOLEAN
RunModel_PartialFilter(EarObjectPtr data)
{
	int		i;
	double	attenuationFactor[MAX_CHANNELS], intensity[MAX_CHANNELS];
	GammaToneCoeffs	*narrowGammaTone[MAX_CHANNELS];
	GammaToneCoeffs *wideGammaTone[MAX_CHANNELS];
	TwoPoleCoeffs	*butterworthLP[MAX_CHANNELS], *besselLP[MAX_CHANNELS];
				
	if (data == NULL) {
		NotifyError("RunModel_BasilarM_Stone3: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_BasilarM_Stone3())
		return(FALSE);
		
	/* Initialise Variables and coefficients */
	
	SetProcessName_EarObject(data, "Stone3 Basilar Membrane Filtering");
	if (!CheckRamp_SignalData(data->inSignal)) {
		NotifyError("RunModel_BasilarM_Stone3: Input signal not correctly " \
		  "initialised.");
		return(FALSE);
	}
	InitOutFromInSignal_EarObject(data, basilarM3.theCFs->numChannels);
	InitCoeffs_BasilarM_Stone3(narrowGammaTone, wideGammaTone, butterworthLP,
	  besselLP, data);	
	
	/* Filter signal */
	
	GammaTone_Filters(data->outSignal, narrowGammaTone);
	IIR2_Filters(data->outSignal, butterworthLP);
	
	/* Compression_Filters(data->outSignal, RMS_AMP(basilarM3.compressionThreshold) *
	  SQRT_2, basilarM3.compressionExponent);
	
	GammaTone_Filters(data->outSignal, narrowGammaTone);
	IIR2_Filters(data->outSignal, butterworthLP); */
		
	FreeCoeffs_BasilarM_Stone3(narrowGammaTone, wideGammaTone, butterworthLP,
	  besselLP, data);
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

}

/******************************************************************************/
/****************************** Main Body *************************************/
/******************************************************************************/

int main()
{
	int		i;
	double	intensity1, intensity2, frequency1, frequency2;
	double	centreFreq[NUM_CHANNELS], primaryIntensity[NUM_CHANNELS];
	double	preCompressionIntensity[NUM_CHANNELS], refRMSAmplitude;
	FILE	*fp;
	EarObjectPtr	stimulus = NULL, filter = NULL, fourierTransform = NULL;
	EarObjectPtr	partialFilter = NULL;
	ModulePtr		stModule = NULL, bMModule = NULL;
	
	printf("Starting Test Harness...\n");
	
	ReadParsFromFile(PARAMETERS_FILE);
	printf("\nThis test generates a multiple pure tone stimulus which is\n");
	printf("presented to the '%s' model.\n", bMModuleName);
	printf("The initial frequencies and intensities are read from the\n");
	printf("stimulus parameter file, but only the first frequency (f1) will ");
	printf("be used (f2 = f1 * %g).\n", frequencyRatio);
	printf("The intensities of the tones are incremented by %g and %g dB\n",
	  intensityInc1, intensityInc2);
	printf("for each of %d readings.\n", numReadings);
	printf("The '%s' filter is set with CF = %d * f1 - %d * f2 Hz.\n",
	  bMModuleName, multi1, multi2);
	printf("A rise time of %g ms is applied to the stimulus.\n",
	  MSEC(rampInterval));
	printf("\n");
	
	/* Initialise EarObjects. */
	
	stimulus = Init_EarObject("'stimulus' for DistortProdsII");
	filter = Init_EarObject("'filter' for DistortProdsII");
	partialFilter = Init_EarObject("'partialFilter' for DistortProdsII");
	fourierTransform = Init_EarObject("'fourierTransform' for DistortProdsII");

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject(stimulus, filter);
	ConnectOutSignalToIn_EarObject(stimulus, partialFilter);
	ConnectOutSignalToIn_EarObject(filter, fourierTransform);

	/* Initialise modules. */

	if ((stModule = Init_ModuleMgr("PureTone_Multi")) == NULL)
		exit(1);
	if (!Do1(stModule->ReadPars, stParFile))
		exit(1);
	frequency1 = Do1(stModule->GetIndividualFreq, 0);
	frequency2 = frequency1 * frequencyRatio;
	intensity1 = Do1(stModule->GetIndividualIntensity, 0);
	intensity2 = Do1(stModule->GetIndividualIntensity, 1);
	Do2(stModule->SetIndividualFreq, 1, frequency2);
	Do0(stModule->PrintPars);

	if ((bMModule = Init_ModuleMgr(bMModuleName)) == NULL)
		exit(1);
	if (!Do1(bMModule->ReadPars, bMParFile))
		exit(1);
	centreFreq[CHANNEL] = frequency1 * multi1 - frequency2 * multi2;
	Do2(bMModule->SetCentreFreq, centreFreq, sizeof (centreFreq));
	Do0(bMModule->PrintPars);
	
	/* Start main process. */
	
	printf("\n");
	if ((fp = fopen(outputFile, "w")) == NULL) {
		NotifyError("DistortProdsII: Could not open file '%s' for output.",
		  outputFile);
		exit(1);
	}
	fprintf(fp, "int. (dB SPL)\tf1\tf2\t%df1 - %df2\n", multi1, multi2);
	for (i = 0; i < numReadings; i++) {
		printf("Reading %d of %d...\n", i + 1, numReadings);
		Do1(stModule->GenerateSignal, stimulus);
		RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval);
		CalcIntensities_GenAnalysis(primaryIntensity, stimulus->outSignal,
		  rampInterval);
		if (!Do1(bMModule->RunModel, filter))
			exit(1);
		RunModel_PartialFilter(partialFilter);
		CalcIntensities_GenAnalysis(preCompressionIntensity,
		  partialFilter->outSignal, rampInterval);
		if (!RealFT_FourierT(fourierTransform, TRUE))
			exit(1);
		switch (intensityOutputMode) {
		case	INPUT_INTENSITY_REFERENCE:
			refRMSAmplitude = RMS_AMP(primaryIntensity[CHANNEL]);
			break;
		default:
			refRMSAmplitude = 1.0;
		} /* Switch */
		fprintf(fp, "%10.5g\t%10.5g\t%10.5g\t%10.5g\n",
		  preCompressionIntensity[CHANNEL], 
		  OutputdB(fourierTransform, frequency1, refRMSAmplitude),
		  OutputdB(fourierTransform, frequency2, refRMSAmplitude),
		  OutputdB(fourierTransform, centreFreq[CHANNEL], refRMSAmplitude));
		Do2(stModule->SetIndividualIntensity, 0, (intensity1 += intensityInc1));
		Do2(stModule->SetIndividualIntensity, 1, (intensity2 += intensityInc2));
	}
	fclose(fp);
	PrintProcessName_EarObject("Stimulus signal generated by the '%s' "\
	  "paradigm.\n", stimulus);
	PrintProcessName_EarObject("The '%s' was used.\n", filter);
	FreeAll_EarObject();

	printf("Finished test.\n");
    exit(0);
	
}
