/*******
 *
 * FTAnalysis.c
 *
 * This is a simple test harness to check the Basilar Membrane module.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/

#include <stdlib.h> 
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "CRLHeaders.h"

/******************************************************************************/
/****************************** Constant definitions **************************/
/******************************************************************************/

#define	PARAMETERS_FILE		"FTAnalysis.par"	/* Name of paramters file.*/
#define	NUM_CHANNELS		1			/* No. of channels for the filter. */
#define CHANNEL				0			/* Channel no. for test. */

/******************************************************************************/
/****************************** Global variables ******************************/
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], stModuleName[MAXLINE];
char	bMParFile[MAXLINE], bMModuleName[MAXLINE];
double	rampInterval;

/******************************************************************************/
/****************************** Functions and subroutines *********************/
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	fclose(fp);
	Free_ParFile();
	
}

/******************************************************************************/
/****************************** Main Body *************************************/
/******************************************************************************/

int main()
{
	double	intensity[NUM_CHANNELS];
	EarObjectPtr	stimulus = NULL, filter = NULL, fourierTransform = NULL;
	ModulePtr		stModule = NULL, bMModule = NULL;
	
	
	printf("Starting Test Harness...\n");
	
	ReadParsFromFile(PARAMETERS_FILE);
	printf("\nThis test generates a %s stimulus which is presented to a\n",
	  stModuleName);
	printf("%s model,\n", bMModuleName);
	printf("If necessary, a rise time of %g ms is applied to the stimulus.\n",
	  MSEC(rampInterval));
	printf("\n");
	
	/* Initialise EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((filter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);
	fourierTransform = Init_EarObject("null");

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject(stimulus, filter);
	ConnectOutSignalToIn_EarObject(filter, fourierTransform);

	/* Initialise modules. */

	if (!DoFun1(ReadPars, stimulus, stParFile))
		exit(1);
	DoFun(PrintPars, stimulus);
	
	if (!DoFun1(ReadPars, filter, bMParFile))
		exit(1);
	DoFun(PrintPars, filter);
	
	/* Start main process and print diagonstics. */
	
	printf("\n");
    DoProcess(GenerateSignal, stimulus);
	if (!stimulus->outSignal->rampFlag) {
		RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval);
		printf("FTAnalysis: Stimulus has been ramped.\n");
	}
	CalcIntensities_GenAnalysis(intensity, stimulus->outSignal, 0.0);
	printf("The calculated intensity = %g dB (SPL)\n", intensity[CHANNEL]);
	if (!DoProcess(RunModel, filter))
		exit(1);
	PrintProcessName_EarObject("Stimulus signal generated by the '%s' "\
	  "paradigm.\n", stimulus);
	PrintProcessName_EarObject("The '%s' was used.\n", filter);
	Modulus_FourierT(fourierTransform);
	WriteOutSignal_DataFile(outputFile, fourierTransform);
	FreeAll_EarObject();

	printf("Finished test.\n");
    exit(0);
	
}
