/***************
 *
 * IORatiovsFreq.c
 *
 * This is a test harness for the Stone Composite filter basilar membrane model
 * module.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ****************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "CRLHeaders.h"

/******************************************************************************/ 
/****************************** Constant definitions **************************/ 
/******************************************************************************/ 

#define	PARAMETERS_FILE		"IORatiovsFreq.par" /* Name of paramters file. */
#define DUMMY_FREQUENCY		0.0			/* Initial dummy frequency. */
#define NUM_CHANNELS		1			/* The number of filter channels */
#define CHANNEL				0			/* Work channel. */

/******************************************************************************/ 
/****************************** Global variables ******************************/ 
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], stModuleName[MAXLINE];
char	bMParFile[MAXLINE], bMModuleName[MAXLINE];
int		numReadings;
double	initialCentreFrequency;
double	rampInterval;

/******************************************************************************/ 
/****************************** Functions and subroutines *********************/ 
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%d", &numReadings);
	GetPars_ParFile(fp, "%lf", &initialCentreFrequency);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	fclose(fp);
	Free_ParFile();
	
}

/******************************************************************************/ 
/****************************** Main Body *************************************/ 
/******************************************************************************/ 

int main()
{
	int		i;
	double	outIntensity[NUM_CHANNELS], centreFreq[NUM_CHANNELS];
	double	inputIntensity[NUM_CHANNELS];
	FILE	*fp;
	EarObjectPtr	stimulus = NULL, filter = NULL;
	
	printf("Starting Test Harness...\n");
	
	ReadParsFromFile(PARAMETERS_FILE);
	printf("\nThis test changes the stimulus intensity supplied to the \n");
	printf("gamma-tone filter.\n");
	printf("The output is the characteristic frequency vs the output/input\n");
	printf("intensity ratio.\n");
	printf("%d readings are calculated.\n\n", numReadings);

	/* Initialise EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((filter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject(stimulus, filter);

	/* Initialise modules. */

	if (!DoFun1(ReadPars, stimulus, stParFile))
		exit(1);
	DoFun(PrintPars, stimulus);
	
	if (!DoFun1(ReadPars, filter, bMParFile))
		exit(1);
	DoFun(PrintPars, filter);

	/* Start main process. */
	/* Calculate intensity values for each centre frequency. */
	
	if ((fp = fopen(outputFile, "w")) == NULL) {
		fprintf(stderr, "IORatiovsFreq.c: cannot open '%s'.\n", outputFile);
		exit(1);
	}
	for (i = 0; i < numReadings; i++) {
		printf("Reading %d of %d ...\n", i + 1, numReadings);
		centreFreq[CHANNEL] = initialCentreFrequency * pow(2.0, (double) i);
		DoFun2(SetCentreFreq, filter, centreFreq, sizeof(centreFreq));
		DoFun1(SetFrequency, stimulus, centreFreq[CHANNEL]);
		DoProcess(GenerateSignal, stimulus);
		CalcIntensities_GenAnalysis(inputIntensity, stimulus->outSignal,
		  rampInterval);		
		RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval);
		DoProcess(RunModel, filter);
		CalcIntensities_GenAnalysis(outIntensity, filter->outSignal,
		  rampInterval);
		fprintf(fp, "%10.5lf \t %10.5lf\n", centreFreq[CHANNEL],
		  outIntensity[CHANNEL] / inputIntensity[CHANNEL]);
    }
	fclose(fp);
	PrintProcessName_EarObject("Signal filtered using the '%s'.\n", filter);
	printf("Results output to file '%s'.\n", outputFile);
	FreeAll_EarObject();
	
	printf("Finished test.\n");
	return(0);
	
}

