/***************
 *
 * InputvsOut.c
 *
 * This is a test harness for the basilar membrane model module.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ****************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "CRLHeaders.h"

/******************************************************************************/ 
/****************************** Constant definitions **************************/ 
/******************************************************************************/ 

#define	PARAMETERS_FILE		"InputvsOut.par"	/* Name of paramters file. */
#define NUM_CHANNELS		1				/* The number of filter channels */
#define CHANNEL				0				/* Work Channel. */

/******************************************************************************/ 
/****************************** Global variables ******************************/ 
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], stModuleName[MAXLINE];
char	bMParFile[MAXLINE], bMModuleName[MAXLINE];
int		numReadings;
double	initialIntensity, intensityIncrement, rampInterval;

/******************************************************************************/ 
/****************************** Functions and subroutines *********************/ 
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%d", &numReadings);
	GetPars_ParFile(fp, "%lf", &initialIntensity);
	GetPars_ParFile(fp, "%lf", &intensityIncrement);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	fclose(fp);
	Free_ParFile();
	
}

/******************************************************************************/ 
/****************************** Main Body *************************************/ 
/******************************************************************************/ 

int main()
{
	int		i;
	double	intensity[NUM_CHANNELS], centreFreq[NUM_CHANNELS];
	FILE	*fp;
	EarObjectPtr	stimulus = NULL, filter = NULL;
	
	printf("Starting Test Harness...\n");
	
	ReadParsFromFile(PARAMETERS_FILE);
	printf("\nThis test changes the stimulus intensity supplied to the\n");
	printf("gamma-tone filter.\n");
	printf("The stimulus is a pure tone with intensity values ranging from\n");
	printf("%lg dB (SPL) in increments of %lg dB.\n", initialIntensity,
	  intensityIncrement);
	printf("The stimulus is ramped with a rise time of %lg ms.\n",
	  MSEC(rampInterval));
	printf("The output is the input intensity vs the final output " \
	  "intensity - %d readings.\n\n", numReadings);

	/* Initialise EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((filter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject(stimulus, filter);

	/* Initialise modules. */

	if (!DoFun1(ReadPars, stimulus, stParFile))
		exit(1);
	DoFun(PrintPars, stimulus);
	
	if (!DoFun1(ReadPars, filter, bMParFile))
		exit(1);
	DoFun(PrintPars, filter);
	

	/* Start main process. */

	if ((fp = fopen(outputFile, "w")) == NULL) {
		fprintf(stderr, " InputvsOut.c: cannot open '%s'.\n", outputFile);
		exit(1);
	}
	for (i = 0; i < numReadings; i++) {
		printf("Reading %d of %d...\n", i + 1, numReadings);
		DoFun1(SetIntensity, stimulus, initialIntensity + intensityIncrement
		  * i);
        DoProcess(GenerateSignal, stimulus);
		RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval);
		DoProcess(RunModel, filter);
        CalcIntensities_GenAnalysis(intensity, filter->outSignal, rampInterval); 
		fprintf(fp, "%10.5lf \t %10.5lf\n", initialIntensity +
		  intensityIncrement * i, intensity[CHANNEL]);
     }
	fclose(fp);
	PrintProcessName_EarObject("Signal filtered using the '%s'.\n", filter);
	printf("Results output to file '%s'.\n", outputFile);
	FreeAll_EarObject();

	printf("Finished test.\n");
	return(0);
	
}

