/***************
 *
 * Latency.c
 *
 * This is a test harness for the basilar membrane model module.
 * This test harness requires 1000K of memory to run.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ****************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "CRLHeaders.h"

/******************************************************************************/ 
/****************************** Constant definitions **************************/ 
/******************************************************************************/ 

#define	PARAMETERS_FILE		"Latency.par"	/* Name of paramters file. */

/******************************************************************************/ 
/****************************** Global variables ******************************/ 
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], stModuleName[MAXLINE];
char	bMParFile[MAXLINE], bMModuleName[MAXLINE];
int		latencyPeak;
double	rampInterval;

/******************************************************************************/ 
/****************************** Functions and subroutines *********************/ 
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%d", &latencyPeak);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	fclose(fp);
	Free_ParFile();
	
}

/******************************************************************************/ 
/****************************** Main Body *************************************/ 
/******************************************************************************/ 

int main()
{
	int		i;
	double	firstPeaks[MAX_CHANNELS], binWidth;
	clock_t	startTime;
	FILE	*fp;
	EarObjectPtr	stimulus = NULL, filter = NULL;
	
	printf("Starting Test Harness...\n");
	startTime = clock();
	
	ReadParsFromFile(PARAMETERS_FILE);
	printf("\nThis test routine calculates the latency of the gamma-tone\n");
	printf("filter for a range of CF values.\n");
	printf("A %s stimulus is used.\n", stModuleName);
	printf("If necessary (if a non click stimulus is used), a rise time of \n");
	printf("%g ms is applied to the stimulus.\n",
	  MSEC(rampInterval));
	printf("\n");
	
	/* Initialise EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((filter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject(stimulus, filter);

	/* Initialise modules. */

	if (!DoFun1(ReadPars, stimulus, stParFile))
		exit(1);
	DoFun(PrintPars, stimulus);
	
	if (!DoFun1(ReadPars, filter, bMParFile))
		exit(1);
	DoFun(PrintPars, filter);

	/* Start main process. */

	DoProcess(GenerateSignal, stimulus);
	if (!stimulus->outSignal->rampFlag ) {
		RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval );
		printf("\tAudPerGen: Stimulus has been ramped.\n" );
	}
	DoProcess(RunModel, filter);
	PrintProcessName_EarObject("Signal filtered using the '%s'.\n", filter);
	FindPeakPositions_GenAnalysis(firstPeaks, latencyPeak, filter->outSignal);
	printf("There are %d channels \n", filter->outSignal->numChannels);
	if ((fp = fopen(outputFile, "w")) == NULL) {
		fprintf(stderr, "TestBasilarM3: cannot open '%s'.\n", outputFile);
		exit(1);
	}
	printf("Outputting results to file '%s'.\n", outputFile);
	for (i = 0; i < filter->outSignal->numChannels; i++)
		fprintf(fp, "%lg \t %lg\n", MSEC(firstPeaks[i]), DoFun1(GetCF, filter,
		  i));
	fclose(fp);
	FreeAll_EarObject();

	printf("This program took %lu seconds to run.\n", (clock() - startTime) / 
	  CLOCKS_PER_SEC);
	printf("Finished test.\n");
	return(0);
	
}

