/***************
 *
 * QvsFrequency.c
 *
 * This is a test harness for the basilar membrane model module.
 * It was an example of an insecure EarObject connection.
 * It requires 500k of memory.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ****************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "CRLHeaders.h"

/******************************************************************************/
/****************************** Constant definitions **************************/
/******************************************************************************/

#define	PARAMETERS_FILE		"QvsFrequency.par" /* Name of paramters file. */
#define NUM_CHANNELS		1				/* The number of filter channels */
#define CHANNEL				0				/* Work channel. */
#define INTENSITY_ACCURACY	0.1		/* Accuracy of intensity points calculated*/

/******************************************************************************/
/***************************** Macro definitions ******************************/
/******************************************************************************/

/******************************************************************************/
/****************************** Global variables ******************************/
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], stModuleName[MAXLINE];
char	bMParFile[MAXLINE], bMModuleName[MAXLINE];
int		numberOfFreqs;
double	lowestCFFrequency, highestCFFrequency;
double	variationFromCF, frequencyIncrement, dBDownDiff, rampInterval;
double	minimumFreq;
EarObjectPtr	stimulus = NULL, filter = NULL;

/******************************************************************************/
/****************************** Function prototypes ***************************/
/******************************************************************************/

EarObjectPtr	RunModel(void);

/******************************************************************************/
/****************************** Functions and subroutines *********************/
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%lf", &lowestCFFrequency);
	GetPars_ParFile(fp, "%lf", &highestCFFrequency);
	GetPars_ParFile(fp, "%d", &numberOfFreqs);
	GetPars_ParFile(fp, "%lf", &variationFromCF);
	GetPars_ParFile(fp, "%lf", &frequencyIncrement);
	GetPars_ParFile(fp, "%lf", &minimumFreq);
	GetPars_ParFile(fp, "%lf", &dBDownDiff);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	fclose(fp);
	Free_ParFile();
	
}

/****************************** RunModel **************************************/

/*
 * This routine calculates the model being investigated.
 * It returns a pointer to the model's final EarObject.
 */

EarObjectPtr
RunModel(void)
{
	double	intensity[1];
	
	DoProcess(GenerateSignal, stimulus);
	RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval);
	DoProcess(RunModel, filter);
	return(filter);

}

/******************************************************************************/
/****************************** Main Body *************************************/
/******************************************************************************/

int main()
{
	int		i;
	double	qValue, centreFreq[NUM_CHANNELS];
	FILE	*fp;
	CFList		*frequencies;
	
	printf("Starting Test Harness...\n");
	
	ReadParsFromFile(PARAMETERS_FILE);
	printf("\nThis module shows the Q (%lg dB down) relation to\n", dBDownDiff);
	printf("characteristic frequency for a filter.\n");
	printf("The filter is tested over the range %lg - %lg Hz, \n",
	  lowestCFFrequency, highestCFFrequency);
	printf("using an logarithmic scale with %d different frequency values.\n",
	  numberOfFreqs);
	printf("A pure tone probe stimulus ramped with a rise time of %lg ms\n",
	  MSEC(rampInterval));
	printf("and a sampling interval = 1/ 4.0 * CF is used.\n");
	printf("The sampling interval is limited by the minimum frequency of\n");
	printf("%g Hz.\n", minimumFreq);
	printf("The frequencies are probed +/- %lg of the filter centre\n",
	  variationFromCF);
	printf("frequency in increments of %lg Hz.\n", frequencyIncrement);
	printf("The output is the Q value vs. characteristic frequency.\n");
	printf("\n");
	
	/* Initialise EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((filter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject(stimulus, filter);

	/* Initialise modules. */

	if (!DoFun1(ReadPars, stimulus, stParFile))
		exit(1);
	DoFun(PrintPars, stimulus);
	
	if (!DoFun1(ReadPars, filter, bMParFile))
		exit(1);
	DoFun(PrintPars, filter);
	
	/* Start main process. */

	/* GenerateLog is used to create a frequency list, which is not directly
	 * used for the filter itself. */
	frequencies = GenerateLog_CFList(lowestCFFrequency, highestCFFrequency,
	  numberOfFreqs);
	if ((fp = fopen(outputFile, "w")) == NULL) {
		fprintf(stderr, "QvsFrequency: Cannot open '%s',.\n", outputFile);
		exit(1);
	}
	for (i = 0; i < frequencies->numChannels; i++) {
		printf("Reading %d of %d (%lg Hz)...\n", i + 1,
		  frequencies->numChannels, frequencies->channel[i]);
		centreFreq[CHANNEL] = frequencies->channel[i];
		DoFun2(SetCentreFreq, filter, centreFreq, sizeof centreFreq);
		/* Set the sampling interval to min for Freq. */
		if (centreFreq[CHANNEL] >= minimumFreq)
			DoFun1(SetSamplingInterval, stimulus, 1.0 / (4.0 *
			  (centreFreq[CHANNEL] + frequencyIncrement)));
		else
			DoFun1(SetSamplingInterval, stimulus, 1.0 / (4.0 *
			  minimumFreq));
		qValue = CalcQValue_ExpAnalysis(RunModel,
		  stimulus->module->SetFrequency, centreFreq[CHANNEL], variationFromCF, dBDownDiff, frequencyIncrement,
		  INTENSITY_ACCURACY, rampInterval);
		fprintf(fp, "%10.5lf \t %10.5lf \t %5.2lf\n", centreFreq[CHANNEL],
		  qValue, INTENSITY_ACCURACY * 100.0 / (centreFreq[CHANNEL] / qValue));
	}
	fclose(fp);
	PrintProcessName_EarObject("Stimulus signal generated by the '%s' "\
	  "paradigm.\n", stimulus);
	PrintProcessName_EarObject("Signal filtered using the '%s'.\n", filter);
	printf("Results output to file '%s'.\n", outputFile);	

	FreeAll_EarObject();
	
	printf("Finished test.\n");
	return(0);
	
}

