/*******
 *
 * PLockvsInt.c
 *
 * This is a simple test harness to make sure that things are working.
 * This test harness investigates synchronisation index of the hair-cell
 * model, against the intensity of the signal.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/

#include <stdlib.h> 
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "CRLHeaders.h"

/******************************************************************************/ 
/****************************** Constant definitions **************************/ 
/******************************************************************************/ 

#define	PARAMETERS_FILE		"PLockvsInt.par" /* Name of paramters file. */
#define	NUM_CHANNELS		1			/* Number of signal channels. */
#define	CHANNEL				0			/* Channel for testing. */

/******************************************************************************/ 
/****************************** Global variables ******************************/ 
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], pEParFile[MAXLINE];
char	bMParFile[MAXLINE], rPParFile[MAXLINE], hCParFile[MAXLINE];
char	sGParFile[MAXLINE];

char	stModuleName[MAXLINE], pEModuleName[MAXLINE], bMModuleName[MAXLINE];
char	rPModuleName[MAXLINE], hCModuleName[MAXLINE], sGModuleName[MAXLINE];

int		numReadings;
double	stimulusFrequency, initialIntensity, intensityIncrement;
double	rampInterval, binWidthPH, binWidthPSTH;

EarObjectPtr	stimulus = NULL, pEFilter = NULL, bMFilter = NULL;
EarObjectPtr	recpPotn = NULL, modelIHC = NULL, spikeGeneration = NULL;

/******************************************************************************/ 
/****************************** Functions and subroutines *********************/ 
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", pEParFile, pEModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%s %s", rPParFile, rPModuleName);
	GetPars_ParFile(fp, "%s %s", hCParFile, hCModuleName);
	GetPars_ParFile(fp, "%s %s", sGParFile, sGModuleName);
	GetPars_ParFile(fp, "%d", &numReadings);
	GetPars_ParFile(fp, "%lf", &stimulusFrequency);
	GetPars_ParFile(fp, "%lf", &initialIntensity);
	GetPars_ParFile(fp, "%lf", &intensityIncrement);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	GetPars_ParFile(fp, "%lf", &binWidthPH);
	GetPars_ParFile(fp, "%lf", &binWidthPSTH);
	fclose(fp);
	Free_ParFile();
	
}

/****************************** RunModel **************************************/

/*
 * This routine calculates the model being investigated.
 * It returns a pointer to the model's final EarObject.
 */

EarObjectPtr
RunSimulation(void)
{
	double	intensity[1];
	
	DoProcess(GenerateSignal, stimulus);
	if (!stimulus->outSignal->rampFlag )
		RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval );
	DoProcess(RunModel, pEFilter);
	DoProcess(RunModel, bMFilter);
	DoProcess(RunModel, recpPotn);
	DoProcess(RunModel, modelIHC);
	DoProcess(RunModel, spikeGeneration);
	return(spikeGeneration);

}
/******************************************************************************/ 
/****************************** Main Body *************************************/ 
/******************************************************************************/ 

int main()
{
	int		i;
	double	intensity, synchIndex, maxAdaptedRate;
	double	adaptedRate, minAdaptedRate = 0.0, periodHOffset;
	FILE	*fp;
	EarObjectPtr	histogram = NULL;
	
	printf("Starting Test Harness...\n");
	
	ReadParsFromFile(PARAMETERS_FILE);
	printf("This test routine calculates the synchronisation index (vector\n");
	printf("strength).\n");
	printf("The intensity is varied in this test from %lg dB in increments of "\
	  "%lg dB (SPL).\n", initialIntensity, intensityIncrement);
	printf("The stimulus is ramped with a %lg ms rise time, if necessary.\n",
	  MSEC(rampInterval));
	printf("The equation is from M. J. Hewitt's thesis (1992).\n\n");
	printf("The model process contains the following modules:\n\n");
	printf("\tStimulus generation:\t%s\n", stModuleName);
	printf("\tOuter-/middle-ear:\t%s\n", pEModuleName);
	printf("\tBasilar membrane:\t%s\n", bMModuleName);
	printf("\tIHC receptor pot.:\t%s\n", rPModuleName);
	printf("\tInner hair cell (IHC):\t%s\n", hCModuleName);
	printf("\tAuditory nerve spike generation:\t%s\n", sGModuleName);
	printf("\n");
	
	/* Initialise EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((pEFilter = Init_EarObject(pEModuleName)) == NULL)
		exit(1);
	if ((bMFilter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);
	if ((recpPotn = Init_EarObject(rPModuleName)) == NULL)
		exit(1);
	if ((modelIHC = Init_EarObject(hCModuleName)) == NULL)
		exit(1);
	if ((spikeGeneration = Init_EarObject(sGModuleName)) == NULL)
		exit(1);
	histogram = Init_EarObject("null");

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject( stimulus, pEFilter );
	ConnectOutSignalToIn_EarObject( pEFilter, bMFilter );
	ConnectOutSignalToIn_EarObject( bMFilter, recpPotn );
	ConnectOutSignalToIn_EarObject( recpPotn, modelIHC );
	ConnectOutSignalToIn_EarObject( modelIHC, spikeGeneration );
 	ConnectOutSignalToIn_EarObject( spikeGeneration, histogram );
   
	/* Initialise modules. */

	printf("Module parameters...\n\n" );

	if (!DoFun1( ReadPars, stimulus, stParFile))
		exit(1);
	DoFun( PrintPars, stimulus );

	if (!DoFun1( ReadPars, pEFilter, pEParFile))
		exit(1);
	DoFun( PrintPars, pEFilter );

	if (!DoFun1( ReadPars, bMFilter, bMParFile))
		exit(1);
	DoFun( PrintPars, bMFilter );

	if (!DoFun1( ReadPars, recpPotn, rPParFile))
		exit(1);
	DoFun( PrintPars, recpPotn );
	
	if (!DoFun1( ReadPars, modelIHC, hCParFile))
		exit(1);
	DoFun( PrintPars, modelIHC );

	if (!DoFun1( ReadPars, spikeGeneration, sGParFile))
		exit(1);
	DoFun( PrintPars, spikeGeneration );
	printf("\n");

	/* Start main process. */
	/* Calculate the maximum adapted rate and connect appropriate signals. */

	RunSimulation();
	CalcPSTH_ExpAnalysis(histogram, binWidthPSTH);
	maxAdaptedRate =
	  histogram->outSignal->channel[CHANNEL][histogram->outSignal->length - 1];
	periodHOffset = histogram->outSignal->length * histogram->outSignal->dt /
	  2.0;

	if ((fp = fopen(outputFile, "w")) == NULL) {
		fprintf(stderr, "PLockvsInt: Cannot open file '%s'.\n", outputFile);
		exit(1);
	}
	for (i = 0; i < numReadings; i++) {
		printf("Calculating reading %d of %d...\n", i + 1, numReadings);
		intensity = initialIntensity + (i * intensityIncrement);
		DoFun1(SetIntensity, stimulus, intensity);
		RunSimulation();
		synchIndex = CalcSynchIndex_ExpAnalysis(histogram, binWidthPH, 1.0 /
		  stimulusFrequency, periodHOffset);
		CalcPSTH_ExpAnalysis(histogram, binWidthPSTH);
		adaptedRate =
		  histogram->outSignal->channel[CHANNEL][histogram->outSignal->length -
		  1];
		if (minAdaptedRate == 0.0)
        	minAdaptedRate = adaptedRate;
		fprintf(fp, "%10.5lf \t %10.5lf \t %10lf\n", intensity, synchIndex,
		  (adaptedRate - minAdaptedRate) / (maxAdaptedRate - minAdaptedRate));
	}
	fclose(fp);

	FreeAll_EarObject();
	
	printf("Finished test.\n");
	return(0);
	
}
