/*******
 *
 * RateIntensity.c
 *
 * This is a simple test harness to make sure that things are working.
 * This harness calculates the rate-intensity functions for the hair cell model.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/

#include <stdlib.h> 
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "CRLHeaders.h"

/******************************************************************************/ 
/****************************** Constant definitions **************************/ 
/******************************************************************************/ 

#define	PARAMETERS_FILE		"RateIntensity.par" /* Name of paramters file. */
#define	NUM_CHANNELS		1			/* Number of signal channels. */
#define	CHANNEL				0			/* Channel for testing. */

/******************************************************************************/
/****************************** Global variables ******************************/
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], pEParFile[MAXLINE];
char	bMParFile[MAXLINE], rPParFile[MAXLINE], hCParFile[MAXLINE];
char	sGParFile[MAXLINE];

char	stModuleName[MAXLINE], pEModuleName[MAXLINE], bMModuleName[MAXLINE];
char	rPModuleName[MAXLINE], hCModuleName[MAXLINE], sGModuleName[MAXLINE];

int		numReadings;
double	initialIntensity, intensityIncrement;
double	onsetBinWidth, onsetSearchRange, rampInterval;

/******************************************************************************/
/****************************** Functions and subroutines *********************/
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", pEParFile, pEModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%s %s", rPParFile, rPModuleName);
	GetPars_ParFile(fp, "%s %s", hCParFile, hCModuleName);
	GetPars_ParFile(fp, "%s %s", sGParFile, sGModuleName);
	GetPars_ParFile(fp, "%d", &numReadings);
	GetPars_ParFile(fp, "%lf", &initialIntensity);
	GetPars_ParFile(fp, "%lf", &intensityIncrement);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	GetPars_ParFile(fp, "%lf", &onsetBinWidth);
	GetPars_ParFile(fp, "%lf", &onsetSearchRange);
	fclose(fp);
	Free_ParFile();
	
}

/******************************************************************************/
/****************************** Main Body *************************************/
/******************************************************************************/ 

int main()
{
	int		i;
	clock_t	startTime;
	double	intensity, onsetPeakRate[NUM_CHANNELS], sStateAverage[NUM_CHANNELS];
	double	signalDuration;
	FILE	*fp;
	EarObjectPtr	stimulus = NULL, pEFilter = NULL, bMFilter = NULL;
	EarObjectPtr	recpPotn = NULL, modelIHC = NULL, spikeGen = NULL;
	
	printf("Starting Test Harness...\n");
	
	ReadParsFromFile(PARAMETERS_FILE);
	printf("\nThis test routine calculates the rate-intensity function for\n");
	printf("the %s.\n", hCModuleName);
	printf("The %s stimulus intensity ranges from %lg dB SPL: with\n",
	  initialIntensity, stModuleName);
	printf("%d readings in increments of %lg dB.\n", numReadings,
	  intensityIncrement);
	printf("The onset bin width is %lg ms, and is searched for over the\n",
	  MSEC(onsetBinWidth));
	printf("first %lg ms of the signal", MSEC(onsetSearchRange));
	printf("The model process contains the following modules:\n\n");
	printf("\tStimulus generation:\t%s\n", stModuleName);
	printf("\tOuter-/middle-ear:\t%s\n", pEModuleName);
	printf("\tBasilar membrane:\t%s\n", bMModuleName);
	printf("\tIHC receptor pot.:\t%s\n", rPModuleName);
	printf("\tInner hair cell (IHC):\t%s\n", hCModuleName);
	printf("\tAuditory nerve spike generation:\t%s\n", sGModuleName);
	printf("\n");
	/* Initialise EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((pEFilter = Init_EarObject(pEModuleName)) == NULL)
		exit(1);
	if ((bMFilter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);
	if ((recpPotn = Init_EarObject(rPModuleName)) == NULL)
		exit(1);
	if ((modelIHC = Init_EarObject(hCModuleName)) == NULL)
		exit(1);
	if ((spikeGen = Init_EarObject(sGModuleName)) == NULL)
		exit(1);

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject( stimulus, pEFilter );
	ConnectOutSignalToIn_EarObject( pEFilter, bMFilter );
	ConnectOutSignalToIn_EarObject( bMFilter, recpPotn );
	ConnectOutSignalToIn_EarObject( recpPotn, modelIHC );
	ConnectOutSignalToIn_EarObject( modelIHC, spikeGen );

	/* Initialise modules. */

	printf("Module parameters...\n\n" );

	if (!DoFun1( ReadPars, stimulus, stParFile))
		exit(1);
	DoFun( PrintPars, stimulus );

	if (!DoFun1( ReadPars, pEFilter, pEParFile))
		exit(1);
	DoFun( PrintPars, pEFilter );

	if (!DoFun1( ReadPars, bMFilter, bMParFile))
		exit(1);
	DoFun( PrintPars, bMFilter );

	if (!DoFun1( ReadPars, recpPotn, rPParFile))
		exit(1);
	DoFun( PrintPars, recpPotn );
	
	if (!DoFun1( ReadPars, modelIHC, hCParFile))
		exit(1);
	DoFun( PrintPars, modelIHC );

	if (!DoFun1( ReadPars, spikeGen, sGParFile))
		exit(1);
	DoFun( PrintPars, spikeGen );

    /* Main simulation process */
	
	printf("\n");
	if ((fp = fopen(outputFile, "w")) == NULL) {
		fprintf(stderr, "RateIntensity: Cannot open file '%s'.\n",
		  outputFile);
		exit(1);
	}
	startTime = clock();
	for (i = 0; i < numReadings; i++) {
		printf("Calculating reading %d of %d...\n", i + 1, numReadings);
		intensity = initialIntensity + (i * intensityIncrement);
		DoFun1(SetIntensity, stimulus, intensity);
		GenerateSignal_PureTone(stimulus);
		if (!stimulus->outSignal->rampFlag )
			RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval );
		DoProcess(RunModel, pEFilter);
		DoProcess(RunModel, bMFilter);
		DoProcess(RunModel, recpPotn);
		DoProcess(RunModel, modelIHC);
		DoProcess(RunModel, spikeGen);
		CalcOnSetPeakRates_ExpAnalysis(onsetPeakRate, spikeGen->outSignal,
		 onsetBinWidth, onsetSearchRange);
		signalDuration = GetDuration_SignalData(spikeGen->outSignal);
	    CalcAverages_GenAnalysis(sStateAverage, spikeGen->outSignal,
		  signalDuration / 2.0, signalDuration);
	    fprintf(fp, "%10.5lf \t %10.5lf \t %10.5lf\n", intensity,
	      onsetPeakRate[CHANNEL], sStateAverage[CHANNEL] /
		  spikeGen->outSignal->dt);
	}
	fclose(fp);
	printf("The process took %lu seconds to run.\n", (clock() -
	  startTime) / CLOCKS_PER_SEC);
	FreeAll_EarObject();
	
	printf("Finished test.\n");
	return(0);
	
}
