/**********************
 *
 * PeakTroughs.c
 *
 * This program investigates the hair cell receptor potential model.
 * 
 *
 **********************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "CRLHeaders.h"

/******************************************************************************/
/****************************** Constant definitions **************************/
/******************************************************************************/ 

#define	PARAMETERS_FILE		"PeakTroughs.par" /* Name of paramters file.*/
#define NUM_CHANNELS		1		/* No. of filter channels. */
#define CHANNEL				0		/* Work filter channel. */
#define OFFSET_DURATION_SCALE	0.5	/* Point of stimulus for calculations. */

/******************************************************************************/ 
/****************************** Global variables ******************************/ 
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], pEParFile[MAXLINE];
char	bMParFile[MAXLINE], rPParFile[MAXLINE];

char	stModuleName[MAXLINE], pEModuleName[MAXLINE], bMModuleName[MAXLINE];
char	rPModuleName[MAXLINE];

double	initialIntensity, intensityIncrement, finalIntensity;
double	rampInterval;

/******************************************************************************/ 
/****************************** Functions and subroutines *********************/ 
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", pEParFile, pEModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%s %s", rPParFile, rPModuleName);
	GetPars_ParFile(fp, "%lf", &initialIntensity);
	GetPars_ParFile(fp, "%lf", &finalIntensity);
	GetPars_ParFile(fp, "%lf", &intensityIncrement);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	fclose(fp);
	Free_ParFile();
	
}

/******************************************************************************/ 
/****************************** Main Body *************************************/ 
/******************************************************************************/ 

int main()
{
	int			i;
	double		centreFrequencies[NUM_CHANNELS], intensity, offsetTime;
	ChanLen		troughs[NUM_CHANNELS], peaks[NUM_CHANNELS];
	FILE		*fp;
	EarObjectPtr	stimulus = NULL, pEFilter = NULL, bMFilter = NULL;
	EarObjectPtr	recpPotn = NULL;
	
	printf("Starting test program...\n\n");
	
	ReadParsFromFile(PARAMETERS_FILE);
	printf("This test routine investigates the hair cell receptor potential\n");
	printf("peak/trough response.\n");
	printf("Output is received from the Stone 2 composite filter.\n");
	printf("The stimulus intensity varies over the range %lg - %lg dB SPL,\n",
	  initialIntensity, finalIntensity);
	printf("in increments of %lg dB SPL.\n", intensityIncrement);
	printf("The stimulus is ramped with a %lg ms rise time.\n",
	  MSEC(rampInterval));
	printf("The trough and peak receptor potentials are calculated from the\n");
	printf("signal offset of %lg * the signal duration.\n",
	  OFFSET_DURATION_SCALE);
	printf("The model process contains the following modules:\n\n");
	printf("\tStimulus generation:\t%s\n", stModuleName);
	printf("\tOuter-/middle-ear:\t%s\n", pEModuleName);
	printf("\tBasilar membrane:\t%s\n", bMModuleName);
	printf("\tIHC receptor pot.:\t%s\n", rPModuleName);
	printf("\n");
		
	/* Initialising EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((pEFilter = Init_EarObject(pEModuleName)) == NULL)
		exit(1);
	if ((bMFilter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);
	if ((recpPotn = Init_EarObject(rPModuleName)) == NULL)
		exit(1);

	/* Set up EarObject connections. */
	
	ConnectOutSignalToIn_EarObject( stimulus, pEFilter );
	ConnectOutSignalToIn_EarObject( pEFilter, bMFilter );
	ConnectOutSignalToIn_EarObject( bMFilter, recpPotn );

	/* Initialising Modules */
	
	if (!DoFun1( ReadPars, stimulus, stParFile))
		exit(1);
	DoFun( PrintPars, stimulus );

	if (!DoFun1( ReadPars, pEFilter, pEParFile))
		exit(1);
	DoFun( PrintPars, pEFilter );

	if (!DoFun1( ReadPars, bMFilter, bMParFile))
		exit(1);
	DoFun( PrintPars, bMFilter );

	if (!DoFun1( ReadPars, recpPotn, rPParFile))
		exit(1);
	DoFun( PrintPars, recpPotn );
	
	/* Start main process. */

	printf("\n");
	DoProcess(GenerateSignal, stimulus);
	if (!stimulus->outSignal->rampFlag )
		RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval );
	offsetTime = GetDuration_SignalData(stimulus->outSignal) *
	  OFFSET_DURATION_SCALE;
	if ((fp = fopen(outputFile, "w")) == NULL) {
		fprintf(stderr, "PeakTroughs: Cannot open file '%s'\n", outputFile);
		exit(1);
	}
	fprintf(fp, "Amp. (uPa)\tp(K+) trough\t Amp. (uPa)\tp(K+) peak\n");
	for (intensity = initialIntensity; intensity <= finalIntensity;
	  intensity += intensityIncrement) {
		printf("Intensity = %lg/%lg dB SPL...\n", intensity, finalIntensity);
		DoFun1(SetIntensity, stimulus, intensity);
		DoProcess(GenerateSignal, stimulus);
		if (!stimulus->outSignal->rampFlag )
			RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval );
		DoProcess(RunModel, pEFilter);
		DoProcess(RunModel, bMFilter);
		DoProcess(RunModel, recpPotn);
		FindNextMinIndexes_GenAnalysis(troughs, recpPotn->outSignal,
		  offsetTime);
		FindNextPeakIndexes_GenAnalysis(peaks, recpPotn->outSignal,
		  offsetTime);
		fprintf(fp, "%6.3lf\t%lg\t%6.2lf\t%lg\n", -RMS_AMP(intensity),
		  recpPotn->outSignal->channel[CHANNEL][troughs[CHANNEL]],
		  RMS_AMP(intensity),
		  recpPotn->outSignal->channel[CHANNEL][peaks[CHANNEL]]);
	}
	fclose(fp);
	FreeAll_EarObject();
	
	exit(0);
	
}
