/*****************
 *
 * CellTest.c
 *
 * This program tests the auditory periphery with the MacGregor neural
 * cell model.
 * The auditory periphery consists of:
 *
 *	Gamma Tone basilar membrane filter model;
 *	Meddis86 inner hair cell model;
 *	Spike generation model;
 *	Dendrite model low-pass filter;
 *
 *****************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "CRLHeaders.h"

/******************************************************************************/ 
/****************************** Constant definitions **************************/ 
/******************************************************************************/ 

#define	PARAMETERS_FILE		"CellTest.par"	/* Name of paramters file. */
#define NUM_CHANNELS		1			/* The number of filter channels */
#define CHANNEL				0			/* Channel number. */

/******************************************************************************/ 
/****************************** Global variables ******************************/ 
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], pEParFile[MAXLINE];
char	bMParFile[MAXLINE], rPParFile[MAXLINE], hCParFile[MAXLINE];
char	sGParFile[MAXLINE], dEParFile[MAXLINE], mGParFile[MAXLINE];

char	stModuleName[MAXLINE], pEModuleName[MAXLINE], bMModuleName[MAXLINE];
char	rPModuleName[MAXLINE], hCModuleName[MAXLINE], sGModuleName[MAXLINE];
char	dEModuleName[MAXLINE], mGModuleName[MAXLINE];

double	rampInterval, binWidth;

/******************************************************************************/ 
/****************************** Functions and subroutines *********************/ 
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", pEParFile, pEModuleName);
	GetPars_ParFile(fp, "%s %s", bMParFile, bMModuleName);
	GetPars_ParFile(fp, "%s %s", rPParFile, rPModuleName);
	GetPars_ParFile(fp, "%s %s", hCParFile, hCModuleName);
	GetPars_ParFile(fp, "%s %s", sGParFile, sGModuleName);
	GetPars_ParFile(fp, "%s %s", dEParFile, dEModuleName);
	GetPars_ParFile(fp, "%s %s", mGParFile, mGModuleName);
	GetPars_ParFile(fp, "%lf", &rampInterval);
	GetPars_ParFile(fp, "%lf", &binWidth);
	fclose(fp);
	Free_ParFile();
	
}

/******************************************************************************/ 
/****************************** Main Body *************************************/ 
/******************************************************************************/ 

int main()
{
	EarObjectPtr	stimulus = NULL, pEFilter = NULL, bMFilter = NULL;
	EarObjectPtr	recpPotn = NULL, hairCell = NULL, spikeGeneration = NULL;
	EarObjectPtr	dendrite = NULL, neuron = NULL, pSTHistogram = NULL;
	
	ReadParsFromFile(PARAMETERS_FILE);
	printf("\nThis test program incorporates an auditory periphery up to\n");
	printf("the neural cell model\n");
	printf("If necessary the stimulus is sine-ramped with a rise time of "\
	  "%lg ms.\n", MSEC(rampInterval));
	printf("The output is a post stimulus time histogram with a bin width\n");
	printf("bin width of %lg ms.\n", MSEC(binWidth));
	printf("The model process contains the following modules:\n\n");
	printf("\tStimulus generation:\t%s\n", stModuleName);
	printf("\tOuter-/middle-ear:\t%s\n", pEModuleName);
	printf("\tBasilar membrane:\t%s\n", bMModuleName);
	printf("\tIHC receptor pot.:\t%s\n", rPModuleName);
	printf("\tInner hair cell (IHC):\t%s\n", hCModuleName);
	printf("\tAuditory nerve spike generation:\t%s\n", sGModuleName);
	printf("\tDendritic filtering:\t%s\n", dEModuleName);
	printf("\tNeural cell model:\t%s\n", mGModuleName);
	printf("\n");
	
	/* Initialise EarObjects. */
	
	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((pEFilter = Init_EarObject(pEModuleName)) == NULL)
		exit(1);
	if ((bMFilter = Init_EarObject(bMModuleName)) == NULL)
		exit(1);
	if ((recpPotn = Init_EarObject(rPModuleName)) == NULL)
		exit(1);
	if ((hairCell = Init_EarObject(hCModuleName)) == NULL)
		exit(1);
	if ((spikeGeneration = Init_EarObject(sGModuleName)) == NULL)
		exit(1);
	if ((dendrite = Init_EarObject(dEModuleName)) == NULL)
		exit(1);
	if ((neuron = Init_EarObject(mGModuleName)) == NULL)
		exit(1);
	pSTHistogram = Init_EarObject("null");

	/* Set up the EarObject connections. */
	
	ConnectOutSignalToIn_EarObject( stimulus, pEFilter );
	ConnectOutSignalToIn_EarObject( pEFilter, bMFilter );
	ConnectOutSignalToIn_EarObject( bMFilter, recpPotn );
	ConnectOutSignalToIn_EarObject( recpPotn, hairCell );
	ConnectOutSignalToIn_EarObject( hairCell, spikeGeneration );
 	ConnectOutSignalToIn_EarObject( spikeGeneration, dendrite );
 	ConnectOutSignalToIn_EarObject( dendrite, neuron );
	ConnectOutSignalToIn_EarObject( neuron, pSTHistogram );

	/* Initialise modules. */
	
	printf("Module parameters...\n\n" );

	if (!DoFun1( ReadPars, stimulus, stParFile))
		exit(1);
	DoFun( PrintPars, stimulus );

	if (!DoFun1( ReadPars, pEFilter, pEParFile))
		exit(1);
	DoFun( PrintPars, pEFilter );

	if (!DoFun1( ReadPars, bMFilter, bMParFile))
		exit(1);
	DoFun( PrintPars, bMFilter );

	if (!DoFun1( ReadPars, recpPotn, rPParFile))
		exit(1);
	DoFun( PrintPars, recpPotn );
	
	if (!DoFun1( ReadPars, hairCell, hCParFile))
		exit(1);
	DoFun( PrintPars, hairCell );

	if (!DoFun1( ReadPars, spikeGeneration, sGParFile))
		exit(1);
	DoFun( PrintPars, spikeGeneration );

	if (!DoFun1( ReadPars, dendrite, sGParFile))
		exit(1);
	DoFun( PrintPars, dendrite );

	if (!DoFun1( ReadPars, neuron, mGParFile))
		exit(1);
	DoFun( PrintPars, neuron );

	/* Main Process. */
	
	printf("\n");
	DoProcess(GenerateSignal, stimulus);
	if (!stimulus->outSignal->rampFlag )
		RampUpOutSignal_Ramp(stimulus, Sine_Ramp, rampInterval );
	DoProcess(RunModel, pEFilter);
	DoProcess(RunModel, bMFilter);
	DoProcess(RunModel, recpPotn);
	DoProcess(RunModel, hairCell);
	DoProcess(RunModel, spikeGeneration);
	DoProcess(RunModel, dendrite);
	DoProcess(RunModel, neuron);
	CalcPSTH_ExpAnalysis(pSTHistogram, binWidth);
	WriteOutSignal_DataFile(outputFile, neuron);

	/* Tidy up memory space before leaving. */
		
	FreeAll_EarObject();
	
	return(0);
	
}