/********************
 *
 * IOFunction.c
 *
 * This program tests the basic behaviour of the McGregor2 neuron model.
 * With regard to changing parameters.  The average level of the output
 * from the function is tested for divergence, when parameters become too
 * great in value.  Investigations proceed from a valid set of parameters, e.g.
 * stellate, bushy or fusiform.
 *
 ********************/

#include <stdio.h>
#include <stdlib.h>

#include "CRLHeaders.h"

/******************************************************************************/
/****************************** Constant definitions **************************/
/******************************************************************************/ 

#define	PARAMETERS_FILE		"IOFunction.par"	/* Name of paramters file. */
#define	NUM_CHANNELS		1			/* Number of signal channels. */
#define	CHANNEL				0			/* Channel for testing. */

/******************************************************************************/
/****************************** Global variables ******************************/
/******************************************************************************/

char	outputFile[MAXLINE], stParFile[MAXLINE], mGParFile[MAXLINE];
char	stModuleName[MAXLINE], mGModuleName[MAXLINE];
int		numberOfReadings;
double	parameterValue, parameterIncrement, searchOffset;

/******************************************************************************/
/****************************** Functions and subroutines *********************/
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 * It expects there to be one parameter per line.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%s %s", stParFile, stModuleName);
	GetPars_ParFile(fp, "%s %s", mGParFile, mGModuleName);
	GetPars_ParFile(fp, "%d", &numberOfReadings);
	GetPars_ParFile(fp, "%lf", &parameterValue);
	GetPars_ParFile(fp, "%lf", &parameterIncrement);
	GetPars_ParFile(fp, "%lf", &searchOffset);
	fclose(fp);
	Free_ParFile();
	
}

/******************************************************************************/
/****************************** Main Body *************************************/
/******************************************************************************/ 

int main()
{
	int				i, j;
	double			potential;
	ChanLen			index[NUM_CHANNELS];
	FILE			*fp;
	EarObjectPtr	stimulus = NULL, neuron = NULL;

	ReadParsFromFile(PARAMETERS_FILE);
	printf("\nThis program is used to test the general behaviour of the\n");
	printf("McGregor2 neural cell module with respect to its parameters.\n");
	printf("%d readings are taken with the respective parameter being\n",
	  numberOfReadings);
	printf("changed in increments of %lg, from the initial reading, set in\n",
	  parameterIncrement);
	printf("parameter file '%s'.\n\n", mGParFile);

	/* Initialise EarObjects. */

	if ((stimulus = Init_EarObject(stModuleName)) == NULL)
		exit(1);
	if ((neuron = Init_EarObject(mGModuleName)) == NULL)
		exit(1);

	/* Connect EarObjects. */

	ConnectOutSignalToIn_EarObject(stimulus, neuron);

	/* Initialise modules. */

	if (!DoFun1(ReadPars, stimulus, stParFile))
		exit(1);
	DoFun(PrintPars, stimulus);
	
	if (!DoFun1(ReadPars, neuron, mGParFile))
		exit(1);
	DoFun(PrintPars, neuron);

	/* Main process. */
	
	printf("\n");
	if ((fp = fopen(outputFile, "w")) == NULL) {
		fprintf(stderr, "IOFunction: Cannot open file '%s'\n", outputFile);
		exit(1);
	}
	for (i = 0;  i < numberOfReadings; i++) {
		printf("Reading %d of %d (value = %g)...\n", i + 1,
		  numberOfReadings, parameterValue);
		DoFun1(SetAmplitude, stimulus, parameterValue);		
		DoProcess(GenerateSignal, stimulus);
		DoProcess(RunModel, neuron);
		if (parameterValue < 0.0) {
			FindNextMinIndexes_GenAnalysis(index, neuron->outSignal,
			  searchOffset);
			potential = neuron->outSignal->channel[CHANNEL][index[CHANNEL]];
		} else {
			FindNextPeakIndexes_GenAnalysis(index, neuron->outSignal,
			  searchOffset);
			potential = DoFun1(GetPotentialResponse, neuron,
			  neuron->outSignal->channel[CHANNEL][index[CHANNEL]]);
		}
		fprintf(fp, "%lg\t%lg\n", parameterValue, potential);
		parameterValue += parameterIncrement;
	}
	fclose(fp);
	printf("Test results output to file '%s'.\n", outputFile);

	/* Tidying up memory. */

	FreeAll_EarObject();

	return(0);

}
