/*****************
 *
 * GeomDist.c
 *
 * This routine investigates the behaviour of the
 * Geometric Distribution function and random number generator in the
 * The Chi square test is used.
 * UtRandom module.
 *
 *****************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "CRLHeaders.h"

/******************************************************************************/ 
/****************************** Constant definitions **************************/ 
/******************************************************************************/ 

#define	PARAMETERS_FILE		"GeomDist.par"	/* Name of paramters file. */

/******************************************************************************/ 
/****************************** Global variables ******************************/ 
/******************************************************************************/

char	outputFile[MAXLINE];
int		numberOfCategories, numberOfObservations;
double	minProb, maxProb, probIncrement; 

/******************************************************************************/ 
/****************************** Function Prototypes ***************************/ 
/******************************************************************************/

void	CalculateProbabilityTable(double theTable[], double p, int n);
long	Factorial(long number);

/******************************************************************************/ 
/****************************** Functions and subroutines *********************/ 
/******************************************************************************/

/****************************** ReadParsFromFile ******************************/

/*
 * This program reads a specified number of parameters from a file.
 */
 
void
ReadParsFromFile(char *fileName)
{
	char	line[MAXLINE];
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadTestPars: Cannot open data file '%s'.\n", fileName);
		exit(1);
	}
	printf("Reading parameters from file: %s\n", fileName);
	Init_ParFile();
	GetPars_ParFile(fp, "%s", outputFile);
	GetPars_ParFile(fp, "%d", &numberOfCategories);
	GetPars_ParFile(fp, "%d", &numberOfObservations);
	GetPars_ParFile(fp, "%lf", &minProb);
	GetPars_ParFile(fp, "%lf", &maxProb);
	GetPars_ParFile(fp, "%lf", &probIncrement);
	fclose(fp);
	Free_ParFile();
	if (maxProb >= 1.0) {
		NotifyError("GeomDist: Probabilities greater than or equal to 1.0\n"\
		  "are not allowed (a maximum of %lg was set in the parameters file).",
		  maxProb);
		exit(1);
	}
	
}

/*************************** Factorial ****************************************/

/*
 * This function returns the factorial of an integer.
 *
 */

long
Factorial(long number)
{
	long	i, fact;
	
	for (i = 1, fact = 1; i < number; i++)
		fact *= (i + 1);
	
	return(fact);
	
}
	
/*************************** CalculateProbabilityTable ************************/

/*
 * This routine calculates the a probability table, using the binomial
 * distribution.
 */

void
CalculateProbabilityTable(double theTable[], double p, int n)
{
	int		i;
	long	factN;

	factN = Factorial(n);
	for (i = 0; i < n; i++)
		theTable[i] = factN / (Factorial(i) * Factorial(n - i)) * pow(p, i) *
		  pow(1 - p, n - i);

}

/******************************************************************************/
/*************************** Main Body ****************************************/
/******************************************************************************/

int main()
{
	long	i, *categories;
	clock_t	startTime;
	double	prob, sum, chiSquare, *probTable;
	FILE	*fp;
	
	printf("Starting Test Harness...\n");
	startTime = clock();

	ReadParsFromFile(PARAMETERS_FILE);
	printf("This program investigates the behaviour of the geometric\n");
	printf("distribution algorithm.\n");
	printf("It tests the 'Chi-square' statistic for %d observations\n",
	  numberOfObservations);
	printf("of %d categories of events.\n", numberOfCategories);
	printf("The statistic is calculated over the probability range\n");
	printf("%lg - %lg in increments of %lg.\n\n", minProb, maxProb,
	  probIncrement);
	  
	if ((categories = (long *) calloc(sizeof(long), numberOfCategories)) ==
	  NULL) {
		NotifyError("GeomDist: Out of memory (categories).");
		exit(1);
	}
	if ((probTable = (double *) calloc(sizeof(double), numberOfCategories)) ==
	  NULL) {
		NotifyError("GeomDist: Out of memory (probTable).");
		exit(1);
	}

	if ((fp = fopen(outputFile, "w")) == NULL) {
		fprintf(stderr, "GeomDist: Cannot open file '%s'\n", outputFile);
		exit(1);
	}
	fprintf(fp, "Prob.\tChi Square\n");
	
	for (prob = minProb; prob < maxProb + probIncrement; prob +=
	  probIncrement) {
		printf("Probability %lg/%lg...\n", prob, maxProb);
		for (i = 0; i < numberOfCategories; i++)
			categories[i] = 0;
		CalculateProbabilityTable(probTable, prob, numberOfCategories);
		/* Probability table test */
		for (i = 0, sum = 0.0; i < numberOfCategories; i++)
			sum += probTable[i];
		printf("The sum of probabilities is %lg.\n", sum);
		for (i = 0; i < numberOfObservations; i++)
			categories[GeomDist_Random(prob, numberOfCategories)]++;
		for (i = 0, sum = 0.0; i < numberOfCategories; i++)
			sum += categories[i] * categories[i] / probTable[i];
		chiSquare = sum / numberOfObservations - numberOfObservations;
		fprintf(fp, "%5.3lg\t%8.6lg\n", prob, chiSquare);
	}
	fclose(fp);
	printf("This program took %lu seconds to run.\n", (clock() - startTime) / 
	  CLOCKS_PER_SEC);
	free(categories);
	
	return(0);
}