/*********************
 *
 * AnExperiment.c
 *
 * This module contains various analysis routines, which have been used in
 * real experiments.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 *********************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "AnGeneral.h"
#include "AnExperiment.h"

/**************************** CalcOnsetPeakRates *******************************/

/*
 * This function calculates the Onset peak firing rate for a signal, expected to
 * be the result of a hair cell model.
 * It finds the signal bin which has the largest value, over the specified time
 * range.
 */

BOOLEAN
CalcOnSetPeakRates_ExpAnalysis(double onsetPeakRate[], SignalDataPtr theSignal,
 double binWidth, double timeRange)
{
	int		chan;
	register	double		sum, maxBin;
	ChanLen		i, j, binWidthIndex, timeRangeIndex;
	register	ChanData	*dataPtr, *binPtr;

	if (!CheckPars_SignalData(theSignal)) {
		NotifyError("CalcOnSetPeakRates_ExpAnalysis: Signal not set.");
		return(FALSE);
	}
	if ((binWidthIndex = binWidth / theSignal->dt) > theSignal->length) {
		NotifyError("CalcOnSetPeakRates_ExpAnalysis: Illegal bin width = %g s!",
		  binWidth);
		return(FALSE);
	}
	if (timeRange < 0.0 ||
	  ((timeRangeIndex = timeRange / theSignal->dt) > theSignal->length)) {
		NotifyError("CalcOnSetPeakRates_ExpAnalysis: End time must be shorter "\
		  "then the signal length %g s.", theSignal->length * theSignal->dt);
		return(FALSE);
	}
	for (chan = 0; chan < theSignal->numChannels; chan++) {
		dataPtr = theSignal->channel[chan];
	    for (i = 0, maxBin = -LARGE_VALUE; i < timeRangeIndex - binWidthIndex;
	      i++, dataPtr++) {
	    	for (j = 0, sum = 0.0, binPtr = dataPtr; j < binWidthIndex; j++,
	    	  binPtr++)
	    		sum += *binPtr;
	    	if (sum > maxBin)
	    		maxBin = sum;
	    }
		onsetPeakRate[chan] = maxBin / binWidthIndex / theSignal->dt;
    }
    return(TRUE);

}

/***************************** CalcPSTH ***************************************/

/*
 * This routine generates a post_stimulus histogram from the EarObject's input
 * signal.  It overwrites previous data if the output signal has already been
 * initialised.
 *
 */

BOOLEAN
CalcPSTH_ExpAnalysis(EarObjectPtr data, double binWidth)
{
	int		chan;
	register	double		binSum;
	register	ChanData	*inPtr, *outPtr;
	ChanLen	i, binWidthIndex;
	
	if (data == NULL) {
		NotifyError("CalcPSTH_ExpAnalysis: EarObject not initialised.");
		return(FALSE);
	}
	if (!CheckPars_SignalData(data->inSignal)) {
		NotifyError("CalcPSTH_ExpAnalysis: Input signal not correctly set.");		
		return(FALSE);
	}
	binWidthIndex = (int) (binWidth / data->inSignal->dt);
	if ((binWidth < data->inSignal->dt) || (binWidthIndex >
	  data->inSignal->length)) {
		NotifyError("CalcPSTH_ExpAnalysis: Invalid binWidth (%g ms).",
		  MSEC(binWidth));
		return(FALSE);
	}
	SetProcessName_EarObject(data, "Post Stimulus Time Histogram (PSTH) "\
	  "analysis");
	if (!InitOutSignal_EarObject(data, data->inSignal->numChannels,
	  GetDuration_SignalData(data->inSignal) / binWidth, binWidth)) {
		NotifyError("CalcPSTH_ExpAnalysis: Cannot initialise output channels.");
		return(FALSE);
	}	
	for (chan = 0; chan < data->inSignal->numChannels; chan++) {
		inPtr = data->inSignal->channel[chan];
		outPtr = data->outSignal->channel[chan];
		for (i = 0, binSum = 0.0; i < data->inSignal->length; i++) {
			binSum += *(inPtr++);
			if (((i + 1) % binWidthIndex ) == 0) {
				if ( (outPtr - data->outSignal->channel[chan]) <
				  data->outSignal->length)
					*(outPtr++) = binSum / binWidth;
				binSum = 0.0;
			}
		}
	}
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
    return(TRUE);
	
}

/***************************** CalcPH **************************************************/

/*
 * This routine generates a period histogram from the EarObject's input signal.
 * It overwrites previous data if the output signal has already been initialised.
 * Accurate binning must be achieved, so real values mut be used in determining the
 * bin widths.
 * The period is calculated from the first minimum after the offset.
 *
 */

BOOLEAN
CalcPH_ExpAnalysis(EarObjectPtr data, double binWidth, double period, double offset)
{
	int		chan;
	double	nextCutOff, numberOfPeriods;
	register	double		binSum;
	register	ChanData	*inPtr, *outPtr;
	ChanLen	i, offsetIndex[MAX_CHANNELS], binWidthIndex, indexWithOffset;
	
	if (data == NULL) {
		NotifyError("CalcPH_ExpAnalysis: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_SignalData(data->inSignal)) {
		NotifyError("CalcPH_ExpAnalysis: Input signal not correctly set.");		
		return(FALSE);
	}
	if ((binWidth < data->inSignal->dt) || (binWidth > period)) {
		NotifyError("CalcPH_ExpAnalysis: Invalid bin width = %g ms!",
		  MSEC(binWidth));
		return(FALSE);
	}
	binWidthIndex = (ChanLen) (binWidth / data->inSignal->dt);
	if ((period < data->inSignal->dt) || ((ChanLen) (period / data->inSignal->dt)
	  > data->inSignal->length)) {
		NotifyError("CalcPH_ExpAnalysis: Invalid period = %g ms!", MSEC(period));
		return(FALSE);
	}
	SetProcessName_EarObject(data, "Period Histogram (PH) analysis");
	if (!InitOutSignal_EarObject(data, data->inSignal->numChannels,
	  (ChanLen) (period / binWidth), binWidth)) {
		NotifyError("CalcPH_ExpAnalysis: Cannot initialise output channels.");
		return(FALSE);
	}
	FindNextMinIndexes_GenAnalysis(offsetIndex, data->inSignal, offset);
	for (chan = 0; chan < data->inSignal->numChannels; chan++) {
		inPtr = data->inSignal->channel[chan];
		outPtr = data->outSignal->channel[chan];
		nextCutOff = period;
		numberOfPeriods = (data->inSignal->length - offsetIndex[chan]) *
		  data->inSignal->dt / period;
		for (i = offsetIndex[chan], binSum = 0.0; (i < data->inSignal->length); i++) {
			binSum += *(inPtr++);
			indexWithOffset = i - offsetIndex[chan];
			if ( (indexWithOffset * data->inSignal->dt) >= nextCutOff) {
				outPtr = data->outSignal->channel[chan];
				nextCutOff += period;
			}
			if ( (indexWithOffset % binWidthIndex) == 0 ) {
				if ((outPtr - data->outSignal->channel[chan]) <
				  data->outSignal->length)	/* - needed because of rounding errors. */
					*(outPtr++) += binSum / binWidth / numberOfPeriods;
				binSum = 0.0;
			}
		}
	}
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
    return(TRUE);
	
}

/***************************** CalcACF *************************************************/

/*
 * This routine generates an auto-correlation function from the EarObject's input signal.
 * It overwrites previous data if the output signal has already been initialised.
 * The function is calculated backwards from the time specified.  An exponential
 * decay function is included.
 * The same binning as for the input signal is used.
 *
 */

BOOLEAN
CalcACF_ExpAnalysis(EarObjectPtr data, double timeOffset, double timeConstant,
  double period)
{
	int		chan;
	double	*exponentDt;
	register	double	*expDtPtr;
	register	ChanData	*inPtr, *outPtr;
	ChanLen	i, timeOffsetIndex, periodIndex, deltaT;
	
	if (data == NULL) {
		NotifyError("CalcACF_ExpAnalysis: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_SignalData(data->inSignal)) {
		NotifyError("CalcACF_ExpAnalysis: Input signal not correctly set.");
		return(FALSE);
	}
	timeOffsetIndex = (ChanLen) (timeOffset / data->inSignal->dt);
	if ((timeOffsetIndex > data->inSignal->length) || (timeOffset < 0.0)) {
		NotifyError("CalcACF_ExpAnalysis: Invalid time offset!");
		return(FALSE);
	}
	if ((timeConstant < 0.0)) {
		NotifyError("CalcACF_ExpAnalysis: Invalid time constant!");
		return(FALSE);
	}
	periodIndex = (ChanLen) (period / data->inSignal->dt);
	if ((periodIndex << 1 > data->inSignal->length) || (period < 0.0)) {
		NotifyError("CalcACF_ExpAnalysis: Time period must be in range 0.0 " \
		  "- %g ms!", MSEC((data->inSignal->length >> 1) * data->inSignal->dt));
		return(FALSE);
	}
	SetProcessName_EarObject(data, "Auto Correlation Function (ACF) analysis");
	if (!InitOutSignal_EarObject(data, data->inSignal->numChannels, 
	  periodIndex, data->inSignal->dt)) {
		NotifyError("CalcACF_ExpAnalysis: Cannot initialise output channels.");
		return(FALSE);
	}
	/* Create exponent lookup table */
	if ((exponentDt = (double *) calloc(periodIndex, sizeof(double))) == NULL) {
		NotifyError("CalcACF_ExpAnalysis: Out of memory for exponent lookup table.");
		return(FALSE);
	}
	for (i = 0, expDtPtr = exponentDt; i < periodIndex; i++, expDtPtr++)
		*expDtPtr = exp( -(i * data->inSignal->dt) / timeConstant) *
		  data->inSignal->dt;
		  		
	for (chan = 0; chan < data->inSignal->numChannels; chan++) {
		outPtr = data->outSignal->channel[chan];
		for (deltaT = 0; deltaT < periodIndex; deltaT++, outPtr++) {
			inPtr = &data->inSignal->channel[chan][timeOffsetIndex - 1];
			for (i = 0, expDtPtr = exponentDt; i < periodIndex; i++, expDtPtr++,
			  inPtr--) {
				*outPtr += *inPtr * *(inPtr - deltaT) * *expDtPtr;
			}
			*outPtr /= periodIndex;
		}
	}
	free(exponentDt);
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

}

/****************************** FindThresholdIntensity ************************/

/*
 * This general routine calculates the threshold intensity of model, starting
 * from an initial 'guess' intensity.
 * It returns the respective intensity that produces the threshold intensity.
 * It works only on a single channel.
 */
 
double
FindThresholdIntensity_ExpAnalysis(EarObjectPtr (* RunModel)(void),
  BOOLEAN (* SetIntensity)(double), double targetThreshold,
  double targetAccuracy, double calcOffset, BOOLEAN *firstPass)
{
	int		i;
	double	intensityDiff, gradient, presentInputIntensity, lastOutIntensity;
	static double	lastInputIntensity;
	static double	presentOutIntensity[ANALYSIS_SINGLE_CHANNEL];
	
	if (targetAccuracy <= 0.0) {
		NotifyWarning("FindThresholdIntensity_ExpAnalysis: Illegal target "\
		  "threshold accuracy (= %g).", targetAccuracy);
		return(0.0);
	}
	if (*firstPass) {
		*firstPass = FALSE;
		lastInputIntensity = ANALYSIS_FIRST_INTENSITY_GUESS;
		if (!(* SetIntensity)(lastInputIntensity)) {
			NotifyWarning("FindThresholdIntensity_ExpAnalysis: Intensity not set.");
			return(0.0);
		}
		if (!CalcIntensities_GenAnalysis(presentOutIntensity,
		  (* RunModel)()->outSignal, calcOffset)) {
			NotifyWarning("FindThresholdIntensity_ExpAnalysis: Could not "\
			  "calculate present intensity (first pass).  Returning zero.");
			return(0,0);
		}
	}
	presentInputIntensity = 1.0 + lastInputIntensity;
	for (i = 0; i < MAXIMUM_THRESHOLD_LOOP; i++) {
		lastOutIntensity = presentOutIntensity[ANALYSIS_CHANNEL];
		if (!(* SetIntensity)(presentInputIntensity)) {
			NotifyWarning("FindThresholdIntensity_ExpAnalysis: Intensity not set.");
			return(0.0);
		}
		if (!CalcIntensities_GenAnalysis(presentOutIntensity,
		  (* RunModel)()->outSignal, calcOffset)) {
			NotifyWarning("FindThresholdIntensity_ExpAnalysis: Could not "\
			  "calculate present output intensity.  Returning zero.");
			return(0,0);
		}
		intensityDiff = targetThreshold - presentOutIntensity[ANALYSIS_CHANNEL];
		if (fabs(intensityDiff) <= targetAccuracy)
			return(presentInputIntensity);
		gradient = (presentOutIntensity[ANALYSIS_CHANNEL] - lastOutIntensity) /
		  (presentInputIntensity - lastInputIntensity);
		lastInputIntensity = presentInputIntensity;
		presentInputIntensity = intensityDiff / gradient + presentInputIntensity;
		
	}
	NotifyWarning("FindThresholdIntensity_ExpAnalysis: Looped %d times and "\
	  "couldn't\n\tcome any closer to the threshold than %g dB.\n",
	  MAXIMUM_THRESHOLD_LOOP, intensityDiff);
	return(presentInputIntensity);

}

/****************************** CalcSynchIndex ********************************/

/*
 * This routine calculates the synchronisation index/vector strength, by
 * calculating the period histogram of a model's signal.
 * This indicates the ability of a system to follow the periodicity of a
 * stimulus. The equation is from M. J. Hewitt's thesis (1992).
 */

double
CalcSynchIndex_ExpAnalysis(EarObjectPtr histogram, double pHBinWidth,
  double pHPeriod, double pHOffset)
{
	double	theta, sineSum, cosineSum, rSum;
	ChanLen	k;
	register ChanData	*r;
	
	if (!CalcPH_ExpAnalysis(histogram, pHBinWidth, pHPeriod, pHOffset)) {
		NotifyError("CalcSynchIndex_ExpAnalysis: Could not calculate period "\
		  "histogram, returning zero.");
		return(0.0);
	}
	sineSum = cosineSum = rSum = 0.0;
	r = histogram->outSignal->channel[ANALYSIS_CHANNEL];
	for (k = 0; k < histogram->outSignal->length; k++) {
        theta = PIx2 * k / histogram->outSignal->length; 
		sineSum += r[k] * cos(theta);
		cosineSum += r[k] * sin(theta);
		rSum += r[k];
	}
	if (rSum == 0.0) {
		NotifyError("CalcSynchIndex_ExpAnalysis: Zero sum, returning zero.");
		return(0.0);
	}
	return(sqrt(sineSum * sineSum + cosineSum * cosineSum) / rSum);

}

/****************************** CalcQValue ************************************/
  
/*
 * This function calculates the Q-value for a model.
 * It works only on a single channel.
 */

double
CalcQValue_ExpAnalysis(EarObjectPtr (* RunModel)(void),
  BOOLEAN (* SetStimulusFreq)(double), double centreFreq,
  double variationFromCF, double dBDownDiff, double initialFreqIncrement,
  double accuracy, double calcOffset)
{
	BOOLEAN	forewards;
	double	intensity[ANALYSIS_SINGLE_CHANNEL], maxIntensity, frequency;
	double	minTestFreq, maxTestFreq, minFreq, intensityDiff, freqIncrement;
	double	freqVariation;

	if ( !(* SetStimulusFreq)(centreFreq) ) {
		NotifyError("CalcQValue_ExpAnalysis: Could not set model frequency "\
		  "parameter (CF).  Returning Q = -1.0.");
		return(-1.0);
	}
	if (!CalcIntensities_GenAnalysis(intensity, (* RunModel)()->outSignal,
	  calcOffset) ) {
		NotifyError("CalcQValue_ExpAnalysis: Could not calculate model "\
		  "intensity (at CF).  Returning Q = -1.0.");
		return(-1.0);
	}
	maxIntensity = intensity[ANALYSIS_CHANNEL];
	initialFreqIncrement = fabs(initialFreqIncrement);
	freqVariation = centreFreq * variationFromCF;
	if ((initialFreqIncrement - freqVariation) <= 0.1)
		initialFreqIncrement = freqVariation - 1.0;
	minTestFreq = centreFreq - freqVariation;
	maxTestFreq = centreFreq + freqVariation;
	frequency = centreFreq;
	freqIncrement = -initialFreqIncrement;
	forewards = FALSE;

	while ((frequency > minTestFreq) && (frequency < maxTestFreq) ) {
		frequency += freqIncrement;
		if ( !(* SetStimulusFreq)(frequency) ) {
			NotifyError("CalcQValue_ExpAnalysis: Could not set model frequency "\
			  "parameter.  Returning Q = -1.0.");
			return(-1.0);
		}
		if (!CalcIntensities_GenAnalysis(intensity, (* RunModel)()->outSignal,
		  calcOffset) ) {
			NotifyError("CalcQValue_ExpAnalysis: Could not calculate model "\
			  "intensity.  Returning Q = -1.0.");
			return(-1.0);
		}
		if ( (intensityDiff = (maxIntensity - intensity[ANALYSIS_CHANNEL])) >=
		  dBDownDiff ) {
			if ((intensityDiff - dBDownDiff) < accuracy) {
				if (forewards)
					return(centreFreq / (frequency - minFreq));
				forewards = TRUE;
				minFreq = frequency;
				/* Now search forwards. */
				freqIncrement = initialFreqIncrement;
				frequency = centreFreq;
			} else {
				frequency -= freqIncrement;
				freqIncrement /= 2.0;
			}
		}
	}
	NotifyError("CalcQValue_ExpAnalysis: Q value %g dB down not in frequency\n"\
	  "range %g - %g Hz (variation from CF = %g).  Returning Q = -1.0.",
	  dBDownDiff, minTestFreq, maxTestFreq, variationFromCF);
	return(-1.0);
		
}

