/*********************
 *
 * AnGeneral.c
 *
 * This module contains various general analysis routines.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 *********************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "AnGeneral.h"

/***************************** FindNextMinIndexes **************************************/

/*
 * This routine finds the next minimum after a specified time and returns the index --
 * the position in the channel.
 * The agorithm looks for the first minimum after a negative gradient.  A negative
 * gradient is where cPtr(i + 1) - cPtr(i) < 0.
 * The results are returned in the index[] function argument.
 * If the search is successful, then the function will return TRUE.
 */

BOOLEAN
FindNextMinIndexes_GenAnalysis(ChanLen index[], SignalDataPtr theSignal, double offset)
{
	BOOLEAN	found, negativeGradient;
	int		chan;
	ChanLen	i, offsetIndex;
	register	ChanData	*cPtr, lastValue;

	if (!CheckPars_SignalData(theSignal)) {
		NotifyError("FindNextMinIndexes_GenAnalysis: Signal not set correctly.");
		return(FALSE);
	}
	offsetIndex = (ChanLen) (offset /  theSignal->dt);
	if (offsetIndex > theSignal->length) {
		NotifyError("FindNextMinIndexes_GenAnalysis: Offset value is longer than " \
		  "the signal duration (Offset = %u).", offsetIndex);
		return(FALSE);
    }
    for (chan = 0; chan < theSignal->numChannels; chan++) {
	    cPtr = theSignal->channel[chan];
	    lastValue = cPtr[offsetIndex];
	    negativeGradient = FALSE;
	    for (i = offsetIndex + 1, found = FALSE; (i < theSignal->length - 1) && !found;
	      i++) {
		    if (!negativeGradient)
		    	negativeGradient = (cPtr[i] - lastValue < 0);
	    	if (negativeGradient && (cPtr[i] < cPtr[i + 1]) ) {
	    		index[chan] = i;
	    		found = TRUE;
	    		break;
	    	}
	    	lastValue = cPtr[i];
	    }
	    if (!found) {
	    	NotifyWarning("FindNextMinIndexes_GenAnalysis: Minimum not found. "\
	    	  "Returning offset index = %u.", offsetIndex);
	    	index[chan] = offsetIndex;
	    	return(FALSE);
	    }
	}
	return(TRUE);

}

/***************************** FindNextPeakIndexes *************************************/

/*
 * This routine finds the next peak after a specified time and returns the index --
 * the position in the channel.
 * The agorithm looks for the first maximum after a positive gradient.  A positive
 * gradient is where cPtr(i + 1) - cPtr(i) > 0.
 * The results are returned in the index[] function argument.
 * If the search is successful, then the function will return TRUE.
 */

BOOLEAN
FindNextPeakIndexes_GenAnalysis(ChanLen index[], SignalDataPtr theSignal, double offset)
{
	BOOLEAN	found, positiveGradient;
	int		chan;
	ChanLen	i, offsetIndex;
	register	ChanData	lastValue, *cPtr;

	if (!CheckPars_SignalData(theSignal)) {
		NotifyError("FindNextPeakIndexes_GenAnalysis: Signal not set correctly.");
		return(FALSE);
	}
	offsetIndex = (ChanLen) (offset /  theSignal->dt);
	if (offsetIndex > theSignal->length) {
		NotifyError("FindNextPeakIndexes_GenAnalysis: Offset value is longer than " \
		  "the signal duration (Offset = %u).", offsetIndex);
		return(FALSE);
	}
	for (chan = 0; chan < theSignal->numChannels; chan++) {
	    cPtr = theSignal->channel[chan];
	    lastValue = cPtr[offsetIndex];
	    positiveGradient = FALSE;
	    for (i = offsetIndex + 1, found = FALSE; (i < theSignal->length - 1) && !found;
	      i++) {
	      	if (!positiveGradient)
	      		positiveGradient = (cPtr[i] - lastValue > 0.0);
	    	if (positiveGradient && (cPtr[i] > cPtr[i + 1]) ) {
	    		index[chan] = i;
	    		found = TRUE;
	    		break;
	    	}
	    	lastValue = cPtr[i];
	    }
		if (!found) {
	    	NotifyWarning("FindNextPeakIndexes_GenAnalysis: Peak not found. Returning "\
	    	  "offset index = %u, channel[%d].", offsetIndex, chan);
	    	index[chan] = offsetIndex;
	    	return(FALSE);
	    }
	}
	return(TRUE);

}

/***************************** FindPeakPositions ***************************************/

/*
 * This routine calculates the positions of the n'th positive peaks of a signal
 * data set.
 */

void
FindPeakPositions_GenAnalysis(double firstPeaks[], int peakNumber,
  SignalDataPtr theSignal)
{
	BOOLEAN	positiveGradient;
	int		chan, peakCount;
	ChanLen	i;
	register	ChanData	lastValue, *cPtr;

	if (!CheckPars_SignalData(theSignal)) {
		NotifyError("FindPeakPositions_GenAnalysis: Signal not set correctly.");
		exit(1);
	}
	for (chan = 0; chan < theSignal->numChannels-1; chan++) {
		firstPeaks[chan] = 0.0;
		cPtr = theSignal->channel[chan];
	    positiveGradient = FALSE;
		for (i = 1, lastValue = cPtr[i], peakCount = 0; i < theSignal->length - 1; i++) {
	      	if (!positiveGradient)
	      		positiveGradient = (cPtr[i] - lastValue > 0.0);
	    	if (positiveGradient && (cPtr[i] > cPtr[i + 1]) )
				if (++peakCount == peakNumber) {
					firstPeaks[chan] = i * theSignal->dt;
					break;
				}
			lastValue = cPtr[i];
		}

    }

}

/**************************** CalcIntensities ******************************************/

/*
 * This function returns the intensities in the array intensityValue passed as a
 * function argument.
 * The intensity is calculated starting from the offset position supplied.
 */
 
BOOLEAN
CalcIntensities_GenAnalysis(double intensityValue[], SignalDataPtr theSignal,
  double offsetTime)
{
	int			i;
	ChanLen		j, offsetIndex;
	register	double		sum;
	register	ChanData	*dataPtr;
	
	if (!CheckPars_SignalData(theSignal)) {
		NotifyError("CalcIntensities_GenAnalysis: Signal not set.");
		return(FALSE);
	}
	offsetIndex = (int) (offsetTime / theSignal->dt);
	if (offsetIndex >= theSignal->length) {
		NotifyError("CalcIntensities_GenAnalysis: Illegal offset = %g (s).", offsetTime);
		return(FALSE);
	}
	for (i = 0; i < theSignal->numChannels; i++) {
		dataPtr = &theSignal->channel[i][offsetIndex];
		for (j = offsetIndex, sum = 0.0; j < theSignal->length; j++, dataPtr++)
			sum += *dataPtr * *dataPtr;
		intensityValue[i] = DB_SPL(sqrt(sum / (theSignal->length - offsetIndex)));
	}
	return(TRUE);
	
}

/***************************** AveChannels *********************************************/

/*
 * This routine sums the channels for each time step of a signal data structure, then
 * calculates the average by dividing by the number of channels.
 * Effectively it reduces multiple channel data to a single channel.
 * It overwrites previous data if the output signal has already been initialised.
 *
 */

BOOLEAN
AveChannels_GenAnalysis(EarObjectPtr data)
{
	int		chan;
	ChanLen	i;
	ChanData	*outPtr;
	
	if (data == NULL) {
		NotifyError("AveChannels_GenAnalysis: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_SignalData(data->inSignal)) {
		NotifyError("CalcPH_GenAnalysis: Input signal not correctly set.");		
		return(FALSE);
	}
	SetProcessName_EarObject(data, "Average accross channels analysis");
	if (!InitOutSignal_EarObject(data, ANALYSIS_SINGLE_CHANNEL,
	  data->inSignal->length, data->inSignal->dt)) {
		NotifyError("CalcPH_GenAnalysis: Cannot initialise output channel.");
		return(FALSE);
	}
	;
	for (i = 0, outPtr = data->outSignal->channel[ANALYSIS_CHANNEL]; i <
	  data->inSignal->length; i++, outPtr++) {
		for (chan = 0; chan < data->inSignal->numChannels; chan++)
			*outPtr += data->inSignal->channel[chan][i];
		*outPtr /= data->inSignal->numChannels;
	}
			
   if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
    return(TRUE);
	
}

/*************************** CalcAverages *************************************/

/*
 * This routine calculates the average value for each channel.
 */

BOOLEAN
CalcAverages_GenAnalysis(double average[], SignalDataPtr theSignal,
  double startPeriod, double endPeriod)
{
	int		chan;
	register	double	sum;
	ChanLen	i, startIndex, endIndex;
	register	ChanData	*dataPtr;
	
	if (!CheckPars_SignalData(theSignal)) {
		NotifyError("CalcAverages_GenAnalysis: Signal not set.");
		return(FALSE);
	}
	if (startPeriod >= endPeriod) {
		NotifyError("CalcAverages_GenAnalysis: Start time must be earlier than "\
		 "end time for the averaging period.");
		return(FALSE);
	}
	if (startPeriod < 0.0 ||
	  ((startIndex = startPeriod / theSignal->dt) > theSignal->length)) {
		NotifyError("CalcAverages_GenAnalysis: Start time must be in range %g - "\
		  "%g s.", 0.0, theSignal->length * theSignal->dt);
		return(FALSE);
	}
	if (endPeriod < 0.0 ||
	  ((endIndex = endPeriod / theSignal->dt) > theSignal->length)) {
		NotifyError("CalcAverages_GenAnalysis: End time mus be in range %g - "\
		  "%g s.", 0.0, theSignal->length * theSignal->dt);
		return(FALSE);
	}
	for (chan = 0; chan < theSignal->numChannels; chan++) {
		dataPtr = &theSignal->channel[chan][startIndex];
		for (i = startIndex, sum = 0.0; i < endIndex; i++)
			sum += *dataPtr++;
		average[chan] = sum / (endIndex - startIndex);
	}
	return(TRUE);
	
}

/*************************** CalcHalfCompAverages *****************************/ 

/*
 * This routine calculates the half-wave rectified average of a wave-form.
 * (Only the positive sections of the wave are summed).
 * It was written by Trevor Shackleton, Sussex, U.K.
 */

BOOLEAN
CalcHalfCompAverages_GenAnalysis(double average[], SignalDataPtr theSignal,
  double startPeriod, double endPeriod)
{
	int		chan;
	double  sum, x;
	ChanLen i, startIndex, endIndex;
	ChanData	*dataPtr;
	
	if (!CheckPars_SignalData(theSignal)) {
		NotifyError("CalcHalfCompAverages_GenAnalysis: Signal not set.");
		return(FALSE);
	}
	if (startPeriod >= endPeriod) {
		NotifyError("CalcHalfCompAverages_GenAnalysis: Start time must be "\
		  "earlier than end time for the averaging period.");
		return(FALSE);
	}
	if (startPeriod < 0.0 ||
	  ((startIndex = startPeriod / theSignal->dt) > theSignal->length)) {
		NotifyError("CalcHalfCompAverages_GenAnalysis: Start time must be in "\
		  "range %lg - %lg s.", 0.0, theSignal->length * theSignal->dt);
		return(FALSE);
	}
	if (endPeriod < 0.0 ||
	  ((endIndex = endPeriod / theSignal->dt) > theSignal->length)) {
		NotifyError("CalcHalfCompAverages_GenAnalysis: End time mus be in "\
		  "range %lg - %lg s.", 0.0, theSignal->length * theSignal->dt);
		return(FALSE);
	}
	for (chan = 0; chan < theSignal->numChannels; chan++) {
		dataPtr = &theSignal->channel[chan][startIndex];
		for (i = startIndex, sum = 0.0; i < endIndex; i++)
			sum += ((x = *dataPtr++) > 0) ? x : 0;
		average[chan] = sum / (endIndex - startIndex);
	}
	return(TRUE);
	
}

/*************************** CalcHalfRareAverages *****************************/ 

/*
 * This routine calculates the half-wave rectified average of a wave-form.
 * (Only the negative sections of the wave are summed).
 * It was written by Trevor Shackleton, Sussex, U.K.
 */

BOOLEAN
CalcHalfRareAverages_GenAnalysis(double average[], SignalDataPtr theSignal,
  double startPeriod, double endPeriod)
{
	int		chan;
	double  sum, x;
	ChanLen i, startIndex, endIndex;
	register ChanData	*dataPtr;
	
	if (!CheckPars_SignalData(theSignal)) {
		NotifyError("CalcHalfRareAverages_GenAnalysis: Signal not set.");
		return(FALSE);
	}
	if (startPeriod >= endPeriod) {
		NotifyError("CalcHalfRareAverages_GenAnalysis: Start time must be "\
		  "earlier than end time for the averaging period.");
		return(FALSE);
	}
	if (startPeriod < 0.0 ||
	  ((startIndex = startPeriod / theSignal->dt) > theSignal->length)) {
		NotifyError("CalcHalfRareAverages_GenAnalysis: Start time must be in "\
		  "range %lg - %lg s.", 0.0, theSignal->length * theSignal->dt);
		return(FALSE);
	}
	if (endPeriod < 0.0 ||
	  ((endIndex = endPeriod / theSignal->dt) > theSignal->length)) {
		NotifyError("CalcHalfRareAverages_GenAnalysis: End time mus be in "\
		  "range %lg - %lg s.", 0.0, theSignal->length * theSignal->dt);
		return(FALSE);
	}
	for (chan = 0; chan < theSignal->numChannels; chan++) {
		dataPtr = &theSignal->channel[chan][startIndex];
		for (i = startIndex, sum = 0.0; i < endIndex; i++)
			sum += ((x = *dataPtr++) < 0) ? x : 0;
		average[chan] = sum / (endIndex - startIndex);
	}
	return(TRUE);
	
}


