/******************
 *
 * FiDataFile.c
 *
 * This Filing module deals with the writing reading of data files.
 *
 * The sound format file routines were revised from sources from Lyon's
 * Cochlear Model, The Program, Malcolm Slaney, Advanced Technology Group,
 * Apple Computer, Inc., malcom@apple.com 1988.
 *
 * - except for the MicroSoft Wave file format routines.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ******************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "UtUIEEEFloat.h"
#include "UtUPortableIO.h"
#include "FiParFile.h"
#include "FiDataFile.h"

/******************************************************************************/
/****************************** Constant definitions **************************/
/******************************************************************************/

/******************************************************************************/
/****************************** Global variables ******************************/
/******************************************************************************/

DataFile dataFile = {

	FALSE,		/* mustSetMaxSamplesFlag */
	"",			/* name */
	2,			/* wordSize */
	0,			/* endian */
	8000.0,		/* defaultSampleRate */
	-1,			/* maxSamples */
	Write16BitsHighLow,
	Write32BitsHighLow,
	Read16BitsHighLow,
	Read32BitsHighLow
	
}, *dataFilePtr = &dataFile;

/******************************************************************************/
/****************************** Subroutines & functions ***********************/
/******************************************************************************/

/**************************** Init ********************************************/

/*
 * This routine initialises the DataFile module parameters to default values.
 */

void
Init_DataFile(void)
{
	dataFilePtr->mustSetMaxSamplesFlag = FALSE;
	dataFilePtr->name[0] = '\0';
	dataFilePtr->wordSize = 2;
	dataFilePtr->endian = 0;
	dataFilePtr->defaultSampleRate = 8000.0;
	dataFilePtr->maxSamples = -1;
	dataFilePtr->Write16Bits = Write16BitsHighLow;
	dataFilePtr->Write32Bits = Write32BitsHighLow;
	dataFilePtr->Read16Bits = Read16BitsHighLow;
	dataFilePtr->Read32Bits = Read32BitsHighLow;
	
}

/**************************** SetWordSize_DataFile ****************************/

/*
 * This routine sets the word size for the binary file data writing.
 * Only the values 1 or 2 a valid.
 */

BOOLEAN
SetWordSize_DataFile(int wordSize)
{
	if (wordSize < 1 || wordSize > 2) {
		NotifyError("SetWordSize_DataFile: Illegal value - must be 1 or 2, not "\
		  "%d.", wordSize);
		return(FALSE);
	}
	dataFilePtr->wordSize = wordSize;
	return(TRUE);
	
}

/**************************** SetEndianRW_DataFile ****************************/

/*
 * Sets the reading and writing routines of the 'dataFile' structure.
 */

void
SetEndianRW_DataFile(int endian)
{
	switch (endian) {
	case BIG_ENDIAN_DATA_FILE:
	default:
		dataFilePtr->Write16Bits = Write16BitsHighLow;
		dataFilePtr->Write32Bits = Write32BitsHighLow;
		dataFilePtr->Read16Bits = Read16BitsHighLow;
		dataFilePtr->Read32Bits = Read32BitsHighLow;
		break;
	case LITTLE_ENDIAN_DATA_FILE:
		dataFilePtr->Write16Bits = Write16BitsLowHigh;
		dataFilePtr->Write32Bits = Write32BitsLowHigh;
		dataFilePtr->Read16Bits = Read16BitsLowHigh;
		dataFilePtr->Read32Bits = Read32BitsLowHigh;
		break;
	} /* Switch */

}

/**************************** SetEndian_DataFile ******************************/

/*
 * This routine sets the endedness of binary data file reading/writing to
 * big- or little-endian format.
 * It assigns the appropriate routine names to the 'dataFile' global
 * structure.
 * Can be set to default (big-), big- or little-endian.
 *
 */

BOOLEAN
SetEndian_DataFile(int endian)
{
	if (endian < 0 || endian > 2) {
		NotifyError("SetEndian_DataFile: Illegal value - must be 0, 1 or 2.\n");
		return(FALSE);
	}
	dataFilePtr->endian = endian;
	SetEndianRW_DataFile(endian);
	return(TRUE);
}

/**************************** Format_DataFile *********************************/

/*
 * This routine returns the pointer to a data file format.
 *
 */

char *
Format_DataFile(int i)
{
	static char	*soundFormat[] = {
		"AIF",
		"AIFF",
		"DAT",
		"RAW",
		"WAV",
		"WAVE",
		""
	};
	
	return(soundFormat[i]);

}

/**************************** FileLength **************************************/

/*
 * This function returns with the length of a file.
 * It assumes that the file has alread been opened.
 */

ChanLen
FileLength_DataFile(FILE *fp)
{
	ChanLen	pos, end;

	pos = ftell(fp);
	fseek(fp, SEEK_SET, SEEK_END);
	end = ftell(fp);
	fseek(fp, pos, 0);
	return(end);
	
}

/**************************** WriteOutSignal **********************************/

/*
 * This routine sends the output signal of an EarObject to file.
 * It assumes that the EarObject has been initialised.
 * It uses the OutputToFile_SignalData routine, which checks that the
 * SignalData field has been correctly initialised.
 */
 
BOOLEAN
WriteASCII_DataFile(char *fileName, EarObjectPtr data)
{
	if (!OutputToFile_SignalData(fileName, data->outSignal)) {
		if (data->processName != NULL)
			NotifyWarning("WriteASCII_DataFile: Data from EarObject: "\
			  "%s, has not been output to file.", data->processName);
		else
			NotifyWarning("WriteASCII_DataFile: Data from EarObject: "\
			  "<no name>, has not been output to file.");
		return(FALSE);
	} else {
		return(TRUE);
	}
		
}

/**************************** ReadAIFFParams **********************************/

/*
 * This routine reads the parameters from the chunks of an AIFF file (stream).
 * It does not read the sound data itself, but does return its position
 * in the file.
 * At present only the common chunk and sound data chunk are used, all other
 * chunks are ignored.
 */

void
ReadAIFFParams_DataFile(FILE *fp, AIFFParamsPtr p)
{
	int32	subSize;
	
	while (p->chunkSize > 0){
		p->chunkSize -= 4;
		switch(dataFilePtr->Read32Bits(fp)){
		case	AIFF_COMM:
			p->chunkSize -= subSize = dataFilePtr->Read32Bits(fp);
			p->numChannels = dataFilePtr->Read16Bits(fp);
			subSize -= 2;
			p->numSampleFrames = dataFilePtr->Read32Bits(fp);
			subSize -= 4;
			p->sampleSize = dataFilePtr->Read16Bits(fp);
			subSize -= 2;
			p->sampleRate = ReadIeeeExtendedHighLow(fp);
			subSize-=10;
			while (subSize > 0){
				getc(fp);
				subSize--;
			}
			break;
		case	AIFF_SSND:
			p->chunkSize -= subSize = dataFilePtr->Read32Bits(fp);
			p->offset = dataFilePtr->Read32Bits(fp);
			subSize -= 4;
			p->blockSize = dataFilePtr->Read32Bits(fp);
			subSize -= 4;
			p->soundPosition = ftell(fp) + p->offset;
			while (subSize > 0){
				getc(fp);
				subSize--;
			}
			break;
		default:
			p->chunkSize -= subSize = dataFilePtr->Read32Bits(fp);
			while (subSize > 0){
				getc(fp);
				subSize--;
			}
			break;
		} /* switch */
	}


}

/**************************** ReadSoundChunkData ******************************/

/*
 * This routine reads the sound data from an AIFF file stream.
 * It expects the soundPosition for the file stream to have been
 * set to the start of the sound data.
 * It sets the dataFilePtr->wordSize field.
 */

BOOLEAN
ReadSoundChunkData_DataFile(FILE *fp, EarObjectPtr data, AIFFParamsPtr p)
{
	ChanLen	i, j, length;

	fseek(fp, p->soundPosition, 0);
	length = p->numSampleFrames;

	if (dataFilePtr->maxSamples >= 0 && length > dataFilePtr->maxSamples)
		length = dataFilePtr->maxSamples;	/* The original code had this */
										/* in, so I left it. */		
	if (!InitOutSignal_EarObject(data, p->numChannels, length, 1.0 /
	  p->sampleRate) ) {
		NotifyError("ReadSoundChunkData_DataFile: Cannot initialise output "\
		  "signal");
		return(FALSE);
	}
	switch (dataFilePtr->wordSize = (p->sampleSize + 7) / 8){
	case 1:
		for (i=0; i < length; i++)
			for (j = 0; j < p->numChannels; j++)
				data->outSignal->channel[j][i] = ReadByte(fp) / 127.0;
		break;
	case 2:
		for (i=0; i < length; i++)
			for (j = 0; j < p->numChannels; j++)
				data->outSignal->channel[j][i] =
				  dataFilePtr->Read16Bits(fp) / 32768.0;
		break;
	case 4:
		for (i=0; i < length; i++)
			for (j = 0; j < p->numChannels; j++)
				data->outSignal->channel[j][i] =
				  dataFilePtr->Read32Bits(fp) / 32768.0 / 65536.0;
		break;
	default:
		NotifyError("ReadSoundChunk_DataFile: Unsupported sample size.");
		return(FALSE);
		break;
	}
	return(TRUE);

}

/**************************** ReadAIFF ****************************************/

/*
 * This function reads a file in AIFF format.
 * The data is stored in the output signal of an EarObject which will have the
 * same number of channels as the input signal.
 * If a dash, '-.sufffix', is given as the file name, then the data will be 
 * read from the standard input.
 * It returns TRUE if successful.
 */

BOOLEAN
ReadAIFF_DataFile(char *fileName, EarObjectPtr data)
{
	FILE	*fp;
	AIFFParams	pars;

	if (fileName[0] != '-') {	
		if ((fp = fopen(fileName, "rb")) == NULL) {
			NotifyError("ReadAIFF_DataFile: Couldn't open %s for reading.",
			  fileName);
			return(FALSE);
		}
	} else {
		fp = stdin;
	}
	SetProcessName_EarObject(data, "'%s' AIFF file", fileName);
	if (dataFilePtr->endian == 0)
		SetEndianRW_DataFile(BIG_ENDIAN_DATA_FILE);
	if (dataFilePtr->Read32Bits(fp) != AIFF_FORM) {
		NotifyError("ReadAIFF_DataFile: Couldn't find initial chunk in "\
		  "file '%s'.", fileName);
		return(FALSE);
	}
	pars.chunkSize = dataFilePtr->Read32Bits(fp);
	if (dataFilePtr->Read32Bits(fp) != AIFF_AIFF){
		NotifyError("ReadAIFF_DataFile: Couldn't find AIFF chunk in file '%s'.",
		  fileName);
		return(FALSE);
	}

	ReadAIFFParams_DataFile(fp, &pars);

	if (!pars.soundPosition) {
		NotifyError("ReadAIFF_DataFile: Didn't find a SSND chunk in file '%s'.",
		  fileName);
		return(FALSE);
	}

	if (!ReadSoundChunkData_DataFile(fp, data, &pars)) {
		NotifyError("ReadAIFF_DataFile: Couldn't read the SSND chunk in "\
		  "file '%s'.",fileName);
		return(FALSE);
	}
	if (fp != stdin)
		fclose(fp);
	return(TRUE);

}

/*************************** GetASCIISamplingInterval *************************/

/*
 * This routine determines the sampling interval from an ascii file.
 * It expects the file to have at least two data columns, and assumes.
 * that the first column has equi-spaced time values.
 * The sampling interval returned is the difference the first two column 1
 * entries.
 * The file is rewound after interogation.
 */

double
GetASCIISamplingInterval_DataFile(FILE *fp)
{
	char	line[MAXLINE];
	double	initialTime, nextTime;

	if (fgets(line, MAXLINE, fp) == NULL) {
		NotifyError("GetASCIISamplingInterval_DataFile: Couldn't read first "\
		  "line from file.");
		return(-1.0);
	}
	sscanf(line, "%lf", &initialTime);
	if (fgets(line, MAXLINE, fp) == NULL) {
		NotifyError("GetASCIISamplingInterval_DataFile: Couldn't read second "\
		  "line from file.");
		return(-1.0);
	}
	sscanf(line, "%lf", &nextTime);
	rewind(fp);
	return(nextTime - initialTime);

}

/*************************** NumberOfColumns **********************************/

/*
 * This function determines the number of data columns in a file.
 */

int
NumberOfColumns_DataFile(FILE *fp)
{
	BOOLEAN	number;
	char	*c, line[MAXLINE];
	int		numcols;

	if (fgets(line, MAXLINE, fp) == NULL)
		return(0);
	/* Find the start of the first number */
	for(c = line; !isdigit(*c) && (*c != '.') && (*c != '-'); c++);
	for (numcols = 0, number = FALSE; *c != '\0'; c++) {
		if (isdigit(*c) || (*c == '.') || (*c == '-') ||
		  (*c == '+') || (*c == 'E') || (*c == 'e')) {
			if (!number) {
				numcols++;
				number = TRUE;
			} 
		} else
			number = FALSE;
	}
	rewind(fp);
	return(numcols);
	
} /* NumberOfColumns_DataFile */

/**************************** ReadASCII ***************************************/

/*
 * This function reads a file in ASCII stream format.
 * The data is stored in the output signal of an EarObject.
 * If the file has more than one data column, then the first column will be
 * used to determine the sampling interval, thereafter it will be ignored.
 * This is the default file, and is assumed if no suffix is given
 * If a dash, '-.suffix', is given as the file name, then the data will be 
 * read from the standard input.
 * It returns TRUE if successful.
 */

BOOLEAN
ReadASCII_DataFile(char *fileName, EarObjectPtr data)
{
	BOOLEAN	singleColumn;
	int		chan, numColumns;
	double	dt, dummy;
	ChanLen	i, length;
	FILE	*fp;

	dataFilePtr->mustSetMaxSamplesFlag = TRUE;
	if (!CheckPars_DataFile()) {
		NotifyError("ReadASCII_DataFile: Parameters have not been "\
		  "correctly set.");
		return(FALSE);
	}
	dataFilePtr->mustSetMaxSamplesFlag = FALSE;
	if (fileName[0] != '-') {	
		if ((fp = fopen(fileName, "r")) == NULL) {
			NotifyError("ReadASCII_DataFile: Couldn't open '%s' for reading.",
			  fileName);
			return(FALSE);
		}
	} else {
		fp = stdin;
	}
	if ((numColumns = NumberOfColumns_DataFile(fp)) == 0) {
		NotifyError("ReadASCII_DataFile: Illegal no. of columns: '%s'.",
		  fileName);
		return(FALSE);
	}
	singleColumn = (numColumns == 1);
	if (singleColumn)
		dt = 1.0 / dataFilePtr->defaultSampleRate;
	else {
		numColumns--;
		if ((dt = GetASCIISamplingInterval_DataFile(fp)) < 0.0) {
			NotifyError("ReadASCII_DataFile: Illegal sampling interval: '%s'.",
			  fileName);
			return(FALSE);
		}
	}
	SetProcessName_EarObject(data, "'%s' ASCII (dat) file", fileName);
	length = dataFilePtr->maxSamples;
	if (!InitOutSignal_EarObject(data, numColumns, length, dt) ) {
		NotifyError("ReadASCII_DataFile: Cannot initialise output signal");
		return(FALSE);
	}
	for (i = 0; i < length; i++)
		if (singleColumn)
			fscanf(fp, "%lf", &data->outSignal->channel[0][i]);
		else {
			fscanf(fp, "%lf", &dummy);
			for (chan = 0; chan < data->outSignal->numChannels; chan++)
				fscanf(fp, "%lf", &data->outSignal->channel[chan][i]);
		}
	if (fp != stdin)
		fclose(fp);
	return(TRUE);

}

/**************************** ReadRaw *****************************************/

/*
 * This function reads a file in raw byte-stream format.
 * The data is stored in the output signal of an EarObject which will have
 * one channel.
 * This is the default file, and is assumed if no suffix is given
 * If a dash, '-.suffix', is given as the file name, then the data will be 
 * read from the standard input.
 * It returns TRUE if successful.
 */

BOOLEAN
ReadRaw_DataFile(char *fileName, EarObjectPtr data)
{
	ChanLen	i, length;
	FILE	*fp;

	dataFilePtr->mustSetMaxSamplesFlag = TRUE;
	if (!CheckPars_DataFile()) {
		NotifyError("ReadRaw_DataFile: Parameters have not been "\
		  "correctly set.");
		return(FALSE);
	}
	dataFilePtr->mustSetMaxSamplesFlag = FALSE;
	if (fileName[0] != '-') {	
		if ((fp = fopen(fileName, "rb")) == NULL) {
			NotifyError("ReadRaw_DataFile: Couldn't open '%s' for reading.",
			  fileName);
			return(FALSE);
		}
	} else {
		fp = stdin;
	}
	SetProcessName_EarObject(data, "'%s' byte (raw) file", fileName);
	length = dataFilePtr->maxSamples;
	if (!InitOutSignal_EarObject(data, 1, length, 1.0 /
	  dataFilePtr->defaultSampleRate) ) {
		NotifyError("ReadRaw_DataFile: Cannot initialise output signal");
		return(FALSE);
	}
	for (i = 0; i < length; i++)
		if (dataFilePtr->wordSize == 1)
			data->outSignal->channel[0][i] = ReadByte(fp) / 127.0;
		else
			data->outSignal->channel[0][i] = dataFilePtr->Read16Bits(fp) / 32767.0;
	if (fp != stdin)
		fclose(fp);
	return(TRUE);

}

/**************************** ReadWaveHeader **********************************/

/*
 * This function reades a waver header structure from file.
 * It returns FALSE if it fails.
 */

BOOLEAN
ReadWaveHeader_DataFile(FILE *fp, WaveHeaderPtr p)
{
	if ((p->identifier = dataFilePtr->Read32Bits(fp)) != WAVE_RIFF) {
		NotifyError("ReadWaveHeader: Could not find RIFF identifier.");
		return(FALSE);
	}
	p->length = dataFilePtr->Read32Bits(fp);
	if ((p->chunkType = dataFilePtr->Read32Bits(fp)) != WAVE_WAVE) {
		NotifyError("ReadWaveHeader: Could not find chunk identifier.");
		return(FALSE);
	}
	if ((p->subChunkType = dataFilePtr->Read32Bits(fp)) != WAVE_FMT) {
		NotifyError("ReadWaveHeader: Could not find sub-chunk identifier.");
		return(FALSE);
	}
	p->subChunkLength = dataFilePtr->Read32Bits(fp);
	p->format = dataFilePtr->Read16Bits(fp);
	p->numChannels = dataFilePtr->Read16Bits(fp);
	p->sampleRate = dataFilePtr->Read32Bits(fp);
	p->byteRate = dataFilePtr->Read32Bits(fp);
	p->sampleBytes = dataFilePtr->Read16Bits(fp);
	p->sampleSize = dataFilePtr->Read16Bits(fp);
	if ((p->dataChunk = dataFilePtr->Read32Bits(fp)) != WAVE_DATA) {
		NotifyError("ReadWaveHeader: Could not find data chunk identifier.");
		return(FALSE);
	}
	p->dataChunkLength = dataFilePtr->Read32Bits(fp);
	return(TRUE);

}

/**************************** ReadWave ****************************************/

/*
 * This function reads a file in Microsoft Wave format.
 * The data is stored in the output signal of an EarObject which will have
 * one channel.
 * This is the default file, and is assumed if no suffix is given
 * If a dash, '-.suffix', is given as the file name, then the data will be 
 * read from the standard input.
 * I am not quite certain about the stereo format.
 * It returns TRUE if successful.
 */

BOOLEAN
ReadWave_DataFile(char *fileName, EarObjectPtr data)
{
	ChanLen	i, j, length;
	double	scale;
	FILE	*fp;
	WaveHeader	pars;
	int16	(* ReadSample)(FILE *);

	if (fileName[0] != '-') {	
		if ((fp = fopen(fileName, "rb")) == NULL) {
			NotifyError("ReadWave_DataFile: Couldn't open '%s' for reading.",
			  fileName);
			return(FALSE);
		}
	} else {
		fp = stdin;
	}
	if (dataFilePtr->endian == 0)
		SetEndianRW_DataFile(LITTLE_ENDIAN_DATA_FILE);
	if (!ReadWaveHeader_DataFile(fp, &pars)) {
		NotifyError("ReadWave_DataFile: Could not reade file '%s'.", fileName);
		return(FALSE);
	}
	SetProcessName_EarObject(data, "'%s' Microsoft Wave file", fileName);
	length = (ChanLen) pars.dataChunkLength;
	if (dataFilePtr->maxSamples >= 0 && length > dataFilePtr->maxSamples)
		length = dataFilePtr->maxSamples;	/* The original code had this */
										/* in, so I left it. */
	switch (pars.sampleSize) {
	case	8:
		ReadSample = (int16 (*)(FILE *)) ReadByte;
		scale = (double) (0xff >> 1);
		break;
	case	16:
		ReadSample = dataFilePtr->Read16Bits;
		scale = (double) (0xffff >> 1);
		break;
	default:
		NotifyError("ReadWave_DataFile: Unsupported sample bit size = %d.",
		  pars.sampleSize);
		return(FALSE);
	} /* switch */		
	if (!InitOutSignal_EarObject(data, (int) pars.numChannels, length, 1.0 /
	  pars.sampleRate)) {
		NotifyError("ReadWave_DataFile: Cannot initialise output signal");
		return(FALSE);
	}
	for (i = 0; i < length; i++)
		for (j = 0; j < pars.numChannels; j++) {
			data->outSignal->channel[j][i] = (ChanData) ((* ReadSample)(fp) /
			  scale);
		}
	if (fp != stdin)
		fclose(fp);
	return(TRUE);

}

/**************************** WriteSignal **************************************/

/*
 * This function writes out the channel data, dependent upon the
 * wordSize setting.
 * It assumes that the signal has been initialised.
 */

BOOLEAN
WriteSignal_DataFile(FILE *fp, SignalDataPtr signal, int wordSize)
{
	int		j;
	double	normalise, maxValue;
	ChanLen	i;
	
	maxValue = -LARGE_VALUE;
	for (i = 0; i < signal->length; i++)
		for (j = 0; j < signal->numChannels; j++)
			if (fabs(signal->channel[j][i]) > maxValue)
				maxValue = fabs(signal->channel[j][i]);
	normalise = (int16) (0xffff >> (17 - dataFilePtr->wordSize * 8)) / maxValue;
	switch (wordSize) {
	case	1:
		for (i = 0; i < signal->length; i++)
			for (j = 0; j < signal->numChannels; j++)
				Write8Bits(fp, (int) (signal->channel[j][i] * normalise + 0.5));
		break;
	case	2:
		for (i = 0; i < signal->length; i++)
			for (j = 0; j < signal->numChannels; j++)
				dataFilePtr->Write16Bits(fp, (int) (signal->channel[j][i] *
				  normalise + 0.5));
		break;
	case	4:
		for (i = 0; i < signal->length; i++)
			for (j = 0; j < signal->numChannels; j++)
				dataFilePtr->Write32Bits(fp, (int) (signal->channel[j][i] *
				  normalise + 0.5));
		break;
	default:
		NotifyError("WriteSignal_DataFile: Unsupported sample size.");
		return(FALSE);
		break;
	} /* Switch */
	return(TRUE);

}

/**************************** WriteWaveHeader *********************************/

/*
 * This routine writes the header structure for the Wave format to a file
 * (stream).
 * It does not write the sound data itself.
 */

void
WriteWaveHeader_DataFile(FILE *fp, EarObjectPtr data)
{
	int32		totalLength, dataLength;
	
	dataLength = data->outSignal->length * data->outSignal->numChannels;
	totalLength = 44 + dataLength;

	dataFilePtr->Write32Bits(fp, WAVE_RIFF);
	dataFilePtr->Write32Bits(fp, totalLength);		/* Total file length */
	dataFilePtr->Write32Bits(fp,  WAVE_WAVE);	/* Type of Chunk */

	dataFilePtr->Write32Bits(fp, WAVE_FMT);	/* Sub chunk type */
	dataFilePtr->Write32Bits(fp, 16);			/* Sub chunk Size */
	dataFilePtr->Write16Bits(fp, WAVE_PCM_CODE); /* No idea what this is for. */
	dataFilePtr->Write16Bits(fp, data->outSignal->numChannels);
	dataFilePtr->Write32Bits(fp,(int32) (1.0 / data->outSignal->dt + 0.5));
											/*Sample rate*/
	dataFilePtr->Write32Bits(fp, (int32) (dataFilePtr->wordSize / data->outSignal->dt +
	  0.5));
	dataFilePtr->Write16Bits(fp, dataFilePtr->wordSize);		/* Sample bytes */
	dataFilePtr->Write16Bits(fp, dataFilePtr->wordSize * 8);	/* Sample size in bits*/

	dataFilePtr->Write32Bits(fp, WAVE_DATA);
	dataFilePtr->Write32Bits(fp, dataLength);	/* Sample count */

}

/**************************** WriteWave ***************************************/

/*
 * This function writes the data from the output signal of an EarObject.
 * It returns FALSE if it fails in any way.
 * If the fileName is a dash, '-.suffix', then the file will be written to the
 * standard output.
 */

BOOLEAN
WriteWave_DataFile(char *fileName, EarObjectPtr data)
{
	BOOLEAN	ok;
	FILE	*fp;

	if (fileName[0] != '-') {
		if ((fp = fopen(fileName, "wb")) == NULL) {
			NotifyError("WriteWave_DataFile: Couldn't open '%s' for writing.",
				fileName);
			return(FALSE);
		}
	} else {
		fp = stdout;
	}
	if (dataFilePtr->endian == 0)
		SetEndianRW_DataFile(LITTLE_ENDIAN_DATA_FILE);
	NotifyWarning("Writing Micosoft Wave file %s with %lu samples.\n", fileName,
	  data->outSignal->length);
	WriteWaveHeader_DataFile(fp, data);
	ok = WriteSignal_DataFile(fp, data->outSignal, dataFilePtr->wordSize);
	if (fp != stdin)
		fclose(fp);
	return(ok);
	
}

/**************************** WriteAIFFParams *********************************/

/*
 * This routine writes the chunks of the AIFF format to a file (stream).
 * It does not write the sound data itself, but does write the other parameters
 * for the sound chunk.
 * At present only the common chunk and sound data chunk are written.  It is
 * intended to write a chunk specific to the CRL.
 */

void
WriteAIFFParams_DataFile(FILE *fp, EarObjectPtr data)
{
	int32		chunkSize = 0;
	
	chunkSize = 30 + data->outSignal->length * data->outSignal->numChannels * 
	  dataFilePtr->wordSize;

	dataFilePtr->Write32Bits(fp, AIFF_FORM);	/* Magic Number */
	dataFilePtr->Write32Bits(fp, chunkSize);	/* Chunk Size */
	dataFilePtr->Write32Bits(fp, AIFF_AIFF);	/* Type of Chunk */

						/* AIFF Common Chunk */
	dataFilePtr->Write32Bits(fp, AIFF_COMM);
	dataFilePtr->Write32Bits(fp, 18);		/* Chunk Size */
	dataFilePtr->Write16Bits(fp, data->outSignal->numChannels);
	dataFilePtr->Write32Bits(fp, (int32) data->outSignal->length);	
	dataFilePtr->Write16Bits(fp, 8 * dataFilePtr->wordSize);	/* Sample Size */
	WriteIeeeExtendedHighLow(fp, 1.0 / data->outSignal->dt);

	chunkSize = 8 + data->outSignal->length * data->outSignal->numChannels *
	  dataFilePtr->wordSize;
	dataFilePtr->Write32Bits(fp, AIFF_SSND);
	dataFilePtr->Write32Bits(fp, chunkSize);	/* Chunk Size */
	dataFilePtr->Write32Bits(fp, 0);		/* Offset */
	dataFilePtr->Write32Bits(fp, 0);		/* Block Size*/

}

/**************************** WriteAIFF ***************************************/

/*
 * This function writes the data from the output signal of an EarObject.
 * It returns FALSE if it fails in any way.
 * If the fileName is a dash, '-.suffix', then the file will be written to the
 * standard output.
 */

BOOLEAN
WriteAIFF_DataFile(char *fileName, EarObjectPtr data)
{
	BOOLEAN	ok;
	FILE	*fp;

	if (fileName[0] != '-') {
		if ((fp = fopen(fileName, "wb")) == NULL) {
			NotifyError("WriteAIFF_DataFile: Couldn't open '%s' for writing.",
				fileName);
			return(FALSE);
		}
	} else {
		fp = stdout;
	}
	if (dataFilePtr->endian == 0)
		SetEndianRW_DataFile(BIG_ENDIAN_DATA_FILE);
	NotifyWarning("Writing AIFF file %s with %lu frames.\n", fileName,
	  data->outSignal->length);
	WriteAIFFParams_DataFile(fp, data);
	ok = WriteSignal_DataFile(fp, data->outSignal, dataFilePtr->wordSize);
	if (fp != stdin)
		fclose(fp);
	return(ok);
	
}

/**************************** WriteRaw ****************************************/

/*
 * This function writes a signal to file in the raw word format.
 * It will average the signal across all channels.
 * It returns FALSE if it fails in any way.
 * If the fileName is a dash, '-.suffix', then the file will be written to the
 * standard output.
 */

BOOLEAN
WriteRaw_DataFile(char *fileName, EarObjectPtr data)
{
	register ChanLen	i, j;
	FILE	*fp;
	double	maxValue, normalise, sum;

	if (fileName[0] != '-') {
		if ((fp = fopen(fileName, "wb")) == NULL) {
			NotifyError("WriteAIFF_DataFile: Couldn't open '%s' for writing.",
				fileName);
			return(FALSE);
		}
	} else {
		fp = stdout;
	}
	maxValue = -LARGE_VALUE;
	for (i = 0; i < data->outSignal->length; i++)
		for (j = 0; j < data->outSignal->numChannels; j++)
			if (fabs(data->outSignal->channel[j][i]) > maxValue)
				maxValue = fabs(data->outSignal->channel[j][i]);
	NotifyWarning("Writing RAW/Byte file %s with %lu bytes.\n", fileName,
	  data->outSignal->length);
	normalise = (int16) (0xffff >> (17 - dataFilePtr->wordSize * 8)) / maxValue /
	  data->outSignal->numChannels;
	for (i = 0; i < data->outSignal->length; i++) {
		for (j = 0, sum = 0.0; j < data->outSignal->numChannels; j++)
			sum += data->outSignal->channel[j][i];
		switch (dataFilePtr->wordSize) {
		case	1:
			Write8Bits(fp, (int) (sum * normalise));
			break;
		case	2:
			dataFilePtr->Write16Bits(fp, (int) (sum * normalise));
			break;
		case	4:
			dataFilePtr->Write32Bits(fp, (int) (sum * normalise));
		default:
			if (fp != stdin)
				fclose(fp);
			NotifyError("WriteRaw_DataFile: Unsupported sample size.");
			return(FALSE);
			break;
		} /* switch */
	}
	if (fp != stdin)
		fclose(fp);
	return(TRUE);

}

/**************************** GetSuffix ***************************************/

/*
 * This routine returns the suffix of a file name i.e. any characters
 * after a ".".
 * It returns the entire file name if no suffix is returned.
 * 
 */
 
char *
GetSuffix_DataFile(char *fileName)
{
	static char scanLine[MAXLINE]; /* - must be static so that memory is kept.*/
	char	*token;
	
	strcpy(scanLine, fileName);	/* Remember, strtok clobbers scanLine. */
	token = strtok(scanLine, ".");
	token = strtok(NULL, " ");
	if (token != NULL)
		return(token);
	else 
		return(fileName);
		
}

/**************************** ReadSignal **************************************/

/*
 * This function reads a file in one of the supported file types.
 * It distinguishes sound file formats by the file suffix used.
 */

BOOLEAN
ReadSignal_DataFile(char *fileName, EarObjectPtr data)
{
	BOOLEAN	ok;
	char	*suffix, *p;
	int		i;

	suffix = GetSuffix_DataFile(fileName);
	for (p = suffix; *p != '\0'; p++)		/* Convert to upper case. */
		*p = toupper(*p);
	for (i = 0; strcmp(suffix, Format_DataFile(i)) != 0 && Format_DataFile(i) !=
	 '\0'; i++);
	switch (i) {
	case	AIF_DATA_FILE:
	case	AIFF_DATA_FILE:
		ok = ReadAIFF_DataFile(fileName, data);
		break;
	case	DAT_DATA_FILE:
		ok = ReadASCII_DataFile(fileName, data);
		break;
	case	WAV_DATA_FILE:
	case	WAVE_DATA_FILE:
		ok = ReadWave_DataFile(fileName, data);
		break;
	case	RAW_DATA_FILE:
	default	:
		ok = ReadRaw_DataFile(fileName, data);
		break;
	} /* switch */
	if (!ok)
    	NotifyError("ReadSignal_DataFile: Could not read file '%s'.", fileName);
	if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
    return(ok);
	
}

/**************************** WriteOutSignal **********************************/

/*
 * This function writes a file in one of the supported file types.
 * It distinguishes sound file formats by the file suffix used.
 */

BOOLEAN
WriteOutSignal_DataFile(char *fileName, EarObjectPtr data)
{
	BOOLEAN	ok;
	char	*suffix, *p;
	int		i;
	
	ok = FALSE;
	if (data == NULL) {
		NotifyError("WriteOutSignal_DataFile: EarObject not initialised.");
		return(FALSE);
	}	
	suffix = GetSuffix_DataFile(fileName);
	for (p = suffix; *p != '\0'; p++)		/* Convert to upper case. */
		*p = toupper(*p);
	for (i = 0; strcmp(suffix, Format_DataFile(i)) != 0 && Format_DataFile(i) !=
	  '\0'; i++);
	switch (i) {
	case	AIF_DATA_FILE:
	case	AIFF_DATA_FILE:
		ok = WriteAIFF_DataFile(fileName, data);
		break;
	case	DAT_DATA_FILE:
		ok = WriteASCII_DataFile(fileName, data);
		break;
	case	WAV_DATA_FILE:
	case	WAVE_DATA_FILE:
		ok = WriteWave_DataFile(fileName, data);
		break;
	case	RAW_DATA_FILE:
	default	:
		ok = WriteRaw_DataFile(fileName, data);
		break;
	}
	if (!ok)
		NotifyError("WriteOutSignal_DataFile: Could not write file '%s'.",
		  fileName);
	return(ok);

}

/********************************* SetDefaultSampleRate ***********************/

/*
 * This function sets the module's DefaultSampleRate parameter.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetDefaultSampleRate_DataFile(double theDefaultSampleRate)
{
	dataFilePtr->defaultSampleRate = theDefaultSampleRate;
	return(TRUE);

}

/********************************* SetMaxSamples ******************************/

/*
 * This function sets the module's DefaultSampleRate parameter.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetMaxSamples_DataFile(double theMaxSamples)
{
	dataFilePtr->maxSamples = (int32) theMaxSamples;
	return(TRUE);

}

/**************************** SetFileName *************************************/

/*
 * This routine sets the 'name' field of the dataFile structure.
 * It is called as the ReadPars routine from the module manager.
 * The fileName field is used by the ReadSignal_DataFile_Named routine.
 */

BOOLEAN
SetFileName_DataFile(char *fileName)
{
	if (*fileName == (char) NULL) {
		NotifyError("SetFileName_DataFile: Illegal file name.");
		return(FALSE);
	}
	strcpy(dataFilePtr->name, fileName);
	return(TRUE);

}

/********************************* CheckPars **********************************/

/*
 * This routine checks that the necessary parameters for the module have been
 * correctly initialised.
 * It returns TRUE if there are no problems.
 */
 
BOOLEAN
CheckPars_DataFile(void)
{
	if (dataFilePtr->name[0] == (char) NULL) {
		NotifyError("CheckPars_DataFile: File name not set.");
		return(FALSE);
	}
	if (dataFilePtr->mustSetMaxSamplesFlag && (dataFilePtr->maxSamples < 1)) {
		NotifyError("CheckPars_DataFile: Maximum samples parameter "\
		  "incorrectly set.");
		return(FALSE);
	}
	return(TRUE);
	
}	

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetPars_DataFile(char *theFileName, int theWordSize, int theEndian,
  double theDefaultSampleRate, double theMaxSamples)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetFileName_DataFile(theFileName))
		ok = FALSE;
	if (!SetWordSize_DataFile(theWordSize))
		ok = FALSE;
	if (!SetEndian_DataFile(theEndian))
		ok = FALSE;
	if (!SetDefaultSampleRate_DataFile(theDefaultSampleRate))
		ok = FALSE;
	if (!SetMaxSamples_DataFile(theMaxSamples))
		ok = FALSE;
	if (!ok)
		NotifyError("SetPars_DataFile: Failed to set all module parameters.");
	return(ok);
	
}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_DataFile(void)
{
	if (!CheckPars_DataFile()) {
		NotifyError("PrintPars_DataFile: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("DataFile Module Parameters:-\n");
	printf("\tFile name = %s\n", dataFilePtr->name);
	if (dataFilePtr->endian == 1)
		printf("\tBig ");
	else
		printf("\tLittle ");
	printf("endian format,\tDefault word size = %d\n", dataFilePtr->wordSize);
	printf("\tDefault sample rate = %g,\tMaximum samples = ",
	  dataFilePtr->defaultSampleRate);
	if (dataFilePtr->maxSamples < 0)
		printf("No limit\n");
	else
		printf("%lu\n", dataFilePtr->maxSamples);

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_DataFile(char *parFileName)
{
	BOOLEAN	ok;
	char	fileName[MAXLINE];
	int		wordSize, endian;
	double  defaultSampleRate, maxSamples;
    FILE    *fp;
    
    if ((fp = fopen(parFileName, "r")) == NULL) {
        NotifyError("ReadPars_DataFile: Cannot open data file '%s'.\n",
		  parFileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_DataFile: Reading from '%s'", parFileName);
    Init_ParFile();
	ok = TRUE;
	if (!GetPars_ParFile(fp, "%s", fileName))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%d", &wordSize))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%d", &endian))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &defaultSampleRate))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &maxSamples))
		ok = FALSE;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_DataFile:  Not enough, lines or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_DataFile(fileName, wordSize, endian,defaultSampleRate,
	  maxSamples)) {
		NotifyError("ReadPars_DataFile: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/**************************** ReadSignal...Named ******************************/

/*
 * This routine calls the ReadSignal_DataFile routine using the named file
 * in the dataFile structure.
 */

BOOLEAN
ReadSignal_DataFile_Named(EarObjectPtr data)
{
	if (!CheckPars_DataFile()) {
		NotifyError("ReadSignal_DataFile_Named: Parameters have not been "\
		  "correctly set.");
		return(FALSE);
	}
	return(ReadSignal_DataFile(dataFilePtr->name, data));

}
