/*****************
 *
 * FiParFile.c
 *
 * This module contains the general file utilities for the LUTEar Core Routines
 * Library.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 *****************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>

#include "LUTEar.h"
#include "FiParFile.h"
 
/******************************************************************************/
/*************************** Global Variables *********************************/
/******************************************************************************/

ParFile		*parFile = NULL;

/******************************************************************************/
/*************************** Functions and subroutines ************************/
/******************************************************************************/

/*
 * This routine initialises the parFile data structure.  It allocates its
 * memory dynamically, and sets the respective fields to their default.
 * values.
 */

/****************************** Init ******************************************/

BOOLEAN
Init_ParFile(void)
{
	if (parFile != NULL) {
		NotifyError("Init_ParFile: Module already initialised!");
		return(FALSE);
	}
	if ((parFile = (ParFile *) malloc(sizeof(ParFile))) == NULL) {
		NotifyError("Init_ParFile: Cannot allocate memory for parameters.");
		return(FALSE);
	}
	parFile->diagnosticsFlag = FALSE;
	parFile->parLineCount = 0;
	return(TRUE);

}

/********************************* CheckInit **********************************/

/*
 * This routine checks that the the module) has been initialised.
 * It prevents attempts to use uninitailised pointers which will cause
 * the module to crash the system.
 * It returns FALSE if the gobal variable has not been initialised.
 */
 
BOOLEAN
CheckInit_ParFile(void)
{
	if (parFile == NULL) {
		NotifyError("CheckInit_ParFile: General file routines module not " \
		  "initialised.");
		return(FALSE);
	}
	return(TRUE);
	
}

/********************************* Free ***************************************/

/*
 * This routine frees the space allocated for the global variable parFile.
 */

BOOLEAN
Free_ParFile(void)
{
	if (!CheckInit_ParFile()) {
		NotifyError("Free_ParFile: Module not initialised!");
		return(FALSE);
	}
	free(parFile);
	parFile = NULL;
	return(TRUE);
	
}

/****************************** GetLine ***************************************/

/*
 * This function reads a line from a file.  If the maximum number of characters
 * is exceeded, then it flushes the rest of the line in the file.
 * It returns a pointer to the string store if successful, and a NULL if not.
 */

char *
GetLine_ParFile(char *line, int maxChars, FILE *fp)
{
	int		i, c;
	
	for (i = 0; ((c = fgetc(fp)) != EOF) && (i < maxChars) && (c != '\n'); i++)
		line[i] = c;
	line[i] = '\0';
	if (c == EOF && i == 0)
		return(NULL);
	if (c != '\n')
		while ((c = fgetc(fp)) != '\n' && c != EOF);
	return(line);

}	

/****************************** CommentOrBlankLine ****************************/

/*
 * This routine checks if a line read is a comment line;
 * it returns TRUE if it is.
 * It also checks for blank lines.
 * A comment line is any line that begins with white space followed by a '#'.
 */

BOOLEAN
CommentOrBlankLine_ParFile(char *line)
{
	char	*p;
	
	for (p = line; (*p != '\0') && isspace(*p); p++);
	if (strlen(p) == 0)
		return(TRUE);
	if (*p == '#')
		return(TRUE);
	return(FALSE);

}
			
/****************************** IdentifyFormat ********************************/

/*
 * This subroutine returns a type character as defined by the format string
 * passed as an argument.  Only a single value is looked for.
 * It assumes that only doubles are used.
 * It returns EOF if a '%' is not found at the start of the string, and a '\0'
 * if an invalid character is found.
 */

char
IdentifyFormat_ParFile(char *fmt)
{
	char	*p;

	p = fmt;
	if (*p++ == '%') {
		if (*p == 'l')	/* This assumes that doubles are always used. */
			p++;
		switch (*p) {
			case 'u':
			case 'd':
			case 'f':
			case 's':
			case 'i':	/* Identifier. */
				return(*p);
			default:
				NotifyError("IdentifyFormat_ParFile: Illegal format string, "\
				  "'%s'.", fmt);
				return('\0');
		}
	} else
		return(EOF);

}

/****************************** GetPars ***************************************/

/*
 * This function reads respective variables from a line in a file.  It uses a
 * format similar to the standard C printf statement.
 * If it fails then it returns FALSE.
 */
 
BOOLEAN
GetPars_ParFile(FILE *fp, char *fmt, ...)
{
	static char	line[MAXLINE], fmtScanLine[MAXLINE];
	char	c, formatType, *formatToken, *restOfLine;
	va_list	args;
	
	if (!CheckInit_ParFile())
		exit(1);
	do {
		if (GetLine_ParFile(line, MAXLINE, fp) == NULL) {
			c = (parFile->parLineCount > 1)? 's': ' ';
			NotifyError("GetPars_ParFile: Premature end of parameter " \
			  "file!  Only %d line%c read.", parFile->parLineCount, c);
			return(FALSE);
		}
	} while(CommentOrBlankLine_ParFile(line));
	for (restOfLine = line; isspace(*restOfLine); restOfLine++); 
	parFile->parLineCount++;
	va_start(args, fmt);
	strcpy(fmtScanLine, fmt);	/* This is needed because strtok bashes
								 * its string argument. */
	formatToken = strtok(fmtScanLine, FORMAT_DELIMITERS);
	while (formatToken != NULL && restOfLine != NULL &&
	  (formatType = IdentifyFormat_ParFile(formatToken)) != EOF) {
		switch (formatType) {
			case 'u':
				sscanf(restOfLine, formatToken, va_arg(args, unsigned long *));
				break;
			case 'd':
				sscanf(restOfLine, formatToken, va_arg(args, int *));
				break;
			case 'f':
				sscanf(restOfLine, formatToken, va_arg(args, double *));
				break;
			case 's':
				sscanf(restOfLine, formatToken, va_arg(args, char *));
				break;
			default:
				NotifyError("GetPars_ParFile: Could not identify format "\
				  "string.");
				return(FALSE);
		}
		formatToken = strtok(NULL, FORMAT_DELIMITERS);
		/* This next section gets rid of the white space */
		for (restOfLine = strpbrk(restOfLine, FORMAT_DELIMITERS);
		  (restOfLine != NULL) && isspace(*restOfLine); restOfLine++);
	}
	va_end(args);
	if (parFile->diagnosticsFlag)
		printf("GetPars_ParFile: Read line: '%s'\n", line);
	return(TRUE);

}
