/******************
 *
 * EarObject.c
 *
 * This module contains the methods for the EarObject class.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ******************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
 
#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"

/******************************************************************************/
/*************************** Global Variables *********************************/
/******************************************************************************/

EarObjRefPtr	mainEarObjectList = NULL;

/******************************************************************************/
/********************************* Subroutines and functions ******************/
/******************************************************************************/

/**************************** Init ********************************************/

/*
 * This routine takes care of the correct initialisation of an EarObject: the
 * input and output signals are set to NULL, as required later, and the
 * pointer to the prepared object is returned.
 * The Module connected with the process is supplied as an argument.  This
 * is a change from the previous version.
 * The routine returns NULL if an error occurs.
 * All initialised EarObjects are registered in the main list.
 * 
 */

EarObjectPtr
Init_EarObject(char *moduleName)
{
	static EarObjHandle	handleNumber = 0;
	EarObjectPtr	data;

	if ((data = (EarObjectPtr) malloc(sizeof(EarObject))) == NULL) {
		NotifyError("Init_EarObject: Out of memory for EarObject " \
		  "(used with the '%s' module).", moduleName);
		return(NULL);
	}
	data->processName = NULL;
	data->localInSignalFlag = FALSE;
	data->localOutSignalFlag = FALSE;
	data->updateCustomersFlag = TRUE;
	data->inSignal = NULL;
	data->outSignal = NULL;
	data->customerList = NULL;
	data->supplierList = NULL;
	data->handle = handleNumber++;		/* Unique handle for each EarObject. */
	if (!AddEarObjRef_EarObject(&mainEarObjectList, data)) {
		NotifyError("Init_EarObject: Could not register new EarObject.");
		Free_EarObject(&data);
		return(NULL);
	}
	if ((data->module = Init_ModuleMgr(moduleName)) == NULL) {
		NotifyError("Init_EarObject: Could not set module type.");
		Free_EarObject(&data);
		return(NULL);
	}
	return(data);

} /* Init_EarObject */

/**************************** Free ********************************************/

/*
 * This routine ensures that all pointers of the object have their memory
 * released before memory of the object itself is released.
 * Only locally create signals will be free'd, i.e. each EarObject must release
 * the signals which it created.
 */
 
void
Free_EarObject(EarObjectPtr *theObject)
{
	if (*theObject == NULL)
		return;
	if ((*theObject)->localInSignalFlag)
		Free_SignalData(&(*theObject)->inSignal);
	if ((*theObject)->localOutSignalFlag)
		Free_SignalData(&(*theObject)->outSignal);
	/* This next line unresgisters the object from the main list. */
	FreeEarObjRef_EarObject(&mainEarObjectList, (*theObject)->handle);
	/* This next line will set the customer input signals to NULL. */
	UpdateCustomers_EarObject(*theObject);
	RemoveEarObjRefs_EarObject(*theObject);
	free((*theObject)->processName);
	Free_ModuleMgr(&(*theObject)->module);
	
	FreeEarObjRefList_EarObject(&(*theObject)->customerList);
	FreeEarObjRefList_EarObject(&(*theObject)->supplierList);
	(*theObject)->processName = NULL;
	free(*theObject);
	*theObject = NULL;		
	
} /* Free_EarObject */

/**************************** FreeAll *****************************************/

/*
 * This routine frees the memory for all of the initialised EarObjects from
 * the main register list.
 * It does not, however, set the pointer variables to NULL - watch this.
 */

void
FreeAll_EarObject(void)
{
	while (mainEarObjectList != NULL)
		FreeEarObjRef_EarObject(&mainEarObjectList,
		  mainEarObjectList->earObject->handle);

}

/**************************** SetProcessName **********************************/

/*
 * Sets a process name for an EarObject.  This sort of information will be used
 * in process titles and dislpays.
 * Space is allocated for the name, if it has not already been initialised.
 * The name can only be set once: repeated calls will do nothing.
 * This routine uses the variable argument format - like 'printf' does.
 */
 
void
SetProcessName_EarObject(EarObjectPtr theObject, char *format, ...)
{
	char	string[MAXLINE];
	va_list	args;
	
	if (theObject->processName == NULL) {
		va_start(args, format);
		vsprintf(string, format, args);
		va_end(args);
		if ((theObject->processName = (char *) malloc(strlen(string) + 1)) ==
		   NULL) {
			NotifyError("EarObject_SetProcessName: Out of memory, cannot "\
			  "allocate name");
			NotifyError("EarObject_SetProcessName: string = '%s'.", string);
			return;
		}
		strcpy(theObject->processName, string);
	}

} /* SetProcessName_EarObject */

/**************************** DestroyOldOutSignal *****************************/

/*
 * This function checks the parameters of a SignalData data structure and
 * decides whether or not to create a new data signal element, i.e. if the
 * dimensions of the SignalData object have change.
 * It destroys the old SignalData object if it is to be re-created, and returns
 * TRUE, otherwise FALSE is returned and no action is taken.
 */

BOOLEAN
DestroyOldOutSignal_EarObject(EarObjectPtr data, int numChannels,
  ChanLen length, double samplingInterval)
{
	if ( !data->outSignal->dtFlag || (data->outSignal->dt !=
	  samplingInterval) || !data->outSignal->lengthFlag ||
	  (data->outSignal->length != length) || !data->outSignal->numChannels ||
	  (data->outSignal->numChannels != numChannels)) {
		Free_SignalData(&data->outSignal);
		data->updateCustomersFlag = TRUE;
		return(TRUE);
	} else
		return(FALSE);
}

/**************************** InitOutSignal ***********************************/

/*
 * This routine sets the appriopriate fields of a signal and initialises the
 * channels.
 * Signal setting is usually carried out automatically by the module routines
 * using this routine.
 * The signal channels are initialised to zero.
 * It will overwrite any previous signal information, and also allows the
 * signal length to be re-set.
 * It will copy the input signal's the rampFlag setting if the signal is set.
 * 20-1-94 L.P.O'Mard this routine has been changed to accept the signal length
 * rather than the signal duration.
 */
 
BOOLEAN
InitOutSignal_EarObject(EarObjectPtr data, int numChannels,
  ChanLen length, double samplingInterval)
{
	BOOLEAN	createNewSignal;
	int		i;
	register	ChanData	*dataPtr;
	ChanLen	j;
	
	if (samplingInterval == 0.0) {
		NotifyError("InitOutSignal_EarObject: Time step is zero for "\
		  "EarObject '%s'!", data->processName);
		return(FALSE);
	}
	createNewSignal = TRUE;
	if (data->outSignal != NULL)
		createNewSignal = DestroyOldOutSignal_EarObject(data, numChannels,
		  length, samplingInterval);
	else
		data->updateCustomersFlag = TRUE;	/* Customers must be updated anyway. */
		
	if (createNewSignal) {
		data->outSignal = Init_SignalData("InitOutSignal_EarObject");
		SetLength_SignalData(data->outSignal, length);
		SetSamplingInterval_SignalData(data->outSignal, samplingInterval);
		for (i = 0; i < numChannels; i++)
			if (!InitChannel_SignalData(data->outSignal)) {
				NotifyError("InitOutSignal_EarObject: Cannot initialise "\
				  "output channels for EarObject '%s'.", data->processName);
				Free_SignalData(&data->outSignal);
				data->localOutSignalFlag = FALSE;
				return(FALSE);
			}
	} else {
		for (i = 0; i < numChannels; i++)
			for (j = 0, dataPtr = data->outSignal->channel[i]; j < length;
			  j++)
				*(dataPtr++) = 0.0;
	}
	data->outSignal->rampFlag = (data->inSignal == NULL)? FALSE:
	  data->inSignal->rampFlag;
	data->localOutSignalFlag = TRUE;
	return(TRUE);
	
}

/**************************** InitOutFromInSignal *****************************/

/*
 * This routine initialises or re-initialises the output signal of data set,
 * from the input signal.  Signal setting is usually carried out automatically
 * by the module routines using this routine.
 * It will overwrite any previous signal information, and also allows the
 * signal length to be re-set.
 * If the argument numChannels is zero, then the routine initialises the same
 * number of channels as for the input signal.
 * If the input signal has more than one channel, then the output will be an
 * interleaving of numChannel multiple channels.
 */
 
BOOLEAN
InitOutFromInSignal_EarObject(EarObjectPtr data, int numChannels)
{
	BOOLEAN	createNewSignal;
	int		i, channelsToSet;
	
	if (!CheckPars_SignalData(data->inSignal))
		exit(1);
	createNewSignal = TRUE;
    channelsToSet = (numChannels == 0) ? data->inSignal->numChannels:
      numChannels;
      
	if (data->outSignal != NULL)
		createNewSignal = DestroyOldOutSignal_EarObject(data, numChannels,
		  data->inSignal->length, data->inSignal->dt);
	else
		data->updateCustomersFlag = TRUE;/* Customers must be updated anyway. */
		
	if (createNewSignal) {
		data->outSignal = Init_SignalData("InitOutFromInSignal_EarObject");
		SetSamplingInterval_SignalData(data->outSignal, data->inSignal->dt);
		SetLength_SignalData(data->outSignal, data->inSignal->length);
		data->outSignal->rampFlag = data->inSignal->rampFlag;
		for (i = 0; i < channelsToSet; i++)
			if (!InitChannel_SignalData(data->outSignal)) {
				NotifyError("InitOutFromInSignal_EarObject: Cannot "\
				  "initialise output channels for EarObject '%s'.",
				  data->processName);
				return(FALSE);
			}
	}
	data->localOutSignalFlag = TRUE;
	if (!SetChannelsFromSignal_SignalData(data->outSignal, data->inSignal)) {
		NotifyError("InitOutFromInSignal_EarObject: Cannot set output "\
		  "Channels for EarObject: '%s'.", data->processName);
		return(FALSE);
	}
	return(TRUE);
	
}

/**************************** PrintProcessName ********************************/

/*
 * This routine outputs the process name to the standard output, provided that
 * it has been set, otherwise it will output "<no name set>".
 * The message string should be in the form "text....%s.....more text (if 
 * required).".
 * It assumes that all EarObjects are initialised to NULL when declared.
 */
 
void
PrintProcessName_EarObject(char *message, EarObjectPtr data)
{
	if (data->processName != NULL)
		printf(message, data->processName);
	else
		printf(message, "<no name set>");
}

/**************************** ConnectOutSignalToIn ****************************/

/*
 * This routine connects the output signal of the supplier to the input signal
 * of the customer.  It does this by merely putting the customer into the
 * supplier's customer list, and adding the supplier to the customer's
 * supplier list.
 * This routine tests for NULL EarObjects.
 */
 
BOOLEAN
ConnectOutSignalToIn_EarObject(EarObjectPtr supplier, EarObjectPtr customer)
{
	if (supplier == NULL) {
		NotifyError("ConnectOutSignalToIn_EarObject: Supplier EarObject not "\
		  "initialised.");
		return(FALSE);
	}
	if (customer == NULL) {
		NotifyError("ConnectOutSignalToIn_EarObject: Customer EarObject not "\
		  "initialised.");
		return(FALSE);
	}
	if (supplier == customer) {
		NotifyError("ConnectOutSignalToIn_EarObject: Attempted to connect "\
		  "EarObject to itself!");
		return(FALSE);
	}
	if (!AddEarObjRef_EarObject(&supplier->customerList, customer))
		return(FALSE);
	if (!AddEarObjRef_EarObject(&customer->supplierList, supplier))
		return(FALSE);
	if (supplier->outSignal != NULL)
		customer->inSignal = supplier->outSignal;
	return(TRUE);
	
}

/**************************** CreateEarObjRef *********************************/

/*
 * This function returns a pointer to a earObjRef node.
 * It returns NULL if it fails.
 */
 
EarObjRefPtr
CreateEarObjRef_EarObject(EarObjectPtr theObject)
{
	EarObjRef	*newNode;
	
	if ((newNode = (EarObjRef *) malloc(sizeof (EarObjRef))) == NULL) {
		NotifyError("CreateEarObjRef_EarObject: Out of memory.");
		return(NULL);
	}
	newNode->earObject = theObject;
	newNode->next = NULL;
	return(newNode);

}

/**************************** AddEarObjRef ************************************/

/*
 * This routine adds an EarObject reference to an EarObject's reference list.
 * It places the earObject handles in ascending order.
 * This function assumes that the supplier and EarObjects have been
 * initialised.
 */

BOOLEAN
AddEarObjRef_EarObject(EarObjRefPtr *theList, EarObjectPtr theObject)
{
	EarObjRefPtr	temp, p;

    if (*theList == NULL) {
    	if ((*theList = CreateEarObjRef_EarObject(theObject)) ==
    	  NULL) {
    		NotifyError("AddEarObjRef_EarObject: Could not create new node.");
    		return(FALSE);
    	}
    	
		return(TRUE);
	}
	for (p = *theList; (p->next != NULL) && (p->earObject->handle <
	  theObject->handle); p = p->next);
	if (p->earObject->handle == theObject->handle) {
		NotifyWarning("AddEarObjRef_EarObject: EarObject already in list.");
		return(TRUE);
	}
	if ((temp = CreateEarObjRef_EarObject(theObject)) == NULL) {
		NotifyError("AddEarObjRef_EarObject: Could not create temp node.");
		return(FALSE);
	}
	if (p->earObject->handle > theObject->handle) {
		temp->earObject = theObject;
		temp->next = p;
		if (p == *theList)	/* Test for head of list. */
			*theList = temp;	
		else
			p = temp;
	} else
		p->next = temp;
	return(TRUE);

}

/**************************** FreeEarObjRef ***********************************/

/*
 * This routine removes an EarObjRef from a list.
 */

void
FreeEarObjRef_EarObject(EarObjRefPtr *theList, EarObjHandle theHandle)
{
	EarObjRefPtr	p, lastRef;

    if (*theList == NULL)
    	return;
	for (p = *theList; (p != NULL) && (p->earObject->handle !=
	  theHandle); p = p->next)
		lastRef = p;
	if (p->earObject->handle == theHandle) {
		if (p == *theList) /* Test for head of list. */
			*theList = p->next;
		else
			lastRef->next = p->next;
		free(p);
	}

}

/**************************** FreeEarObjRefList *******************************/

/*
 * This routine frees the space allocated for a EarObjRef linked list.
 */
 
void
FreeEarObjRefList_EarObject(EarObjRefPtr *theList)
{
	EarObjRefPtr	temp;
	
	while (*theList != NULL) {
		temp = *theList;		/* Point to top earObjRef node */
		*theList = temp->next;	/* Point earObjRefList to next node*/
		free(temp);
	}

}

/**************************** UpdateCustomers *********************************/

/*
 * This routine updates all of an EarObject's customers, i.e. all customers
 * pointing to its output signal are reset.
 */

void
UpdateCustomers_EarObject(EarObjectPtr theObject)
{
	EarObjRefPtr	p;
	
	for (p = theObject->customerList; p != NULL; p = p->next) {
		p->earObject->inSignal = theObject->outSignal;
	}
	theObject->updateCustomersFlag = FALSE;
	
}

/**************************** RemoveEarObjRefs ********************************/

/*
 * This routine updates all references to an EarObject, i.e. this routine is
 * only used when deleting an EarObject, when all customers must delete it
 * from their supplier lists, and all suppliers must delete it from their
 * customer lists.
 */

void
RemoveEarObjRefs_EarObject(EarObjectPtr theObject)
{
	EarObjRefPtr	p;
	
	for (p = theObject->supplierList; p != NULL; p = p->next) {
		FreeEarObjRef_EarObject(&p->earObject->customerList, theObject->handle);
#		ifdef DEBUG
		printf("Removing EarObject[%d] from EarObject[%d] customer list.\n",
		  theObject->handle, p->earObject->handle);
#		endif
	}
	for (p = theObject->customerList; p != NULL; p = p->next) {
		FreeEarObjRef_EarObject(&p->earObject->supplierList, theObject->handle);
#		ifdef DEBUG
		printf("Removing EarObject[%d] from EarObject[%d] supplier list.\n",
		  theObject->handle, p->earObject->handle);
#		endif
	}
	theObject->updateCustomersFlag = FALSE;
	
}

