/**************
 *
 * LUTEar.c
 *
 * This file contains the common system routines:
 *  - error routines.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 **************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "LUTEar.h"

/******************************************************************************/
/************************ Global variables ************************************/
/******************************************************************************/

FILE	*warningsFile = stdout;		/* File to which warnings should be sent. */
FILE	*errorsFile = stderr;		/* File to which errors should be sent. */

/******************************************************************************/
/************************ Subroutines and functions ***************************/
/******************************************************************************/

/***************************** NotifyError ************************************/

/*
 * This routine prints out an error message, preceded by a bell sound.
 * It is used in the same way as the printf statement.
 * It does not take any further action, as responses to errors differ: some
 * errors are recoverable while others are fatal.
 * This routine will be incorporated in an alert dialog in the LUTEar
 * application.
 */

void
NotifyError(char *format, ...)
{
	va_list	args;
	
	va_start(args, format);
	if (errorsFile == stderr)
		fprintf(stderr, "\07");
	vfprintf(errorsFile, format, args);
	fprintf(errorsFile, "\n");
	va_end(args);

} /* NotifyError */

/*************************** NotifyWarning ************************************/

/*
 * This routine prints out a warning message, preceded by a bell sound.  It is
 * used in the same way as the printf statement.  It does not take any further
 * action.
 */
 
void
NotifyWarning(char *format, ...)
{
	va_list	args;
	
	va_start(args, format);
	vfprintf(warningsFile, format, args);
	fprintf(warningsFile, "\n");
	va_end(args);

} /* NotifyWarning */

/*************************** SetWarningsFile **********************************/

/*
 * This function sets the file to which warnings should be sent.
 * The default is the standard output, which can be reset by sending
 * a null string as the function argument.
 * The function returns TRUE if successful.
 */

BOOLEAN
SetWarningsFile(char *fileName)
{
	if (fileName[0] == '\0') {
		warningsFile = stdout;
		return(TRUE);
	}
	if ((warningsFile = fopen(fileName, "w")) == NULL) {
		NotifyError("SetWarningsFile: Could not open file '%s'.", fileName);
		return(FALSE); 
	}
	return(TRUE);

}

/*************************** SetErrorsFile ************************************/

/*
 * This function sets the file to which errors should be sent.
 * The default is the standard error, which can be reset by sending
 * a null string as the function argument.
 * The function returns TRUE if successful.
 */

BOOLEAN
SetErrorsFile(char *fileName)
{
	if (fileName[0] == '\0') {
		errorsFile = stderr;
		return(TRUE);
	}
	if ((errorsFile = fopen(fileName, "w")) == NULL) {
		NotifyError("SetErrorsFile: Could not open file '%s'.", fileName);
		return(FALSE); 
	}
	return(TRUE);

}
