/*******
 *
 * SignalData.c
 *
 * This module contains the methods for the SignalData class.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/

#include <stdio.h> 
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h" 
#include "SignalData.h"

/***************************************************************************************/
/********************************* Subroutines and functions ***************************/
/***************************************************************************************/

/**************************** Init *****************************************************/

/*
 * This function allocates memory for the signal data-set and sets the default values.
 * The numChannels field should never be changed by the user.  It is only incremented
 * when an additional channel is initialised.
 * The function returns a pointer to the prepared structure.
 */

SignalData *
Init_SignalData(char *callingFunctionName)
{
	SignalData *theData;
	
	if ((theData = (SignalData *) malloc(sizeof(SignalData))) == NULL ) {
		NotifyError("Init_SignalData: Out of Memory (called by %s).",
		  callingFunctionName);
		exit(1);
	}
	theData->numChannels = 0;
	theData->length = 0;
	theData->dt = 0.0;
	theData->dtFlag = FALSE;
	theData->lengthFlag = FALSE;
	theData->rampFlag = FALSE;
	return(theData);
	
} /* Init_SignalData */

/**************************** Free *****************************************************/

/*
 * This routine frees the memory allocated to a SignalData pointer.
 * It also sets the pointer to NULL, hence the address of the pointer is passed to this
 * function.
 */
 
void
Free_SignalData(SignalDataPtr *theData)
{
	int	i;
	
	if (*theData != NULL) {
		for (i = 0; i < (*theData)->numChannels; i++)
			free((*theData)->channel[i]);
		free(*theData);
		*theData = NULL;
	}
	
} /* Free_SignalData */

/**************************** CheckInit ************************************************/

/*
 * This routine checks whether or not a signal has been initialised.
 *
 */
 
BOOLEAN
CheckInit_SignalData(SignalDataPtr theSignal, char *callingFunction)
{
	if (theSignal != NULL)
		return(TRUE);
	else {
		NotifyError("CheckInit_SignalData: Signal not set in %s.", callingFunction);
		return(FALSE);
	}
	
}

/**************************** InitChannel **********************************************/

/*
 * This method allocates memory space for a channel.
 * It automatically updates the numChannel counter, and returns TRUE if successful.
 *
 */

BOOLEAN
InitChannel_SignalData(SignalDataPtr theData)
{
	ChanData	*p;
	
	if (!CheckInit_SignalData(theData, "InitChannel_SignalData"))
		return(FALSE);
	if (theData->length < 1) {
		NotifyError("InitChannel_SignalData: Invalid signal length = %u.",
		  theData->length);
		return(FALSE);
	}
	if (theData->numChannels >= MAX_CHANNELS - 1) {
		NotifyError("InitChannel_SignalData: Attempted to exceed channel the "\
		  "limit.");
		return(FALSE);
	}
	if ((p = (ChanData *) calloc(theData->length, sizeof(ChanData))) == NULL) {
		NotifyError("InitChannel_SignalData: Out of memory");
		return(FALSE);
	}
	theData->channel[theData->numChannels++] = p;
	return(TRUE);

} /* InitChannel_SignalData */

/**************************** CheckPars ************************************************/

/*
 * This routine checks that the necessary parameters for a signal have been
 * correctly initialised.
 * It returns TRUE if there are no problems.
 */

BOOLEAN
CheckPars_SignalData(SignalDataPtr theSignal)
{
	if (!CheckInit_SignalData(theSignal, "CheckPars_SignalData"))
		return(FALSE);
	if (theSignal->numChannels == 0) {
		NotifyError("CheckPars_SignalData: No Channels have been set.");
		return(FALSE);
	}
	if (theSignal->length == 0) {
		NotifyError("CheckPars_SignalData: Invalid zero signal length.");
		return(FALSE);
	}
	if (theSignal->dt <= 0.0) {
		NotifyError("CheckPars_SignalData: Invalid signal sampling interval = %lg.",
		  theSignal->dt);
		return(FALSE);
	}
	return(TRUE);
	
}

/**************************** CheckRamp ************************************************/

/*
 * This function tests if the ramp flag of a signal has been set.  Some processes require
 * that signals should be ramped, so this facility allows a warning to be given if
 * a signal has not been ramped.  The flag is set in the UtRamp module.
 * It returns TRUE if the signal has been ramped.
 */

BOOLEAN
CheckRamp_SignalData(SignalDataPtr theSignal)
{
	if (!CheckInit_SignalData(theSignal, "CheckRamp_SignalData"))
		return(FALSE);
	if (!theSignal->rampFlag)
		NotifyWarning("CheckRamp_SignalData: Signal is not ramped (see UtRamp.c).");
	return(TRUE);
	
}

/**************************** SetChannelsFromSignal ************************************/

/*
 * This routine copies the data from the orignal channel into all of the
 * channels of the supplied signal.  If the supplier signal is multi-channelled, then
 * the output signal will be an interleave of the supplier channels.
 * It returns true if the operation was successful,
 */

BOOLEAN
SetChannelsFromSignal_SignalData(SignalDataPtr theSignal, SignalDataPtr supplier)
{
	int		i;
	ChanLen	j;
	register	ChanData	*copyChannel, *supplierChannel;
	
	if (!CheckInit_SignalData(theSignal, "SetChannelsFromSignal_SignalData"))
		return(FALSE);
	if (!CheckInit_SignalData(supplier, "SetChannelsFromSignal_SignalData "\
	  "(supplier)"))
		return(FALSE);
	for (i = 0; i < theSignal->numChannels; i++) {
		copyChannel = theSignal->channel[i];
		supplierChannel = supplier->channel[i % supplier->numChannels];
		for (j = 0; j < theSignal->length; j++)
			*(copyChannel++) = *(supplierChannel++);
	}
	return(TRUE);
	
}

/**************************** SameType *************************************************/

/*
 * Checks that two signal data sets have the same dimensions, i.e. length, sampling
 * interval etc.
 * It returns TRUE if they have the same dimensions.
 */

BOOLEAN
SameType_SignalData(SignalDataPtr a, SignalDataPtr b)
{

	if (!a->lengthFlag || !b->lengthFlag) {
		NotifyError("SameType_SignalData: Both data set length values must be set! "\
		  "(lengths:- a = %u, b = %u).", a->length, b->length);
		return(FALSE);
	}
	if  ((a->length != b->length) || (a->numChannels != b->numChannels)) {
		NotifyError("SameType_SignalData: Both data sets must have the same "\
		  "length and number of channels! (lengths:- a = %u, b = %u; channels:- "\
		  "a = %d, b = %d.)", a->length, b->length, a->numChannels, b->numChannels);
		return(FALSE);
	}
    return(TRUE);

}

/**************************** SameType...NoDiagnostics ***********************************/

/*
 * Checks that two signal data sets have the same dimensions, i.e. length, sampling
 * interval etc.
 * It returns TRUE if they have the same dimensions.
 */

BOOLEAN
SameType_SignalData_NoDiagnostics(SignalDataPtr a, SignalDataPtr b)
{

	if (!a->lengthFlag || !b->lengthFlag)
		return(FALSE);
	if  ((a->length != b->length) || (a->numChannels != b->numChannels) ||
	  (a->dt != b->dt))
		return(FALSE);
    return(TRUE);

}

/**************************** InitCopy *************************************************/

/*
 * Initialises a data signal as a replica of an orignal data-set.
 * It returns a pointer to the new signal.
 */
 
SignalDataPtr
InitCopy_SignalData(SignalDataPtr original)
{
	int		i;
	ChanLen	j;
	ChanData	*copyPtr, *originalPtr;
	SignalData	*copy;
	
	if (!CheckInit_SignalData(original, "InitCopy_SignalData"))
		exit(1);
	copy = Init_SignalData("InitCopy_SignalData");
	copy->dtFlag = original->dtFlag;
	copy->lengthFlag = original->lengthFlag;
	copy->rampFlag = original->rampFlag;
	copy->length = original->length;
	copy->dt = original->dt;
	for (i = 0; i < original->numChannels; i++)
		if (InitChannel_SignalData(copy)) {
			copyPtr = copy->channel[i];
			originalPtr = original->channel[i];
			for (j = 0; j < copy->length; j++)
				*(copyPtr++) = *(originalPtr++);
		}
	return(copy);
	
}

/**************************** Scale ***************************************************/

/*
 * This routine scales data values by a specified multiplier.
 *
 */

BOOLEAN
Scale_SignalData(SignalDataPtr d, double multiplier)
{
	int		i;
	ChanLen	j;
	ChanData	*dataPtr;
	
	if (!CheckInit_SignalData(d, "Scale_SignalData (theSignal)"))
		return(FALSE);
	for (i = 0; i < d->numChannels; i++)
		for (j = 0, dataPtr = d->channel[i]; j < d->length; j++)
			*(dataPtr++) *= multiplier;
	return(TRUE);
	
}

/**************************** AddScaledData ********************************************/

/*
 * Adds two data structures together, with a supplied scaling difference.
 * The result is put into the first of the data sets: a = a + b * scale;
 * This version can apply a different scale to each channel.
 *
 */

BOOLEAN
AddScaledData_SignalData(SignalDataPtr a, SignalDataPtr b, double *scale)
{
	int		i;
	ChanLen	j;
	ChanData	*aPtr, *bPtr;
	
	if (!CheckInit_SignalData(a, "AddScaledData_SignalData (a)") ||
	  !CheckInit_SignalData(b, "AddScaledData_SignalData (b)"))
		return(FALSE);
	if (!SameType_SignalData(a, b))
		return(FALSE);
	for (i = 0; i < a->numChannels; i++, scale++) {
		aPtr = a->channel[i];
		bPtr = b->channel[i];
		for (j = 0; j < a->length; j++)
			*(aPtr++) += *(bPtr++) * *scale;
	}
	return(TRUE);
	
}

/**************************** AddScaledData...Single ***********************************/

/*
 * Adds two data structures together, with a supplied scaling difference.
 * The result is put into the first of the data sets: a = a + b * scale;
 * This version expects a single scaling factor to be applied to all the channels.
 *
 */

BOOLEAN
AddScaledData_SignalData_Single(SignalDataPtr a, SignalDataPtr b, double scale)
{
	int		i;
	ChanLen	j;
	ChanData	*aPtr, *bPtr;
	
	if (!CheckInit_SignalData(a, "AddScaledData_SignalData (a)") ||
	  !CheckInit_SignalData(b, "AddScaledData_SignalData (b)"))
		return(FALSE);
	if (!SameType_SignalData(a, b))
		return(FALSE);
	for (i = 0; i < a->numChannels; i++) {
		aPtr = a->channel[i];
		bPtr = b->channel[i];
		for (j = 0; j < a->length; j++)
			*(aPtr++) += *(bPtr++) * scale;
	}
	return(TRUE);
	
}

/**************************** SetSamplingInterval **************************************/

/*
 * The sampling interval can be set any number of times.
 */
 
void
SetSamplingInterval_SignalData(SignalDataPtr theData, double theSamplingInterval)
{
	if (!CheckInit_SignalData(theData, "SetSamplingInterval_SignalData"))
		exit(1);
	theData->dt = theSamplingInterval;
	theData->dtFlag = TRUE;

}

/**************************** SetLength ************************************************/

/*
 * The length of a data set can only be set for a particular data set once.
 */

void
SetLength_SignalData(SignalDataPtr theData, ChanLen theLength)
{
	if (!CheckInit_SignalData(theData, "SetLength_SignalData"))
		exit(1);
	if (!theData->lengthFlag) {
		theData->length = theLength;
		theData->lengthFlag = TRUE;
	} else
		NotifyError("SetLength_SignalData: Data set length cannot be changed.");
		
}

/**************************** OutputToFile *********************************************/

/*
 * This routine outputs a signal to a specied file, using the simple LUTEar format.
 */

BOOLEAN
OutputToFile_SignalData(char *fileName, SignalDataPtr theData)
{
	int		j;
    ChanLen	i;
	FILE	*fp;
	
	if (!CheckPars_SignalData(theData)) {
		NotifyError("OutputToFile_SignalData: Parameters not set for data-set.");
		return(FALSE);
	}
	printf("Output sent to file '%s'.\n", fileName);
	if ((fp = fopen(fileName, "w")) == NULL) {
		NotifyError("OutputToFile_SignalData: Cannot open file '%s'", fileName);
		return(FALSE);
	}
			
	for (i = 0; i < theData->length; i++) {
		fprintf(fp, "%8.5e", i * theData->dt);
		for (j = 0; j < theData->numChannels; j++)
			fprintf(fp, "\t %g", theData->channel[j][i]);
		fprintf(fp, "\n");
	}
	fclose(fp);
	return(TRUE);

}

/**************************** GetDuration **********************************************/

/*
 * This function returns the duration of a signal.
 */

double
GetDuration_SignalData(SignalDataPtr theSignal)
{
	if (!CheckPars_SignalData(theSignal)) {
		NotifyError("GetDuration_SignalData: Parameters not set for data-set, returned "\
		  "zero.");
		return(0.0);
	}
	return(theSignal->length * theSignal->dt);
	
}

/**************************** GaindB ******************************************/

/*
 * This routine increases the signal intensity by a specified number of dB.
 *
 */

BOOLEAN
GaindB_SignalData(SignalDataPtr d, double gaindB)
{
	int		i;
	double	scale;
	ChanLen	j;
	ChanData	*dataPtr;
	
	if (!CheckInit_SignalData(d, "Scale_SignalData (theSignal)"))
		return(FALSE);
	scale = pow(10.0, gaindB / 20.0);
	for (i = 0; i < d->numChannels; i++)
		for (j = 0, dataPtr = d->channel[i]; j < d->length; j++)
			*(dataPtr++) *= scale;
	return(TRUE);
	
}


