/*******
 *
 * MoHairCell.c
 *
 * This module contains the routines for the Meddis 93 Hair cell model module.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "Mo2HairCell.h"
#include "UtRandom.h"
#include "FiParFile.h"

/******************************************************************************/
/*************************** Global Variables *********************************/
/******************************************************************************/

HairCell2		hairCell2 = {

					FALSE,	/* numberOfRunsFlag */
					FALSE,	/* replenishRateFlag */
					FALSE,	/* lossRateFlag */
					FALSE,	/* reprocessRateFlag */
					FALSE,	/* recoveryRateFlag */
					FALSE,	/* maxFreePoolFlag */
					FALSE,	/* permeabilityPHFlag */
					FALSE,	/* permeabilityPZFlag */
					FALSE,	/* refractoryPeriodFlag */
					
					0,		/* numberOfRuns */
					0,		/* maxFreePool_M */
					0.0,	/* replenishRate_y  */ 
					0.0,	/* lossRate_l */
					0.0,	/* reprocessRate_x */
					0.0,	/* recoveryRate_r */
					0.0,	/* permeabilityPH_h */
					0.0,	/* permeabilityPZ_z */
					0.0		/* refractoryPeriod */
			
				};

/******************************************************************************/
/*************************** Subroutines and Functions ************************/
/******************************************************************************/

/********************************* SetNumberOfRuns ****************************/

/*
 * This function sets the number of times the model is run for a particular
 * input.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetNumberOfRuns_IHC_Meddis93(int theNumberOfRuns)
{
	if (theNumberOfRuns < 0) {
		NotifyError("SetNumberOfRuns_IHC_Meddis93: The value '%d' is illegal.");
		return(FALSE);
	}
	hairCell2.numberOfRunsFlag = TRUE;
	hairCell2.numberOfRuns = theNumberOfRuns;
	return(TRUE);

}

/********************************* SetReplenishRate ***************************/

/*
 * This function sets the module's replenish rate, y, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetReplenishRate_IHC_Meddis93(double theReplenishRate)
{
	hairCell2.replenishRateFlag = TRUE;
	hairCell2.replenishRate_y = theReplenishRate;
	return(TRUE);

}

/********************************* SetLossRate ********************************/

/*
 * This function sets the module's loss rate, l, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetLossRate_IHC_Meddis93(double theLossRate)
{
	hairCell2.lossRateFlag = TRUE;
	hairCell2.lossRate_l = theLossRate;
	return(TRUE);

}

/********************************* SetReprocessRate ***************************/

/*
 * This function sets the module's reprocess rate, x  parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetReprocessRate_IHC_Meddis93(double theReprocessRate)
{
	hairCell2.reprocessRateFlag = TRUE;
	hairCell2.reprocessRate_x = theReprocessRate;
	return(TRUE);

}

/********************************* SetRecoveryRate ****************************/

/*
 * This function sets the module's recovery rate, r parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetRecoveryRate_IHC_Meddis93(double theRecoveryRate)
{
	hairCell2.recoveryRateFlag = TRUE;
	hairCell2.recoveryRate_r = theRecoveryRate;
	return(TRUE);

}

/********************************* SetMaxFreePool *****************************/

/*
 * This function sets the module's maximum free transmitter pool, M parameter.
 * This is a double variable to maintain transparency with Meddis 86 model.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetMaxFreePool_IHC_Meddis93(double theMaxFreePool)
{
	if (theMaxFreePool < 1) {
		NotifyError("SetMaxFreePool_IHC_Meddis93: Illegal initial value: M = "\
		  "%d.", theMaxFreePool);
		return(FALSE);
	}
	hairCell2.maxFreePoolFlag = TRUE;
	hairCell2.maxFreePool_M = theMaxFreePool;
	return(TRUE);

}

/********************************* SetPermeabilityPH **************************/

/*
 * This function sets the module's permeabilityP constant H parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetPermeabilityPH_IHC_Meddis93(double thePermeabilityPH)
{
	hairCell2.permeabilityPHFlag = TRUE;
	hairCell2.permeabilityPH_h = thePermeabilityPH;
	return(TRUE);

}

/********************************* SetPermeabilityPZ **************************/

/*
 * This function sets the module's permeability constant Z parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetPermeabilityPZ_IHC_Meddis93(double thePermeabilityPZ)
{
	hairCell2.permeabilityPZFlag = TRUE;
	hairCell2.permeabilityPZ_z = thePermeabilityPZ;
	return(TRUE);

}

/********************************* SetRefractoryPeriod ************************/

/*
 * This function sets the module's refractory period parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetRefractoryPeriod_IHC_Meddis93(double theRefractoryPeriod)
{
	if (theRefractoryPeriod <= 0.0) {
		NotifyError("SetRefractoryPeriod_IHC_Meddis93: Refractory period "\
		  "cannot be negative!");
		return(FALSE);
	}
	hairCell2.refractoryPeriodFlag = TRUE;
	hairCell2.refractoryPeriod = theRefractoryPeriod;
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 */

BOOLEAN
SetPars_IHC_Meddis93(int theNumberOfRuns, double y, double l, double x,
  double r, double M, double h, double z, double refractoryPeriod)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetNumberOfRuns_IHC_Meddis93(theNumberOfRuns))
		ok = FALSE;
	if (!SetReplenishRate_IHC_Meddis93(y))
		ok = FALSE;
	if (!SetLossRate_IHC_Meddis93(l))
		ok = FALSE;
	if (!SetReprocessRate_IHC_Meddis93(x))
		ok = FALSE;
	if (!SetRecoveryRate_IHC_Meddis93(r))
		ok = FALSE;
	if (!SetMaxFreePool_IHC_Meddis93(M))
		ok = FALSE;
	if (!SetPermeabilityPH_IHC_Meddis93(h))
		ok = FALSE;
	if (!SetPermeabilityPZ_IHC_Meddis93(z))
		ok = FALSE;
	if (!SetRefractoryPeriod_IHC_Meddis93(refractoryPeriod))
		ok = FALSE;
	return(ok);
	  
}

/********************************* CheckPars **********************************/

/*
 * This routine checks that all of the parameters for the module are set.
 * Because this module is defined by its coeffients, default values are always
 * set if none of the parameters are set, otherwise all of them must be
 * individually set.
 */

BOOLEAN
CheckPars_IHC_Meddis93(void)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!hairCell2.replenishRateFlag) {
		NotifyError("CheckPars_IHC_Meddis93: Replenish rate constant, y, "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!hairCell2.lossRateFlag) {
		NotifyError("CheckPars_IHC_Meddis93: Loss rate constant, l, "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!hairCell2.reprocessRateFlag) {
		NotifyError("CheckPars_IHC_Meddis93: Reprocessing rate constant, "\
		  "x, not correctly set.");
		ok = FALSE;
	}
	if (!hairCell2.recoveryRateFlag) {
		NotifyError("CheckPars_IHC_Meddis93: Recovery rate constant, r, "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!hairCell2.maxFreePoolFlag) {
		NotifyError("CheckPars_IHC_Meddis93: Max. free transmitter in pool "\
		  "constant, M, not correctly set.");
		ok = FALSE;
	}
	if (!hairCell2.permeabilityPHFlag) {
		NotifyError("CheckPars_IHC_Meddis93: Permeability constant, h, "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!hairCell2.permeabilityPZFlag) {
		NotifyError("CheckPars_IHC_Meddis93: Permeability constant, z, "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!hairCell2.refractoryPeriodFlag) {
		NotifyError("CheckPars_IHC_Meddis93: Refractory period not , "\
		  "correctly set.");
		ok = FALSE;
	}
	if (!hairCell2.numberOfRunsFlag) {
		NotifyError("CheckPars_IHC_Meddis93: The number of model runs has "\
		  "not been set.");
		ok = FALSE;
	}
	return(ok);
		
}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_IHC_Meddis93(void)
{
	if (!CheckPars_IHC_Meddis93()) {
		NotifyError("PrintPars_IHC_Meddis93: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("Meddis 93 Inner Hair Cell Module Parameters:-\n");
	printf("\tReplenishment rate, y = %g /s,\tPermeability constant, z = "\
	  "%g,\n", hairCell2.replenishRate_y, hairCell2.permeabilityPZ_z);
	printf("\tLoss rate, l = %g /s,\tReprocessing rate, x = %g /s\n",
	  hairCell2.lossRate_l, hairCell2.reprocessRate_x);
	printf("\tRecovery rate, r = %g /s,\tPermeability constant, h = %g\n",
	  hairCell2.recoveryRate_r, hairCell2.permeabilityPH_h);
	printf("\tMaximum transmitter packets in free pool, M = %d,\n",
	  hairCell2.maxFreePool_M);
	printf("\tRefractory period = %g ms,\tNumber of repeat runs = %d.\n",
	  MSEC(hairCell2.refractoryPeriod), hairCell2.numberOfRuns);

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_IHC_Meddis93(char *fileName)
{
	BOOLEAN	ok;
	int		numberOfRuns;
	double	replenishRate_y, lossRate_l, reprocessRate_x, recoveryRate_r;
	double	permeabilityPH_h, permeabilityPZ_z, refractoryPeriod, maxFreePool_M;
	FILE	*fp;
	
	if ((fp = fopen(fileName, "r")) == NULL) {
		NotifyError("ReadPars_IHC_Meddis93: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
	}
	NotifyWarning("ReadPars_IHC_Meddis93: Reading from '%s", fileName);
	Init_ParFile();
	ok = TRUE;
	if (!GetPars_ParFile(fp, "%d", &numberOfRuns))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &maxFreePool_M))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &replenishRate_y))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &lossRate_l))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &reprocessRate_x))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &recoveryRate_r))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &permeabilityPH_h))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &permeabilityPZ_z))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &refractoryPeriod))
		ok = FALSE;
	fclose(fp);
	Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_IHC_Meddis93: Not enough lines, or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_IHC_Meddis93(numberOfRuns, replenishRate_y, lossRate_l,
	  reprocessRate_x, recoveryRate_r, maxFreePool_M, permeabilityPH_h,
	  permeabilityPZ_z, refractoryPeriod)) {
		NotifyError("ReadPars_IHC_Meddis93: Could not set parameters.");
		return(FALSE);
	}
	return(ok);
	
}

/********************************* CheckData **********************************/

/*
 * This routine checks that the input signal is correctly initialised
 * (using CheckRamp_SignalData), and determines whether the parameter values
 * are valid for the signal sampling interval.
 */

BOOLEAN
CheckData_IHC_Meddis93(EarObjectPtr data)
{
	BOOLEAN	ok;
	double	dt, maximumH;
	
	if (data == NULL) {
		NotifyError("CheckData_IHC_Meddis93: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckRamp_SignalData(data->inSignal))
		return(FALSE);
	dt = data->inSignal->dt;
	if (dt > MEDDIS93_MAX_DT) {
		NotifyError("CheckData_IHC_Meddis93: Maximum sampling interval "\
		  "exceeded.");
		return(FALSE);
	}
	ok = TRUE;
	if (hairCell2.replenishRate_y * dt >= 1.0) {
		NotifyError("CheckData_IHC_Meddis93: Replenishment rate, y = %g /s "\
		  "is too high for the sampling interval.",
		  hairCell2.replenishRate_y);
		ok = FALSE;
	}
	if (hairCell2.lossRate_l * dt >= 1.0) {
		NotifyError("CheckData_IHC_Meddis93: Loss rate, l = %g /s "\
		  "is too high for the sampling interval.",
		  hairCell2.lossRate_l);
		ok = FALSE;
	}
	if (hairCell2.recoveryRate_r * dt >= 1.0) {
		NotifyError("CheckData_IHC_Meddis93: Recovery rate, r = %g /s "\
		  "is too high for the sampling interval.",
		  hairCell2.recoveryRate_r);
		ok = FALSE;
	}
	if (hairCell2.reprocessRate_x * dt >= 1.0) {
		NotifyError("CheckData_IHC_Meddis93: Reprocess rate, x = %g /s "\
		  "is too high for the sampling interval.",
		  hairCell2.reprocessRate_x);
		ok = FALSE;
	}
	maximumH = log(1.0 / (hairCell2.permeabilityPZ_z * dt));
	/* if (hairCell2.permeabilityPH_h > maximumH) {
		NotifyError("CheckData_IHC_Meddis93: The permeability constant, h = "\
		  "%g, is too high for\nthe signal sampling interval, dt = %g ms, "\
		  "and the permeability constant, z = %g (h < %g).",
		  hairCell2.permeabilityPH_h, dt, hairCell2.permeabilityPZ_z);
		ok = FALSE;
	} */
	return(ok);
	
}

/********************************* InitSpontOut *******************************/

/*
 * This routine initialises the hair cells to the spontaneous firing rate,
 * and sets the refractory timers to a large value.
 * The Spontaneous firing rate is determined from the equilbrium state of the
 * system with no input: the reservoir contents do not change in time.
 * For the p'(0) value it uses the first value from the input signal, which it
 * assumes to contain the appropriate value.  This must be carefully accounted
 * when extending this code for 'continuous' mode.
 */

void
InitSpontOut_IHC_Meddis93(EarObjectPtr data, HairCellVars2 hCChannels[])
{
	ChanLen	i;
	double	spontPerm_k0, spontCleft_c0, spontFreePool_q0, spontReprocess_w0;
	HairCell2	*hC;
	
	hC = &hairCell2;		/* Shorter variable for long formulae. */
	
	spontPerm_k0 = hC->permeabilityPZ_z * exp(hC->permeabilityPH_h *
	 data->inSignal->channel[0][0]); /* p'(0) is the same for all channels. */
	spontCleft_c0 = hC->maxFreePool_M * hC->replenishRate_y *
	   spontPerm_k0 / (hC->replenishRate_y * (hC->lossRate_l +
	   hC->recoveryRate_r) + spontPerm_k0 * hC->lossRate_l);
	spontFreePool_q0 = spontCleft_c0 * (hC->lossRate_l + hC->recoveryRate_r) /
	  spontPerm_k0;
	spontReprocess_w0 = spontCleft_c0 * hC->recoveryRate_r /
	  hC->reprocessRate_x;
	
	for (i = 0; i < data->outSignal->numChannels; i++) {
		hCChannels[i].cleftC = (int) spontCleft_c0;
		hCChannels[i].reservoirQ = (int) spontFreePool_q0;
		hCChannels[i].reprocessedW = (int) spontReprocess_w0;
		hCChannels[i].refractoryTimer = hC->refractoryPeriod;
	}
	
}

/********************************* RunModel ***********************************/

/*
 * This routine runs the input signal through the model and puts the result
 * into the output signal.  It checks that all initialisation has been
 * correctly carried out by calling the appropriate checking routines.
 */

BOOLEAN
RunModel_IHC_Meddis93(EarObjectPtr data)
{
	BOOLEAN	clipped;
	int		i, k;
	register int	replenish, reprocessed, reUptake, lost, ejected;
	ChanLen	j;
	double	dt, ydt, xdt, ldt, rdt, kdt, zdt;
	ChanData	*inPtr, *outPtr;
	HairCell2	*hC;
	HairCellVars2	hCChannels[MAX_CHANNELS];
	
	if (!CheckPars_IHC_Meddis93())		
		return(FALSE);
	if (!CheckData_IHC_Meddis93(data)) {
		NotifyError("RunModel_IHC_Meddis93: Process data invalid.");
		return(FALSE);
	}
	if (!InitOutSignal_EarObject(data, data->inSignal->numChannels,
	  data->inSignal->length, data->inSignal->dt)) {
		NotifyError("RunModel_IHC_Meddis93: Could not initialise output "\
		  "signal.");
		return(FALSE);
	}
	SetProcessName_EarObject(data, "Meddis 93 probabilistic hair cell");
	dt = data->outSignal->dt;
	hC = &hairCell2;		/* Shorter name for use with long equations. */
	
	ydt = hC->replenishRate_y * dt;
	ldt = hC->lossRate_l * dt;
	rdt = hC->recoveryRate_r * dt;
	xdt = hC->reprocessRate_x * dt;
	zdt = hC->permeabilityPZ_z * dt;
	clipped = FALSE;
	
	for (k = 0; k < hC->numberOfRuns; k++) {
		InitSpontOut_IHC_Meddis93(data, hCChannels);
		for (i = 0; i < data->outSignal->numChannels; i++) {
			inPtr = data->inSignal->channel[i];
			for (j = 0, outPtr = data->outSignal->channel[i]; j <
			  data->outSignal->length; j++, outPtr++) {
				kdt = zdt * exp(hC->permeabilityPH_h * *inPtr++);
				if (kdt >= 1.0) {
					kdt = 0.99;
					clipped = TRUE;
				}
				replenish = (hCChannels[i].reservoirQ < hC->maxFreePool_M)?
				  GeomDist_Random(ydt, hC->maxFreePool_M -
				  hCChannels[i].reservoirQ): 0.0;
				if (((ejected = GeomDist_Random(kdt,
				  hCChannels[i].reservoirQ)) > 0)  &&
				  (hCChannels[i].refractoryTimer >= hC->refractoryPeriod)) {
						hCChannels[i].refractoryTimer = 0.0;
						*outPtr += 1.0;
				} else
					hCChannels[i].refractoryTimer += dt;
				lost = GeomDist_Random(ldt, hCChannels[i].cleftC);
				reUptake = GeomDist_Random(rdt, hCChannels[i].cleftC - lost);
				reprocessed = GeomDist_Random(xdt, hCChannels[i].reprocessedW);
	
				hCChannels[i].reservoirQ += replenish - ejected + reprocessed;
				hCChannels[i].cleftC += ejected - lost - reUptake;
				hCChannels[i].reprocessedW += reUptake - reprocessed;
			}
		}
	}
	if (data->updateCustomersFlag)
		UpdateCustomers_EarObject(data);
	if (clipped)
		NotifyWarning("RunModel_IHC_Meddis93: Clipping has occurred.");
	return(TRUE);
		
} /* RunModel_iHC_Meddis93 */
