/*******************
 *
 * Mo3BasilarM.c
 *
 * This is an implementation of a Gamma Tone filter using the filters.c module.
 * The Stone Composite filter is used to model the response of the Basilar
 * Membrane.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 *******************/
 
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "UtFilters.h"
#include "UtCFList.h"
#include "Mo3BasilarM.h"
#include "FiParFile.h"

/******************************************************************************/
/*************************** Global Variables *********************************/
/******************************************************************************/

BasilarM3	basilarM3 = {

				FALSE,	/* orderFlag */
				FALSE,	/* narrowBroadeningCoeffFlag */
				FALSE,	/* wideBroadeningCoeffFlag */
				FALSE,	/* compressionThresholdFlag */
				FALSE,	/* compressionExponentFlag */
				FALSE,	/* wideFCFFlag */
				FALSE,	/* baseWideFAttenFlag */
				FALSE,	/* wideFAttenScaleFlag */
				FALSE,	/* wideFRefCFFlag */
				FALSE,	/* narrowFLPCutOffScaleFlag */
				FALSE,	/* wideFLPCutOffScaleFlag */
				FALSE,	/* theCFsFlag */
				0,		/* order */
				0.0,	/* narrowBroadeningCoeff */
				0.0,	/* wideBroadeningCoeff */
				0.0,	/* compressionThreshold */
				0.0,	/* compressionExponent */
				0.0,	/* wideFCF */
				0.0,	/* wideFRefCF */
				0.0,	/* baseWideFAtten */
				0.0,	/* wideFAttenScale */
				0.0,	/* narrowFLPCutOffScale */
				0.0,	/* wideFLPCutOffScale */
				NULL	/* Wide Filter */
				
			};

/******************************************************************************/
/**************************** Subroutines & functions *************************/
/******************************************************************************/

/********************************* CheckPars **********************************/

/*
 * This routine checks that the necessary parameters for the module have been
 * correctly initialised.
 * It returns TRUE if there are no problems.
 */
 
BOOLEAN
CheckPars_BasilarM_Stone3(void)
{
	BOOLEAN	ok;

	ok = TRUE;
	if (!basilarM3.orderFlag) {
		NotifyError("CheckPars_BasilarM_Stone3: Filter order not set.");
		ok = FALSE;
	}
	if (!basilarM3.narrowBroadeningCoeffFlag) {
		NotifyError("CheckPars_BasilarM_Stone3: Filter narrow broadening "
		  "coefficient not set.");
		ok = FALSE;
	}
	if (!basilarM3.wideBroadeningCoeffFlag) {
		NotifyError("CheckPars_BasilarM_Stone3: Filter wide broadening "
		  "coefficient not set.");
		ok = FALSE;
	}
	if (!basilarM3.compressionThresholdFlag) {
		NotifyError("CheckPars_BasilarM_Stone3: Filter compression threshold "
		  "not set.");
		ok = FALSE;
	}
	if (!basilarM3.compressionExponentFlag) {
		NotifyError("CheckPars_BasilarM_Stone3: Filter compression ratio not "\
		  "set.");
		ok = FALSE;
	}
	if (!basilarM3.wideFCFFlag) {
		NotifyError("CheckPars_BasilarM_Stone3: Wide filter centre frequency "\
		  "not set.");
		ok = FALSE;
	}
	if (!basilarM3.wideFRefCFFlag) {
		NotifyError("CheckPars_BasilarM_Stone3: Wide filter reference centre "\
		  "frequency not set.");
		ok = FALSE;
	}
	if (!basilarM3.baseWideFAttenFlag) {
		NotifyError("CheckPars_BasilarM_Stone3: Wide filter base attenuation "\
		  "not set.");
		ok = FALSE;
	}
	if (!basilarM3.wideFAttenScaleFlag) {
		NotifyError("CheckPars_BasilarM_Stone3: Wide filter attenuation scale "\
		  "not set.");
		ok = FALSE;
	}
	if (!basilarM3.narrowFLPCutOffScaleFlag) {
		NotifyError("CheckPars_BasilarM_Stone3: Narrow filter low-pass cut-off "\
		  "scale not set.");
		ok = FALSE;
	}
	if (!basilarM3.wideFLPCutOffScaleFlag) {
		NotifyError("CheckPars_BasilarM_Stone3: Wide filter low pass cut-off "\
		  "scale not set.");
		ok = FALSE;
	}
	if (!basilarM3.theCFsFlag) {
		NotifyError("CheckPars_BasilarM_Stone3: Centre frequencies for the "\
		  "filter bank not set.");
		ok = FALSE;
	} else {
		if (!CheckPars_CFList(basilarM3.theCFs)) {
			NotifyError("CheckPars_BasilarM_Stone3: Centre frequency list "\
			  " parameters not correctly set.");
			ok = FALSE;
		}
	}
	return(ok);

}

/********************************* SetOrder ***********************************/

/*
 * This function sets the module's order parameter.  It first checks that the
 * module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetOrder_BasilarM_Stone3(int theOrder)
{
	basilarM3.orderFlag = TRUE;
	basilarM3.order = theOrder;
	return(TRUE);
	
}

/************************* SetNarrowBroadeningCoeff ***************************/

/*
 * This function sets the module's narrow broadening coefficient parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetNarrowBroadeningCoeff_BasilarM_Stone3(double theNarrowBroadeningCoeff)
{
	basilarM3.narrowBroadeningCoeffFlag = TRUE;
	basilarM3.narrowBroadeningCoeff = theNarrowBroadeningCoeff;
	return(TRUE);
	
}

/*********************** SetWideBroadeningCoeff *******************************/

/*
 * This function sets the module's wide broadening coefficient parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetWideBroadeningCoeff_BasilarM_Stone3(double theWideBroadeningCoeff)
{
	basilarM3.wideBroadeningCoeffFlag = TRUE;
	basilarM3.wideBroadeningCoeff = theWideBroadeningCoeff;
	return(TRUE);
	
}

/********************** SetCompressionThreshold *******************************/

/*
 * This function sets the module's compression threshold parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetCompressionThreshold_BasilarM_Stone3(double theCompressionThreshold)
{
	basilarM3.compressionThresholdFlag = TRUE;
	basilarM3.compressionThreshold = theCompressionThreshold;
	return(TRUE);
	
}

/********************** SetCompressionExponent ********************************/

/*
 * This function sets the module's compression exponent parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetCompressionExponent_BasilarM_Stone3(double theCompressionExponent)
{
	basilarM3.compressionExponentFlag = TRUE;
	basilarM3.compressionExponent = theCompressionExponent;
	return(TRUE);
	
}

/********************** SetWideFCF ********************************************/

/*
 * This function sets the module's wide filter centre frequency parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetWideFCF_BasilarM_Stone3(double theWideFCF)
{
	basilarM3.wideFCFFlag = TRUE;
	basilarM3.wideFCF = theWideFCF;
	return(TRUE);
	
}

/********************** SetWideFRefCF *****************************************/

/*
 * This function sets the module's wide filter reference centre frequency
 * parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetWideFRefCF_BasilarM_Stone3(double theWideFRefCF)
{
	if (theWideFRefCF == 0.0) {
		NotifyError("SetWideFRefCF_BasilarM_Stone3: Illegal value!");
		return(FALSE);
	}
	basilarM3.wideFRefCFFlag = TRUE;
	basilarM3.wideFRefCF = theWideFRefCF;
	return(TRUE);
	
}

/********************** SetBaseWideFAtten *************************************/

/*
 * This function sets the module's base wide filter centre attenuation
 * parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetBaseWideFAtten_BasilarM_Stone3(double theBaseWideFAtten)
{
	basilarM3.baseWideFAttenFlag = TRUE;
	basilarM3.baseWideFAtten = theBaseWideFAtten;
	return(TRUE);
	
}

/********************** SetWideFAttenScale ************************************/

/*
 * This function sets the module's wide filter attenuation scale parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetWideFAttenScale_BasilarM_Stone3(double theWideFAttenScale)
{
	basilarM3.wideFAttenScaleFlag = TRUE;
	basilarM3.wideFAttenScale = theWideFAttenScale;
	return(TRUE);
	
}

/********************** SetNarrowFLPCutOffScale *******************************/

/*
 * This function sets the module's narrow filter low-pass cut-off scale
 * parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetNarrowFLPCutOffScale_BasilarM_Stone3(double theNarrowFLPCutOffScale)
{
	basilarM3.narrowFLPCutOffScaleFlag = TRUE;
	basilarM3.narrowFLPCutOffScale = theNarrowFLPCutOffScale;
	return(TRUE);
	
}

/********************** SetWideFLPCutOffScale *********************************/

/*
 * This function sets the module's wide filter low-pass cut-off scale
 * parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetWideFLPCutOffScale_BasilarM_Stone3(double theWideFLPCutOffScale)
{
	basilarM3.wideFLPCutOffScaleFlag = TRUE;
	basilarM3.wideFLPCutOffScale = theWideFLPCutOffScale;
	return(TRUE);
	
}

/*********************** SetCFList ********************************************/

/*
 * This function sets the CFList data structure for the filter bank.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetCFList_BasilarM_Stone3(CFList *theCFList)
{
	if (!CheckPars_CFList(theCFList)) {
		NotifyError("SetCFList_BasilarM_Stone3: Centre frequency structure "\
		  "not initialised correctly set.");
		return(FALSE);
	}
	basilarM3.theCFsFlag = TRUE;
	basilarM3.theCFs = theCFList;
	return(TRUE);

}

/*********************** SetNumChannels ***************************************/

/*
 * This function sets the number of channels for the filter bank.
 * It also initialises the CFList data structure, if it has not already been
 * initialised.
 * It must be set before the number of centre frequencies is set.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetNumChannels_BasilarM_Stone3(int theChannels)
{
	if ((theChannels > MAX_CHANNELS) || (theChannels < 0)) {
		NotifyError("SetNumChannels_BasilarM_Stone3: %d is not in the range "\
		  "0-%d.", theChannels, MAX_CHANNELS);
		return(FALSE);
	}
	if (!basilarM3.theCFsFlag) {
		if ((basilarM3.theCFs =
		  Init_CFList("SetNumChannels_BasilarM_Stone3")) == NULL) {
			NotifyError("SetNumChannels_BasilarM_Stone3: Could not " \
			  "initialise centre frequency list structure.");
			return(FALSE);
		}
		basilarM3.theCFsFlag = TRUE;
	}
	basilarM3.theCFs->numChannels = theChannels;
	return(TRUE);

}

/*********************** SetCentreFreq ****************************************/

/*
 * This function sets the array of centre frequencies for the filter bank.
 * It first checks that the module has first been initialised.
 * It also initialises the CFList data structure, if it has not already been
 * initialised.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetCentreFreq_BasilarM_Stone3(double *theFrequencies, size_t arraySize)
{
	int		i, arrayDimension;
	
	if (!basilarM3.theCFs->numChannels) {
		NotifyError("SetCentreFreq_BasilarM_Stone3: The number of channels "\
		  "must be set first! (Use SetNumChannels_BasilarM_Stone3.)");
		return(FALSE);
	}
	arrayDimension = arraySize / sizeof (double);
	if ( arrayDimension < basilarM3.theCFs->numChannels) {
		NotifyError("SetCentreFreq_BasilarM_Stone3: Array dimension, %d, is "\
		  "less than the number of channels, %d, specified.", arrayDimension,
		  basilarM3.theCFs->numChannels);
		return(FALSE);
	}
	/* The check on the CFList initialisation is not necessarily needed, as the
	 * setting of the number of channels will take care of this, and must, at
	 * present, be done first.  However, this has been included, just in case
	 * this policy is changed later and I forget about doing it.
	 */
	if (!basilarM3.theCFsFlag) {
		if ((basilarM3.theCFs =
		  Init_CFList("SetCentreFreq_BasilarM_Stone3")) == NULL) {
			NotifyError("SetCentreFreq_BasilarM_Stone3: Could not initialise "\
			  "centre frequency list structure.");
			return(FALSE);
		}
		basilarM3.theCFsFlag = TRUE;
	}
	for (i = 0; i < basilarM3.theCFs->numChannels; i++) {
		if (theFrequencies[i] < 0.0) {
			NotifyError("SetCentreFreq_BasilarM_Stone3: Illegal frequency[%d] = "\
			  "%g Hz.", i, theFrequencies[i]);
			return(FALSE);
		}
		basilarM3.theCFs->channel[i] = theFrequencies[i];
	}
	return(TRUE);

}

/****************************** SetPars ***************************************/

/*
 * This function sets all the module's parameters.
 * array, It uses the variable argument format to accept the (i) order,
 * numChannels, CF array size, and; (ii) order, numChannels = -1, CFList data
 * structure - formats.
 * Each format only reads the number of arguments required.
 * At present, the error testing for the correct argument formats is somewhat
 * tenuous.
 * It returns TRUE if the operation is successful.
 */
  
BOOLEAN
SetPars_BasilarM_Stone3(int theOrder, double theNarrowBroadeningCoeff,
  double theWideBroadeningCoeff, double theCompressionThreshold,
  double theCompressionExponent, double wideFCF, double wideFRefCF,
  double baseWideFAtten, double wideFAttenScale, double narrowFLPCutOffScale,
  double wideFLPCutOffScale, int theChannels, ...)
{
	size_t	arraySize, arrayDimension;
	double	*theFrequencies;
	va_list	ap;
	BOOLEAN	ok;
	CFList	*theCFList;
	
	ok = TRUE;
	if (!SetOrder_BasilarM_Stone3(theOrder))
		ok = FALSE;
	if (!SetNarrowBroadeningCoeff_BasilarM_Stone3(theNarrowBroadeningCoeff))
		ok = FALSE;
	if (!SetWideBroadeningCoeff_BasilarM_Stone3(theWideBroadeningCoeff))
		ok = FALSE;
	if (!SetCompressionThreshold_BasilarM_Stone3(theCompressionThreshold))
		ok = FALSE;
	if (!SetCompressionExponent_BasilarM_Stone3(theCompressionExponent))
		ok = FALSE;
	if (!SetWideFCF_BasilarM_Stone3(wideFCF))
		ok = FALSE;
	if (!SetWideFRefCF_BasilarM_Stone3(wideFRefCF))
		ok = FALSE;
	if (!SetBaseWideFAtten_BasilarM_Stone3(baseWideFAtten))
		ok = FALSE;
	if (!SetWideFAttenScale_BasilarM_Stone3(wideFAttenScale))
		ok = FALSE;
	if (!SetNarrowFLPCutOffScale_BasilarM_Stone3(narrowFLPCutOffScale))
		ok = FALSE;
	if (!SetWideFLPCutOffScale_BasilarM_Stone3(wideFLPCutOffScale))
		ok = FALSE;
	va_start(ap, theChannels);		/* Make ap point to the 1st unnamed arg. */
	if (theChannels >= 0) {
		theFrequencies = va_arg(ap, double *);
		arraySize = va_arg(ap, size_t);
		arrayDimension = arraySize / sizeof(double);
		if (arrayDimension > MAX_CHANNELS) {
			NotifyError("SetPars_BasilarM_Stone3: Illegal use of function! "\
			  "Check arguments.");
			ok = FALSE;
		} else {
			if (!SetNumChannels_BasilarM_Stone3(theChannels))
				ok = FALSE;
			if (!SetCentreFreq_BasilarM_Stone3(theFrequencies, arraySize))
				ok = FALSE;
		}
	} else {
		theCFList = (CFList *) va_arg(ap, CFList *);
		if (!SetCFList_BasilarM_Stone3(theCFList))
			ok = FALSE;
	}
	va_end(ap);		/* Clean up when done. */
	if (!ok)
		NotifyError("SetPars_BasilarM_Stone3: Failed to set all module " \
		  "parameters.");
	return(ok);
	
}

/****************************** GetCF *****************************************/

/*
 * This routine returns the respective centre frequency value for a module.
 *
 */

double
GetCF_BasilarM_Stone3(int channel)
{
	if (!CheckPars_BasilarM_Stone3()) {
		NotifyError("GetCF_BasilarM_Stone3: Parameters have not been "\
		  "correctly set, zero returned.");
		return(0.0);
	}
	if (channel < 0 || channel >= basilarM3.theCFs->numChannels) {
		NotifyError("GetCF_BasilarM_Stone3: Channel number must be in the "\
		  "range 0 to %d, zero returned.", basilarM3.theCFs->numChannels - 1);
		return(0.0);
	}
	return(basilarM3.theCFs->channel[channel]);

}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_BasilarM_Stone3(void)
{
	int		i;
	
	if (!CheckPars_BasilarM_Stone3()) {
		NotifyError("PrintPars_BasilarM_Stone3: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("Stone 2 Composite Basilar Membrane Filter Module Parameters:-\n");
	printf("\t%10s\t%10s\n", "Filter No.", "Frequency");
	printf("\t%10s\t%10s\n", "          ", "   (Hz)  ");
	for (i = 0; i < basilarM3.theCFs->numChannels; i++)
		printf("\t%10d\t%10g\n", i, basilarM3.theCFs->channel[i]);
	printf("\tFilter order = %d.\n", basilarM3.order);
	printf("\tNarrow filter broadening coefficient = %g,\n",
	  basilarM3.narrowBroadeningCoeff);
	printf("\tWide filter broadening coefficient = %g,\n",
	  basilarM3.wideBroadeningCoeff);
	printf("\tCompression Threshold = %g dB SPL,\tCompression Exponent = "\
	  "%g.\n", basilarM3.compressionThreshold,
	  basilarM3.compressionExponent);
	printf("\tWide filter CF = %g Hz,\tWide filter reference CF = %g Hz,\n",
	  basilarM3.wideFCF, basilarM3.wideFRefCF);
	printf("\tBase wide filter attenuation = %g dB SPL,\n",
	  basilarM3.baseWideFAtten);
	printf("\tWide filter attenuation scale = %g,\n",
	  basilarM3.wideFAttenScale);
	printf("\tNarrow/wide filter low-pass cut-off scale = %g / %g "\
	  "respectively.\n", basilarM3.narrowFLPCutOffScale,
	  basilarM3.wideFLPCutOffScale);

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_BasilarM_Stone3(char *fileName)
{
	BOOLEAN	ok;
	char	theFrequencyMode[MAXLINE];
	int		filterOrder;
	double	narrowBroadeningCoeff, wideBroadeningCoeff;
	double	compressionThreshold, compressionExponent;
	double	wideFCF, wideFRefCF, baseWideFAtten, wideFAttenScale;
	double	narrowFLPCutOffScale, wideFLPCutOffScale;
    FILE    *fp;
    CFListPtr	frequencies;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_BasilarM_Stone3: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_BasilarM_Stone3: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
	if (!GetPars_ParFile(fp, "%d", &filterOrder))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%s", theFrequencyMode))
		ok = FALSE;
	else
		ok = ((frequencies = Read_CFList(fp, theFrequencyMode)) != NULL);
	if (!GetPars_ParFile(fp, "%lf", &narrowBroadeningCoeff))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &wideBroadeningCoeff))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &compressionThreshold))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &compressionExponent))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &wideFCF))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &wideFRefCF))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &baseWideFAtten))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &wideFAttenScale))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &narrowFLPCutOffScale))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &wideFLPCutOffScale))
		ok = FALSE;
	fclose(fp);
	Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_BasilarM_Stone3: Not enough lines, or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_BasilarM_Stone3(filterOrder, narrowBroadeningCoeff,
	  wideBroadeningCoeff, compressionThreshold, compressionExponent,
	  wideFCF, wideFRefCF, baseWideFAtten, wideFAttenScale,
	  narrowFLPCutOffScale, wideFLPCutOffScale, USE_CFLIST, frequencies)) {
		NotifyError("ReadPars_BasilarM_Stone3: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
   
}

/**************************** InitCoeffs **************************************/

/*
 * This routine initialises the coefficients for all the filters involved in
 * the Stone composite filter.
 * It assumes that the centre frequency list is set up correctly.
 * At present there are two constants used in this routine i.e. 0.8333 and 0.9.
 * I will formalise their definition later (huh!).
 */

void
InitCoeffs_BasilarM_Stone3(GammaToneCoeffs *narrowGammaTone[],
  GammaToneCoeffs *wideGammaTone[], TwoPoleCoeffs *butterworthLP[],
  TwoPoleCoeffs *besselLP[], EarObjectPtr data)
{
	int		i, auralChan, subFilterOrder;
	double	sampleRate;

	sampleRate = 1.0 / data->inSignal->dt;
	subFilterOrder = basilarM3.order / 2;
	for (auralChan = 0; auralChan < data->inSignal->numChannels; auralChan++)
		for (i = 0; i < basilarM3.theCFs->numChannels; i++) {
			narrowGammaTone[i + auralChan] =
			  InitCoeffs_Filters_GammaTVWidth(basilarM3.theCFs->channel[i],
			  sampleRate, subFilterOrder, basilarM3.narrowBroadeningCoeff);
			wideGammaTone[i + auralChan] = InitCoeffs_Filters_GammaTVWidth(
			  basilarM3.wideFCF, sampleRate, subFilterOrder,
			  basilarM3.wideBroadeningCoeff);
			butterworthLP[i + auralChan] =
			  InitCoeffs_Filters_IIR2(Filters_butt2Poly,
			  basilarM3.theCFs->channel[i] * basilarM3.narrowFLPCutOffScale,
			  sampleRate, LOWPASS);
			besselLP[i + auralChan] = InitCoeffs_Filters_IIR2(Filters_bess2Poly,
			  basilarM3.wideFCF * basilarM3.wideFLPCutOffScale,
			  sampleRate, LOWPASS);
		}

}

/**************************** FreeCoeffs **************************************/

/*
 * This routine frees the memory allocated for the Stone composite filter
 * coefficients.
 */

void
FreeCoeffs_BasilarM_Stone3(GammaToneCoeffs *narrowGammaTone[],
  GammaToneCoeffs *wideGammaTone[], TwoPoleCoeffs *butterworthLP[],
  TwoPoleCoeffs *besselLP[], EarObjectPtr data)
{
	int		i;

	for (i = 0; i < data->outSignal->numChannels; i++) {
		free(narrowGammaTone[i]);
		free(wideGammaTone[i]);
		free(butterworthLP[i]);
		free(besselLP[i]);
	}
}

/**************************** Simulate_Model **********************************/

/*
 * This routine filters the input signal and puts the result into the output
 * signal.
 * It checks that all initialisation has been correctly carried out by calling
 * the appropriate checking routines.
 * The appropriate coefficients are calculated at the beginning, then the memory
 * used is released before the program returns.
 * The wide filter signal is initialised by running this routine.  Thereafter
 * memory is only re-allocated if the signal changes in dimension.  The memory
 * for the wide filter signal is free'd when the module is un-initialised.
 */

BOOLEAN
RunModel_BasilarM_Stone3(EarObjectPtr data)
{
	int		i, totalChannels;
	double	attenuationFactor[MAX_CHANNELS];
	GammaToneCoeffs	*narrowGammaTone[MAX_CHANNELS];
	GammaToneCoeffs *wideGammaTone[MAX_CHANNELS];
	TwoPoleCoeffs	*butterworthLP[MAX_CHANNELS], *besselLP[MAX_CHANNELS];
				
	if (data == NULL) {
		NotifyError("RunModel_BasilarM_Stone3: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_BasilarM_Stone3())
		return(FALSE);
		
	/* Initialise Variables and coefficients */
	
	SetProcessName_EarObject(data, "Stone3 Basilar Membrane Filtering");
	if (!CheckRamp_SignalData(data->inSignal)) {
		NotifyError("RunModel_BasilarM_Stone3: Input signal not correctly " \
		  "initialised.");
		return(FALSE);
	}
	totalChannels = basilarM3.theCFs->numChannels * data->inSignal->numChannels;
	InitOutFromInSignal_EarObject(data, totalChannels);
	basilarM3.wideF = Init_EarObject("NULL");
	/* "Manual" connection of original input signal to the wide filter. */
	basilarM3.wideF->inSignal = data->inSignal;
	InitOutFromInSignal_EarObject(basilarM3.wideF, totalChannels);
	InitCoeffs_BasilarM_Stone3(narrowGammaTone, wideGammaTone, butterworthLP,
	  besselLP, data);	
	
	/* Filter signal */
	
	GammaTone_Filters(data->outSignal, narrowGammaTone);
	IIR2_Filters(data->outSignal, butterworthLP);
	
	Compression_Filters(data->outSignal, RMS_AMP(basilarM3.compressionThreshold) *
	  SQRT_2, basilarM3.compressionExponent);
	
	GammaTone_Filters(data->outSignal, narrowGammaTone);
	IIR2_Filters(data->outSignal, butterworthLP);
	
	GammaTone_Filters(basilarM3.wideF->outSignal, wideGammaTone);
	IIR2_Filters(basilarM3.wideF->outSignal, besselLP);
	
	/* Addition of signals */
	
	for (i = 0; i < basilarM3.theCFs->numChannels; i++)
		attenuationFactor[i] = RMS_AMP(basilarM3.baseWideFAtten +
		  basilarM3.wideFAttenScale *
		  log10(basilarM3.theCFs->channel[i / data->inSignal->numChannels] /
		  basilarM3.wideFRefCF));
	AddScaledData_SignalData(data->outSignal, basilarM3.wideF->outSignal,
	  attenuationFactor);
	
	FreeCoeffs_BasilarM_Stone3(narrowGammaTone, wideGammaTone, butterworthLP,
	  besselLP, data);
	Free_EarObject(&basilarM3.wideF);
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

}
