/*******************
 *
 * Mo4BasilarM.c
 *
 * This is an implementation of a Gamma Tone filter using the filters.c module.
 * The Stone Composite filter is used to model the response of the Basilar
 * Membrane.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 *******************/
 
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "UtFilters.h"
#include "UtCFList.h"
#include "Mo4BasilarM.h"
#include "FiParFile.h"

/******************************************************************************/
/*************************** Global Variables *********************************/
/******************************************************************************/

BasilarM4	basilarM4 = {

				FALSE,	/* orderFlag */
				FALSE,	/* compressionThresholdFlag */
				FALSE,	/* compressionExponentFlag */
				FALSE,	/* wideFCFFlag */
				FALSE,	/* narrowFLPCutOffScaleFlag */
				FALSE,	/* wideFLPCutOffScaleFlag */
				FALSE,	/* theCFsFlag */
				0,		/* order */
				0.0,	/* compressionThreshold */
				0.0,	/* compressionExponent */
				0.0,	/* wideFCF */
				0.0,	/* narrowFLPCutOffScale */
				0.0,	/* wideFLPCutOffScale */
				NULL,	/* theCFs */
				NULL	/* Wide Filter */
				
			};

/******************************************************************************/
/**************************** Subroutines & functions *************************/
/******************************************************************************/

/********************************* CheckPars **********************************/

/*
 * This routine checks that the necessary parameters for the module have been
 * correctly initialised.
 * It returns TRUE if there are no problems.
 */
 
BOOLEAN
CheckPars_BasilarM_Stone4(void)
{
	BOOLEAN	ok;

	ok = TRUE;
	if (!basilarM4.orderFlag) {
		NotifyError("CheckPars_BasilarM_Stone4: Filter order not set.");
		ok = FALSE;
	}
	if (!basilarM4.compressionThresholdFlag) {
		NotifyError("CheckPars_BasilarM_Stone4: Filter compression threshold "
		  "not set.");
		ok = FALSE;
	}
	if (!basilarM4.compressionExponentFlag) {
		NotifyError("CheckPars_BasilarM_Stone4: Filter compression ratio not "\
		  "set.");
		ok = FALSE;
	}
	if (!basilarM4.wideFCFFlag) {
		NotifyError("CheckPars_BasilarM_Stone4: Wide filter centre frequency "\
		  "not set.");
		ok = FALSE;
	}
	if (!basilarM4.narrowFLPCutOffScaleFlag) {
		NotifyError("CheckPars_BasilarM_Stone4: Narrow filter low-pass cut-off "\
		  "scale not set.");
		ok = FALSE;
	}
	if (!basilarM4.wideFLPCutOffScaleFlag) {
		NotifyError("CheckPars_BasilarM_Stone4: Wide filter low pass cut-off "\
		  "scale not set.");
		ok = FALSE;
	}
	if (!basilarM4.theCFsFlag) {
		NotifyError("CheckPars_BasilarM_Stone4: Centre frequencies for the "\
		  "filter bank not set.");
		ok = FALSE;
	} else {
		if (!CheckPars_CFList(basilarM4.theCFs)) {
			NotifyError("CheckPars_BasilarM_Stone4: Centre frequency list "\
			  " parameters not correctly set.");
			ok = FALSE;
		}
	}
	return(ok);

}

/********************************* SetOrder ***********************************/

/*
 * This function sets the module's order parameter.  It first checks that the
 * module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetOrder_BasilarM_Stone4(int theOrder)
{
	basilarM4.orderFlag = TRUE;
	basilarM4.order = theOrder;
	return(TRUE);
	
}

/********************** SetCompressionThreshold *******************************/

/*
 * This function sets the module's compression threshold parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetCompressionThreshold_BasilarM_Stone4(double theCompressionThreshold)
{
	basilarM4.compressionThresholdFlag = TRUE;
	basilarM4.compressionThreshold = theCompressionThreshold;
	return(TRUE);
	
}

/********************** SetCompressionExponent ********************************/

/*
 * This function sets the module's compression exponent parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetCompressionExponent_BasilarM_Stone4(double theCompressionExponent)
{
	basilarM4.compressionExponentFlag = TRUE;
	basilarM4.compressionExponent = theCompressionExponent;
	return(TRUE);
	
}

/********************** SetWideFCF ********************************************/

/*
 * This function sets the module's wide filter centre frequency parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetWideFCF_BasilarM_Stone4(double theWideFCF)
{
	basilarM4.wideFCFFlag = TRUE;
	basilarM4.wideFCF = theWideFCF;
	return(TRUE);
	
}

/********************** SetNarrowFLPCutOffScale *******************************/

/*
 * This function sets the module's narrow filter low-pass cut-off scale
 * parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetNarrowFLPCutOffScale_BasilarM_Stone4(double theNarrowFLPCutOffScale)
{
	basilarM4.narrowFLPCutOffScaleFlag = TRUE;
	basilarM4.narrowFLPCutOffScale = theNarrowFLPCutOffScale;
	return(TRUE);
	
}

/********************** SetWideFLPCutOffScale *********************************/

/*
 * This function sets the module's wide filter low-pass cut-off scale
 * parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetWideFLPCutOffScale_BasilarM_Stone4(double theWideFLPCutOffScale)
{
	basilarM4.wideFLPCutOffScaleFlag = TRUE;
	basilarM4.wideFLPCutOffScale = theWideFLPCutOffScale;
	return(TRUE);
	
}

/*********************** SetCFList ********************************************/

/*
 * This function sets the CFList data structure for the filter bank.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetCFList_BasilarM_Stone4(CFList *theCFList)
{
	if (!CheckPars_CFList(theCFList)) {
		NotifyError("SetCFList_BasilarM_Stone4: Centre frequency structure "\
		  "not initialised correctly set.");
		return(FALSE);
	}
	basilarM4.theCFsFlag = TRUE;
	basilarM4.theCFs = theCFList;
	return(TRUE);

}

/*********************** SetNumChannels ***************************************/

/*
 * This function sets the number of channels for the filter bank.
 * It also initialises the CFList data structure, if it has not already been
 * initialised.
 * It must be set before the number of centre frequencies is set.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetNumChannels_BasilarM_Stone4(int theChannels)
{
	if ((theChannels > MAX_CHANNELS) || (theChannels < 0)) {
		NotifyError("SetNumChannels_BasilarM_Stone4: %d is not in the range "\
		  "0-%d.", theChannels, MAX_CHANNELS);
		return(FALSE);
	}
	if (!basilarM4.theCFsFlag) {
		if ((basilarM4.theCFs =
		  Init_CFList("SetNumChannels_BasilarM_Stone4")) == NULL) {
			NotifyError("SetNumChannels_BasilarM_Stone4: Could not " \
			  "initialise centre frequency list structure.");
			return(FALSE);
		}
		basilarM4.theCFsFlag = TRUE;
	}
	basilarM4.theCFs->numChannels = theChannels;
	return(TRUE);

}

/*********************** SetCentreFreq ****************************************/

/*
 * This function sets the array of centre frequencies for the filter bank.
 * It first checks that the module has first been initialised.
 * It also initialises the CFList data structure, if it has not already been
 * initialised.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetCentreFreq_BasilarM_Stone4(double *theFrequencies, size_t arraySize)
{
	int		i, arrayDimension;
	
	if (!basilarM4.theCFs->numChannels) {
		NotifyError("SetCentreFreq_BasilarM_Stone4: The number of channels "\
		  "must be set first! (Use SetNumChannels_BasilarM_Stone4.)");
		return(FALSE);
	}
	arrayDimension = arraySize / sizeof (double);
	if ( arrayDimension < basilarM4.theCFs->numChannels) {
		NotifyError("SetCentreFreq_BasilarM_Stone4: Array dimension, %d, is "\
		  "less than the number of channels, %d, specified.", arrayDimension,
		  basilarM4.theCFs->numChannels);
		return(FALSE);
	}
	/* The check on the CFList initialisation is not necessarily needed, as the
	 * setting of the number of channels will take care of this, and must, at
	 * present, be done first.  However, this has been included, just in case
	 * this policy is changed later and I forget about doing it.
	 */
	if (!basilarM4.theCFsFlag) {
		if ((basilarM4.theCFs =
		  Init_CFList("SetCentreFreq_BasilarM_Stone4")) == NULL) {
			NotifyError("SetCentreFreq_BasilarM_Stone4: Could not initialise "\
			  "centre frequency list structure.");
			return(FALSE);
		}
		basilarM4.theCFsFlag = TRUE;
	}
	for (i = 0; i < basilarM4.theCFs->numChannels; i++) {
		if (theFrequencies[i] < 0.0) {
			NotifyError("SetCentreFreq_BasilarM_Stone4: Illegal frequency[%d] = "\
			  "%g Hz.", i, theFrequencies[i]);
			return(FALSE);
		}
		basilarM4.theCFs->channel[i] = theFrequencies[i];
	}
	return(TRUE);

}


/****************************** SetPars ***************************************/

/*
 * This function sets all the module's parameters.
 * array, It uses the variable argument format to accept the (i) order,
 * numChannels, CF array size, and; (ii) order, numChannels = -1, CFList data
 * structure - formats.
 * Each format only reads the number of arguments required.
 * At present, the error testing for the correct argument formats is somewhat
 * tenuous.
 * It returns TRUE if the operation is successful.
 */
  
BOOLEAN
SetPars_BasilarM_Stone4(int theOrder, double theCompressionThreshold,
  double theCompressionExponent, double wideFCF, double narrowFLPCutOffScale,
  double wideFLPCutOffScale, int theChannels, ...)
{
	size_t	arraySize, arrayDimension;
	double	*theFrequencies;
	va_list	ap;
	BOOLEAN	ok;
	CFList	*theCFList;
	
	ok = TRUE;
	if (!SetOrder_BasilarM_Stone4(theOrder))
		ok = FALSE;
	if (!SetCompressionThreshold_BasilarM_Stone4(theCompressionThreshold))
		ok = FALSE;
	if (!SetCompressionExponent_BasilarM_Stone4(theCompressionExponent))
		ok = FALSE;
	if (!SetWideFCF_BasilarM_Stone4(wideFCF))
		ok = FALSE;
	if (!SetNarrowFLPCutOffScale_BasilarM_Stone4(narrowFLPCutOffScale))
		ok = FALSE;
	if (!SetWideFLPCutOffScale_BasilarM_Stone4(wideFLPCutOffScale))
		ok = FALSE;
	va_start(ap, theChannels);		/* Make ap point to the 1st unnamed arg. */
	if (theChannels >= 0) {
		theFrequencies = va_arg(ap, double *);
		arraySize = va_arg(ap, size_t);
		arrayDimension = arraySize / sizeof(double);
		if (arrayDimension > MAX_CHANNELS) {
			NotifyError("SetPars_BasilarM_Stone4: Illegal use of function! "\
			  "Check arguments.");
			ok = FALSE;
		} else {
			if (!SetNumChannels_BasilarM_Stone4(theChannels))
				ok = FALSE;
			if (!SetCentreFreq_BasilarM_Stone4(theFrequencies, arraySize))
				ok = FALSE;
		}
	} else {
		theCFList = (CFList *) va_arg(ap, CFList *);
		if (!SetCFList_BasilarM_Stone4(theCFList))
			ok = FALSE;
	}
	va_end(ap);		/* Clean up when done. */
	if (!ok)
		NotifyError("SetPars_BasilarM_Stone4: Failed to set all module " \
		  "parameters.");
	return(ok);
	
}

/****************************** GetCF *****************************************/

/*
 * This routine returns the respective centre frequency value for a module.
 *
 */

double
GetCF_BasilarM_Stone4(int channel)
{
	if (!CheckPars_BasilarM_Stone4()) {
		NotifyError("GetCF_BasilarM_Stone4: Parameters have not been "\
		  "correctly set, zero returned.");
		return(0.0);
	}
	if (channel < 0 || channel >= basilarM4.theCFs->numChannels) {
		NotifyError("GetCF_BasilarM_Stone4: Channel number must be in the "\
		  "range 0 to %d, zero returned.", basilarM4.theCFs->numChannels - 1);
		return(0.0);
	}
	return(basilarM4.theCFs->channel[channel]);

}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_BasilarM_Stone4(void)
{
	int		i;
	
	if (!CheckPars_BasilarM_Stone4()) {
		NotifyError("PrintPars_BasilarM_Stone4: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("Stone 2 Composite Basilar Membrane Filter Module Parameters:-\n");
	printf("\t%10s\t%10s\n", "Filter No.", "Frequency");
	printf("\t%10s\t%10s\n", "          ", "   (Hz)  ");
	for (i = 0; i < basilarM4.theCFs->numChannels; i++)
		printf("\t%10d\t%10g\n", i, basilarM4.theCFs->channel[i]);
	printf("\tFilter order = %d.\n", basilarM4.order);
	printf("\tCompression Threshold = %g dB SPL,\tCompression Exponent = "\
	  "%g.\n", basilarM4.compressionThreshold,
	  basilarM4.compressionExponent);
	printf("\tWide filter CF = %g Hz,\n", basilarM4.wideFCF);
	printf("\tNarrow/wide filter low-pass cut-off scale = %g / %g "\
	  "respectively.\n", basilarM4.narrowFLPCutOffScale,
	  basilarM4.wideFLPCutOffScale);

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_BasilarM_Stone4(char *fileName)
{
	BOOLEAN	ok;
	char	theFrequencyMode[MAXLINE];
	int		filterOrder;
	double	compressionThreshold, compressionExponent;
	double	wideFCF, narrowFLPCutOffScale, wideFLPCutOffScale;
    FILE    *fp;
    CFListPtr	frequencies;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_BasilarM_Stone4: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_BasilarM_Stone4: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
	if (!GetPars_ParFile(fp, "%d", &filterOrder))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%s", theFrequencyMode))
		ok = FALSE;
	else
		ok = ((frequencies = Read_CFList(fp, theFrequencyMode)) != NULL);
	if (!GetPars_ParFile(fp, "%lf", &compressionThreshold))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &compressionExponent))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &wideFCF))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &narrowFLPCutOffScale))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &wideFLPCutOffScale))
		ok = FALSE;
	fclose(fp);
	Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_BasilarM_Stone4: Not enough lines, or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_BasilarM_Stone4(filterOrder, compressionThreshold,
	  compressionExponent, wideFCF, narrowFLPCutOffScale, wideFLPCutOffScale,
	    USE_CFLIST, frequencies)) {
		NotifyError("ReadPars_BasilarM_Stone4: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
   
}

/**************************** InitCoeffs **************************************/

/*
 * This routine initialises the coefficients for all the filters involved in
 * the Stone composite filter.
 * It assumes that the centre frequency list is set up correctly.
 * At present there are two constants used in this routine i.e. 0.8333 and 0.9.
 * I will formalise their definition later (huh!).
 */

void
InitCoeffs_BasilarM_Stone4(GammaToneCoeffs *narrowGammaTone[],
  GammaToneCoeffs *wideGammaTone[], TwoPoleCoeffs *butterworthLP[],
  TwoPoleCoeffs *besselLP[], EarObjectPtr data)
{
	int		i, auralChan, subFilterOrder;
	double	sampleRate, frequency;

	sampleRate = 1.0 / data->inSignal->dt;
	subFilterOrder = basilarM4.order / 2;
	for (auralChan = 0; auralChan < data->inSignal->numChannels; auralChan++)
		for (i = 0; i < basilarM4.theCFs->numChannels; i++) {
			frequency = basilarM4.theCFs->channel[i];
			narrowGammaTone[i + auralChan] =
			  InitCoeffs_Filters_GammaTVWidth(frequency, sampleRate,
			  subFilterOrder, BM4_NARROW_BROADENING);
			wideGammaTone[i + auralChan] =
			  InitCoeffs_Filters_GammaTVWidth(basilarM4.wideFCF, sampleRate, 
			  subFilterOrder, BM4_BROADENING_K1 + BM4_BROADENING_K2 *
			  frequency + BM4_BROADENING_K3 * frequency * frequency);
			butterworthLP[i + auralChan] =
			  InitCoeffs_Filters_IIR2(Filters_butt2Poly,
			  basilarM4.theCFs->channel[i] * basilarM4.narrowFLPCutOffScale,
			  sampleRate, LOWPASS);
			besselLP[i + auralChan] = InitCoeffs_Filters_IIR2(Filters_bess2Poly,
			  basilarM4.wideFCF * basilarM4.wideFLPCutOffScale, sampleRate,
			  LOWPASS);
		}

}

/**************************** FreeCoeffs **************************************/

/*
 * This routine frees the memory allocated for the Stone composite filter
 * coefficients.
 */

void
FreeCoeffs_BasilarM_Stone4(GammaToneCoeffs *narrowGammaTone[],
  GammaToneCoeffs *wideGammaTone[], TwoPoleCoeffs *butterworthLP[],
  TwoPoleCoeffs *besselLP[], EarObjectPtr data)
{
	int		i;

	for (i = 0; i < data->outSignal->numChannels; i++) {
		free(narrowGammaTone[i]);
		free(wideGammaTone[i]);
		free(butterworthLP[i]);
		free(besselLP[i]);
	}
}

/**************************** Simulate_Model **********************************/

/*
 * This routine filters the input signal and puts the result into the output
 * signal.
 * It checks that all initialisation has been correctly carried out by calling
 * the appropriate checking routines.
 * The appropriate coefficients are calculated at the beginning, then the memory
 * used is released before the program returns.
 * The wide filter signal is initialised by running this routine.  Thereafter
 * memory is only re-allocated if the signal changes in dimension.  The memory
 * for the wide filter signal is free'd when the module is un-initialised.
 */

BOOLEAN
RunModel_BasilarM_Stone4(EarObjectPtr data)
{
	int		i, totalChannels;
	double	attenuationFactor[MAX_CHANNELS], frequency;
	GammaToneCoeffs	*narrowGammaTone[MAX_CHANNELS];
	GammaToneCoeffs *wideGammaTone[MAX_CHANNELS];
	TwoPoleCoeffs	*butterworthLP[MAX_CHANNELS], *besselLP[MAX_CHANNELS];
				
	if (data == NULL) {
		NotifyError("RunModel_BasilarM_Stone4: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_BasilarM_Stone4())
		return(FALSE);
		
	/* Initialise Variables and coefficients */
	
	SetProcessName_EarObject(data, "Stone4 Basilar Membrane Filtering");
	if (!CheckRamp_SignalData(data->inSignal)) {
		NotifyError("RunModel_BasilarM_Stone4: Input signal not correctly " \
		  "initialised.");
		return(FALSE);
	}
	totalChannels = basilarM4.theCFs->numChannels * data->inSignal->numChannels;
	InitOutFromInSignal_EarObject(data, totalChannels);
	if (floor(1.0 / (2.0 * data->outSignal->dt)) <= basilarM4.wideFCF) {
		NotifyError("RunModel: Sampling rate (dt = %6.3g ms) is " \
		  "too low for the wide filter centre frequency.",
		  MSEC(data->outSignal->dt));
		return(FALSE);
	} /* Nyquist critical frequency test. */

	basilarM4.wideF = Init_EarObject("NULL");
	/* "Manual" connection of original input signal to the wide filter. */
	basilarM4.wideF->inSignal = data->inSignal;
	InitOutFromInSignal_EarObject(basilarM4.wideF, totalChannels);
	InitCoeffs_BasilarM_Stone4(narrowGammaTone, wideGammaTone, butterworthLP,
	  besselLP, data);
	  	
	/* Filter signal */
	
	GammaTone_Filters(data->outSignal, narrowGammaTone);
	IIR2_Filters(data->outSignal, butterworthLP);
	
	Compression_Filters(data->outSignal,
	  RMS_AMP(basilarM4.compressionThreshold) * SQRT_2,
	  basilarM4.compressionExponent);
	
	GammaTone_Filters(data->outSignal, narrowGammaTone);
	IIR2_Filters(data->outSignal, butterworthLP);
	
	GammaTone_Filters(basilarM4.wideF->outSignal, wideGammaTone);
	IIR2_Filters(basilarM4.wideF->outSignal, besselLP);
	
	/* Addition of signals */
	
	for (i = 0; i < data->outSignal->numChannels; i++) {
		frequency = basilarM4.theCFs->channel[i / data->inSignal->numChannels];
		attenuationFactor[i] = RMS_AMP(BM4_ATTENUATION_K1 +
		  BM4_ATTENUATION_K2 * frequency + BM4_ATTENUATION_K3 * frequency *
		  frequency);
	}
	AddScaledData_SignalData(data->outSignal, basilarM4.wideF->outSignal,
	  attenuationFactor);
	
	FreeCoeffs_BasilarM_Stone4(narrowGammaTone, wideGammaTone, butterworthLP,
	  besselLP, data);
	Free_EarObject(&basilarM4.wideF);
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

}
