/******************
 *
 * MoArleKim.c
 *
 * This module implements Arle & Kim's version of the MacGregor Model.
 * The changes implemented in "Neural Modeling of instrinsic and spike-discharge
 * properties of cochlear nucleus neurons", J. E. Erle and D. O. Kim,
 * Biological Cybernetics, Springer Verlag 1991. - have been added to this
 * module version of the McGregor model.
 *
 * It was revised from code written by Trevor Shackleton and M. J. Hewitt.
 *
 ******************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "FiParFile.h"
#include "MoArleKim.h"

/******************************************************************************/
/*************************** Global Variables *********************************/
/******************************************************************************/

ArleKim		arleKim = {

					FALSE,		/* membraneTConstFlag */
					FALSE,		/* thresholdTConstFlag */
					FALSE,		/* accomConstFlag */
					FALSE,		/* delayedRectKCondFlag */
					FALSE,		/* restingThresholdFlag */
					FALSE,		/* actionPotentialFlag */
					FALSE,		/* kDecayTConstFlag */
					FALSE,		/* nonLinearVConstFlag */
					FALSE,		/* kReversalPotenFlag */
					FALSE,		/* bReversalPotenFlag */
					FALSE,		/* kRestingCondFlag */
					FALSE,		/* bRestingCondFlag */
					0.0,		/* membraneTConst_Tm */
					0.0,		/* kDecayTConst_TGk */
					0.0,		/* thresholdTConst_TTh */
					0.0,		/* accomConst_c */
					0.0,		/* delayedRectKCond_b */
					0.0,		/* restingThreshold_Th0 */
					0.0,		/* actionPotential */
					0.0,		/* nonLinearVConst_Vnl */
					0.0,		/* kReversalPoten_Ek */
					0.0,		/* bReversalPoten_Eb */
					0.0,		/* kRestingCond_gk */
					0.0,		/* bRestingCond_gb */
					
				};

/******************************************************************************/
/*************************** Subroutines and Functions ************************/
/******************************************************************************/

/********************************* SetMembraneTConst **************************/

/*
 * This function sets the module's membrane time constant, Tm, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetMembraneTConst_Neuron_ArleKim(double theMembraneTConst)
{
	if (theMembraneTConst == 0.0) {
		NotifyError("SetMembraneTConst_Neuron_ArleKim: Invalid value = "\
		  "%g mV\n", theMembraneTConst);
		return(FALSE);
	} 
	arleKim.membraneTConstFlag = TRUE;
	arleKim.membraneTConst_Tm = theMembraneTConst;
	return(TRUE);

}

/********************************* SetKDecayTConst ****************************/

/*
 * This function sets the module's potassium conductance decay time constant,
 * TGk, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetKDecayTConst_Neuron_ArleKim(double theKDecayTConst)
{
	if (theKDecayTConst == 0.0) {
		NotifyError("SetKDecayTConst_Neuron_ArleKim: Invalid value = "\
		  "%g mV\n", theKDecayTConst);
		return(FALSE);
	} 
	arleKim.kDecayTConstFlag = TRUE;
	arleKim.kDecayTConst_TGk = theKDecayTConst;
	return(TRUE);

}

/********************************* SetThresholdTConst *************************/

/*
 * This function sets the module's threshold time constant, TTh, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetThresholdTConst_Neuron_ArleKim(double theThresholdTConst)
{
	if (theThresholdTConst == 0.0) {
		NotifyError("SetThresholdTConst_Neuron_ArleKim: Invalid value = "\
		  "%g mV\n", theThresholdTConst);
		return(FALSE);
	} 
	arleKim.thresholdTConstFlag = TRUE;
	arleKim.thresholdTConst_TTh = theThresholdTConst;
	return(TRUE);

}

/********************************* SetAccomConst ******************************/

/*
 * This function sets the module's accomodation constant, c, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetAccomConst_Neuron_ArleKim(double theAccomConst)
{
	arleKim.accomConstFlag = TRUE;
	arleKim.accomConst_c = theAccomConst;
	return(TRUE);

}

/********************************* SetDelayedRectKCond ************************/

/*
 * This function sets the module's delayed rectifier potassium conductance
 * strength, b, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetDelayedRectKCond_Neuron_ArleKim(double theDelayedRectKCond)
{
	arleKim.delayedRectKCondFlag = TRUE;
	arleKim.delayedRectKCond_b = theDelayedRectKCond;
	return(TRUE);

}

/********************************* SetRestingThreshold ************************/

/*
 * This function sets the module's resting threhold, Th0, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetRestingThreshold_Neuron_ArleKim(double theRestingThreshold)
{
	arleKim.restingThresholdFlag = TRUE;
	arleKim.restingThreshold_Th0 = theRestingThreshold;
	return(TRUE);

}

/********************************* SetActionPotential *************************/

/*
 * This function sets the module's action potential parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetActionPotential_Neuron_ArleKim(double theActionPotential)
{
	arleKim.actionPotentialFlag = TRUE;
	arleKim.actionPotential = theActionPotential;
	return(TRUE);

}

/********************************* SetNonLinearVConst *************************/

/*
 * This function sets the module's non-linear voltage constant, Vnl, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetNonLinearVConst_Neuron_ArleKim(double theNonLinearVConst)
{
	if (theNonLinearVConst == 0.0) {
		NotifyError("SetNonLinearVConst_Neuron_ArleKim: Invalid value = "\
		  "%g mV\n", theNonLinearVConst);
		NotifyWarning("SetNonLinearVConst_Neuron_ArleKim: To set the term "\
		  "involving this parameter to zero set gb to zero.");
		return(FALSE);
	} 
	arleKim.nonLinearVConstFlag = TRUE;
	arleKim.nonLinearVConst_Vnl = theNonLinearVConst;
	return(TRUE);

}

/********************************* SetKReversalPoten **************************/

/*
 * This function sets the module's reversal potential of the K conductance
 * parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetKReversalPoten_Neuron_ArleKim(double theKReversalPoten)
{
	arleKim.kReversalPotenFlag = TRUE;
	arleKim.kReversalPoten_Ek = theKReversalPoten;
	return(TRUE);

}

/********************************* SetBReversalPoten **************************/

/*
 * This function sets the module's reversal potential of all other conductances
 * parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetBReversalPoten_Neuron_ArleKim(double theBReversalPoten)
{
	arleKim.bReversalPotenFlag = TRUE;
	arleKim.bReversalPoten_Eb = theBReversalPoten;
	return(TRUE);

}

/********************************* SetKRestingCond ****************************/

/*
 * This function sets the module's resting potassium conductance parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetKRestingCond_Neuron_ArleKim(double theKRestingCond)
{
	arleKim.kRestingCondFlag = TRUE;
	arleKim.kRestingCond_gk = theKRestingCond;
	return(TRUE);

}

/********************************* SetBRestingCond ****************************/

/*
 * This function sets the module's resting component of all other conductances
 * parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetBRestingCond_Neuron_ArleKim(double theBRestingCond)
{
	arleKim.bRestingCondFlag = TRUE;
	arleKim.bRestingCond_gb = theBRestingCond;
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 */

BOOLEAN
SetPars_Neuron_ArleKim(double tm, double tGk, double tTh, double c, double b,
  double tH0, double aP, double vnl, double ek, double eb, double gk,
  double gb)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetMembraneTConst_Neuron_ArleKim(tm))
		ok = FALSE;
	if (!SetKDecayTConst_Neuron_ArleKim(tGk))
		ok = FALSE;
	if (!SetThresholdTConst_Neuron_ArleKim(tTh))
		ok = FALSE;
	if (!SetAccomConst_Neuron_ArleKim(c))
	    ok = FALSE;
	if (!SetDelayedRectKCond_Neuron_ArleKim(b))
		ok = FALSE;
	if (!SetRestingThreshold_Neuron_ArleKim(tH0))
		ok = FALSE;
	if (!SetActionPotential_Neuron_ArleKim(aP))
		ok = FALSE;
	if (!SetNonLinearVConst_Neuron_ArleKim(vnl))
		ok = FALSE;
	if (!SetKReversalPoten_Neuron_ArleKim(ek))
		ok = FALSE;
	if (!SetBReversalPoten_Neuron_ArleKim(eb))
		ok = FALSE;
	if (!SetKRestingCond_Neuron_ArleKim(gk))
		ok = FALSE;
	if (!SetBRestingCond_Neuron_ArleKim(gb))
		ok = FALSE;
	return(ok);
	  
}

/********************************* CheckPars **********************************/

/*
 * This routine checks that all of the parameters for the module are set.
 * Because this module is defined by its coeffients, default values are always
 * set if none of the parameters are set, otherwise all of them must be
 * individually set.
 */

BOOLEAN
CheckPars_Neuron_ArleKim(void)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!arleKim.membraneTConstFlag) {
		NotifyError("CheckPars_Neuron_ArleKim: Membrane time constant, Tm, "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!arleKim.kDecayTConstFlag) {
		NotifyError("CheckPars_Neuron_ArleKim: Potassium decay time  "\
		  "constant, TGk, not correctly set.");
		ok = FALSE;
	}
	if (!arleKim.thresholdTConstFlag) {
		NotifyError("CheckPars_Neuron_ArleKim: Threshold time constant, TTh, "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!arleKim.accomConstFlag) {
		NotifyError("CheckPars_Neuron_ArleKim: Accomodation constant, c, "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!arleKim.delayedRectKCondFlag) {
		NotifyError("CheckPars_Neuron_ArleKim: Delayed rectifier potassium "\
		  "conductance, b, not correctly set.");
		ok = FALSE;
	}
	if (!arleKim.restingThresholdFlag) {
		NotifyError("CheckPars_Neuron_ArleKim: Resting threshold constant, "\
		  "Th0, not correctly set.");
		ok = FALSE;
	}
	if (!arleKim.actionPotentialFlag) {
		NotifyError("CheckPars_Neuron_ArleKim: Action potential constant "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!arleKim.nonLinearVConstFlag) {
		NotifyError("CheckPars_Neuron_ArleKim: Non-linear voltage constant, "\
		  "Vnl, not correctly set.");
		ok = FALSE;
	}
	if (!arleKim.kReversalPotenFlag) {
		NotifyError("CheckPars_Neuron_ArleKim: Reversal potential of the "\
		  "potassium conductance, Ek, not correctly set.");
		ok = FALSE;
	}
	if (!arleKim.bReversalPotenFlag) {
		NotifyError("CheckPars_Neuron_ArleKim: Reversal potential of all "\
		  "other conductances, Eb, not correctly set.");
		ok = FALSE;
	}
	if (!arleKim.kRestingCondFlag) {
		NotifyError("CheckPars_Neuron_ArleKim: Resting potassium "\
		  "conductance, gk, not correctly set.");
		ok = FALSE;
	}
	if (!arleKim.bRestingCondFlag) {
		NotifyError("CheckPars_Neuron_ArleKim: Resting component "\
		  "for all other conductances, gb,  not correctly set.");
		ok = FALSE;
	}
	/*
	This needs some work.
	dEExpression = (arleKim.kRestingCond_gk +
	  arleKim.bRestingCond_gb) / arleKim.membraneTConst_Tm;
	if (dEExpression < ARLEKIM_DE_LIMIT) {
		NotifyError("CheckPars_Neuron_ArleKim:  The total conductance, gk + "\
		  "gb, is too high or the membrane time-constant is too low.\n  This "\
		  "may cause the model output to produce a floating-point error.\n");
	} */
	return(ok);
		
}

/****************************** GetPotentialResponse **************************/

/*
 * This routine returns the action if the threshold is exceeded.
 *
 */

double
GetPotentialResponse_Neuron_ArleKim(double potential)
{
	if (!CheckPars_Neuron_ArleKim()) {
		NotifyError("GetPotentialResponse_Neuron_ArleKim: Parameters have "\
		  "not been correctly set, zero returned.");
		return(0.0);
	}
	if (potential > arleKim.restingThreshold_Th0 + arleKim.accomConst_c)
		potential -= arleKim.actionPotential;

	return(potential);

}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_Neuron_ArleKim(void)
{
	if (!CheckPars_Neuron_ArleKim()) {
		NotifyError("PrintPars_Neuron_ArleKim: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("ArleKim Neural Cell Module Parameters:-\n");
	printf("\tMembrane time constant, Tm = %g ms,\n",
	  MSEC(arleKim.membraneTConst_Tm));
	printf("\tPotassium decay time constant, TGk = %g ms,\n",
	  MSEC(arleKim.kDecayTConst_TGk));
	printf("\tThreshold rise time constant, TTh = %g ms,\n",
	  MSEC(arleKim.thresholdTConst_TTh));
	printf("\tAccommodation constant, c = %g,\n", arleKim.accomConst_c);
	printf("\tDelayed rectifier potassium conductance, b = %g nano Siemens.\n",
	  arleKim.delayedRectKCond_b);
	printf("\tResting threshold, Th0 = %g mV,\tAction potential = %g mV.\n",
	  arleKim.restingThreshold_Th0, arleKim.actionPotential);
	printf("\tNon-linear voltage constant, Vnl = %g mV,\n",
	  arleKim.nonLinearVConst_Vnl);
	printf("\tReversal potential of the potassium conductance, Ek = %g mV,\n",
	  arleKim.kReversalPoten_Ek);
	printf("\tReversal potential of all other conductances, Eb = %g mV,\n",
	  arleKim.bReversalPoten_Eb);
	printf("\tResting potassium conductance, gk = %g nS,\n",
	  arleKim.kRestingCond_gk);
	printf("\tAll other resting conductances, gb = %g nS,\n",
	  arleKim.bRestingCond_gb);

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_Neuron_ArleKim(char *fileName)
{
	BOOLEAN	ok;
	double	membraneTConst_Tm, kDecayTConst_TGk, thresholdTConst_TTh;
	double	accomConst_c, delayedRectKCond_b, restingThreshold_Th0;
	double	actionPotential, nonLinearVConst_Vnl, kReversalPoten_Ek;
	double	bReversalPoten_Eb, kRestingCond_gk, bRestingCond_gb;
    FILE    *fp;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_Neuron_ArleKim: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_Neuron_ArleKim: Reading from '%s'", fileName);

    Init_ParFile();
	ok = TRUE;
	if (!GetPars_ParFile(fp, "%lf", &membraneTConst_Tm))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &kDecayTConst_TGk))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &thresholdTConst_TTh))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &accomConst_c))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &delayedRectKCond_b))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &restingThreshold_Th0))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &actionPotential))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &nonLinearVConst_Vnl))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &kReversalPoten_Ek))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &bReversalPoten_Eb))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &kRestingCond_gk))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &bRestingCond_gb))
		ok = FALSE;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_Neuron_ArleKim: Not enough lines, or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_Neuron_ArleKim(membraneTConst_Tm, kDecayTConst_TGk,
	  thresholdTConst_TTh, accomConst_c, delayedRectKCond_b, 
	  restingThreshold_Th0, actionPotential, nonLinearVConst_Vnl,
	  kReversalPoten_Ek, bReversalPoten_Eb, kRestingCond_gk, bRestingCond_gb)) {
		NotifyError("ReadPars_Neuron_ArleKim: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/********************************* RunModel ***********************************/

/*
 * This routine runs the input signal through the model and puts the result
 * into the output signal.  It checks that all initialisation has been
 * correctly carried out by calling the appropriate checking routines.
 * potential_V = Cell-membrane pot. above resting level (mV);
 * kEquilibriumPot_Vk = Equlibrium potential of K conductance. (mV);
 * kConductance_Gk = Potassium conductance (ns);
 * threshold_Th = Time-varying cell threshold (mV);
 * spikeVariable_s = spike variable for cell.
 */

BOOLEAN
RunModel_Neuron_ArleKim(EarObjectPtr data)
{
	int			i, spikeState_s, lastSpikeState;
	double		dt, potential_V, kEquilibriumPot_Vk, kConductance_Gk;
	double		threshold_Th;
	double		bConductance_Gb, totalConductance_G, restingPotential_Er;
	double		bEquilibriumPot_Vb;
	register double	dGk, dTh, dV, tmOverDt, tGkOverDt, tThOverDt, bOverDt;
	ChanLen		j;
	ChanData	*inPtr, *outPtr;
	ArleKim	*c;
	
	if (data == NULL) {
		NotifyError("RunModel_Neuron_ArleKim: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_Neuron_ArleKim())		
		return(FALSE);
	if (!CheckRamp_SignalData(data->inSignal))
		return(FALSE);
	if (!InitOutSignal_EarObject(data, data->inSignal->numChannels,
	  data->inSignal->length, data->inSignal->dt)) {
		NotifyError("RunModel_Neuron_ArleKim: Could not initialise output "\
		  "signal.");
		return(FALSE);
	}
	SetProcessName_EarObject(data, "ArleKim neural cell");
	dt = data->outSignal->dt;
	c = &arleKim;
	
	potential_V = kConductance_Gk = bConductance_Gb = 0.0;
	totalConductance_G = c->kRestingCond_gk + c->bRestingCond_gb;
	restingPotential_Er = (c->kRestingCond_gk * c->kReversalPoten_Ek +
	  c->bRestingCond_gb * c->bReversalPoten_Eb) / totalConductance_G;
	kEquilibriumPot_Vk = c->kReversalPoten_Ek - restingPotential_Er;
	bEquilibriumPot_Vb = c->bReversalPoten_Eb - restingPotential_Er;
	
	threshold_Th = c->restingThreshold_Th0;
	tmOverDt  = c->membraneTConst_Tm / dt;
	tGkOverDt = c->kDecayTConst_TGk / dt;
	tThOverDt = c->thresholdTConst_TTh / dt;
	bOverDt = c->delayedRectKCond_b / dt;
	lastSpikeState = spikeState_s = 0;
	
	for (i = 0; i < data->outSignal->numChannels; i++) {
		inPtr = data->inSignal->channel[i];
		outPtr = data->outSignal->channel[i];
		for (j = 0; j < data->outSignal->length; j++, outPtr++) {
			bConductance_Gb = c->bRestingCond_gb * (exp(potential_V /
			  c->nonLinearVConst_Vnl) - 1.0);
			dV =  (-potential_V + (*(inPtr++) + kConductance_Gk *
			   (kEquilibriumPot_Vk - potential_V) + bConductance_Gb *
			   (bEquilibriumPot_Vb - potential_V)) / totalConductance_G) /
			   tmOverDt;
			dTh = (-(threshold_Th - c->restingThreshold_Th0) + c->accomConst_c *
			  potential_V ) / tThOverDt;	
			*outPtr = (ChanData) potential_V + restingPotential_Er;
			/* add increment to gk if at start of spike, otherwise allow decay*/
			if (lastSpikeState < spikeState_s) {
				dGk = (-kConductance_Gk + bOverDt) / tGkOverDt;
				*outPtr += (ChanData) c->actionPotential;
			} else
				dGk = -kConductance_Gk / tGkOverDt;
			threshold_Th += dTh;
			potential_V += dV;
			kConductance_Gk += dGk;
			lastSpikeState = spikeState_s;
			spikeState_s = (potential_V < threshold_Th)? 0: 1;
		}
	}
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);
		
} /* RunModel_Neuron_ArleKim */
