/*******************
 *
 * MoBasilarM.c
 *
 * This is an implementation of a Gamma Tone filter using the filters.c module.
 * The Gamma Tone filter is used to model the response of the Basilar Membrane.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 *******************/
 
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "UtFilters.h"
#include "UtCFList.h"
#include "MoBasilarM.h"
#include "FiParFile.h"

/**************************** Constant Definitions *************************************/

/**************************** Global subroutines ***************************************/

BasilarM	basilarM = {

					FALSE,		/* orderFlag */
					FALSE,		/* theCFsFlag */
					0,			/* order */
					NULL		/* theCFs */

			};

/***************************************************************************************/
/********************************* Subroutines and functions ***************************/
/***************************************************************************************/

/********************************* CheckPars *******************************************/

/*
 * This routine checks that the necessary parameters for the module have been
 * correctly initialised.
 * It returns TRUE if there are no problems.
 */
 
BOOLEAN
CheckPars_BasilarM_GammaT(void)
{
	if (!basilarM.orderFlag) {
		NotifyError("CheckPars_BasilarM_GammaT: Filter order not set.");
		return(FALSE);
	}
	if (!basilarM.theCFsFlag) {
		NotifyError("CheckPars_BasilarM_GammaT: Centre frequencies for the "\
		  "filter bank not set.");
		return(FALSE);
	} else {
		if (!CheckPars_CFList(basilarM.theCFs)) {
			NotifyError("CheckPars_BasilarM_GammaT: Centre frequency list "\
			  " parameters not correctly set.");
			return(FALSE);
		}
	}
	return(TRUE);

}

/********************************* SetOrder ********************************************/

/*
 * This function sets the module's order parameter.  It first checks that the module
 * has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetOrder_BasilarM_GammaT(int theOrder)
{
	basilarM.orderFlag = TRUE;
	basilarM.order = theOrder;
	return(TRUE);
	
}

/********************************* SetCFList *******************************************/

/*
 * This function sets the CFList data structure for the filter bank.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetCFList_BasilarM_GammaT(CFList *theCFList)
{
	if (!CheckPars_CFList(theCFList)) {
		NotifyError("SetCFList_BasilarM_GammaT: Centre frequency structure not "\
		  "initialised correctly set.");
		return(FALSE);
	}
	basilarM.theCFsFlag = TRUE;
	basilarM.theCFs = theCFList;
	return(TRUE);

}

/********************************* SetNumChannels **************************************/

/*
 * This function sets the number of channels for the filter bank.
 * It also initialises the CFList data structure, if it has not already been initialised.
 * It must be set before the number of centre frequencies is set.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetNumChannels_BasilarM_GammaT(int theChannels)
{
	if ((theChannels > MAX_CHANNELS) || (theChannels < 0)) {
		NotifyError("SetNumChannels_BasilarM_GammaT: %d is not in the range 0-%d.",
		  theChannels, MAX_CHANNELS);
		return(FALSE);
	}
	if (!basilarM.theCFsFlag) {
		if ((basilarM.theCFs = Init_CFList("SetNumChannels_BasilarM_GammaT")) == NULL) {
			NotifyError("SetNumChannels_BasilarM_GammaT: Could not initialise " \
			  "centre frequency list structure.");
			return(FALSE);
		}
		basilarM.theCFsFlag = TRUE;
	}
	basilarM.theCFs->numChannels = theChannels;
	return(TRUE);

}

/********************************* SetCentreFreq ***************************************/

/*
 * This function sets the array of centre frequencies for the filter bank.
 * It first checks that the module has first been initialised.
 * It also initialises the CFList data structure, if it has not already been initialised.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetCentreFreq_BasilarM_GammaT(double *theFrequencies, size_t arraySize)
{
	int		i, arrayDimension;
	
	if (!basilarM.theCFs->numChannels) {
		NotifyError("SetCentreFreq_BasilarM_GammaT: The number of channels "\
		  "must be set first! (Use SetNumChannels_BasilarM_GammaT.)");
		return(FALSE);
	}
	arrayDimension = arraySize / sizeof (double);
	if ( arrayDimension < basilarM.theCFs->numChannels) {
		NotifyError("SetCentreFreq_BasilarM_GammaT: Array dimension, %d, is less "\
		  "than the number of channels, %d, specified.", arrayDimension,
		  basilarM.theCFs->numChannels);
		return(FALSE);
	}
	/* The check on the CFList initialisation is not necessarily needed, as the
	 * setting of the number of channels will take care of this, and must, at present,
	 * be done first.  However, this has been included, just in case this policy is
	 * changed later and I forget about doing it.
	 */
	if (!basilarM.theCFsFlag) {
		if ((basilarM.theCFs = Init_CFList("SetCentreFreq_BasilarM_GammaT")) == NULL) {
			NotifyError("SetCentreFreq_BasilarM_GammaT: Could not initialise " \
			  "centre frequency list structure.");
			return(FALSE);
		}
		basilarM.theCFsFlag = TRUE;
	}
	for (i = 0; i < basilarM.theCFs->numChannels; i++) {
		if (theFrequencies[i] < 0.0) {
			NotifyError("SetCentreFreq_BasilarM_GammaT: Illegal frequency[%d] = %g Hz.",
			  i, theFrequencies[i]);
			return(FALSE);
		}
		basilarM.theCFs->channel[i] = theFrequencies[i];
	}
	return(TRUE);

}

/********************************* SetPars *********************************************/

/*
 * This function sets all the module's parameters.
 * It uses the variable argument format to accept the (i) order, numChannels, CF array,
 * CF array size, and; (ii) order, numChannels = -1, CFList data structure - formats.
 * Each format only reads the number of arguments required.
 * At present, the error testing for the correct argument formats is somewhat tenuous.
 * It returns TRUE if the operation is successful.
 */
  
BOOLEAN
SetPars_BasilarM_GammaT(int theOrder, int theChannels, ...)
{
	size_t	arraySize, arrayDimension;
	double	*theFrequencies;
	va_list	ap;
	BOOLEAN	ok;
	CFList	*theCFList;
	
	va_start(ap, theChannels);		/* Make ap point to the 1st unnamed arg. */
	ok = TRUE;
	if (!SetOrder_BasilarM_GammaT(theOrder))
		ok = FALSE;
	if (theChannels >= 0) {
		theFrequencies = va_arg(ap, double *);
		arraySize = va_arg(ap, size_t);
		arrayDimension = arraySize / sizeof(double);
		if (arrayDimension > MAX_CHANNELS) {
			NotifyError("SetPars_BasilarM_GammaT: Illegal use of function!  "\
			  "Check arguments.");
			ok = FALSE;
		} else {
			if (!SetNumChannels_BasilarM_GammaT(theChannels))
				ok = FALSE;
			if (!SetCentreFreq_BasilarM_GammaT(theFrequencies, arraySize))
				ok = FALSE;
		}
	} else {
		theCFList = (CFList *) va_arg(ap, CFList *);
		if (!SetCFList_BasilarM_GammaT(theCFList))
			ok = FALSE;
	}
	va_end(ap);		/* Clean up when done. */
	if (!ok)
		NotifyError("SetPars_BasilarM_GammaT: Failed to set all module " \
		  "parameters.");
	return(ok);
	
}

/****************************** GetCF *****************************************/

/*
 * This routine returns the respective centre frequency value for a module.
 *
 */

double
GetCF_BasilarM_GammaT(int channel)
{
	if (!CheckPars_BasilarM_GammaT()) {
		NotifyError("GetCF_BasilarM_GammaT: Parameters have not been "\
		  "correctly set, zero returned.");
		return(0.0);
	}
	if (channel < 0 || channel >= basilarM.theCFs->numChannels) {
		NotifyError("GetCF_BasilarM_GammaT: Channel number must be in the "\
		  "range 0 to %d, zero returned.", basilarM.theCFs->numChannels - 1);
		return(0.0);
	}
	return(basilarM.theCFs->channel[channel]);

}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_BasilarM_GammaT(void)
{
	int		i;
	
	if (!CheckPars_BasilarM_GammaT()) {
		NotifyError("PrintPars_BasilarM_GammaT: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("Gamma Tone Basilar Membrane Filter Module Parameters:-\n");
	printf("\t%10s\t%10s\n", "Filter No.", "Frequency");
	printf("\t%10s\t%10s\n", "          ", "   (Hz)  ");
	for (i = 0; i < basilarM.theCFs->numChannels; i++)
		printf("\t%10d\t%10g\n", i, basilarM.theCFs->channel[i]);
	printf("\tFilter order = %d.\n", basilarM.order);

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_BasilarM_GammaT(char *fileName)
{
	BOOLEAN	ok;
	char	theFrequencyMode[MAXLINE];
	int		filterOrder;
    FILE    *fp;
    CFListPtr	frequencies;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_BasilarM_GammaT: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_BasilarM_GammaT: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
	if (!GetPars_ParFile(fp, "%d", &filterOrder))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%s", theFrequencyMode))
		ok = FALSE;
	else
		ok = ((frequencies = Read_CFList(fp, theFrequencyMode)) != NULL);
    fclose(fp);
    Free_ParFile();
	if (!SetPars_BasilarM_GammaT(filterOrder, USE_CFLIST, frequencies)) {
		NotifyError("ReadPars_BasilarM_GammaT: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
   
}

/**************************** Simulate_Model *******************************************/

/*
 * This routine filters the input signal and puts the result into the output signal.
 * It checks that all initialisation has been correctly carried out by calling the
 * appropriate checking routines.
 * The appropriate coefficients are calculated at the beginning, then the memory used is
 * released before the program returns.
 */

BOOLEAN
RunModel_BasilarM_GammaT(EarObjectPtr data)
{
	int		i, totalChannels;
	double	sampleRate;
    GammaToneCoeffs	*coefficients[MAX_CHANNELS];
				
	if (data == NULL) {
		NotifyError("RunModel_BasilarM_GammaT: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_BasilarM_GammaT())
		return(FALSE);
		
	/* Initialise Variables and coefficients */
	
	SetProcessName_EarObject(data, "Gamma tone basilar membrane filtering");
	if (!CheckRamp_SignalData(data->inSignal)) {
		NotifyError("RunModel_BasilarM_GammaT: Input signal not correctly " \
		  "initialised.");
		return(FALSE);
	}
	totalChannels = basilarM.theCFs->numChannels * data->inSignal->numChannels;
	InitOutFromInSignal_EarObject(data, totalChannels);
	sampleRate = 1.0 / data->inSignal->dt;
	for (i = 0; i < data->outSignal->numChannels; i++)
		coefficients[i] = InitCoeffs_Filters_GammaTone(
		  basilarM.theCFs->channel[i / data->inSignal->numChannels],
		  sampleRate, basilarM.order);

	/* Filter signal */
	
	GammaTone_Filters(data->outSignal, coefficients);
	for (i = 0; i < data->outSignal->numChannels; i++)
    	free(coefficients[i]);
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

}
