/*******************
 *
 * MoDendrite.c
 *
 * This is an implementation of a 2nd order Butterworth low-pass filter.
 * The filter is used to approximate the effects of dendrite filtering.
 * It uses the filter IIR filter from the UtFilter module.
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 *******************/
 
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "UtFilters.h"
#include "MoDendrite.h"
#include "FiParFile.h"

/******************************************************************************/
/**************************** Global subroutines ******************************/
/******************************************************************************/

Dendrite	dendrite = {

					FALSE,		/* cutOffFrequencyFlag */
					FALSE,		/* signalMultiplierFlag */
					0.0,		/* cutOffFrequency */
					0.0			/* signalMultiplier */
					
			};

/******************************************************************************/
/**************************** Subroutines and functions ***********************/
/******************************************************************************/

/********************************* SetCutOffFrequency *************************/

/*
 * This function sets the module's lower, 3dB down cut off frequency parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetCutOffFrequency_Neuron_DendriteLP(double theCutOffFrequency)
{
	dendrite.cutOffFrequencyFlag = TRUE;
	dendrite.cutOffFrequency = theCutOffFrequency;
	return(TRUE);

}

/********************************* SetSignalMultiplier ************************/

/*
 * This function sets the module's signal multiplying factor.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetSignalMultiplier_Neuron_DendriteLP(double theSignalMultiplier)
{
	dendrite.signalMultiplierFlag = TRUE;
	dendrite.signalMultiplier = theSignalMultiplier;
	return(TRUE);

}

/********************************* CheckPars **********************************/

/*
 * This routine checks that the necessary parameters for the module have been
 * correctly initialised.
 * It returns TRUE if there are no problems.
 */
 
BOOLEAN
CheckPars_Neuron_DendriteLP(void)
{
	if (!dendrite.cutOffFrequencyFlag) {
		NotifyError("CheckPars_Neuron_DendriteLP: Cut-off frequency "\
		  "not set.");
		return(FALSE);
	}
	if (!dendrite.signalMultiplierFlag) {
		NotifyError("CheckPars_Neuron_DendriteLP: Signal multiplying factor "\
		  "not set.");
		return(FALSE);
	}
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetPars_Neuron_DendriteLP(double theLowerCutOffFreq, double theUpperCutOffFreq)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetCutOffFrequency_Neuron_DendriteLP(theLowerCutOffFreq))
		ok = FALSE;
	if (!SetSignalMultiplier_Neuron_DendriteLP(theUpperCutOffFreq))
		ok = FALSE;
	if (!ok)
		NotifyError("SetPars_Neuron_DendriteLP: Failed to set all module "\
		  "parameters.");
	return(ok);
	
}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_Neuron_DendriteLP(void)
{
	if (!CheckPars_Neuron_DendriteLP()) {
		NotifyError("PrintPars_Neuron_DendriteLP: Parameters have not "\
		  "been correctly set.");
		return;
	}
	printf("Dendrite Low-Pass Filter Module Parameters:-\n");
	printf("\tCut-off Frequency = %g Hz,\tSignal multiplying factor = %g.\n",
	  dendrite.cutOffFrequency, dendrite.signalMultiplier);

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_Neuron_DendriteLP(char *fileName)
{
	BOOLEAN	ok;
	double	cutOffFrequency, signalMultiplier;
    FILE    *fp;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_Neuron_DendriteLP: Cannot open data file "\
          "'%s'.\n", fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_Neuron_DendriteLP: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
	if (!GetPars_ParFile(fp, "%lf", &cutOffFrequency))
		ok = FALSE;;
	if (!GetPars_ParFile(fp, "%lf", &signalMultiplier))
		ok = FALSE;;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_Neuron_DendriteLP: Not enough lines, or "\
		  "invalid parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_Neuron_DendriteLP(cutOffFrequency, signalMultiplier)) {
		NotifyError("ReadPars_Neuron_DendriteLP: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/**************************** RunModel ****************************************/

/*
 * This routine filters the input signal and puts the result into the output
 * signal.
 * It checks that all initialisation has been correctly carried out by calling
 * the appropriate checking routines.
 * The appropriate coefficients are calculated at the beginning, then the 
 * memory used is released before the program returns.
 */

BOOLEAN
RunModel_Neuron_DendriteLP(EarObjectPtr data)
{
	ContButt1CoeffsPtr	coefficients;
	
	if (data == NULL) {
		NotifyError("RunModel_Neuron_DendriteLP: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_Neuron_DendriteLP())
		return(FALSE);
	SetProcessName_EarObject(data, "Dendritic low-pass filter");
	if (!CheckRamp_SignalData(data->inSignal)) {
		NotifyError("RunModel_Neuron_DendriteLP: Input signal not " \
		  "correctly initialised.");
		return(FALSE);
	}
	if (!InitOutFromInSignal_EarObject(data, 0)) {
		NotifyError("RunModel_Neuron_DendriteLP: Couldn't initialse output "\
		  "signal.");
		return(FALSE);
	}
	Scale_SignalData(data->outSignal, dendrite.signalMultiplier);
	coefficients = InitCoeffs_Filters_IIR1Cont(dendrite.cutOffFrequency,
	  data->inSignal->dt, LOWPASS);

	/* Filter signal */
	
	IIR1Cont_Filters_Single(data->outSignal, coefficients);
	free(coefficients);
	
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

}
