/******************
 *
 * MoMcGregor.c
 *
 * This module implements  the MacGregor Model.
 * It was revised from code written by Trevor Shackleton and M. J. Hewitt.
 *
 ******************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "FiParFile.h"
#include "MoMcGregor.h"

/******************************************************************************/
/*************************** Global Variables *********************************/
/******************************************************************************/

McGregor		mcGregor = {

					FALSE,		/* membraneTConstFlag */
					FALSE,		/* thresholdTConstFlag */
					FALSE,		/* accomConstFlag */
					FALSE,		/* delayedRectKCondFlag */
					FALSE,		/* restingThresholdFlag */
					FALSE,		/* actionPotentialFlag */
					FALSE,		/* kDecayTConstFlag */
					FALSE,		/* kEquilibriumPotFlag */
					0.0,		/* membraneTConst_Tm */
					0.0,		/* kDecayTConst_TGk */
					0.0,		/* thresholdTConst_TTh */
					0.0,		/* accomConst_c */
					0.0,		/* delayedRectKCond_b */
					0.0,		/* restingThreshold_Th0 */
					0.0,		/* actionPotential */
					0.0,		/* kEquilibriumPot_Ek */
					0.0,		/* cellRestingPot_Er */
					
				};

/******************************************************************************/
/*************************** Subroutines and Functions ************************/
/******************************************************************************/

/********************************* SetMembraneTConst **************************/

/*
 * This function sets the module's membrane time constant, Tm, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetMembraneTConst_Neuron_McGregor(double theMembraneTConst)
{
	if (theMembraneTConst == 0.0) {
		NotifyError("SetMembraneTConst_Neuron_McGregor: Invalid value = "\
		  "%g mV\n", theMembraneTConst);
		return(FALSE);
	} 
	mcGregor.membraneTConstFlag = TRUE;
	mcGregor.membraneTConst_Tm = theMembraneTConst;
	return(TRUE);

}

/********************************* SetKDecayTConst ****************************/

/*
 * This function sets the module's potassium conductance decay time constant,
 * TGk, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetKDecayTConst_Neuron_McGregor(double theKDecayTConst)
{
	if (theKDecayTConst == 0.0) {
		NotifyError("SetKDecayTConst_Neuron_McGregor: Invalid value = "\
		  "%g mV\n", theKDecayTConst);
		return(FALSE);
	} 
	mcGregor.kDecayTConstFlag = TRUE;
	mcGregor.kDecayTConst_TGk = theKDecayTConst;
	return(TRUE);

}

/********************************* SetThresholdTConst *************************/

/*
 * This function sets the module's threshold time constant, TTh, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetThresholdTConst_Neuron_McGregor(double theThresholdTConst)
{
	if (theThresholdTConst == 0.0) {
		NotifyError("SetThresholdTConst_Neuron_McGregor: Invalid value = "\
		  "%g mV\n", theThresholdTConst);
		return(FALSE);
	} 
	mcGregor.thresholdTConstFlag = TRUE;
	mcGregor.thresholdTConst_TTh = theThresholdTConst;
	return(TRUE);

}

/********************************* SetAccomConst ******************************/

/*
 * This function sets the module's accomodation constant, c, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetAccomConst_Neuron_McGregor(double theAccomConst)
{
	mcGregor.accomConstFlag = TRUE;
	mcGregor.accomConst_c = theAccomConst;
	return(TRUE);

}

/********************************* SetDelayedRectKCond ************************/

/*
 * This function sets the module's delayed rectifier potassium conductance
 * strength, b, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetDelayedRectKCond_Neuron_McGregor(double theDelayedRectKCond)
{
	mcGregor.delayedRectKCondFlag = TRUE;
	mcGregor.delayedRectKCond_b = theDelayedRectKCond;
	return(TRUE);

}

/********************************* SetRestingThreshold ************************/

/*
 * This function sets the module's resting threhold, Th0, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetRestingThreshold_Neuron_McGregor(double theRestingThreshold)
{
	mcGregor.restingThresholdFlag = TRUE;
	mcGregor.restingThreshold_Th0 = theRestingThreshold;
	return(TRUE);

}

/********************************* SetActionPotential *************************/

/*
 * This function sets the module's action potential parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetActionPotential_Neuron_McGregor(double theActionPotential)
{
	mcGregor.actionPotentialFlag = TRUE;
	mcGregor.actionPotential = theActionPotential;
	return(TRUE);

}

/********************************* SetKEquilibriumPot *************************/

/*
 * This function sets the module's reversal potential of the K conductance
 * parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetKEquilibriumPot_Neuron_McGregor(double theKEquilibriumPot)
{
	mcGregor.kEquilibriumPotFlag = TRUE;
	mcGregor.kEquilibriumPot_Ek = theKEquilibriumPot;
	return(TRUE);

}

/********************************* SetCellRestingPot **************************/

/*
 * This function sets the module's cell resting potential parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetCellRestingPot_Neuron_McGregor(double theCellRestingPot)
{
	mcGregor.cellRestingPotFlag = TRUE;
	mcGregor.cellRestingPot_Er = theCellRestingPot;
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 */

BOOLEAN
SetPars_Neuron_McGregor(double tm, double tGk, double tTh, double c, double b,
  double tH0, double aP, double ek, double er)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetMembraneTConst_Neuron_McGregor(tm))
		ok = FALSE;
	if (!SetKDecayTConst_Neuron_McGregor(tGk))
		ok = FALSE;
	if (!SetThresholdTConst_Neuron_McGregor(tTh))
		ok = FALSE;
	if (!SetAccomConst_Neuron_McGregor(c))
	    ok = FALSE;
	if (!SetDelayedRectKCond_Neuron_McGregor(b))
		ok = FALSE;
	if (!SetRestingThreshold_Neuron_McGregor(tH0))
		ok = FALSE;
	if (!SetActionPotential_Neuron_McGregor(aP))
		ok = FALSE;
	if (!SetKEquilibriumPot_Neuron_McGregor(ek))
		ok = FALSE;
	if (!SetCellRestingPot_Neuron_McGregor(er))
		ok = FALSE;
	return(ok);
	  
}

/********************************* CheckPars **********************************/

/*
 * This routine checks that all of the parameters for the module are set.
 * Because this module is defined by its coeffients, default values are always
 * set if none of the parameters are set, otherwise all of them must be
 * individually set.
 */

BOOLEAN
CheckPars_Neuron_McGregor(void)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!mcGregor.membraneTConstFlag) {
		NotifyError("CheckPars_Neuron_McGregor: Membrane time constant, Tm, "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor.kDecayTConstFlag) {
		NotifyError("CheckPars_Neuron_McGregor: Potassium decay time  "\
		  "constant, TGk, not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor.thresholdTConstFlag) {
		NotifyError("CheckPars_Neuron_McGregor: Threshold time constant, TTh, "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor.accomConstFlag) {
		NotifyError("CheckPars_Neuron_McGregor: Accomodation constant, c, "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor.delayedRectKCondFlag) {
		NotifyError("CheckPars_Neuron_McGregor: Delayed rectifier potassium "\
		  "conductance, b, not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor.restingThresholdFlag) {
		NotifyError("CheckPars_Neuron_McGregor: Resting threshold constant, "\
		  "Th0, not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor.actionPotentialFlag) {
		NotifyError("CheckPars_Neuron_McGregor: Action potential constant "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor.kEquilibriumPotFlag) {
		NotifyError("CheckPars_Neuron_McGregor: Equilibrium potential of the "\
		  "potassium conductance, Ek, not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor.cellRestingPotFlag) {
		NotifyError("CheckPars_Neuron_McGregor: Cell resting potential, "\
		  "Er, not correctly set.");
		ok = FALSE;
	}
	return(ok);
		
}

/****************************** GetPotentialResponse **************************/

/*
 * This routine returns the action if the threshold is exceeded.
 *
 */

double
GetPotentialResponse_Neuron_McGregor(double potential)
{
	if (!CheckPars_Neuron_McGregor()) {
		NotifyError("GetPotentialResponse_Neuron_McGregor: Parameters have "\
		  "not been correctly set, zero returned.");
		return(0.0);
	}
	if (potential > mcGregor.restingThreshold_Th0 + mcGregor.accomConst_c)
		potential -= mcGregor.actionPotential;

	return(potential);

}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_Neuron_McGregor(void)
{
	if (!CheckPars_Neuron_McGregor()) {
		NotifyError("PrintPars_Neuron_McGregor: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("McGregor Neural Cell Module Parameters:-\n");
	printf("\tMembrane time constant, Tm = %g ms,\n",
	  MSEC(mcGregor.membraneTConst_Tm));
	printf("\tPotassium decay time constant, TGk = %g ms,\n",
	  MSEC(mcGregor.kDecayTConst_TGk));
	printf("\tThreshold rise time constant, TTh = %g ms,\n",
	  MSEC(mcGregor.thresholdTConst_TTh));
	printf("\tAccommodation constant, c = %g,\n", mcGregor.accomConst_c);
	printf("\tDelayed rectifier potassium conductance, b = %g nano Siemens.\n",
	  mcGregor.delayedRectKCond_b);
	printf("\tResting threshold, Th0 = %g mV,\tAction potential = %g mV.\n",
	  mcGregor.restingThreshold_Th0, mcGregor.actionPotential);
	printf("\tEquilibrium potential of the potassium conductance, Ek = "\
	  "%g mV,\n", mcGregor.kEquilibriumPot_Ek);
	printf("\tCell resting potential, Er = %g mV,\n",
	  mcGregor.cellRestingPot_Er);

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_Neuron_McGregor(char *fileName)
{
	BOOLEAN	ok;
	double	membraneTConst_Tm, kDecayTConst_TGk, thresholdTConst_TTh;
	double	accomConst_c, delayedRectKCond_b, restingThreshold_Th0;
	double	actionPotential, kEquilibriumPot_Ek, cellRestingPot_Er;
	FILE    *fp;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_Neuron_McGregor: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_Neuron_McGregor: Reading from '%s'", fileName);

    Init_ParFile();
	ok = TRUE;
	if (!GetPars_ParFile(fp, "%lf", &membraneTConst_Tm))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &kDecayTConst_TGk))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &thresholdTConst_TTh))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &accomConst_c))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &delayedRectKCond_b))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &restingThreshold_Th0))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &actionPotential))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &kEquilibriumPot_Ek))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &cellRestingPot_Er))
		ok = FALSE;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_Neuron_McGregor: Not enough lines, or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_Neuron_McGregor(membraneTConst_Tm, kDecayTConst_TGk,
	  thresholdTConst_TTh, accomConst_c, delayedRectKCond_b, 
	  restingThreshold_Th0, actionPotential, kEquilibriumPot_Ek,
	  cellRestingPot_Er)) {
		NotifyError("ReadPars_Neuron_McGregor: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/********************************* RunModel ***********************************/

/*
 * This routine runs the input signal through the model and puts the result
 * into the output signal.  It checks that all initialisation has been
 * correctly carried out by calling the appropriate checking routines.
 * potential_V = Cell-membrane pot. above resting level (mV);
 * kEquilibriumPot_Vk = Equlibrium potential of K conductance. (mV);
 * kConductance_Gk = Potassium conductance (ns);
 * threshold_Th = Time-varying cell threshold (mV);
 * spikeVariable_s = spike variable for cell.
 */

BOOLEAN
RunModel_Neuron_McGregor(EarObjectPtr data)
{
	int			i, spikeState_s, lastSpikeState;
	double		dt, potential_E, kConductance_Gk, threshold_Th;
	register double	dGk, dTh, dE, tmOverDt, tGkOverDt, tThOverDt, bOverDt;
	ChanLen		j;
	ChanData	*inPtr, *outPtr;
	McGregor	*c;
	
	if (data == NULL) {
		NotifyError("RunModel_Neuron_McGregor: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_Neuron_McGregor())		
		return(FALSE);
	if (!CheckRamp_SignalData(data->inSignal))
		return(FALSE);
	if (!InitOutSignal_EarObject(data, data->inSignal->numChannels,
	  data->inSignal->length, data->inSignal->dt)) {
		NotifyError("RunModel_Neuron_McGregor: Could not initialise output "\
		  "signal.");
		return(FALSE);
	}
	SetProcessName_EarObject(data, "McGregor neural cell");
	dt = data->outSignal->dt;
	c = &mcGregor;
	
	potential_E = kConductance_Gk = 0.0;
	threshold_Th = c->restingThreshold_Th0;
	tmOverDt  = c->membraneTConst_Tm / dt;
	tGkOverDt = c->kDecayTConst_TGk / dt;
	tThOverDt = c->thresholdTConst_TTh / dt;
	bOverDt = c->delayedRectKCond_b / dt;
	lastSpikeState = spikeState_s = 0;

	for (i = 0; i < data->outSignal->numChannels; i++) {
		inPtr = data->inSignal->channel[i];
		outPtr = data->outSignal->channel[i];
		for (j = 0; j < data->outSignal->length; j++, outPtr++) {
			dE = (-potential_E + *(inPtr++) + kConductance_Gk *
			  (c->kEquilibriumPot_Ek - potential_E)) / tmOverDt;
			  
			dTh = (-(threshold_Th - c->restingThreshold_Th0) +
			  c->accomConst_c * potential_E ) / tThOverDt;
			*outPtr = potential_E + c->cellRestingPot_Er;
			lastSpikeState = spikeState_s;
			spikeState_s = (potential_E < threshold_Th)? 0: 1;
			if (spikeState_s) {
				dGk = (-kConductance_Gk + bOverDt) / tGkOverDt;
				if (lastSpikeState < spikeState_s)
					*outPtr += c->actionPotential;
			} else
				dGk = -kConductance_Gk / tGkOverDt;			
			threshold_Th += dTh;
			potential_E += dE;
			kConductance_Gk += dGk;
		}
	}
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);
		
} /* RunModel_Neuron_McGregor */
