/******************
 *
 * MoMcGregor2.c
 *
 * This module implements M.J.Hewitt's version of the MacGregor Model.
 * The changes implemented in "Neural Modeling of instrinsic and spike-discharge
 * properties of cochlear nucleus neurons", J. E. Erle and D. O. Kim,
 * Biological Cybernetics, Springer Verlag 1991. - have been added to this
 * module version of the McGregor model.
 *
 * It was revised from code written by Trevor Shackleton and M. J. Hewitt.
 *
 ******************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "FiParFile.h"
#include "MoMcGregor2.h"

/******************************************************************************/
/*************************** Global Variables *********************************/
/******************************************************************************/

McGregor2		mcGregor2 = {

					FALSE,		/* membraneTConstFlag */
					FALSE,		/* thresholdTConstFlag */
					FALSE,		/* accomConstFlag */
					FALSE,		/* delayedRectKCondFlag */
					FALSE,		/* restingThresholdFlag */
					FALSE,		/* actionPotentialFlag */
					FALSE,		/* kDecayTConstFlag */
					FALSE,		/* nonLinearVConstFlag */
					FALSE,		/* kReversalPotenFlag */
					FALSE,		/* bReversalPotenFlag */
					FALSE,		/* kRestingCondFlag */
					FALSE,		/* bRestingCondFlag */
					0.0,		/* membraneTConst_Tm */
					0.0,		/* kDecayTConst_TGk */
					0.0,		/* thresholdTConst_TTh */
					0.0,		/* accomConst_c */
					0.0,		/* delayedRectKCond_b */
					0.0,		/* restingThreshold_Th0 */
					0.0,		/* actionPotential */
					0.0,		/* nonLinearVConst_Vnl */
					0.0,		/* kReversalPoten_Ek0 */
					0.0,		/* bReversalPoten_Eb0 */
					0.0,		/* kRestingCond_gk */
					0.0			/* bRestingCond_gb */
					
				};

/******************************************************************************/
/*************************** Subroutines and Functions ************************/
/******************************************************************************/

/********************************* SetMembraneTConst **************************/

/*
 * This function sets the module's membrane time constant, Tm, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetMembraneTConst_Neuron_McGregor2(double theMembraneTConst)
{
	if (theMembraneTConst == 0.0) {
		NotifyError("SetMembraneTConst_Neuron_McGregor2: Invalid value = "\
		  "%g mV\n", theMembraneTConst);
		return(FALSE);
	} 
	mcGregor2.membraneTConstFlag = TRUE;
	mcGregor2.membraneTConst_Tm = theMembraneTConst;
	return(TRUE);

}

/********************************* SetKDecayTConst ****************************/

/*
 * This function sets the module's potassium conductance decay time constant,
 * TGk, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetKDecayTConst_Neuron_McGregor2(double theKDecayTConst)
{
	if (theKDecayTConst == 0.0) {
		NotifyError("SetKDecayTConst_Neuron_McGregor2: Invalid value = "\
		  "%g mV\n", theKDecayTConst);
		return(FALSE);
	} 
	mcGregor2.kDecayTConstFlag = TRUE;
	mcGregor2.kDecayTConst_TGk = theKDecayTConst;
	return(TRUE);

}

/********************************* SetThresholdTConst *************************/

/*
 * This function sets the module's threshold time constant, TTh, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetThresholdTConst_Neuron_McGregor2(double theThresholdTConst)
{
	if (theThresholdTConst == 0.0) {
		NotifyError("SetThresholdTConst_Neuron_McGregor2: Invalid value = "\
		  "%g mV\n", theThresholdTConst);
		return(FALSE);
	} 
	mcGregor2.thresholdTConstFlag = TRUE;
	mcGregor2.thresholdTConst_TTh = theThresholdTConst;
	return(TRUE);

}

/********************************* SetAccomConst ******************************/

/*
 * This function sets the module's accomodation constant, c, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetAccomConst_Neuron_McGregor2(double theAccomConst)
{
	mcGregor2.accomConstFlag = TRUE;
	mcGregor2.accomConst_c = theAccomConst;
	return(TRUE);

}

/********************************* SetDelayedRectKCond ************************/

/*
 * This function sets the module's delayed rectifier potassium conductance
 * strength, b, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetDelayedRectKCond_Neuron_McGregor2(double theDelayedRectKCond)
{
	mcGregor2.delayedRectKCondFlag = TRUE;
	mcGregor2.delayedRectKCond_b = theDelayedRectKCond;
	return(TRUE);

}

/********************************* SetRestingThreshold ************************/

/*
 * This function sets the module's resting threhold, Th0, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetRestingThreshold_Neuron_McGregor2(double theRestingThreshold)
{
	mcGregor2.restingThresholdFlag = TRUE;
	mcGregor2.restingThreshold_Th0 = theRestingThreshold;
	return(TRUE);

}

/********************************* SetActionPotential *************************/

/*
 * This function sets the module's action potential parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetActionPotential_Neuron_McGregor2(double theActionPotential)
{
	mcGregor2.actionPotentialFlag = TRUE;
	mcGregor2.actionPotential = theActionPotential;
	return(TRUE);

}

/********************************* SetNonLinearVConst *************************/

/*
 * This function sets the module's non-linear voltage constant, Vnl, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetNonLinearVConst_Neuron_McGregor2(double theNonLinearVConst)
{
	if (theNonLinearVConst == 0.0) {
		NotifyError("SetNonLinearVConst_Neuron_McGregor2: Invalid value = "\
		  "%g mV\n", theNonLinearVConst);
		NotifyWarning("SetNonLinearVConst_Neuron_McGregor2: To set the term "\
		  "involving this parameter to zero set gb to zero.");
		return(FALSE);
	} 
	mcGregor2.nonLinearVConstFlag = TRUE;
	mcGregor2.nonLinearVConst_Vnl = theNonLinearVConst;
	return(TRUE);

}

/********************************* SetKReversalPoten **************************/

/*
 * This function sets the module's reversal potential of the K conductance
 * parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetKReversalPoten_Neuron_McGregor2(double theKReversalPoten)
{
	mcGregor2.kReversalPotenFlag = TRUE;
	mcGregor2.kReversalPoten_Ek0 = theKReversalPoten;
	return(TRUE);

}

/********************************* SetBReversalPoten **************************/

/*
 * This function sets the module's reversal potential of all other conductances
 * parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetBReversalPoten_Neuron_McGregor2(double theBReversalPoten)
{
	mcGregor2.bReversalPotenFlag = TRUE;
	mcGregor2.bReversalPoten_Eb0 = theBReversalPoten;
	return(TRUE);

}

/********************************* SetKRestingCond ****************************/

/*
 * This function sets the module's resting potassium conductance parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetKRestingCond_Neuron_McGregor2(double theKRestingCond)
{
	mcGregor2.kRestingCondFlag = TRUE;
	mcGregor2.kRestingCond_gk = theKRestingCond;
	return(TRUE);

}

/********************************* SetBRestingCond ****************************/

/*
 * This function sets the module's resting component of all other conductances
 * parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetBRestingCond_Neuron_McGregor2(double theBRestingCond)
{
	mcGregor2.bRestingCondFlag = TRUE;
	mcGregor2.bRestingCond_gb = theBRestingCond;
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 */

BOOLEAN
SetPars_Neuron_McGregor2(double tm, double tGk, double tTh, double c, double b,
  double tH0, double aP, double vnl, double ek0, double eb0, double gk,
  double gb)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetMembraneTConst_Neuron_McGregor2(tm))
		ok = FALSE;
	if (!SetKDecayTConst_Neuron_McGregor2(tGk))
		ok = FALSE;
	if (!SetThresholdTConst_Neuron_McGregor2(tTh))
		ok = FALSE;
	if (!SetAccomConst_Neuron_McGregor2(c))
	    ok = FALSE;
	if (!SetDelayedRectKCond_Neuron_McGregor2(b))
		ok = FALSE;
	if (!SetRestingThreshold_Neuron_McGregor2(tH0))
		ok = FALSE;
	if (!SetActionPotential_Neuron_McGregor2(aP))
		ok = FALSE;
	if (!SetNonLinearVConst_Neuron_McGregor2(vnl))
		ok = FALSE;
	if (!SetKReversalPoten_Neuron_McGregor2(ek0))
		ok = FALSE;
	if (!SetBReversalPoten_Neuron_McGregor2(eb0))
		ok = FALSE;
	if (!SetKRestingCond_Neuron_McGregor2(gk))
		ok = FALSE;
	if (!SetBRestingCond_Neuron_McGregor2(gb))
		ok = FALSE;
	return(ok);
	  
}

/********************************* CheckPars **********************************/

/*
 * This routine checks that all of the parameters for the module are set.
 * Because this module is defined by its coeffients, default values are always
 * set if none of the parameters are set, otherwise all of them must be
 * individually set.
 */

BOOLEAN
CheckPars_Neuron_McGregor2(void)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!mcGregor2.membraneTConstFlag) {
		NotifyError("CheckPars_Neuron_McGregor2: Membrane time constant, Tm, "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor2.kDecayTConstFlag) {
		NotifyError("CheckPars_Neuron_McGregor2: Potassium decay time  "\
		  "constant, TGk, not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor2.thresholdTConstFlag) {
		NotifyError("CheckPars_Neuron_McGregor2: Threshold time constant, TTh, "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor2.accomConstFlag) {
		NotifyError("CheckPars_Neuron_McGregor2: Accomodation constant, c, "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor2.delayedRectKCondFlag) {
		NotifyError("CheckPars_Neuron_McGregor2: Delayed rectifier potassium "\
		  "conductance, b, not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor2.restingThresholdFlag) {
		NotifyError("CheckPars_Neuron_McGregor2: Resting threshold constant, "\
		  "Th0, not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor2.actionPotentialFlag) {
		NotifyError("CheckPars_Neuron_McGregor2: Action potential constant "\
		  "not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor2.nonLinearVConstFlag) {
		NotifyError("CheckPars_Neuron_McGregor2: Non-linear voltage constant, "\
		  "Vnl, not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor2.kReversalPotenFlag) {
		NotifyError("CheckPars_Neuron_McGregor2: Reversal potential of the "\
		  "potassium conductance, Ek0, not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor2.bReversalPotenFlag) {
		NotifyError("CheckPars_Neuron_McGregor2: Reversal potential of all "\
		  "other conductances, Eb0, not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor2.kRestingCondFlag) {
		NotifyError("CheckPars_Neuron_McGregor2: Resting potassium "\
		  "conductance, gk, not correctly set.");
		ok = FALSE;
	}
	if (!mcGregor2.bRestingCondFlag) {
		NotifyError("CheckPars_Neuron_McGregor2: Resting component "\
		  "for all other conductances, gb,  not correctly set.");
		ok = FALSE;
	}
	/*
	This needs some work.
	dEExpression = (mcGregor2.kRestingCond_gk +
	  mcGregor2.bRestingCond_gb) / mcGregor2.membraneTConst_Tm;
	if (dEExpression < MCGREGOR2_DE_LIMIT) {
		NotifyError("CheckPars_Neuron_McGregor2:  The total conductance, gk + "\
		  "gb, is too high or the membrane time-constant is too low.\n  This "\
		  "may cause the model output to produce a floating-point error.\n");
	} */
	return(ok);
		
}

/****************************** GetPotentialResponse **************************/

/*
 * This routine returns the action if the threshold is exceeded.
 *
 */

double
GetPotentialResponse_Neuron_McGregor2(double potential)
{
	if (!CheckPars_Neuron_McGregor2()) {
		NotifyError("GetPotentialResponse_Neuron_McGregor2: Parameters have "\
		  "not been correctly set, zero returned.");
		return(0.0);
	}
	if (potential > mcGregor2.restingThreshold_Th0 + mcGregor2.accomConst_c)
		potential -= mcGregor2.actionPotential;

	return(potential);

}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_Neuron_McGregor2(void)
{
	if (!CheckPars_Neuron_McGregor2()) {
		NotifyError("PrintPars_Neuron_McGregor2: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("McGregor2 Neural Cell Module Parameters:-\n");
	printf("\tMembrane time constant, Tm = %g ms,\n",
	  MSEC(mcGregor2.membraneTConst_Tm));
	printf("\tPotassium decay time constant, TGk = %g ms,\n",
	  MSEC(mcGregor2.kDecayTConst_TGk));
	printf("\tThreshold rise time constant, TTh = %g ms,\n",
	  MSEC(mcGregor2.thresholdTConst_TTh));
	printf("\tAccommodation constant, c = %g,\n", mcGregor2.accomConst_c);
	printf("\tDelayed rectifier potassium conductance, b = %g nano Siemens.\n",
	  mcGregor2.delayedRectKCond_b);
	printf("\tResting threshold, Th0 = %g mV,\tAction potential = %g mV.\n",
	  mcGregor2.restingThreshold_Th0, mcGregor2.actionPotential);
	printf("\tNon-linear voltage constant, Vnl = %g mV,\n",
	  mcGregor2.nonLinearVConst_Vnl);
	printf("\tReversal potential of the potassium conductance, Ek0 = %g mV,\n",
	  mcGregor2.kReversalPoten_Ek0);
	printf("\tReversal potential of all other conductances, Eb0 = %g mV,\n",
	  mcGregor2.bReversalPoten_Eb0);
	printf("\tResting potassium conductance, gk = %g ns,\n",
	  mcGregor2.kRestingCond_gk);
	printf("\tResting potassium conductance, gb = %g ns,\n",
	  mcGregor2.bRestingCond_gb);

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_Neuron_McGregor2(char *fileName)
{
	BOOLEAN	ok;
	double	membraneTConst_Tm, kDecayTConst_TGk, thresholdTConst_TTh;
	double	accomConst_c, delayedRectKCond_b, restingThreshold_Th0;
	double	actionPotential, nonLinearVConst_Vnl, kReversalPoten_Ek0;
	double	bReversalPoten_Eb0, kRestingCond_gk, bRestingCond_gb;
    FILE    *fp;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_Neuron_McGregor2: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_Neuron_McGregor2: Reading from '%s'", fileName);

    Init_ParFile();
	ok = TRUE;
	if (!GetPars_ParFile(fp, "%lf", &membraneTConst_Tm))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &kDecayTConst_TGk))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &thresholdTConst_TTh))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &accomConst_c))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &delayedRectKCond_b))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &restingThreshold_Th0))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &actionPotential))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &nonLinearVConst_Vnl))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &kReversalPoten_Ek0))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &bReversalPoten_Eb0))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &kRestingCond_gk))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &bRestingCond_gb))
		ok = FALSE;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_Neuron_McGregor2: Not enough lines, or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_Neuron_McGregor2(membraneTConst_Tm, kDecayTConst_TGk,
	  thresholdTConst_TTh, accomConst_c, delayedRectKCond_b, 
	  restingThreshold_Th0, actionPotential, nonLinearVConst_Vnl,
	  kReversalPoten_Ek0, bReversalPoten_Eb0, kRestingCond_gk,
	  bRestingCond_gb)) {
		NotifyError("ReadPars_Neuron_McGregor2: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/********************************* RunModel ***********************************/

/*
 * This routine runs the input signal through the model and puts the result
 * into the output signal.  It checks that all initialisation has been
 * correctly carried out by calling the appropriate checking routines.
 * potential_V = Cell-membrane pot. above resting level (mV);
 * kEquilibriumPot_Vk = Equlibrium potential of K conductance. (mV);
 * kConductance_Gk = Potassium conductance (ns);
 * threshold_Th = Time-varying cell threshold (mV);
 * spikeVariable_s = spike variable for cell.
 */

BOOLEAN
RunModel_Neuron_McGregor2(EarObjectPtr data)
{
	int			i, spikeState_s, lastSpikeState;
	double		dt, potential_E, kEquilibriumPot_Ek, kConductance_Gk;
	double		threshold_Th;
	double		bConductance_Gb, totalConductance_G, restingPotential_Er;
	double		bEquilibriumPot_Eb;
	register double	dGk, dTh, dE, tmOverDt, tGkOverDt, tThOverDt;
	ChanLen		j;
	ChanData	*inPtr, *outPtr;
	McGregor2	*c;
	
	if (data == NULL) {
		NotifyError("RunModel_Neuron_McGregor2: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_Neuron_McGregor2())		
		return(FALSE);
	if (!CheckRamp_SignalData(data->inSignal))
		return(FALSE);
	if (!InitOutSignal_EarObject(data, data->inSignal->numChannels,
	  data->inSignal->length, data->inSignal->dt)) {
		NotifyError("RunModel_Neuron_McGregor2: Could not initialise output "\
		  "signal.");
		return(FALSE);
	}
	SetProcessName_EarObject(data, "McGregor2 neural cell");
	dt = data->outSignal->dt;
	c = &mcGregor2;
	
	potential_E = kConductance_Gk = bConductance_Gb = 0.0;
	totalConductance_G = mcGregor2.kRestingCond_gk +
	  mcGregor2.bRestingCond_gb;
	restingPotential_Er = (mcGregor2.kRestingCond_gk *
	  mcGregor2.kReversalPoten_Ek0 + mcGregor2.bRestingCond_gb *
	  mcGregor2.bReversalPoten_Eb0) / totalConductance_G;
	kEquilibriumPot_Ek = mcGregor2.kReversalPoten_Ek0 - restingPotential_Er;
	bEquilibriumPot_Eb = mcGregor2.bReversalPoten_Eb0 - restingPotential_Er;
	
	threshold_Th = c->restingThreshold_Th0;
	tmOverDt  = c->membraneTConst_Tm / dt;
	tGkOverDt = c->kDecayTConst_TGk / dt;
	tThOverDt = c->thresholdTConst_TTh / dt;
	lastSpikeState = spikeState_s = 0;
	
	for (i = 0; i < data->outSignal->numChannels; i++) {
		inPtr = data->inSignal->channel[i];
		outPtr = data->outSignal->channel[i];
		for (j = 0; j < data->outSignal->length; j++) {
			bConductance_Gb = mcGregor2.bRestingCond_gb * (exp(potential_E /
			  mcGregor2.nonLinearVConst_Vnl) - 1.0);
			dE =  (-potential_E + (*(inPtr++) + kConductance_Gk *
			   (kEquilibriumPot_Ek - potential_E) + bConductance_Gb *
			   (bEquilibriumPot_Eb - potential_E)) / totalConductance_G) /
			   tmOverDt;
			dTh = (-(threshold_Th - c->restingThreshold_Th0) +
			  c->accomConst_c * potential_E ) / tThOverDt;	
			/* add increment to gk if at start of spike, otherwise allow decay*/
			if (lastSpikeState < spikeState_s) {
				dGk = (-kConductance_Gk + c->delayedRectKCond_b) / tGkOverDt;
				*outPtr++ = (ChanData) (potential_E + c->actionPotential);
			} else {
				dGk = -kConductance_Gk / tGkOverDt;
				*outPtr++ = (ChanData) potential_E;
			}
			threshold_Th += dTh;
			potential_E += dE;
			kConductance_Gk += dGk;
			lastSpikeState = spikeState_s;
			spikeState_s = (potential_E < threshold_Th)? 0: 1;
		}
	}
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);
		
} /* RunModel_Neuron_McGregor2 */
