/*******************
 *
 * MoPreEmph.c
 *
 * This is an implementation of a pre-emphasis band pass filter using the filters.c
 * module.
 * The band-pass filter is used to model the effects of the outer and middle ear on
 * an input signal.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 *******************/
 
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "UtFilters.h"
#include "MoPreEmph.h"
#include "FiParFile.h"

/******************************************************************************/
/**************************** Global subroutines ******************************/
/******************************************************************************/

PreEmphasis		preEmphasis = {

					FALSE,		/* upperCutOffFreqFlag */
					FALSE,		/* lowerCutOffFreqFlag */
					0.0,		/* upperCutOffFreq */
					0.0			/* lowerCutOffFreq */
					
				};

/******************************************************************************/
/**************************** Subroutines and functions ***********************/
/******************************************************************************/

/********************************* SetLowerCutOffFreq *************************/

/*
 * This function sets the module's lower, 3dB down cut off frequency parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetLowerCutOffFreq_PreEmphasis_BandPass(double theLowerCutOffFreq)
{
	preEmphasis.lowerCutOffFreqFlag = TRUE;
	preEmphasis.lowerCutOffFreq = theLowerCutOffFreq;
	return(TRUE);

}

/********************************* SetUpperCutOffFreq *************************/

/*
 * This function sets the module's upper, 3dB down cut off frequency parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetUpperCutOffFreq_PreEmphasis_BandPass(double theUpperCutOffFreq)
{
	preEmphasis.upperCutOffFreqFlag = TRUE;
	preEmphasis.upperCutOffFreq = theUpperCutOffFreq;
	return(TRUE);

}

/********************************* CheckPars **********************************/

/*
 * This routine checks that the necessary parameters for the module have been
 * correctly initialised.
 * It returns TRUE if there are no problems.
 */
 
BOOLEAN
CheckPars_PreEmphasis_BandPass(void)
{
	if (!preEmphasis.lowerCutOffFreqFlag) {
		NotifyError("CheckPars_PreEmphasis_BandPass: Lower cut off frequency "\
		  "not set.");
		return(FALSE);
	}
	if (!preEmphasis.upperCutOffFreqFlag) {
		NotifyError("CheckPars_PreEmphasis_BandPass: Upper cut off frequency "\
		  "not set.");
		return(FALSE);
	}
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetPars_PreEmphasis_BandPass(double theLowerCutOffFreq,
  double theUpperCutOffFreq)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetLowerCutOffFreq_PreEmphasis_BandPass(theLowerCutOffFreq))
		ok = FALSE;
	if (!SetUpperCutOffFreq_PreEmphasis_BandPass(theUpperCutOffFreq))
		ok = FALSE;
	if (!ok)
		NotifyError("SetPars_PreEmphasis_BandPass: Failed to set all module "\
		  "parameters.");
	return(ok);
	
}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_PreEmphasis_BandPass(void)
{
	if (!CheckPars_PreEmphasis_BandPass()) {
		NotifyError("PrintPars_PreEmphasis_BandPass: Parameters have not "\
		  "been correctly set.");
		return;
	}
	printf("Preemphasis Band Pass Filter Module Parameters:-\n");
	printf("\tUpper cut-off Frequency = %g Hz,\n",
	  preEmphasis.upperCutOffFreq);
	printf("\tLower cut-off Frequency = %g Hz.\n",
	  preEmphasis.lowerCutOffFreq);

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_PreEmphasis_BandPass(char *fileName)
{
	BOOLEAN	ok;
	double	lowerCutOffFrequency, upperCutOffFrequency;
    FILE    *fp;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_PreEmphasis_BandPass: Cannot open data file "\
          "'%s'.\n", fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_PreEmphasis_BandPass: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
	if (!GetPars_ParFile(fp, "%lf", &lowerCutOffFrequency))
		ok = FALSE;;
	if (!GetPars_ParFile(fp, "%lf", &upperCutOffFrequency))
		ok = FALSE;;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_PreEmphasis_BandPass: Not enough lines, or "\
		  "invalid parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_PreEmphasis_BandPass(lowerCutOffFrequency,
	  upperCutOffFrequency)) {
		NotifyError("ReadPars_PreEmphasis_BandPass: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/**************************** SimulateModel ***********************************/

/*
 * This routine filters the input signal and puts the result into the output
 * signal.
 * It checks that all initialisation has been correctly carried out by calling
 * the appropriate checking routines.
 * The appropriate coefficients are calculated at the beginning, then the 
 * memory used is released before the program returns.
 */

BOOLEAN
RunModel_PreEmphasis_BandPass(EarObjectPtr data)
{
	BandPassCoeffs	*coefficients;
	
	if (data == NULL) {
		NotifyError("RunModel_PreEmphasis_BandPass: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_PreEmphasis_BandPass())
		return(FALSE);
	SetProcessName_EarObject(data, "Pre-emphasis (middle and outer ear) "\
	  "filter");
	if (!CheckRamp_SignalData(data->inSignal)) {
		NotifyError("RunModel_PreEmphasis_BandPass: Input signal not " \
		  "correctly initialised.");
		return(FALSE);
	}
	InitOutFromInSignal_EarObject(data, 0);
	coefficients = InitCoeffs_Filters_BandPass(preEmphasis.lowerCutOffFreq,
	  preEmphasis.upperCutOffFreq, data->inSignal->dt);

	/* Filter signal */
	
	BandPass_Filters(data->outSignal, coefficients);
	free(coefficients);
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

}
