 /****************
 *
 * MoRecptPotn.c
 *
 * This module contains the model for the hair cell receptor potential.
 * It is planned to produce a module in the future which will encompass
 * This process stage and the Meddis93 hair cell model process stage in the
 * same model.  This will use less memory space.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "MoRecptPotn.h"
#include "FiParFile.h"

/******************************************************************************/
/****************************** Constant definitions **************************/
/******************************************************************************/ 

RecptPotn		recptPotn = {

					FALSE,		/* paramQFlag */
					FALSE,		/* paramVFlag */
					FALSE,		/* bMTimeConstTmFlag */
					0.0,		/* param_q */
					0.0,		/* param_v */
					0.0			/* bMTimeConst_tm */
					
				};

/******************************************************************************/
/********************************* Subroutines and functions ******************/
/******************************************************************************/

/********************************* SetParamQ **********************************/

/*
 * This function sets the module's base amplitude, q, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetParamQ_IHC_RecptPotn(double theParamQ)
{
	recptPotn.paramQFlag = TRUE;
	recptPotn.param_q = theParamQ;
	return(TRUE);

}

/********************************* SetParamV **********************************/

/*
 * This function sets the module's amplitude scale, v, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetParamV_IHC_RecptPotn(double theParamV)
{
	recptPotn.paramVFlag = TRUE;
	recptPotn.param_v = theParamV;
	return(TRUE);

}

/********************************* SetBMTimeConstTm ***************************/

/*
 * This function sets the module's basilar membrane time constant, tm, parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetBMTimeConstTm_IHC_RecptPotn(double theBMTimeConstTm)
{
	recptPotn.bMTimeConstTmFlag = TRUE;
	recptPotn.bMTimeConst_tm = theBMTimeConstTm;
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 */

BOOLEAN
SetPars_IHC_RecptPotn(double q, double v, double tm)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetParamQ_IHC_RecptPotn(q))
		ok = FALSE;
	if (!SetParamV_IHC_RecptPotn(v))
		ok = FALSE;
	if (!SetBMTimeConstTm_IHC_RecptPotn(tm))
		ok = FALSE;
	return(ok);
	  
}

/********************************* CheckPars **********************************/

/*
 * This routine checks that all of the parameters for the module are set.
 * Because this module is defined by its coeffients, default values are always
 * set if none of the parameters are set, otherwise all of them must be
 * individually set.
 */

BOOLEAN
CheckPars_IHC_RecptPotn(void)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!recptPotn.paramQFlag) {
		NotifyError("CheckPars_IHC_RecptPotn: Base amplitude, q, not "\
		  "correctly set.");
		ok = FALSE;
	}
	if (!recptPotn.paramVFlag) {
		NotifyError("CheckPars_IHC_RecptPotn: Amplitude scale, v, not "\
		  "correctly set.");
		ok = FALSE;
	}
	if (!recptPotn.bMTimeConstTmFlag) {
		NotifyError("CheckPars_IHC_RecptPotn: Basilar membrane time "\
		  "constant, tm, not correctly set.");
		ok = FALSE;
	}
	return(ok);
		
}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_IHC_RecptPotn(void)
{
	if (!CheckPars_IHC_RecptPotn()) {
		NotifyError("PrintPars_IHC_RecptPotn: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("Receptor Potential Module Parameters:-\n");
	printf("\tBias, q = %g,\tSlope, v = %g,\t", recptPotn.param_q,
	  recptPotn.param_v);
	printf("time constant = %g ms\n",
	  MSEC(recptPotn.bMTimeConst_tm));

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_IHC_RecptPotn(char *fileName)
{
	BOOLEAN	ok;
	double	paramQ, paramV, bMTimeConst;
    FILE    *fp;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_IHC_RecptPotn: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_IHC_RecptPotn: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
	if (!GetPars_ParFile(fp, "%lf", &paramQ))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &paramV))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &bMTimeConst))
		ok = FALSE;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_IHC_RecptPotn: Not enough lines, or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_IHC_RecptPotn(paramQ, paramV, bMTimeConst)) {
		NotifyError("ReadPars_IHC_RecptPotn: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/********************************* RunModel ***********************************/

/*
 * This routine runs the input signal through the model and puts the result
 * into the output signal.  It checks that all initialisation has been
 * correctly carried out by calling the appropriate checking routines.
 */

BOOLEAN
RunModel_IHC_RecptPotn(EarObjectPtr data)
{
	int		i;
	ChanLen	j;
	double	spontProbKChanOpen, probKChanOpen;
	register	double		dtOverTm;
	register	ChanData	*inPtr, *outPtr;
	
	if (!CheckPars_IHC_RecptPotn()) {
		NotifyError("RunModel_IHC_RecptPotn: Parameters have not been "\
		  "correctly set.");
		return(FALSE);
	}
	if (data == NULL) {
		NotifyError("RunModel_IHC_RecptPotn: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckRamp_SignalData(data->inSignal)) {
		NotifyError("RunModel_IHC_RecptPotn: Input signal not correctly " \
		  "initialised.");
		return(FALSE);
	}
	SetProcessName_EarObject(data, "Hair cell receptor potential");
	if (!InitOutSignal_EarObject(data, data->inSignal->numChannels,
	  data->inSignal->length, data->inSignal->dt)) {
		NotifyError("RunModel_IHC_RecptPotn: Could not initialise output "\
		  "signal.");
		return(FALSE);
	}

	/* The spontaneous probability assumes preceeding silence. */
	spontProbKChanOpen = 1.0 / (1.0 + exp(recptPotn.param_q));
	dtOverTm = data->outSignal->dt / recptPotn.bMTimeConst_tm;
	
	for (i = 0; i < data->outSignal->numChannels; i++) {
		inPtr = data->inSignal->channel[i];
		outPtr = data->outSignal->channel[i];
		/* Initial, spontaneous probability calculation */
		probKChanOpen = 1.0 / (1.0 + exp(recptPotn.param_q -
		  recptPotn.param_v * *inPtr++));
		*outPtr++ = (ChanData) (spontProbKChanOpen + (probKChanOpen -
		  spontProbKChanOpen) * dtOverTm);
		
		/* Probability calculation for the rest of the signal. */
		for (j = 1; j < data->outSignal->length; j++) {
			probKChanOpen = 1.0 / (1.0 + exp(recptPotn.param_q -
			  recptPotn.param_v * *inPtr++));
			*outPtr++ = (ChanData) (*(outPtr - 1) + (probKChanOpen -
			  *(outPtr - 1)) * dtOverTm);
		}
	}
	if (data->updateCustomersFlag)
		UpdateCustomers_EarObject(data);
	return(TRUE);
		
} /* RunModel_IHC_RecptPotn */
