/****************
 *
 * MoSpikeGen.c
 *
 * This module contains the model for the hair cell post-synaptic firing.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "UtRandom.h"
#include "MoSpikeGen.h"
#include "FiParFile.h"

/******************************************************************************/
/****************************** Constant definitions **************************/
/******************************************************************************/

SpikeGen		spikeGen = {

					FALSE,		/* refractoryPeriodFlag */
					FALSE,		/* numberOfRunsFlag */
					FALSE,		/* ranSeedFlag */
					1,			/* ranSeed */
					1,			/* numberOfRuns */
					PS_REFRACTORY_PERIOD
					
				};

/******************************************************************************/
/********************************* Subroutines and functions ******************/
/******************************************************************************/

/********************************* SetRanSeed *********************************/

/*
 * This function sets the module's random number seed parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetRanSeed_AuditoryN_SpikeGen(long theRanSeed)
{
	if (theRanSeed == 0) {
		NotifyError("SetRanSeed_AuditoryN_SpikeGen: A non zero value must be "\
		  "used for the random number seed.\n");
		return(FALSE);
	}
	spikeGen.ranSeedFlag = TRUE;
	spikeGen.ranSeed = theRanSeed;
	return(TRUE);

}
/********************************* SetRefractoryPeriod ************************/

/*
 * This function sets the module's refractory period parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetRefractoryPeriod_AuditoryN_SpikeGen(double theRefractoryPeriod)
{
	if (theRefractoryPeriod < 0.0) {
		NotifyError("SetRefractoryPeriod_AuditoryN_SpikeGen: Refractory "\
		  "period must be greater than zero.\n");
		return(FALSE);
	}
	spikeGen.refractoryPeriodFlag = TRUE;
	spikeGen.refractoryPeriod = theRefractoryPeriod;
	return(TRUE);

}

/********************************* SetNumberOfRuns ****************************/

/*
 * This function sets the module's number of runs parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetNumberOfRuns_AuditoryN_SpikeGen(int theNumberOfRuns)
{
	if (theNumberOfRuns < 0) {
		NotifyError("SetNumberOfRuns_AuditoryN_SpikeGen: Illegal number "\
		  "(value = %d).\n", theNumberOfRuns);
		return(FALSE);
	}
	spikeGen.numberOfRunsFlag = TRUE;
	spikeGen.numberOfRuns = theNumberOfRuns;
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 */

BOOLEAN
SetPars_AuditoryN_SpikeGen(int theNumberOfRuns, long theRanSeed,
  double theRefractoryPeriod)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetRanSeed_AuditoryN_SpikeGen(theRanSeed))
		ok = FALSE;
	if (!SetRefractoryPeriod_AuditoryN_SpikeGen(theRefractoryPeriod))
		ok = FALSE;
	if (!SetNumberOfRuns_AuditoryN_SpikeGen(theNumberOfRuns))
		ok = FALSE;
	return(ok);
	  
}

/********************************* CheckPars **********************************/

/*
 * This routine checks that all of the parameters for the module are set.
 * Because this module is defined by its coeffients, default values are always
 * set if none of the parameters are set, otherwise all of them must be
 * individually set.
 */

BOOLEAN
CheckPars_AuditoryN_SpikeGen(void)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!spikeGen.ranSeedFlag) {
		NotifyError("CheckPars_AuditoryN_SpikeGen: Random number seed not "\
		  "correctly set.");
		ok = FALSE;
	}
	if (!spikeGen.refractoryPeriodFlag) {
		NotifyError("CheckPars_AuditoryN_SpikeGen: Refractory period not "\
		  "correctly set.");
		ok = FALSE;
	}
	if (!spikeGen.numberOfRunsFlag) {
		NotifyError("CheckPars_AuditoryN_SpikeGen: NumberOfRuns not "\
		  "correctly set.");
		ok = FALSE;
	}
	return(ok);
		
}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_AuditoryN_SpikeGen(void)
{
	if (!CheckPars_AuditoryN_SpikeGen()) {
		NotifyError("PrintPars_AuditoryN_SpikeGen: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("Post-Synaptic Firing Module Parameters:-\n");
	printf("\tRefractory Period = %g ms\tNumber of runs = %d.\n",
	  MSEC(spikeGen.refractoryPeriod), spikeGen.numberOfRuns);
	printf("\tRandom number seed = %ld\n", spikeGen.ranSeed);

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_AuditoryN_SpikeGen(char *fileName)
{
	BOOLEAN	ok;
	int		numberOfRuns;
	long	ranSeed;
	double	refractoryPeriod;
    FILE    *fp;

	if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_AuditoryN_SpikeGen: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_AuditoryN_SpikeGen: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
	if (!GetPars_ParFile(fp, "%d", &numberOfRuns))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%ld", &ranSeed))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &refractoryPeriod))
		ok = FALSE;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_AuditoryN_SpikeGen: Not enough lines, or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_AuditoryN_SpikeGen(numberOfRuns, ranSeed, refractoryPeriod)) {
		NotifyError("ReadPars_AuditoryN_SpikeGen: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/********************************* RunModel ***********************************/

/*
 * This routine runs the input signal through the model and puts the result
 * into the output signal.  It checks that all initialisation has been
 * correctly carried out by calling the appropriate checking routines.
 */

BOOLEAN
RunModel_AuditoryN_SpikeGen(EarObjectPtr data)
{
	int		i, k;
	ChanLen	j;
	double	timer;
	register	ChanData	*inPtr, *outPtr;
	
	if (data == NULL) {
		NotifyError("RunModel_AuditoryN_SpikeGen: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_AuditoryN_SpikeGen())
		return(FALSE);
	if (!CheckPars_SignalData(data->inSignal)) {
		NotifyError("RunModel_AuditoryN_SpikeGen: Signal not set correctly.");
		return(FALSE);
	}
	SetProcessName_EarObject(data, "Post-Synaptic Spike Firing");
	if (!InitOutSignal_EarObject(data, data->inSignal->numChannels,
	  data->inSignal->length, data->inSignal->dt)) {
		NotifyError("RunModel_AuditoryN_SpikeGen: Could not initialise output "\
		  "signal.");
		return(FALSE);
	}
	for (k = 0; k < spikeGen.numberOfRuns; k++)
		for (i = 0; i < data->outSignal->numChannels; i++) {
			inPtr = data->inSignal->channel[i];
			outPtr = data->outSignal->channel[i];

			for (j = 0, timer = LARGE_VALUE; j < data->outSignal->length; j++) {
				if ((timer > spikeGen.refractoryPeriod) && (*inPtr >
				  Ran01_Random(&spikeGen.ranSeed))) {
					*outPtr += 1.0;
					timer = 0.0;
				} else {
					timer += data->inSignal->dt;
				}
				inPtr++;
				outPtr++;
			}
		}
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);
		
} /* RunModel_AuditoryN_SpikeGen */
