/*******
 *
 * StAMTone.c
 *
 * This module contains the methods for the AM-tone (amplitude modulated)
 *  generation paradigm.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "StAMTone.h"
#include "FiParFile.h"

/********************************* Global variables ***************************/

AMTone		aMTone = {

				FALSE,		/* carrierFrequencyFlag */
				FALSE,		/* modulationFrequencyFlag */
				FALSE,		/* percentAM_Flag */
				FALSE,		/* durationFlag */
				FALSE,		/* dtFlag */
				FALSE,		/* intensityFlag */
				0.0,		/* carrierFrequency */
				0.0,		/* modulationFrequency */
				0.0,		/* percentAM */
				0.0,		/* intensity */
				0.0,		/* duration */
				0.0			/* dt */
					
			};

/******************************************************************************/
/********************************* Subroutines and functions ******************/
/******************************************************************************/

/********************************* CheckPars **********************************/

/*
 * This routine checks that the necessary parameters for the module have been
 * correctly initialised.
 * It also checks that the Nyquist critical frequency is not exceeded.
 * It returns TRUE if there are no problems.
 */
 
BOOLEAN
CheckPars_AMTone(void)
{
	BOOLEAN	ok;
	double	criticalFrequency;
	
	ok = TRUE;
	if (!aMTone.carrierFrequencyFlag) {
		NotifyError("CheckPars_AMTone: Carrier frequency variable not set.");
		ok = FALSE;
	}	
	if (!aMTone.modulationFrequencyFlag) {
		NotifyError("CheckPars_AMTone: Modulation frequency variable not set.");
		ok = FALSE;
	}
	if (!aMTone.percentAM_Flag) {
		NotifyError("CheckPars_AMTone: Percent AM variable not set.");
		ok = FALSE;
	}
		if (!aMTone.intensityFlag) {
		NotifyError("CheckPars_AMTone: intensity variable not set.");
		ok = FALSE;
	}
	if (!aMTone.durationFlag) {
		NotifyError("CheckPars_AMTone: duration variable not set.");
		ok = FALSE;
	}
	if (!aMTone.dtFlag) {
		NotifyError("CheckPars_AMTone: dtFlag variable not set.");
		ok = FALSE;
	}
	criticalFrequency = floor(1.0 / (2.0 * aMTone.dt));
	if (ok && (criticalFrequency <= aMTone.carrierFrequency)) {
		NotifyError("CheckPars_AMTone: Sampling rate (dt = %g ms) is to " \
		  "low for the carrier frequency.", MSEC(aMTone.dt));
		ok = FALSE;
	} 
	if (ok && (criticalFrequency <= aMTone.modulationFrequency)) {
		NotifyError("CheckPars_AMTone: Sampling rate (dt = %g ms) is too "\
		  "low for the modulation frequency.", MSEC(aMTone.dt));
		ok = FALSE;
	} 
	return(ok);
	
}	

/********************************* SetCarrierFrequency ************************/

/*
 * This function sets the module's carrierFrequency parameter.  
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetCarrierFrequency_AMTone(double theCarrierFrequency)
{
	aMTone.carrierFrequencyFlag = TRUE;
	aMTone.carrierFrequency = theCarrierFrequency;
	return(TRUE);

}

/********************************* SetModulationFrequency *********************/

/*
 * This function sets the module's modulationFrequency parameter.  
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetModulationFrequency_AMTone(double theModulationFrequency)
{
	aMTone.modulationFrequencyFlag = TRUE;
	aMTone.modulationFrequency = theModulationFrequency;
	return(TRUE);

}

/********************************* SetPercentAM *******************************/

/*
 * This function sets the module's percentAM parameter.  
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetPercentAM_AMTone(double thePercentAM)
{
	aMTone.percentAM_Flag = TRUE;
	aMTone.percentAM = thePercentAM;
	return(TRUE);

}

/********************************* SetIntensity *******************************/

/*
 * This function sets the module's intensity parameter.
 * It returns TRUE if the operation is successful.
 */


BOOLEAN
SetIntensity_AMTone(double theIntensity)
{
	aMTone.intensityFlag = TRUE;
	aMTone.intensity = theIntensity;
	return(TRUE);

}

/********************************* SetDuration ********************************/

/*
 * This function sets the module's duration parameter.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetDuration_AMTone(double theDuration)
{
	aMTone.durationFlag = TRUE;
	aMTone.duration = theDuration;
	return(TRUE);

}

/********************************* SetSamplingInterval ************************/

/*
 * This function sets the module's samplingInterval parameter.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetSamplingInterval_AMTone(double theSamplingInterval)
{
	if ( theSamplingInterval <= 0.0 ) {
		NotifyError("SetSamplingInterval_AMTone: Illegal sampling interval "\
		  "value = %g!", theSamplingInterval);
		return(FALSE);
	}
	aMTone.dtFlag = TRUE;
	aMTone.dt = theSamplingInterval;
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetPars_AMTone(double theCarrierFrequency, double theModulationFrequency,
  double thePercentAM, double theIntensity, double theDuration,
  double theSamplingInterval)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetCarrierFrequency_AMTone(theCarrierFrequency))
		ok = FALSE;
	if (!SetModulationFrequency_AMTone(theModulationFrequency))
		ok = FALSE;
	if (!SetPercentAM_AMTone(thePercentAM))
		ok = FALSE;
	if (!SetIntensity_AMTone(theIntensity))
		ok = FALSE;
	if (!SetDuration_AMTone(theDuration))
		ok = FALSE;
	if (!SetSamplingInterval_AMTone(theSamplingInterval))
		ok = FALSE;
	if (!ok)
		NotifyError("SetPars_AMTone: Failed to set all module parameters.");
	return(ok);
	
}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_AMTone(void)
{
	if (!CheckPars_AMTone()) {
		NotifyError("PrintPars_AMTone: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("Amplitude Modulated Tone Module Parameters:-\n");
	printf("\tCarrier frequency = %g Hz,\tModulation frequency = %g\n",
	  aMTone.carrierFrequency, aMTone.modulationFrequency);
	printf("\tPecentage Amplitude = %g,\tIntensity = %g dB SPL,\n",
	  aMTone.percentAM, aMTone.intensity);
	printf("\tDuration = %g ms,\tSampling interval = %g ms\n",
	  MSEC(aMTone.duration), MSEC(aMTone.dt));

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_AMTone(char *fileName)
{
	BOOLEAN	ok;
	double  carrierFrequency, modulationFrequency, percentAM, intensity;
	double  duration, samplingInterval;
    FILE    *fp;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_AMTone: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_AMTone: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
	if (!GetPars_ParFile(fp, "%lf", &carrierFrequency))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &modulationFrequency))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &percentAM))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &intensity))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &duration))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &samplingInterval))
		ok = FALSE;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_AMTone:  Not enough, lines or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_AMTone(carrierFrequency, modulationFrequency, percentAM,
	  intensity, duration, samplingInterval)) {
		NotifyError("ReadPars_AMTone: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/********************************* GenerateSignal *****************************/

/*
 * This routine allocates memory for the output signal, if necessary, and
 * generates the signal into channel[0] of the signal data-set.
 * It checks that all initialisation has been correctly carried out by calling
 * the appropriate checking routines.
 * It can be called repeatedly with different parameter values if required.
 * Stimulus generation only sets the output signal, the input signal is not
 * used.
 */
 
BOOLEAN
GenerateSignal_AMTone(EarObjectPtr data)
{
	ChanLen		i;
	double 		modulationIndex;
	register	double		amplitude;
	register	ChanData	*dataPtr;

	if (data == NULL) {
		NotifyError("GenerateSignal_AMTone: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_AMTone())
		return(FALSE);
	SetProcessName_EarObject(data, "AM Tone stimulus");
	if ( !InitOutSignal_EarObject(data, AM_TONE_NUM_CHANNELS,
	  (ChanLen) (aMTone.duration / aMTone.dt), aMTone.dt) ) {
		NotifyError("GenerateSignal_AMTone: Cannot initialise output signal");
		return(FALSE);
	}
	amplitude = RMS_AMP(aMTone.intensity) * SQRT_2;
	modulationIndex = aMTone.percentAM/100.0;
	dataPtr = data->outSignal->channel[0];
	for (i = 0; i < data->outSignal->length; i++)
		*(dataPtr++) = amplitude * (1.0 + modulationIndex * cos(PIx2 *
		  aMTone.modulationFrequency * (i * data->outSignal->dt))) *
		  sin(PIx2 * aMTone.carrierFrequency * (i * data->outSignal->dt));
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

} /* GenerateSignal_AMTone */




