/*******
 *
 * StBPTone.c
 *
 * This module contains the methods for the multiple pure-tone signal
 * generation paradigm.
 * This is a stimulus (St) module.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "StBPTone.h"
#include "FiParFile.h"

/******************************************************************************/
/********************************* Global variables ***************************/
/******************************************************************************/ 

BPureTone	bPureTone = {

				{FALSE, FALSE},	/* frequencyFlag[] */
				{FALSE, FALSE},	/* intensityFlag[] */
				FALSE,			/* phaseDifferenceFlag */
				FALSE,			/* durationFlag */
				FALSE,			/* dtFlag */
				{0.0, 0.0},		/* frequency[] */
				{0.0, 0.0},		/* intensity[] */
				0.0,			/* phaseDifference */
				0.0,		/* duration */
				0.0			/* dt */

			};

/******************************************************************************/
/********************************* Subroutines and functions ******************/
/******************************************************************************/

/********************************* CheckPars **********************************/

/*
 * This routine checks that the necessary parameters for the module have been
 * correctly initialised.
 * It also checks that the Nyquist critical frequency is not exceeded.
 * It returns TRUE if there are no problems.
 */
 
BOOLEAN
CheckPars_PureTone_Binaural(void)
{
	BOOLEAN	ok;
	double	criticalFrequency;
	
	ok = TRUE;
	if (!bPureTone.frequencyFlag[LEFT_BINAURAL_PTONE]) {
		NotifyError("CheckPars_PureTone_Binaural: Left frequency not set.");
		ok = FALSE;
	}
	if (!bPureTone.frequencyFlag[RIGHT_BINAURAL_PTONE]) {
		NotifyError("CheckPars_PureTone_Binaural: Right frequency not set.");
		ok = FALSE;
	}
	if (!bPureTone.intensityFlag[LEFT_BINAURAL_PTONE]) {
		NotifyError("CheckPars_PureTone_Binaural: Left intensity not set.");
		ok = FALSE;
	}
	if (!bPureTone.intensityFlag[RIGHT_BINAURAL_PTONE]) {
		NotifyError("CheckPars_PureTone_Binaural: Right intensity not set.");
		ok = FALSE;
	}
	if (!bPureTone.phaseDifferenceFlag) {
		NotifyError("CheckPars_PureTone_Binaural: Phase difference not set.");
		ok = FALSE;
	}
	if (!bPureTone.durationFlag) {
		NotifyError("CheckPars_PureTone_Binaural: duration variable not set.");
		ok = FALSE;
	}
	if (!bPureTone.dtFlag) {
		NotifyError("CheckPars_PureTone_Binaural: dtFlag variable not set.");
		ok = FALSE;
	}
	criticalFrequency = floor(1.0 / (2.0 * bPureTone.dt));
	if (ok &&((criticalFrequency <= bPureTone.frequency[LEFT_BINAURAL_PTONE]) ||
	  (criticalFrequency <= bPureTone.frequency[RIGHT_BINAURAL_PTONE]))) {
		NotifyError("CheckPars_PureTone_Binaural: Sampling rate (dt = "\
		  "%g ms) is too low for one of the stimulus\nfrequencies.",
		  MSEC(bPureTone.dt));
		ok = FALSE;
	} 
	return(TRUE);
	
}	

/********************************* SetIndividualFreq **************************/

/*
 * This function sets up the module's channel frequency parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetIndividualFreq_PureTone_Binaural(int channel, double theFrequency)
{
	if ((channel != LEFT_BINAURAL_PTONE) && (channel != RIGHT_BINAURAL_PTONE)) {
		NotifyError("SetIndividualFreq_PureTone_Binaural: Illegal channel.");
		return(FALSE);
	}
	bPureTone.frequencyFlag[channel] = TRUE;
	bPureTone.frequency[channel] = theFrequency;
	return(TRUE);

}

/********************************* SetIndividualIntensity *********************/

/*
 * This function sets up the module's channel intensity parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetIndividualIntensity_PureTone_Binaural(int channel, double theIntensity)
{
	if ((channel != LEFT_BINAURAL_PTONE) && (channel != RIGHT_BINAURAL_PTONE)) {
		NotifyError("SetIndividualFreq_PureTone_Binaural: Illegal channel.");
		return(FALSE);
	}
	bPureTone.intensityFlag[channel] = TRUE;
	bPureTone.intensity[channel] = theIntensity;
	return(TRUE);

}

/********************************* SetPhaseDifference *************************/

/*
 * This function sets up the module's phase difference parameter.
 * The pointer to the phases lists is passed to the module.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetPhaseDifference_PureTone_Binaural(double thePhaseDifference)
{

	bPureTone.phaseDifferenceFlag = TRUE;
	bPureTone.phaseDifference = thePhaseDifference;
	return(TRUE);

}

/********************************* SetDuration ********************************/

/*
 * This function sets the module's duration parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetDuration_PureTone_Binaural(double theDuration)
{
	bPureTone.durationFlag = TRUE;
	bPureTone.duration = theDuration;
	return(TRUE);

}

/********************************* SetSamplingInterval ************************/

/*
 * This function sets the module's samplingInterval parameter.  It first checks
 * that the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetSamplingInterval_PureTone_Binaural(double theSamplingInterval)
{
	if ( theSamplingInterval <= 0.0 ) {
		NotifyError("SetSamplingInterval_PureTone_Binaural: Illegal sampling "\
		  "interval value = %g!", theSamplingInterval);
		return(FALSE);
	}
	bPureTone.dtFlag = TRUE;
	bPureTone.dt = theSamplingInterval;
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetPars_PureTone_Binaural(double leftFrequency, double rightFrequency,
  double leftIntensity, double rightIntensity, double phaseDifference,
  double theDuration, double theSamplingInterval)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetIndividualFreq_PureTone_Binaural(LEFT_BINAURAL_PTONE,
	  leftFrequency))
		ok = FALSE;
	if (!SetIndividualFreq_PureTone_Binaural(RIGHT_BINAURAL_PTONE,
	  rightFrequency))
		ok = FALSE;
	if (!SetIndividualIntensity_PureTone_Binaural(LEFT_BINAURAL_PTONE,
	  leftIntensity))
		ok = FALSE;
	if (!SetIndividualIntensity_PureTone_Binaural(RIGHT_BINAURAL_PTONE,
	  rightIntensity))
		ok = FALSE;
	if (!SetPhaseDifference_PureTone_Binaural(phaseDifference))
		ok = FALSE;
	if (!SetDuration_PureTone_Binaural(theDuration))
		ok = FALSE;
	if (!SetSamplingInterval_PureTone_Binaural(theSamplingInterval))
		ok = FALSE;
	if (!ok)
		NotifyError("SetPars_PureTone_Binaural: Failed to set all module "\
		  "parameters.");
	return(ok);
	
}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_PureTone_Binaural(void)
{
	
	if (!CheckPars_PureTone_Binaural()) {
		NotifyError("PrintPars_PureTone_Binaural: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("Multiple Pure Tone Module Parameters:-\n");
	printf("\t\t%10s\t%10s\n", "Frequency ", "Intensity");
	printf("\t\t%10s\t%10s\n", "   (Hz)   ", " (dB SPL)");
	printf("\tLeft:\t%10g\t%10g\n", bPureTone.frequency[LEFT_BINAURAL_PTONE],
	  bPureTone.intensity[LEFT_BINAURAL_PTONE]);
	printf("\tRight:\t%10g\t%10g\n", bPureTone.frequency[RIGHT_BINAURAL_PTONE],
	  bPureTone.intensity[RIGHT_BINAURAL_PTONE]);
	printf("\tPhase difference between channels = %g degrees\n",
	  bPureTone.phaseDifference);
	printf("\tDuration = %g ms,\tSampling interval = %g ms\n",
	  MSEC(bPureTone.duration), MSEC(bPureTone.dt));

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_PureTone_Binaural(char *fileName)
{
	BOOLEAN	ok;
	double  leftFrequency, rightFrequency, leftIntensity, rightIntensity;
	double	phaseDifference, duration, samplingInterval;
	FILE    *fp;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_PureTone_Binaural: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_PureTone_Binaural: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
	if (!GetPars_ParFile(fp, "%lf %lf", &leftFrequency, &leftIntensity))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf %lf", &rightFrequency, &rightIntensity))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &phaseDifference))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &duration))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &samplingInterval))
		ok = FALSE;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_PureTone_Binaural: Not enough lines, or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_PureTone_Binaural(leftFrequency, rightFrequency, leftIntensity,
	  rightIntensity, phaseDifference, duration, samplingInterval)) {
		NotifyError("ReadPars_PureTone_Binaural: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/********************************* GenerateSignal *****************************/

/*
 * This routine allocates memory for the output signal, if necessary, and
 * generates the signal into channel[0] of the signal data-set.
 * It checks that all initialisation has been correctly carried out by calling
 * the appropriate checking routines.
 * It can be called repeatedly with different parameter values if required.
 * Stimulus generation only sets the output signal, the input signal is not
 * used.
 * With repeated calls the Signal memory is only allocated once, then re-used.
 */

BOOLEAN
GenerateSignal_PureTone_Binaural(EarObjectPtr data)
{
	int			j;
	ChanLen		i;
	register	double		amplitude, phase;
	register	ChanData	*dataPtr;

	if (data == NULL) {
		NotifyError("GenerateSignal_PureTone_Binaural: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_PureTone_Binaural())
		return(FALSE);
	SetProcessName_EarObject(data, "Binaural pure tone stimulus");
	if ( !InitOutSignal_EarObject(data, BINAURAL_PTONE_CHANNELS,
	  (ChanLen) (bPureTone.duration / bPureTone.dt), bPureTone.dt) ) {
		NotifyError("GenerateSignal_PureTone_Binaural: Cannot initialise output "\
		  "signal.");
		return(FALSE);
	}
	phase = bPureTone.phaseDifference * PIx2 / 360.0;
	for (j = 0; j < BINAURAL_PTONE_CHANNELS; j++) {
		amplitude = RMS_AMP(bPureTone.intensity[j]) * SQRT_2;
		dataPtr = data->outSignal->channel[j];
		for (i = 0; i < data->outSignal->length; i++)
			*(dataPtr++) = (ChanData) (amplitude * sin(PIx2 *
			   bPureTone.frequency[j] * (i * data->outSignal->dt) + phase * j));
	}
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

} /* GenerateSignal_PureTone_Binaural */

