/******************
 *
 * StClick.c
 *
 * This module contains the methods for the Click GenParadgm sub-class.
 * This is a stimulus (St) module.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ******************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "StClick.h"
#include "FiParFile.h"

/******************************************************************************/ 
/********************************* Global variables ***************************/
/******************************************************************************/ 

Click		click = {

				FALSE,		/* clickTimeFlag */
				FALSE,		/* intensityFlag */
				FALSE,		/* durationFlag */
				FALSE,		/* dtFlag */
				0.0,		/* clickTime */
				0.0,		/* intensity */
				0.0,		/* duration */
				0.0			/* dt */

			};

/******************************************************************************/
/********************************* Subroutines and functions ******************/
/******************************************************************************/

/********************************* CheckPars **********************************/

/*
 * This routine checks that the necessary parameters for the module have been
 * correctly initialised.
 * It returns TRUE if there are no problems.
 */
 
BOOLEAN
CheckPars_Click(void)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!click.clickTimeFlag) {
		NotifyError("CheckPars_Click: clickTime variable not set.");
		ok = FALSE;
	}
	if (!click.intensityFlag) {
		NotifyError("CheckPars_Click: intensity variable not set.");
		ok = FALSE;
	}
	if (!click.durationFlag) {
		NotifyError("CheckPars_Click: duration variable not set.");
		ok = FALSE;
	}
	if (!click.dtFlag) {
		NotifyError("CheckPars_Click: dt (sampling interval) variable not "\
		  "set.");
		ok = FALSE;
	}
	return(ok);
}	

/********************************* SetClickTime *******************************/

/*
 * This function sets the module's clickTime parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetClickTime_Click(double theClickTime)
{
	click.clickTimeFlag = TRUE;
	click.clickTime = theClickTime;
	return(TRUE);

}

/********************************* SetIntensity *******************************/

/*
 * This function sets the module's intensity parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetIntensity_Click(double theIntensity)
{
	click.intensityFlag = TRUE;
	click.intensity = theIntensity;
	return(TRUE);

}

/********************************* SetDuration ********************************/

/*
 * This function sets the module's duration parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetDuration_Click(double theDuration)
{
	click.durationFlag = TRUE;
	click.duration = theDuration;
	return(TRUE);

}

/********************************* SetSamplingInterval ************************/

/*
 * This function sets the module's samplingInterval parameter.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetSamplingInterval_Click(double theSamplingInterval)
{
	if ( theSamplingInterval <= 0.0 ) {
		NotifyError("SetSamplingInterval_Click: Illegal sampling interval "\
		  "value = %g!", theSamplingInterval);
		return(FALSE);
	}
	click.dtFlag = TRUE;
	click.dt = theSamplingInterval;
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetPars_Click(double theClickTime, double theIntensity, double theDuration,
  double theSamplingInterval)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetClickTime_Click(theClickTime))
		ok = FALSE;
	if (!SetIntensity_Click(theIntensity))
		ok = FALSE;
	if (!SetDuration_Click(theDuration))
		ok = FALSE;
	if (!SetSamplingInterval_Click(theSamplingInterval))
		ok = FALSE;
	if (!ok)
		NotifyError("SetPars_Click: Failed to set all module parameters.");
	return(ok);
	
}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_Click(void)
{
	if (!CheckPars_Click()) {
		NotifyError("PrintPars_Click: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("Click Module Parameters:-\n");
	printf("\tClick time = %g ms,\tIntensity = %g dB SPL,\n",
	  MSEC(click.clickTime), click.intensity);
	printf("\tDuration = %g ms,\tSampling interval = %g ms\n",
	  MSEC(click.duration), MSEC(click.dt));

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_Click(char *fileName)
{
	BOOLEAN	ok;
	double  intensity, duration, samplingInterval, clickTime;
    FILE    *fp;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_Click: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_Click: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
    if (!GetPars_ParFile(fp, "%lf", &clickTime))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &intensity))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &duration))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &samplingInterval))
		ok = FALSE;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_Click:  Not enough lines, or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_Click(clickTime, intensity, duration,
	  samplingInterval)) {
		NotifyError("ReadPars_Click: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/********************************* GenerateSignal *****************************/

/*
 * This routine allocates memory for the output signal, if necessary, and
 * generates the signal into channel[0] of the signal data-set.
 * It checks that all initialisation has been correctly carried out by calling
 * the appropriate checking routines.
 * It can be called repeatedly with different parameter values if required.
 * Stimulus generation only sets the output signal, the input signal is not
 * used.
 */

BOOLEAN
GenerateSignal_Click(EarObjectPtr data)
{
	ChanData	*dataPtr;

	if (data == NULL) {
		NotifyError("GenerateSignal_Click: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_Click())
		return(FALSE);
	SetProcessName_EarObject(data, "Click (delta-function) stimulus");

	if (!InitOutSignal_EarObject(data, 1, (ChanLen) (click.duration /
		click.dt), click.dt)) {
		NotifyError("GenerateSignal_Click: Cannot initialise output channels.");
		return(FALSE);
	}
	dataPtr = data->outSignal->channel[0];
	/* The channel is initialised to zero by InitOutSignal_EarObject. */
	dataPtr[(int) (click.clickTime / click.dt)] = SQRT_2 *
	  RMS_AMP(click.intensity);
	data->outSignal->rampFlag = TRUE;	/* Clicks cannot really be ramped. */
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

} /* GenerateSignal_Click */

