/*******
 *
 * StMPPTone.c
 *
 * This module contains the Silence Pure-Tone stimulus generation paradigm.
 * The module generates a pure-tone signal preceded and ended by periods of
 * silence.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "FiParFile.h"
#include "StMPPTone.h"

/********************************* Global variables ***************************/

PureTone4	pureTone4 = {

				FALSE,		/* intensityFlag */
				FALSE,		/* durationFlag */
				FALSE,		/* dtFlag */
				FALSE,		/* beginSilenceFlag */
				FALSE,		/* pulseDurationFlag */
				FALSE,		/* repetitionPeriodFlag */
				FALSE,		/* numPulsesFlag */
				0,			/* numPulses */
				0.0,		/* intensity */
				0.0,		/* duration */
				0.0,		/* dt */
				0.0,		/* beginSilence */
				0.0,		/* pulseDuration */
				0.0,		/* repetitionPeriod */
				NULL		/* *frequencies */

			};

/******************************************************************************/
/********************************* Subroutines and functions ******************/
/******************************************************************************/

/********************************* CheckPars **********************************/

/*
 * This routine checks that the necessary parameters for the module have been
 * correctly initialised.
 * It checks that the Nyquist critical intensity is not exceeded.
 * It returns TRUE if there are no problems.
 */
 
BOOLEAN
CheckPars_PureTone_MultiPulse(void)
{
	BOOLEAN	ok;
	int		i;
	double	criticalFrequency;
	
	ok = TRUE;
	if (!pureTone4.intensityFlag) {
		NotifyError("CheckPars_PureTone_MultiPulse: Intensity variable not set.");
		ok = FALSE;
	}
	if (pureTone4.frequencies == NULL) {
		NotifyError("CheckPars_PureTone_MultiPulse: intensites array not set.");
		ok = FALSE;
	}
	if (!pureTone4.durationFlag) {
		NotifyError("CheckPars_PureTone_MultiPulse: duration variable not set.");
		ok = FALSE;
	}
	if (!pureTone4.beginSilenceFlag) {
		NotifyError("CheckPars_PureTone_MultiPulse: beginSilence variable not set.");
		ok = FALSE;
	}
	if (!pureTone4.pulseDurationFlag) {
		NotifyError("CheckPars_PureTone_MultiPulse: pulseDuration variable not "\
		  "set.");
		ok = FALSE;
	}
	if (!pureTone4.dtFlag) {
		NotifyError("CheckPars_PureTone_MultiPulse: dt variable not set.");
		ok = FALSE;
	}
	if (!pureTone4.repetitionPeriodFlag) {
		NotifyError("CheckPars_PureTone_MultiPulse: repetitionPeriod variable "\
		  "not set.");
		ok = FALSE;
	}
	if (!pureTone4.numPulsesFlag) {
		NotifyError("CheckPars_PureTone_MultiPulse: numPulses variable not "\
		  "set.");
		ok = FALSE;
	}
	if ( (pureTone4.pulseDuration > pureTone4.duration) ) {
	  	NotifyError("CheckPars_PureTone_MultiPulse: The pulse duration  "\
	  	  "is longer than the signal duration.");
	  	ok = FALSE;
	}
	if ( (pureTone4.pulseDuration > pureTone4.repetitionPeriod) ) {
	  	NotifyError("CheckPars_PureTone_MultiPulse: The pulse duration  "\
	  	  "is longer than the repetition period.");
	  	ok = FALSE;
	}
	criticalFrequency = floor(1.0 / (2.0 * pureTone4.dt));
	if (ok)
		for (i = 0; i < pureTone4.numPulses; i++)
			if (criticalFrequency <= pureTone4.frequencies[i]) {
				NotifyError("CheckPars_PureTone_MultiPulse: Sampling rate (dt "\
				  "= %g ms) is too low for frequency No. %d.",
				  MSEC(pureTone4.dt), i);
				ok = FALSE;
			}
	return(ok);

}	

/********************************* SetIntensity *******************************/

/*
 * This function sets the module's intensity parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetIntensity_PureTone_MultiPulse(double theIntensity)
{
	pureTone4.intensityFlag = TRUE;
	pureTone4.intensity = theIntensity;
	return(TRUE);

}

/********************************* SetFrequencies *****************************/

/*
 * This function sets up the module's frequencies array.
 * The pointer to the frequencies lists is passed to the module.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetFrequencies_PureTone_MultiPulse(double *theFrequencies)
{

	pureTone4.frequencies = theFrequencies;
	return(TRUE);

}

/********************************* SetNumPulses *******************************/

/*
 * This function sets the module's numPulses parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetNumPulses_PureTone_MultiPulse(int theNumPulses)
{
	pureTone4.numPulsesFlag = TRUE;
	pureTone4.numPulses = theNumPulses;
	return(TRUE);

}

/********************************* SetIndividualFreq **************************/

/*
 * This function sets the individual frequency values.
 * It first checks if the frequencies have been set.
 */

BOOLEAN
SetIndividualFreq_PureTone_MultiPulse(int theIndex, double theFrequency)
{
	if (pureTone4.frequencies == NULL) {
		NotifyError("SetIndividualFreq_PureTone_MultiPulse: Frequencies "\
		  "not set.");
		return(FALSE);
	}
	if (theIndex > pureTone4.numPulses - 1) {
		NotifyError("SetIndividualFreq_PureTone_MultiPulse: Index value "\
		  "must be in the\nrange 0 - %d (%d).\n", pureTone4.numPulses - 1,
		  theIndex);
		return(FALSE);
	}
	pureTone4.frequencies[theIndex] = theFrequency;
	return(TRUE);

}

/********************************* SetDuration ********************************/

/*
 * This function sets the module's duration parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetDuration_PureTone_MultiPulse(double theDuration)
{
	pureTone4.durationFlag = TRUE;
	pureTone4.duration = theDuration;
	return(TRUE);

}

/********************************* SetBeginSilence ****************************/

/*
 * This function sets the module's beginSilence parameter.  It first checks
 * that the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetBeginSilence_PureTone_MultiPulse(double theBeginSilence)
{
	pureTone4.beginSilenceFlag = TRUE;
	pureTone4.beginSilence = theBeginSilence;
	return(TRUE);

}

/********************************* SetPulseDuration ***************************/

/*
 * This function sets the module's pulseDuration parameter.  It first checks
 * that the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetPulseDuration_PureTone_MultiPulse(double thePulseDuration)
{
	if (thePulseDuration < 0.0) {
		NotifyError("SetPulseDuration_PureTone_MultiPulse: Illegal negative time!");
		return(FALSE);
	}
	pureTone4.pulseDurationFlag = TRUE;
	pureTone4.pulseDuration = thePulseDuration;
	return(TRUE);

}

/********************************* SetRepetitionPeriod ************************/

/*
 * This function sets the module's repetitionPeriod parameter.  It first checks
 * that the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetRepetitionPeriod_PureTone_MultiPulse(double theRepetitionPeriod)
{
	if (theRepetitionPeriod < 0.0) {
		NotifyError("SetRepetitionPeriod_PureTone_MultiPulse: Illegal negative "\
		  "time!");
		return(FALSE);
	}
	pureTone4.repetitionPeriodFlag = TRUE;
	pureTone4.repetitionPeriod = theRepetitionPeriod;
	return(TRUE);

}

/********************************* SetSamplingInterval ************************/

/*
 * This function sets the module's timeStep parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetSamplingInterval_PureTone_MultiPulse(double theSamplingInterval)
{
	if ( theSamplingInterval <= 0.0 ) {
		NotifyError("SetSamplingInterval_PureTone_MultiPulse: Illegal sampling "\
		  "interval value = %g!", theSamplingInterval);
		return(FALSE);
	}
	pureTone4.dtFlag = TRUE;
	pureTone4.dt = theSamplingInterval;
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetPars_PureTone_MultiPulse(int theNumPulses, double theIntensity,
  double *theFrequencies, double theDuration, double theSamplingInterval,
  double theBeginSilence, double thePulseDuration, double theRepetitionPeriod)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetNumPulses_PureTone_MultiPulse(theNumPulses))
		ok = FALSE;
	if (!SetIntensity_PureTone_MultiPulse(theIntensity))
		ok = FALSE;
	if (!SetFrequencies_PureTone_MultiPulse(theFrequencies))
		ok = FALSE;
	if (!SetDuration_PureTone_MultiPulse(theDuration))
		ok = FALSE;
	if (!SetSamplingInterval_PureTone_MultiPulse(theSamplingInterval))
		ok = FALSE;
	if (!SetBeginSilence_PureTone_MultiPulse(theBeginSilence))
		ok = FALSE;
	if (!SetPulseDuration_PureTone_MultiPulse(thePulseDuration))
		ok = FALSE;
	if (!SetRepetitionPeriod_PureTone_MultiPulse(theRepetitionPeriod))
		ok = FALSE;
	if (!ok)
		NotifyError("SetPars_PureTone_MultiPulse: Failed to set all module "\
		  "parameters.");
	return(ok);
	
}

/****************************** GetIndividualFreq *****************************/

/*
 * This routine returns an individual frequency from the frequency list.
 */

double
GetIndividualFreq_PureTone_MultiPulse(int index)
{
	if (!CheckPars_PureTone_MultiPulse()) {
		NotifyError("GetIndividualFreq_PureTone_MultiPulse: Parameters "\
		  "have not been correctly\nset.  Returning zero.");
		return(0.0);
	}
	if (index < 0 || index >= pureTone4.numPulses) {
		NotifyError("GetIndividualFreq_PureTone_MultiPulse: The valid "\
		  "index range is 0 to %d.  Returning zero.\n", pureTone4.numPulses - 1);
		return(0.0);
	}
	return(pureTone4.frequencies[index]);
	
}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_PureTone_MultiPulse(void)
{
	int	i;
	
	if (!CheckPars_PureTone_MultiPulse()) {
		NotifyError("PrintPars_PureTone_MultiPulse: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("Multi-Frequency Pulsed Pure Tone Module Parameters:-\n");
	printf("\tIntensity = %g (Hz)\n", pureTone4.intensity);
	printf("\tDuration = %g ms,\tSampling interval = %g ms\n",
	  MSEC(pureTone4.duration), MSEC(pureTone4.dt));
	printf("\t%10s\t%s\n", "Pulse No.", "Frequency (Hz)");
	for (i = 0; i < pureTone4.numPulses; i++)
		printf("\t%10d\t%10g\n", i, pureTone4.frequencies[i]);
	printf("\tBegin silence = %g ms,\tPulse duration = %g ms\n",
	  MSEC(pureTone4.beginSilence), MSEC(pureTone4.pulseDuration));
	printf("\tRepetition period = %g ms.\n", MSEC(pureTone4.repetitionPeriod));

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_PureTone_MultiPulse(char *fileName)
{
	BOOLEAN	ok;
	int		i, numPulses;
	double  intensity, *frequencies, beginSilence, pulseDuration;
	double  duration, samplingInterval, repetitionPeriod;
    FILE    *fp;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_PureTone_MultiPulse: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_PureTone_MultiPulse: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
     if (!GetPars_ParFile(fp, "%d", &numPulses))
		ok = FALSE;
	if ((frequencies = (double *) calloc(sizeof(double), numPulses)) ==
	  NULL) {
		NotifyError("ReadPars_PureTone_MultiPulse: Cannot allocate memory "\
		  "for frequencies.");
		return(FALSE);
 	}
	for (i = 0; i < numPulses; i++)
		if (!GetPars_ParFile(fp, "%lf", &frequencies[i]))
			ok = FALSE;
   if (!GetPars_ParFile(fp, "%lf", &intensity))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &duration))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &beginSilence))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &pulseDuration))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &repetitionPeriod))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &samplingInterval))
		ok = FALSE;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_PureTone_MultiPulse: Not enough lines, or "\
		  "invalid parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_PureTone_MultiPulse(numPulses, intensity, frequencies, duration,
	  samplingInterval, beginSilence, pulseDuration, repetitionPeriod)) {
		NotifyError("ReadPars_PureTone_MultiPulse: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/********************************* GenerateSignal *****************************/

/*
 * This routine allocates memory for the output signal, if necessary, and
 * generates the signal into channel[0] of the signal data-set.
 * It checks that all initialisation has been correctly carried out by calling
 * the appropriate checking routines.
 * It can be called repeatedly with different parameter values if required.
 * Stimulus generation only sets the output signal, the input signal is not
 * used.
 */

BOOLEAN
GenerateSignal_PureTone_MultiPulse(EarObjectPtr data)
{
	BOOLEAN		pulseOn;
	int			pulseNumber;
	double		amplitude;
	register	ChanData	*dataPtr;
	ChanLen		i, beginIndex, pulseDurationIndex, repetitionPeriodIndex;
	ChanLen		pulseCount;

	if (data == NULL) {
		NotifyError("GenerateSignal_PureTone_MultiPulse: EarObject not "\
		  "initialised.");
		return(FALSE);
	}	
	if (!CheckPars_PureTone_MultiPulse())
		return(FALSE);
	SetProcessName_EarObject(data, "Multi-frequency pulsed pure-tone stimulus");
	if ( !InitOutSignal_EarObject(data, PURE_TONE_3_NUM_CHANNELS,
	  (ChanLen) (pureTone4.duration / pureTone4.dt), pureTone4.dt) ) {
		NotifyError("GenerateSignal_PureTone_MultiPulse: Cannot initialise "\
		  "output signal");
		return(FALSE);
	}
	amplitude = RMS_AMP(pureTone4.intensity) * SQRT_2;
	beginIndex = (ChanLen) (pureTone4.beginSilence /
	  data->outSignal->dt);
	pulseDurationIndex = (ChanLen) (pureTone4.pulseDuration / pureTone4.dt);
	repetitionPeriodIndex = (ChanLen) (pureTone4.repetitionPeriod /
	  pureTone4.dt);
	dataPtr = &data->outSignal->channel[0][beginIndex];
	for (i = 0, pulseCount = 0, pulseOn = FALSE, pulseNumber = 0; i <
	  data->outSignal->length - beginIndex; i++) {
		if (pulseOn) {
			*dataPtr = amplitude * sin(PIx2 *
			  pureTone4.frequencies[pulseNumber] * (pulseCount++ *
			  data->outSignal->dt));
			pulseOn = (pulseCount < pulseDurationIndex);
		} else {
			if (i % repetitionPeriodIndex == 0) {
				pulseOn = TRUE;
				pulseCount = 0;
				pulseNumber++;
				pulseNumber %= pureTone4.numPulses;
			}
		}
		dataPtr++;
	}
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

} /* GenerateSignal_PureTone_MultiPulse */

