/*******
 *
 * StMPTone.c
 *
 * This module contains the methods for the multiple pure-tone signal
 * generation paradigm.
 * This is a stimulus (St) module.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "StMPTone.h"
#include "FiParFile.h"

/******************************************************************************/
/********************************* Global variables ***************************/
/******************************************************************************/ 

MPureTone	mPureTone = {

				FALSE,		/* numPTonesFlag */
				FALSE,		/* durationFlag */
				FALSE,		/* dtFlag */
				0,			/* numPTones */
				0.0,		/* duration */
				0.0,		/* dt */
				NULL,		/* *intensities */
				NULL,		/* *frequencies */
				NULL		/* *phases */

			};

/******************************************************************************/
/********************************* Subroutines and functions ******************/
/******************************************************************************/

/********************************* CheckPars **********************************/

/*
 * This routine checks that the necessary parameters for the module have been
 * correctly initialised.
 * It also checks that the Nyquist critical frequency is not exceeded.
 * It returns TRUE if there are no problems.
 */
 
BOOLEAN
CheckPars_PureTone_Multi(void)
{
	BOOLEAN	ok;
	int		i;
	double	criticalFrequency;
	
	ok = TRUE;
	if (mPureTone.frequencies == NULL) {
		NotifyError("CheckPars_PureTone_Multi: Frequencies array not set.");
		ok = FALSE;
	}
	if (mPureTone.intensities == NULL) {
		NotifyError("CheckPars_PureTone_Multi: intensities array not set.");
		ok = FALSE;
	}
	if (mPureTone.phases == NULL) {
		NotifyError("CheckPars_PureTone_Multi: phases array not set.");
		ok = FALSE;
	}
	if (!mPureTone.numPTones) {
		NotifyError("CheckPars_PureTone_Multi: numPTones variable not set.");
		ok = FALSE;
	}
	if (!mPureTone.durationFlag) {
		NotifyError("CheckPars_PureTone_Multi: duration variable not set.");
		ok = FALSE;
	}
	if (!mPureTone.dtFlag) {
		NotifyError("CheckPars_PureTone_Multi: dtFlag variable not set.");
		ok = FALSE;
	}
	criticalFrequency = floor(1.0 / (2.0 * mPureTone.dt));
	if (ok)
		for (i = 0; i < mPureTone.numPTones; i++)
			if (criticalFrequency <= mPureTone.frequencies[i]) {
				NotifyError("CheckPars_PureTone_Multi: Sampling rate (dt = "\
				  "%g ms) is too low for one of the stimulus\nfrequencies "\
				  "(frequency[%d] = %g).",
				  MSEC(mPureTone.dt), i, mPureTone.frequencies[i]);
				ok = FALSE;
			} 
	return(ok);
	
}	

/********************************* SetNumPTones *******************************/

/*
 * This function sets the module's numPTones parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetNumPTones_PureTone_Multi(int theNumPTones)
{
	mPureTone.numPTonesFlag = TRUE;
	mPureTone.numPTones = theNumPTones;
	return(TRUE);

}

/********************************* SetFrequencies *****************************/

/*
 * This function sets up the module's frequencies array.
 * The pointer to the frequencies lists is passed to the module.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetFrequencies_PureTone_Multi(double *theFrequencies)
{

	mPureTone.frequencies = theFrequencies;
	return(TRUE);

}

/********************************* SetIntensities *****************************/

/*
 * This function sets up the module's intensities array.
 * The pointer to the intensities lists is passed to the module.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetIntensities_PureTone_Multi(double *theIntensities)
{

	mPureTone.intensities = theIntensities;
	return(TRUE);

}

/********************************* SetPhases **********************************/

/*
 * This function sets up the module's phases array.
 * The pointer to the phases lists is passed to the module.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetPhases_PureTone_Multi(double *thePhases)
{

	mPureTone.phases = thePhases;
	return(TRUE);

}

/********************************* SetDuration ********************************/

/*
 * This function sets the module's duration parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetDuration_PureTone_Multi(double theDuration)
{
	mPureTone.durationFlag = TRUE;
	mPureTone.duration = theDuration;
	return(TRUE);

}

/********************************* SetSamplingInterval ************************/

/*
 * This function sets the module's samplingInterval parameter.  It first checks
 * that the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetSamplingInterval_PureTone_Multi(double theSamplingInterval)
{
	if ( theSamplingInterval <= 0.0 ) {
		NotifyError("SetSamplingInterval_PureTone_Multi: Illegal sampling "\
		  "interval value = %g!", theSamplingInterval);
		return(FALSE);
	}
	mPureTone.dtFlag = TRUE;
	mPureTone.dt = theSamplingInterval;
	return(TRUE);

}

/********************************* SetIndividualFreq **************************/

/*
 * This function sets the individual frequency values.
 * It first checks if the frequencies have been set.
 */

BOOLEAN
SetIndividualFreq_PureTone_Multi(int theIndex, double theFrequency)
{
	if (mPureTone.frequencies == NULL) {
		NotifyError("SetIndividualFreq_PureTone_Multi: Frequencies not set.");
		return(FALSE);
	}
	if (theIndex > mPureTone.numPTones - 1) {
		NotifyError("SetIndividualFreq_PureTone_Multi: Index value must be in "\
		  "the\nrange 0 - %d (%d).\n", mPureTone.numPTones - 1, theIndex);
		return(FALSE);
	}
	mPureTone.frequencies[theIndex] = theFrequency;
	return(TRUE);

}

/********************************* SetIndividualIntensity *********************/

/*
 * This function sets the individual intensity values.
 * It first checks if the intensities have been set.
 */

BOOLEAN
SetIndividualIntensity_PureTone_Multi(int theIndex, double theIntensity)
{
	if (mPureTone.intensities == NULL) {
		NotifyError("SetIndividualIntensity_PureTone_Multi: Intensities not "\
		  "set.");
		return(FALSE);
	}
	if (theIndex > mPureTone.numPTones - 1) {
		NotifyError("SetIndividualIntensity_PureTone_Multi: Index value must "\
		  "be in the\nrange 0 - %d (%d).\n", mPureTone.numPTones - 1, theIndex);
		return(FALSE);
	}
	mPureTone.intensities[theIndex] = theIntensity;
	return(TRUE);

}

/********************************* SetIndividualPhase *************************/

/*
 * This function sets the individual phase values.
 * It first checks if the phases have been set.
 */

BOOLEAN
SetIndividualPhase_PureTone_Multi(int theIndex, double thePhase)
{
	if (mPureTone.phases == NULL) {
		NotifyError("SetIndividualPhase_PureTone_Multi: Intensities not "\
		  "set.");
		return(FALSE);
	}
	if (theIndex > mPureTone.numPTones - 1) {
		NotifyError("SetIndividualPhase_PureTone_Multi: Index value must be "\
		  "in the\nrange 0 - %d (%d).\n", mPureTone.numPTones - 1, theIndex);
		return(FALSE);
	}
	mPureTone.phases[theIndex] = thePhase;
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetPars_PureTone_Multi(int theNumPTones, double *theFrequencies,
  double *theIntensities, double *thePhases, double theDuration,
  double theSamplingInterval)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetNumPTones_PureTone_Multi(theNumPTones))
		ok = FALSE;
	if (!SetFrequencies_PureTone_Multi(theFrequencies))
		ok = FALSE;
	if (!SetIntensities_PureTone_Multi(theIntensities))
		ok = FALSE;
	if (!SetPhases_PureTone_Multi(thePhases))
		ok = FALSE;
	if (!SetDuration_PureTone_Multi(theDuration))
		ok = FALSE;
	if (!SetSamplingInterval_PureTone_Multi(theSamplingInterval))
		ok = FALSE;
	if (!ok)
		NotifyError("SetPars_PureTone_Multi: Failed to set all module "\
		  "parameters.");
	return(ok);
	
}

/****************************** GetIndividualFreq *****************************/

/*
 * This routine returns an individual frequency from the frequency list.
 */

double
GetIndividualFreq_PureTone_Multi(int index)
{
	if (!CheckPars_PureTone_Multi()) {
		NotifyError("GetIndividualFreq_PureTone_Multi: Parameters have not "\
		  "been correctly\nset.  Returning zero.");
		return(0.0);
	}
	if (index < 0 || index >= mPureTone.numPTones) {
		NotifyError("GetIndividualFreq_PureTone_Multi: The valid index range "\
		  "is 0 to %d.  Returning zero.\n", mPureTone.numPTones - 1);
		return(0.0);
	}
	return(mPureTone.frequencies[index]);
	
}

/****************************** GetIndividualIntensity ************************/

/*
 * This routine returns an individual intensity from the intensity list.
 */

double
GetIndividualIntensity_PureTone_Multi(int index)
{
	if (!CheckPars_PureTone_Multi()) {
		NotifyError("GetIndividualIntensity_PureTone_Multi: Parameters have "\
		  "not been correctly\nset.  Returning zero.");
		return(0.0);
	}
	if (index < 0 || index >= mPureTone.numPTones) {
		NotifyError("GetIndividualIntensity_PureTone_Multi: The valid index "\
		  "range is 0 to %d.  Returning zero.\n", mPureTone.numPTones - 1);
		return(0.0);
	}
	return(mPureTone.intensities[index]);
	
}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_PureTone_Multi(void)
{
	int		i;
	
	if (!CheckPars_PureTone_Multi()) {
		NotifyError("PrintPars_PureTone_Multi: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("Multiple Pure Tone Module Parameters:-\n");
	printf("\t%10s\t%10s\t%10s\n", "Frequency ", "Intensity", "   Phase  ");
	printf("\t%10s\t%10s\t%10s\n", "   (Hz)   ", " (dB SPL)", " (degrees)");
	for (i = 0; i < mPureTone.numPTones; i++)
		printf("\t%10g\t%10g\t%10g\n", mPureTone.frequencies[i],
		  mPureTone.intensities[i], mPureTone.phases[i]);
	printf("\tDuration = %g ms,\tSampling interval = %g ms\n",
	  MSEC(mPureTone.duration), MSEC(mPureTone.dt));

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_PureTone_Multi(char *fileName)
{
	BOOLEAN	ok;
    int		i, numberOfPTones;
	double  *frequencies, *intensities, *phases, duration, samplingInterval;
	FILE    *fp;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_PureTone_Multi: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_PureTone_Multi: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
    if (!GetPars_ParFile(fp, "%d", &numberOfPTones))
		ok = FALSE;
	if ((frequencies = (double *) calloc(sizeof(double), numberOfPTones)) ==
	  NULL) {
		NotifyError("ReadParsFromFile: Cannot allocate memory for frequencies.");
		return(FALSE);
 	}
	if ((intensities = (double *) calloc(sizeof(double), numberOfPTones)) ==
	  NULL) {
		NotifyError("ReadParsFromFile: Cannot allocate memory for intensities.");
		return(FALSE);
 	}
	if ((phases = (double *) calloc(sizeof(double), numberOfPTones)) == NULL) {
		NotifyError("ReadParsFromFile: Cannot allocate memory for phases.");
		return(FALSE);
 	}
	for (i = 0; i < numberOfPTones; i++)
		if (!GetPars_ParFile(fp, "%lf %lf %lf", &frequencies[i], &intensities[i],
		  &phases[i]))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &duration))
		ok = FALSE;
	if (!GetPars_ParFile(fp, "%lf", &samplingInterval))
		ok = FALSE;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_PureTone_Multi: Not enough lines, or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_PureTone_Multi(numberOfPTones, frequencies, intensities,
	  phases, duration, samplingInterval)) {
		NotifyError("ReadPars_PureTone_Multi: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/********************************* GenerateSignal *****************************/

/*
 * This routine allocates memory for the output signal, if necessary, and
 * generates the signal into channel[0] of the signal data-set.
 * It checks that all initialisation has been correctly carried out by calling
 * the appropriate checking routines.
 * It can be called repeatedly with different parameter values if required.
 * Stimulus generation only sets the output signal, the input signal is not
 * used.
 * With repeated calls the Signal memory is only allocated once, then re-used.
 */

BOOLEAN
GenerateSignal_PureTone_Multi(EarObjectPtr data)
{
	int			j;
	ChanLen		i;
	register	double		amplitude, phase;
	register	ChanData	*dataPtr;

	if (data == NULL) {
		NotifyError("GenerateSignal_PureTone_Multi: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_PureTone_Multi())
		return(FALSE);
	SetProcessName_EarObject(data, "Multiple pure tone stimulus");
	if ( !InitOutSignal_EarObject(data, PURE_TONE_NUM_CHANNELS,
	  (ChanLen) (mPureTone.duration / mPureTone.dt), mPureTone.dt) ) {
		NotifyError("GenerateSignal_PureTone_Multi: Cannot initialise output "\
		  "signal.");
		return(FALSE);
	}
	for (j = 0; j < mPureTone.numPTones; j++) {
		amplitude = RMS_AMP(mPureTone.intensities[j]) * SQRT_2;
		phase = mPureTone.phases[j] * PIx2 / 360.0;
		dataPtr = data->outSignal->channel[0];
		for (i = 0; i < data->outSignal->length; i++)
			*(dataPtr++) += (ChanData) (amplitude * sin(PIx2 *
			   mPureTone.frequencies[j] * (i * data->outSignal->dt) + phase));
	}
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

} /* GenerateSignal_PureTone_Multi */

