/*******
 *
 * StPTone.c
 *
 * This module contains the methods for the simple pure-tone signal generation
 * paradigm.
 * This is a stimulus (St) module.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "StPTone.h"
#include "FiParFile.h"

/********************************* Global variables ***************************/

PureTone	pureTone = {

				FALSE,		/* frequencyFlag */
				FALSE,		/* durationFlag */
				FALSE,		/* dtFlag */
				FALSE,		/* intensityFlag */
				0.0,		/* frequency */
				0.0,		/* intensity */
				0.0,		/* duration */
				0.0			/* dt */

			};

/******************************************************************************/
/********************************* Subroutines and functions ******************/
/******************************************************************************/

/********************************* CheckPars **********************************/

/*
 * This routine checks that the necessary parameters for the module have been
 * correctly initialised.
 * It also checks that the Nyquist critical frequency is not exceeded.
 * It returns TRUE if there are no problems.
 */
 
BOOLEAN
CheckPars_PureTone(void)
{
	BOOLEAN	ok;
	double	criticalFrequency;
	
	ok = TRUE;
	if (!pureTone.frequencyFlag) {
		NotifyError("CheckPars_PureTone: Frequency variable not set.");
		ok = FALSE;
	}
	if (!pureTone.intensityFlag) {
		NotifyError("CheckPars_PureTone: intensity variable not set.");
		ok = FALSE;
	}
	if (!pureTone.durationFlag) {
		NotifyError("CheckPars_PureTone: duration variable not set.");
		ok = FALSE;
	}
	if (!pureTone.dtFlag) {
		NotifyError("CheckPars_PureTone: dt variable not set.");
		ok = FALSE;
	}
	criticalFrequency = floor(1.0 / (2.0 * pureTone.dt));
	if (ok && (criticalFrequency <= pureTone.frequency)) {
		NotifyError("CheckPars_PureTone: Sampling rate (dt = %g ms) is " \
		  "too low for the frequency.", MSEC(pureTone.dt));
		ok = FALSE;
	} 
	return(ok);
	
}	

/********************************* SetFrequency *******************************/

/*
 * This function sets the module's frequency parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetFrequency_PureTone(double theFrequency)
{
	pureTone.frequencyFlag = TRUE;
	pureTone.frequency = theFrequency;
	return(TRUE);

}

/********************************* SetIntensity *******************************/

/*
 * This function sets the module's intensity parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetIntensity_PureTone(double theIntensity)
{
	pureTone.intensityFlag = TRUE;
	pureTone.intensity = theIntensity;
	return(TRUE);

}

/********************************* SetDuration ********************************/

/*
 * This function sets the module's duration parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetDuration_PureTone(double theDuration)
{
	pureTone.durationFlag = TRUE;
	pureTone.duration = theDuration;
	return(TRUE);

}

/********************************* SetSamplingInterval ************************/

/*
 * This function sets the module's samplingInterval parameter.  It first checks
 * that the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetSamplingInterval_PureTone(double theSamplingInterval)
{
	if ( theSamplingInterval <= 0.0 ) {
		NotifyError("SetSamplingInterval_PureTone: Illegal sampling interval "\
		  "value = %g!", theSamplingInterval);
		return(FALSE);
	}
	pureTone.dtFlag = TRUE;
	pureTone.dt = theSamplingInterval;
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetPars_PureTone(double theFrequency, double theIntensity, double theDuration,
  double theSamplingInterval)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetFrequency_PureTone(theFrequency))
		ok = FALSE;
	if (!SetIntensity_PureTone(theIntensity))
		ok = FALSE;
	if (!SetDuration_PureTone(theDuration))
		ok = FALSE;
	if (!SetSamplingInterval_PureTone(theSamplingInterval))
		ok = FALSE;
	if (!ok)
		NotifyError("SetPars_PureTone: Failed to set all module parameters.");
	return(ok);
	
}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_PureTone(void)
{
	if (!CheckPars_PureTone()) {
		NotifyError("PrintPars_PureTone: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("Pure Tone Module Parameters:-\n");
	printf("\tFrequency = %g Hz,\tIntensity = %g dB SPL,\n",
	  pureTone.frequency, pureTone.intensity);
	printf("\tDuration = %g ms,\tSampling interval = %g ms\n",
	  MSEC(pureTone.duration), MSEC(pureTone.dt));

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_PureTone(char *fileName)
{
	BOOLEAN	ok;
	double  frequency, intensity;
	double  duration, samplingInterval;
    FILE    *fp;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_PureTone: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_PureTone: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
    if (!GetPars_ParFile(fp, "%lf", &frequency))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &intensity))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &duration))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &samplingInterval))
		ok = FALSE;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_PureTone: Not enough lines, or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_PureTone(frequency, intensity, duration,
	  samplingInterval)) {
		NotifyError("ReadPars_PureTone: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/********************************* GenerateSignal *****************************/

/*
 * This routine allocates memory for the output signal, if necessary, and
 * generates the signal into channel[0] of the signal data-set.
 * It checks that all initialisation has been correctly carried out by calling
 * the appropriate checking routines.
 * It can be called repeatedly with different parameter values if required.
 * Stimulus generation only sets the output signal, the input signal is not
 * used.
 * With repeated calls the Signal memory is only allocated once, then re-used.
 */

BOOLEAN
GenerateSignal_PureTone(EarObjectPtr data)
{
	ChanLen		i;
	register	double		amplitude;
	register	ChanData	*dataPtr;

	if (data == NULL) {
		NotifyError("GenerateSignal_PureTone: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_PureTone())
		return(FALSE);
	SetProcessName_EarObject(data, "Pure Tone stimulus");
	if ( !InitOutSignal_EarObject(data, PURE_TONE_NUM_CHANNELS,
	  (ChanLen) (pureTone.duration / pureTone.dt), pureTone.dt) ) {
		NotifyError("GenerateSignal_PureTone: Cannot initialise output signal");
		return(FALSE);
	}
	amplitude = RMS_AMP(pureTone.intensity) * SQRT_2;
	dataPtr = data->outSignal->channel[0];
	for (i = 0; i < data->outSignal->length; i++)
		*(dataPtr++) = amplitude * sin(PIx2 * pureTone.frequency *
				  (i * data->outSignal->dt));
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

} /* GenerateSignal_PureTone */

