/*******
 *
 * StPTone2.c
 *
 * This module contains the Silence Pure-Tone stimulus generation paradigm.
 * The module generates a pure-tone signal preceded and ended by periods of
 * silence.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "FiParFile.h"
#include "StPTone2.h"

/********************************* Global variables ***************************/

PureTone2	pureTone2 = {

				FALSE,		/* frequencyFlag */
				FALSE,		/* durationFlag */
				FALSE,		/* dtFlag */
				FALSE,		/* intensityFlag */
				FALSE,		/* beginSilenceFlag */
				FALSE,		/* endSilenceFlag */
				0.0,		/* frequency */
				0.0,		/* intensity */
				0.0,		/* duration */
				0.0,		/* dt */
				0.0,		/* beginSilence */
				0.0			/* endSilence */

			};

/******************************************************************************/
/********************************* Subroutines and functions ******************/
/******************************************************************************/

/********************************* CheckPars **********************************/

/*
 * This routine checks that the necessary parameters for the module have been
 * correctly initialised.
 * It checks that the Nyquist critical frequency is not exceeded.
 * It returns TRUE if there are no problems.
 */
 
BOOLEAN
CheckPars_PureTone_2(void)
{
	BOOLEAN	ok;
	double	criticalFrequency;
	
	ok = TRUE;
	if (!pureTone2.frequencyFlag) {
		NotifyError("CheckPars_PureTone_2: Frequency variable not set.");
		ok = FALSE;
	}
	if (!pureTone2.intensityFlag) {
		NotifyError("CheckPars_PureTone_2: intensity variable not set.");
		ok = FALSE;
	}
	if (!pureTone2.durationFlag) {
		NotifyError("CheckPars_PureTone_2: duration variable not set.");
		ok = FALSE;
	}
	if (!pureTone2.beginSilenceFlag) {
		NotifyError("CheckPars_PureTone_2: beginSilence variable not set.");
		ok = FALSE;
	}
	if (!pureTone2.endSilenceFlag) {
		NotifyError("CheckPars_PureTone_2: endSilence variable not set.");
		ok = FALSE;
	}
	if (!pureTone2.dtFlag) {
		NotifyError("CheckPars_PureTone_2: dt variable not set.");
		ok = FALSE;
	}
	if ( (pureTone2.beginSilence > pureTone2.duration) ||
	  (pureTone2.endSilence > pureTone2.duration) ) {
	  	NotifyError("CheckPars_PureTone_2: A silence length parameters is "\
	  	  "longer than the signal duration.");
	  	ok = FALSE;
	}
	criticalFrequency = floor(1.0 / (2.0 * pureTone2.dt));
	if (ok && (criticalFrequency <= pureTone2.frequency)) {
		NotifyError("CheckPars_PureTone_2: Sampling rate (dt = %g ms) is " \
		  "too low for the frequency.", MSEC(pureTone2.dt));
		ok = FALSE;
	} 
	return(ok);

}	

/********************************* SetFrequency *******************************/

/*
 * This function sets the module's frequency parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetFrequency_PureTone_2(double theFrequency)
{
	pureTone2.frequencyFlag = TRUE;
	pureTone2.frequency = theFrequency;
	return(TRUE);

}

/********************************* SetIntensity *******************************/

/*
 * This function sets the module's intensity parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetIntensity_PureTone_2(double theIntensity)
{
	pureTone2.intensityFlag = TRUE;
	pureTone2.intensity = theIntensity;
	return(TRUE);

}

/********************************* SetDuration ********************************/

/*
 * This function sets the module's duration parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetDuration_PureTone_2(double theDuration)
{
	pureTone2.durationFlag = TRUE;
	pureTone2.duration = theDuration;
	return(TRUE);

}

/********************************* SetBeginSilence ****************************/

/*
 * This function sets the module's beginSilence parameter.  It first checks
 * that the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetBeginSilence_PureTone_2(double theBeginSilence)
{
	pureTone2.beginSilenceFlag = TRUE;
	pureTone2.beginSilence = theBeginSilence;
	return(TRUE);

}

/********************************* SetEndSilence ******************************/

/*
 * This function sets the module's endSilence parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetEndSilence_PureTone_2(double theEndSilence)
{
	pureTone2.endSilenceFlag = TRUE;
	pureTone2.endSilence = theEndSilence;
	return(TRUE);

}

/********************************* SetSamplingInterval ************************/

/*
 * This function sets the module's timeStep parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetSamplingInterval_PureTone_2(double theSamplingInterval)
{
	if ( theSamplingInterval <= 0.0 ) {
		NotifyError("SetSamplingInterval_PureTone_2: Illegal sampling "\
		  "interval value = %g!", theSamplingInterval);
		return(FALSE);
	}
	pureTone2.dtFlag = TRUE;
	pureTone2.dt = theSamplingInterval;
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetPars_PureTone_2(double theFrequency, double theIntensity, double theDuration,
  double theSamplingInterval, double theBeginSilence, double theEndSilence)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetFrequency_PureTone_2(theFrequency))
		ok = FALSE;
	if (!SetIntensity_PureTone_2(theIntensity))
		ok = FALSE;
	if (!SetDuration_PureTone_2(theDuration))
		ok = FALSE;
	if (!SetSamplingInterval_PureTone_2(theSamplingInterval))
		ok = FALSE;
	if (!SetBeginSilence_PureTone_2(theBeginSilence))
		ok = FALSE;
	if (!SetEndSilence_PureTone_2(theEndSilence))
		ok = FALSE;
	if (!ok)
		NotifyError("SetPars_PureTone_2: Failed to set all module parameters.");
	return(ok);
	
}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_PureTone_2(void)
{
	if (!CheckPars_PureTone_2()) {
		NotifyError("PrintPars_PureTone_2: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("Pure Tone 2 (surrounded by silence) Module Parameters:-\n");
	printf("\tFrequency = %g (Hz),\tIntensity = %g dB SPL,\n",
	  pureTone2.frequency, pureTone2.intensity);
	printf("\tDuration = %g ms,\tSampling interval = %g ms\n",
	  MSEC(pureTone2.duration), MSEC(pureTone2.dt));
	printf("\tBegin silence = %g ms,\tEnd silence = %g ms\n",
	  MSEC(pureTone2.beginSilence), MSEC(pureTone2.endSilence));

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_PureTone_2(char *fileName)
{
	BOOLEAN	ok;
	double  frequency, intensity, beginSilence, endSilence;
	double  duration, samplingInterval;
    FILE    *fp;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_PureTone_2: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_PureTone_2: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
    if (!GetPars_ParFile(fp, "%lf", &frequency))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &intensity))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &duration))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &beginSilence))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &endSilence))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &samplingInterval))
		ok = FALSE;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_PureTone_2: Not enough lines, or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_PureTone_2(frequency, intensity, duration,
	  samplingInterval, beginSilence, endSilence)) {
		NotifyError("ReadPars_PureTone_2: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/********************************* GenerateSignal *****************************/

/*
 * This routine allocates memory for the output signal, if necessary, and
 * generates the signal into channel[0] of the signal data-set.
 * It checks that all initialisation has been correctly carried out by calling
 * the appropriate checking routines.
 * It can be called repeatedly with different parameter values if required.
 * Stimulus generation only sets the output signal, the input signal is not
 * used.
 */

BOOLEAN
GenerateSignal_PureTone_2(EarObjectPtr data)
{
	double		totalDuration, amplitude;
	register	ChanData	*dataPtr;
	ChanLen		i, beginSilenceIndex, endSignalIndex;

	if (data == NULL) {
		NotifyError("GenerateSignal_PureTone_2: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_PureTone_2())
		return(FALSE);
	SetProcessName_EarObject(data, "Silence pure-tone stimulus");
	totalDuration = pureTone2.beginSilence + pureTone2.duration +
	  pureTone2.endSilence;
	if ( !InitOutSignal_EarObject(data, PURE_TONE_2_NUM_CHANNELS,
	  (ChanLen) (totalDuration / pureTone2.dt), pureTone2.dt) ) {
		NotifyError("GenerateSignal_PureTone_2: Cannot initialise output "\
		  "signal");
		return(FALSE);
	}
	beginSilenceIndex = (ChanLen) (pureTone2.beginSilence /
	  data->outSignal->dt);
	endSignalIndex = beginSilenceIndex + (ChanLen) (pureTone2.duration /
	  pureTone2.dt);
	amplitude = RMS_AMP(pureTone2.intensity) * SQRT_2;
	dataPtr = data->outSignal->channel[0];
	for (i = 0; i < data->outSignal->length; i++)
		if ( (i > beginSilenceIndex) && (i <= endSignalIndex) )
			*(dataPtr++) = amplitude * sin(PIx2 * pureTone2.frequency *
			  ((i - beginSilenceIndex - 1) * data->outSignal->dt));
		else
			*(dataPtr++) = 0.0;
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

} /* GenerateSignal_PureTone_2 */

