/*******
 *
 * StStepFun.c
 *
 * This module contains the step function stimulus generation paradigm.
 * The module generates a constant signal preceded and ended by periods of
 * silence.
 * The magnitude of the step function has no units, i.e. they must be implicitly
 * defined by the user (Arbitrary units).
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ********/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "FiParFile.h"
#include "StStepFun.h"

/********************************* Global variables ***************************/

StepFunction	stepFunction = {

					FALSE,		/* durationFlag */
					FALSE,		/* dtFlag */
					FALSE,		/* amplitudeFlag */
					FALSE,		/* beginSilenceFlag */
					FALSE,		/* endSilenceFlag */
					0.0,		/* amplitude */
					0.0,		/* duration */
					0.0,		/* dt */
					0.0,		/* beginSilence */
					0.0			/* endSilence */

				};

/******************************************************************************/ 
/********************************* Subroutines and functions ******************/
/******************************************************************************/

/********************************* CheckPars **********************************/

/*
 * This routine checks that the necessary parameters for the module have been
 * correctly initialised.
 * It returns TRUE if there are no problems.
 */
 
BOOLEAN
CheckPars_StepFunction(void)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!stepFunction.amplitudeFlag) {
		NotifyError("CheckPars_StepFunction: amplitude variable not set.");
		ok = FALSE;
	}
	if (!stepFunction.durationFlag) {
		NotifyError("CheckPars_StepFunction: duration variable not set.");
		ok = FALSE;
	}
	if (!stepFunction.beginSilenceFlag) {
		NotifyError("CheckPars_StepFunction: beginSilence variable not set.");
		ok = FALSE;
	}
	if (!stepFunction.endSilenceFlag) {
		NotifyError("CheckPars_StepFunction: endSilence variable not set.");
		ok = FALSE;
	}
	if (!stepFunction.dtFlag) {
		NotifyError("CheckPars_StepFunction: dtFlag variable not set.");
		ok = FALSE;
	}
	if ( ok && ((stepFunction.beginSilence > stepFunction.duration) ||
	  (stepFunction.endSilence > stepFunction.duration)) ) {
	  	NotifyError("CheckPars_StepFunction: A silence length parameters is "\
	  	  "longer than the signal duration.");
	  	ok = FALSE;
	}
	return(ok);

}	

/********************************* SetAmplitude *******************************/

/*
 * This function sets the module's amplitude parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetAmplitude_StepFunction(double theAmplitude)
{
	stepFunction.amplitudeFlag = TRUE;
	stepFunction.amplitude = theAmplitude;
	return(TRUE);

}

/********************************* SetDuration ********************************/

/*
 * This function sets the module's duration parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetDuration_StepFunction(double theDuration)
{
	stepFunction.durationFlag = TRUE;
	stepFunction.duration = theDuration;
	return(TRUE);

}

/********************************* SetBeginSilence ****************************/

/*
 * This function sets the module's beginSilence parameter.  It first checks
 * that the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetBeginSilence_StepFunction(double theBeginSilence)
{
	stepFunction.beginSilenceFlag = TRUE;
	stepFunction.beginSilence = theBeginSilence;
	return(TRUE);

}

/********************************* SetEndSilence ******************************/

/*
 * This function sets the module's endSilence parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetEndSilence_StepFunction(double theEndSilence)
{
	stepFunction.endSilenceFlag = TRUE;
	stepFunction.endSilence = theEndSilence;
	return(TRUE);

}

/********************************* SetSamplingInterval ************************/

/*
 * This function sets the module's timeStep parameter.  It first checks that
 * the module has been initialised.
 * It returns TRUE if the operation is successful.
 */

BOOLEAN
SetSamplingInterval_StepFunction(double theSamplingInterval)
{
	if ( theSamplingInterval <= 0.0 ) {
		NotifyError("SetSamplingInterval_StepFunction: Illegal sampling "\
		  "interval value = %g!", theSamplingInterval);
		return(FALSE);
	}
	stepFunction.dtFlag = TRUE;
	stepFunction.dt = theSamplingInterval;
	return(TRUE);

}

/********************************* SetPars ************************************/

/*
 * This function sets all the module's parameters.
 * It returns TRUE if the operation is successful.
 */
 
BOOLEAN
SetPars_StepFunction(double theAmplitude, double theDuration,
  double theSamplingInterval, double theBeginSilence, double theEndSilence)
{
	BOOLEAN	ok;
	
	ok = TRUE;
	if (!SetAmplitude_StepFunction(theAmplitude))
		ok = FALSE;
	if (!SetDuration_StepFunction(theDuration))
		ok = FALSE;
	if (!SetSamplingInterval_StepFunction(theSamplingInterval))
		ok = FALSE;
	if (!SetBeginSilence_StepFunction(theBeginSilence))
		ok = FALSE;
	if (!SetEndSilence_StepFunction(theEndSilence))
		ok = FALSE;
	if (!ok)
		NotifyError("SetPars_StepFunction: Failed to set all module "\
		  "parameters.");
	return(ok);
	
}

/****************************** PrintPars *************************************/

/*
 * This program prints the parameters of the module to the standard output.
 */
 
void
PrintPars_StepFunction(void)
{
	if (!CheckPars_StepFunction()) {
		NotifyError("PrintPars_StepFunction: Parameters have not been "\
		  "correctly set.");
		return;
	}
	printf("Step Function Module Parameters:-\n");
	printf("\tAmplitude = %g,\t\tDuration = %g ms,\n", stepFunction.amplitude,
	  MSEC(stepFunction.duration));
	printf("\tBegin silence = %g ms,\tEnd silence = %g ms\n",
	  MSEC(stepFunction.beginSilence), MSEC(stepFunction.endSilence));
	printf("\tSampling interval = %g ms\n", MSEC(stepFunction.dt));

}

/****************************** ReadPars **************************************/

/*
 * This program reads a specified number of parameters from a file.
 * It returns FALSE if it fails in any way.
 */
 
BOOLEAN
ReadPars_StepFunction(char *fileName)
{
	BOOLEAN	ok;
	double  amplitude, beginSilence, endSilence;
	double  duration, samplingInterval;
    FILE    *fp;
    
    if ((fp = fopen(fileName, "r")) == NULL) {
        NotifyError("ReadPars_StepFunction: Cannot open data file '%s'.\n",
		  fileName);
		return(FALSE);
    }
    NotifyWarning("ReadPars_StepFunction: Reading from '%s'", fileName);
    Init_ParFile();
	ok = TRUE;
    if (!GetPars_ParFile(fp, "%lf", &amplitude))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &duration))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &beginSilence))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &endSilence))
		ok = FALSE;
    if (!GetPars_ParFile(fp, "%lf", &samplingInterval))
		ok = FALSE;
    fclose(fp);
    Free_ParFile();
	if (!ok) {
		NotifyError("ReadPars_StepFunction: Not enough lines, or invalid "\
		  "parameters, in module parameter file '%s'.", fileName);
		return(FALSE);
	}
	if (!SetPars_StepFunction(amplitude, duration, samplingInterval,
	  beginSilence, endSilence)) {
		NotifyError("ReadPars_StepFunction: Could not set parameters.");
		return(FALSE);
	}
	return(TRUE);
    
}

/********************************* GenerateSignal *****************************/

/*
 * This routine allocates memory for the output signal, if necessary, and
 * generates the signal into channel[0] of the signal data-set.
 * It checks that all initialisation has been correctly carried out by calling
 * the appropriate checking routines.
 * It can be called repeatedly with different parameter values if required.
 * Stimulus generation only sets the output signal, the input signal is not
 * used.
 */

BOOLEAN
GenerateSignal_StepFunction(EarObjectPtr data)
{
	double		totalDuration;
	register	ChanData	*dataPtr;
	ChanLen		i, beginSilenceIndex, endSignalIndex;

	if (data == NULL) {
		NotifyError("GenerateSignal_StepFunction: EarObject not initialised.");
		return(FALSE);
	}	
	if (!CheckPars_StepFunction())
		return(FALSE);
	SetProcessName_EarObject(data, "Silence pure-tone stimulus");
	totalDuration = stepFunction.beginSilence + stepFunction.duration +
	  stepFunction.endSilence;
	if ( !InitOutSignal_EarObject(data, PURE_TONE_2_NUM_CHANNELS,
	  (ChanLen) (totalDuration / stepFunction.dt), stepFunction.dt) ) {
		NotifyError("GenerateSignal_StepFunction: Cannot initialise output "\
		  "signal");
		return(FALSE);
	}
	beginSilenceIndex = (ChanLen) (stepFunction.beginSilence /
	  data->outSignal->dt);
	endSignalIndex = beginSilenceIndex + (ChanLen) (stepFunction.duration /
	  stepFunction.dt);
	dataPtr = data->outSignal->channel[0];
	for (i = 0; i < data->outSignal->length; i++)
		if ( (i > beginSilenceIndex) && (i <= endSignalIndex) )
			*(dataPtr++) = stepFunction.amplitude;
		else
			*(dataPtr++) = 0.0;
	data->outSignal->rampFlag = TRUE;	/* Step functions should not be ramped. */
    if (data->updateCustomersFlag)
    	UpdateCustomers_EarObject(data);
	return(TRUE);

} /* GenerateSignal_StepFunction */

