/**************
 *
 * UtCFList.c
 *
 * This file contains the centre frequency management routines.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 **************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "LUTEar.h"
#include "UtERB.h"
#include "UtCFList.h"
#include "FiParFile.h"

/******************************************************************************/
/****************************** Global Variables ******************************/
/******************************************************************************/


/******************************************************************************/
/****************************** Subroutines & functions ***********************/
/******************************************************************************/

/****************************** Mode_CFList ***********************************/

/*
 * This routine returns the CFList mode, as defined by the string argument.
 * It returns -1 if it fails to find a valid mode.
 */

int
Mode_CFList(char *modeName)
{
	char	*p;
	int		i;
	static char	*modeCFList[] = {
			"SINGLE",
			"ERB",
			"LOG",
			""
			};
	
	for (p = modeName; *p != '\0'; p++) /* Convert to upper case */
		*p = toupper(*p);		
	for (i = 0; strcmp(modeName, modeCFList[i]) != 0 && modeCFList[i] != '\0';
	  i++);
	if (modeCFList[i][0] == '\0')
		return(-1);
	return(i);

}

/****************************** Init ******************************************/

/*
 * This function allocates memory for the centre frequency data structure.
 * The function returns a pointer to the prepared structure.
 */
 
CFListPtr
Init_CFList(char *callingFunctionName)
{
	CFList *theCFList;
	
	if ((theCFList = (CFList *) malloc(sizeof(CFList))) == NULL ) {
		NotifyError("Init_CFList: Out of Memory (called by %s).",
		  callingFunctionName);
		return(NULL);
	}
	theCFList->numChannels = 0;
	theCFList->mode = 0;
	return(theCFList);
	
}


/****************************** CheckInit *************************************/

/*
 * This routine checks whether or not a CFList has been initialised.
 *
 */
 
BOOLEAN
CheckInit_CFList(CFListPtr theCFList, char *callingFunction)
{
	if (theCFList != NULL)
		return(TRUE);
	else {
		NotifyError("CheckInit_CFList: Signal not set in %s.", callingFunction);
		return(FALSE);
	}
	
}

/**************************** CheckPars ***************************************/

/*
 * This routine checks that the necessary parameters for a CFList have been
 * correctly initialised.
 * It returns TRUE if there are no problems.
 */

BOOLEAN
CheckPars_CFList(CFList *theCFList)
{
	int	i;
	
	if (!CheckInit_CFList(theCFList, "CheckPars_CFList"))
		return(FALSE);
	if (theCFList->numChannels <= 0) {
		NotifyError("CheckPars_CFList: Number of Channels not correctly set.");
		return(FALSE);
	}
	for (i = 0; i < theCFList->numChannels; i++)
		if (theCFList->channel[i] < 0.0) {
			NotifyError("CheckPars_CFList: Invalid frequency[%d] = %g Hz.", i,
			  theCFList->channel[i]);
			return(FALSE);
		}
	return(TRUE);
	
}

/****************************** GenerateERB ***********************************/

/*
 * This function generates the centre frequency list from the frequency
 * range and ERB density.
 */

CFListPtr
GenerateERB_CFList(double minCF, double maxCF, double ERBDensity)
{
	int		i, numberOfCFs;
	double	theERBRate;
	CFListPtr	theCFs;

	numberOfCFs = (int) ceil((ERBRateFromF_ERB(maxCF) - 
	  ERBRateFromF_ERB(minCF)) * ERBDensity);
	if (numberOfCFs <= 0) {
		NotifyError("GenerateERB_CFList: Illegal frequency range %g - %g, " \
		  "or ERB density, %g.", minCF, maxCF, ERBDensity);
		return(NULL);
	}
	if ((theCFs = (CFList *) malloc(sizeof(CFList))) == NULL) {
		NotifyError("GenerateERB_CFList: Out of memory.");
		return(NULL);
	}
	theCFs->numChannels = numberOfCFs;
	for (i = 0, theERBRate = ERBRateFromF_ERB(minCF); i < numberOfCFs; i++) {
		theCFs->channel[i] = FFromERBRate_ERB(theERBRate);
		theERBRate += 1.0 / ERBDensity;
	}
	return(theCFs);

}

/****************************** GenerateLog ***********************************/

/*
 * This function generates a logarithmic centre frequency list from the
 * frequency range and the number of filters.
 */

CFListPtr
GenerateLog_CFList(double minCF, double maxCF, int numberOfCFs)
{
	int		i;
	double	theLogRate, logMinCF;
	CFListPtr	theCFs;

	if (maxCF <= minCF) {
		NotifyError("GenerateLog_CFList: Illegal frequency range %g - %g.",
		  minCF, maxCF);
		return(NULL);
	}
	if (numberOfCFs < 1) {
		NotifyError("GenerateLog_CFList: Illegal no. of centre frequencies.");
		return(NULL);
	}
	if ((theCFs = (CFList *) malloc(sizeof(CFList))) == NULL) {
		NotifyError("GenerateLog_CFList: Out of memory.");
		return(NULL);
	}
	theCFs->numChannels = numberOfCFs;
	logMinCF = log10(minCF);
	theLogRate = (log10(maxCF) - logMinCF) / (numberOfCFs - 1);
	for (i = 0; i < numberOfCFs; i++)
		theCFs->channel[i] = pow(10.0, logMinCF + theLogRate * i);
	return(theCFs);

}

/****************************** ReadPars *************************************/

/*
 * This routine routines the CFList parameters from a file.
 * It returns with the address of the CFList, or NULL if it fails.
 */

CFListPtr
Read_CFList(FILE *fp, char *modeName)
{
	BOOLEAN	ok;
	int		mode, numberOfCFs;
	double	singleFrequency, lowestCFFrequency, highestCFFrequency, eRBDensity;
	CFListPtr	theCFs;
	
	theCFs = NULL;
	ok = TRUE;
	switch (mode = Mode_CFList(modeName)) {
	case	CFLIST_SINGLE_MODE:
		if (!GetPars_ParFile(fp, "%lf", &singleFrequency))
			ok = FALSE;
		break;
	case	CFLIST_ERB_MODE:
		if (!GetPars_ParFile(fp, "%lf", &lowestCFFrequency))
			ok = FALSE;
		if (!GetPars_ParFile(fp, "%lf", &highestCFFrequency))
			ok = FALSE;
		if (!GetPars_ParFile(fp, "%lf", &eRBDensity))
			ok = FALSE;
		break;
	case	CFLIST_LOG_MODE:
		if (!GetPars_ParFile(fp, "%lf", &lowestCFFrequency))
			ok = FALSE;
		if (!GetPars_ParFile(fp, "%lf", &highestCFFrequency))
			ok = FALSE;
		if (!GetPars_ParFile(fp, "%d", &numberOfCFs))
			ok = FALSE;
		break;
	default	:
		NotifyError("ReadPars_BasilarM_GammaT: Unknown frequency mode.");
		ok = FALSE;
		break;
	} /* Switch */
	if (!ok) {
		NotifyError("Read_CFList: Failed to read centre frequency list "\
		  "parameters.");
		return(NULL);
	}
	switch (mode) {
	case	CFLIST_SINGLE_MODE:
		if ((theCFs = Init_CFList("Read_CFList")) == NULL) {
			NotifyError("Read_CFList: Could not initialise CF list "\
			  "for single frequency mode.");
			return(NULL);
		}
		theCFs->channel[0] = singleFrequency;
		theCFs->numChannels = 1;
		theCFs->mode = CFLIST_SINGLE_MODE;
		break;
	case	CFLIST_ERB_MODE:
		if ((theCFs = GenerateERB_CFList(lowestCFFrequency,
		  highestCFFrequency, eRBDensity)) == NULL) {
			NotifyError("Read_CFList: Could not set ERB centre "\
			  "frequency list.");
			return(NULL);
		}
		break;
	case	CFLIST_LOG_MODE:
		if ((theCFs = GenerateLog_CFList(lowestCFFrequency,
		  highestCFFrequency, numberOfCFs)) == NULL) {
			NotifyError("Read_CFList: Could not set log centre "\
			  "frequency list.");
			return(NULL);
		}
		break;
	} /* Switch */
	return(theCFs);

}
