/*************
 *
 * UtFilters.c
 *
 * This module contains all the files that initialise and generate the
 * 2-pole GammaTone filters, the 2-pole Butterworth low pass filter, the
 * Bessel low pass filter and the non-linear compresion filter.  They were
 * revised from M.A. Stone's HUMAN.c 8-10-92 (Exp. Psych. Cambridge, England).
 * A digital band pass filter has been added.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 *************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "UtFilters.h"
#include "UtERB.h"
#include "UtCmplxM.h"

/******************************************************************************/ 
/*************************** Global Variables *********************************/ 
/******************************************************************************/ 

/* Bessel coefficents plus scale factor for gain loss. */
												 		 
double	Filters_bess2Poly[4] = {3.0, 3.0, 1.0, 1.2589};

/* Butterworth coefficients plus scale factor for gain loss. */

double	Filters_butt2Poly[4] = {1.0, 1.414213562, 1.0, 1.4142};

/**************************** ZeroArray ***************************************/

/*
 * This routine zero's the elements of an array.
 */

void
ZeroArray_Filters(double *p, int length)
{
	int		i;
	
	for (i = 0; i < length; i++)
		*(p++) = 0.0;
}

/**************************** InitCoeffs...GammaTone **************************/

/*-----------------------------------------------------------------*/
/*---- biquadratic filter table: use latest ERB and rate equations */
/*
 * The function returns a pointer to the coefficients structure.
 */

GammaToneCoeffs *
InitCoeffs_Filters_GammaTone(double freq_at_E, double sampleClk, int order)
{
	double erbT, sin_theta, cos_theta, erb;
	double k0num, k1num;	/* numerator coefficients */
	double k1denom, k2denom;	/* denominator coefficients */
	Complex cf_num, cf_denom, Tf, var1, var2, var3;
    GammaToneCoeffs *p;
	
	erb = ERBFromF_ERB(freq_at_E);
	
	erbT= (TwoPi*erb)/sampleClk;
	cos_theta = cos(TwoPi * freq_at_E/sampleClk);
	sin_theta = sin(TwoPi * freq_at_E/sampleClk);

	k1num   = exp( -erbT) * cos_theta; k0num = 1.;
	k1denom = -2. * k1num; k2denom =  exp(-2.* erbT);
	k1num   = -k1num;

	/* normalise filter to unity gain at centre frequency (needs Complex maths)
	 */
	
	var2.re = k1num; var2.im = 0.0;	var3.re = cos_theta; var3.im = -sin_theta;
	Mult_CmplxM(&var2,&var3,&var2); /* a1*exp(-sT) */
	var1.re = k0num; var1.im = 0.0;
	Add_CmplxM (&var1,&var2,&cf_num); /* a0+a1*exp(-sT) */
	var2.re = k1denom; var2.im = 0.0;
	Mult_CmplxM(&var2,&var3,&var2); /* b1*exp(-sT) */
	var2.re += 1.0; /* 1+b1*exp(-sT) */
	var3.re =  cos(2.*TwoPi*freq_at_E/sampleClk);
	var3.im = -sin(2.*TwoPi*freq_at_E/sampleClk);
	var1.re = k2denom; var1.im = 0.0;
	Mult_CmplxM(&var3,&var1,&var1); /* b2*exp(-2sT) */
	Add_CmplxM (&var1,&var2,&cf_denom); /* sigma(bi*exp(-isT) */

	if(!Div_CmplxM(&cf_denom,&cf_num,&Tf)){ /* invert to get normalised values */
		printf("\n\07***Filter failed to initialise ***\n\n"); exit(0);
	}
	Convert_CmplxM(&Tf,&Tf); k1num *= Tf.re; k0num *= Tf.re;

	if ((p = (GammaToneCoeffs *)
	  malloc(sizeof(GammaToneCoeffs))) == NULL) {
		NotifyError("InitGammaToneCoeffs: Out of memory!");
		exit(1);
    }

	p->a0 = k0num; p->a1 = k1num; p->b1 = k1denom; p->b2 = k2denom;
	p->order = order;
	return(p);

} /* InitCoeffs_Filters_GammaTone */

/**************************** InitCoeffs GammaToneVWidth **********************/

/*-----------------------------------------------------------------*/
/*---- biquadratic filter table: use latest ERB and rate equations */
/*
 * The function returns a pointer to the coefficients structure.
 * It requires an additional parameter which specifies the broadening of the
 * filter width.
 */

GammaToneCoeffs *
InitCoeffs_Filters_GammaTVWidth(double freq_at_E, double sampleClk,
  int order, double widthBroadening)
{
	double erbT, sin_theta, cos_theta, erb;
	double k0num, k1num;		/* numerator coefficients */
	double k1denom, k2denom;	/* denominator coefficients */
	Complex cf_num, cf_denom, Tf, var1, var2, var3;
    GammaToneCoeffs *p;
	
	erb = ERBFromF_ERB(freq_at_E) * widthBroadening;
	
	erbT= (TwoPi*erb)/sampleClk;
	cos_theta = cos(TwoPi * freq_at_E/sampleClk);
	sin_theta = sin(TwoPi * freq_at_E/sampleClk);

	k1num   = exp( -erbT) * cos_theta; k0num = 1.;
	k1denom = -2. * k1num; k2denom =  exp(-2.* erbT);
	k1num   = -k1num;

	/* normalise filter to unity gain at centre frequency (needs Complex maths)
	 */
	
	var2.re = k1num; var2.im = 0.0;	var3.re = cos_theta; var3.im = -sin_theta;
	Mult_CmplxM(&var2,&var3,&var2); /* a1*exp(-sT) */
	var1.re = k0num; var1.im = 0.0;
	Add_CmplxM (&var1,&var2,&cf_num); /* a0+a1*exp(-sT) */
	var2.re = k1denom; var2.im = 0.0;
	Mult_CmplxM(&var2,&var3,&var2); /* b1*exp(-sT) */
	var2.re += 1.0; /* 1+b1*exp(-sT) */
	var3.re =  cos(2.*TwoPi*freq_at_E/sampleClk);
	var3.im = -sin(2.*TwoPi*freq_at_E/sampleClk);
	var1.re = k2denom; var1.im = 0.0;
	Mult_CmplxM(&var3,&var1,&var1); /* b2*exp(-2sT) */
	Add_CmplxM (&var1,&var2,&cf_denom); /* sigma(bi*exp(-isT) */

	if(!Div_CmplxM(&cf_denom,&cf_num,&Tf)){ /* invert to get normalised values */
		printf("\n\07***Filter failed to initialise ***\n\n"); exit(0);
	}
	Convert_CmplxM(&Tf,&Tf); k1num *= Tf.re; k0num *= Tf.re;

	if ((p = (GammaToneCoeffs *)
	  malloc(sizeof(GammaToneCoeffs))) == NULL) {
		NotifyError("InitGammaToneCoeffs: Out of memory!");
		exit(1);
    }

	p->a0 = k0num; p->a1 = k1num; p->b1 = k1denom; p->b2 = k2denom;
	p->order = order;
	return(p);

} /* InitCoeffs_Filters_GammaTVWidth */

/**************************** InitCoeffs...IIR2 *******************************/

/* f3dB is the corner frequency, fs sampling freq, coeffs points to an array of
	 5*double for output, splane contains ascending cfts of the denom
	 polynom. */
 /*
  * Low pass filter, which can be either the ButterWorth or Bessel two pole
  * filters, depending on the splane coefficients.
  * The function returns a pointer to the  coefficients structure.
  */

TwoPoleCoeffs *
InitCoeffs_Filters_IIR2(double *splane, double f3dB, double fs,
  int low_or_high)
{
	double A,A0,A1,A2,B0,B1,B2,C;
	double a0,a1,a2,b1,b2; /* parameters used in z plane transformation (b0 = 1.)*/
	double ai_tot, bi_tot, Tf;
    TwoPoleCoeffs	*p;

	if (low_or_high == HIGHPASS){ /* reverse coefficient order out of splane */
		A0 = 0.; A1 = 0.; A2 = 1.;
		B2 = *splane; B1 = *(splane+1); B0 = *(splane+2); /* general 2 pole */
	}
	else { /* LOWPASS */
		A0 = 1.; A1 = 0.; A2 = 0.;
		B0 = *splane; B1 = *(splane+1); B2 = *(splane+2); /* general 2 pole */
	}
	C = 1./tan((Pi*f3dB)/fs); /* true bilinear transform eq 7-8, p172*/
	
	/* first biquadratic section */
	
	A  = (B0 + B1*C + (B2*C*C)); /* DSP, William D.Stanley, Reston press, pp 174-175 */
	a0 = (A0 + A1*C + A2*C*C)/A;
	a1 = 2.*(A0 - A2*C*C)/A;
	a2 = (A0 - A1*C + A2*C*C)/A;
	b1 = 2.*(B0 - B2*C*C)/A;
	b2 = (B0 - B1*C + B2*C*C)/A;
	if (low_or_high == HIGHPASS){/* now normalise coefficents at z=exp(s0T),
	                              * s0 = PI rads/sec */
		ai_tot = a0; ai_tot -= a1; ai_tot += a2; /* NB a1 and b1 change sign */
		bi_tot = 1.; bi_tot -= b1; bi_tot += b2;
		Tf = bi_tot/ai_tot; /* calculate normalisation factor */
		#ifdef DEBUG
		printf("HP filter gain changed by %7.3f\n",Tf);
		#endif
	}
	else { /* LOWPASS *//* now normalise coefficents at z=exp(s0T), s0 = 0 rads/sec */
		ai_tot = a0; ai_tot += a1; ai_tot += a2;
		bi_tot = 1.; bi_tot += b1; bi_tot += b2;
		Tf = bi_tot/ai_tot; /* calculate normalisation factor */
				#ifdef DEBUG
		printf("LP filter gain changed by %7.3f\n",Tf);
		#endif
	}
	a0 *= Tf; a1 *= Tf; a2 *= Tf; /* perform normalisation */

	if ((p = (TwoPoleCoeffs *)
		malloc(sizeof(TwoPoleCoeffs))) == NULL) {
		NotifyError("InitTwoPoleCoeffs: Out of memory!");
		exit(1);
	}

	p->b1 = b1; p->b2 = b2;
	p->a0 = a0; p->a1 = a1; p->a2 = a2;
	p->gainLossFactor = *(splane+3);
	return(p);

}

/**************************** IIR2 *****************************************************/

/*--- 2 pole biquadratic filter code: coefficient order is d1,d2,n0,n1,n2 ---*/
/*
 * This routine filters the signal, returning the result in the same signal
 * passed to the routine.
 */

void
IIR2_Filters(SignalData *theSignal, TwoPoleCoeffs *p[])
{
	int		i, chan;
	register	double 		temp0, temp1, *state1;
	register	ChanData	*data;
	double		state0[AiBi];
	
	if (!theSignal->lengthFlag) {
		NotifyError("iir2_Filters: Signal data length is not set!");
		exit(1);
	}
	if (theSignal->numChannels == 0) {
		NotifyError("iir2_Filters: No signal channels have been initialised!");
		exit(1);
	}
	state1 = state0 + 1;
	for (chan = 0; chan < theSignal->numChannels; chan++) {
		ZeroArray_Filters(state0, AiBi);
		data = theSignal->channel[chan];
		for (i = 0; i < theSignal->length; i++) {
			temp0  = *data - (p[chan]->b1 * *state0); /* b1 term */
			temp0 -=         (p[chan]->b2 * *state1); /* b2 term */
			temp1  = temp0 * p[chan]->a0; /* a0 term */
			temp1 += (p[chan]->a1 * *state0); /* a1 term */
			temp1 +=         (p[chan]->a2 * *state1); /* a2 term */
			*state1 = *state0;
			*state0 = temp0;
			*(data++) = temp1 * p[chan]->gainLossFactor;
		}
	}

}

/**************************** IIR2...Single ********************************************/

/*--- 2 pole biquadratic filter code: coefficient order is d1,d2,n0,n1,n2 ---*/
/*
 * This routine filters the signal, returning the result in the same signal
 * passed to the routine.
 * This version of the IIR2 filter assumes that there is only one set of filter
 * coefficients which operate on all the channels.
 */

void
II2R_Filters_Single(SignalData *theSignal, TwoPoleCoeffs *p)
{
	int		i, chan;
	register	double 		temp0, temp1, *state1;
	register	ChanData	*data;
	double		state0[AiBi];
	
	if (!theSignal->lengthFlag) {
		NotifyError("iir2_Filters: Signal data length is not set!");
		exit(1);
	}
	if (theSignal->numChannels == 0) {
		NotifyError("iir2_Filters: No signal channels have been initialised!");
		exit(1);
	}
	state1 = state0 + 1;
	for (chan = 0; chan < theSignal->numChannels; chan++) {
		ZeroArray_Filters(state0, AiBi);
		data = theSignal->channel[chan];
		for (i = 0; i < theSignal->length; i++) {
			temp0 = *data - (p->b1 * *state0); /* b1 term */
			temp0 -= (p->b2 * *state1); /* b2 term */
			temp1 = temp0 * p->a0; /* a0 term */
			temp1 += (p->a1 * *state0); /* a1 term */
			temp1 += (p->a2 * *state1); /* a2 term */
			*state1 = *state0;
			*state0 = temp0;
			*(data++) = temp1 /* * p->gainLossFactor */;
		}
	}

}

/**************************** Compression **********************************************/

/*
 * This routine filters the signal, returning the result in the same signal
 * passed to the routine.
 */
 
void
Compression_Filters(SignalData *theSignal, double nrwthr, double nrwcr)
{
	int		i, chan;
	register	ChanData	*data;

	if (!theSignal->lengthFlag) {
		NotifyError("GammaTone_Filters: Signal data length is not set!");
		exit(1);
	}
	if (theSignal->numChannels == 0) {
		NotifyError("Compression_Filters: No signal channels have been initialised!");
		exit(1);
	}
	for (chan = 0; chan <theSignal->numChannels; chan++)	
		for (i = 0, data = theSignal->channel[chan]; i < theSignal->length; i++, data++)
			if(*data >= 0.0) { 			/* positive signal */
				if(*data > nrwthr)		/* linear up to threshold */
					*data = pow((*data - nrwthr), nrwcr) + nrwthr;
					/* compresses excess of signal above threshold */
			} else { 					/* negative signal */
				if(*data < -nrwthr) /* linear up to threshold */
					*data = -pow((-*data - nrwthr), nrwcr) - nrwthr;
			}

}

/**************************** GammaTone ************************************************/

/*
 * This routine filters the signal, returning the result in the same signal
 * passed to the routine.
 * No checks are made here as to whether or not the gamma tone coefficients, passed
 * as an argument array, have been initialised.
 * Each channel has its own filter (coefficients).
 */

void
GammaTone_Filters(SignalData *theSignal, GammaToneCoeffs *p[])
{
	int		i, j, chan, stateVectorLength;
	double		*stateVector;
	register	double		*ptr1, *ptr2, wn;	/* Inner loop variables */
	register	ChanData	*data;

	if (!CheckPars_SignalData(theSignal)) {
		NotifyError("Filter_GammaTone: Signal not correctly initialised.");
		exit(1);
	}	
	/* For the allocation of space to the state vector, the filter for all channels
	 * are assumed to have the same order as the first. */
	 
	stateVectorLength = p[0]->order * 2;	/* 2 for each cascaded filter */
	if ((stateVector = (double *) calloc(stateVectorLength, sizeof(double))) == NULL) {
		NotifyError("GammaTone_Filters: Cannot allocate space for state vector.");
		exit(1);
	}
	for (chan = 0; chan < theSignal->numChannels; chan++) {
		ZeroArray_Filters(stateVector, stateVectorLength);
		data = theSignal->channel[chan];
		for (i = 0; i < theSignal->length; i++, data++)
			for (j = 0, ptr1 = stateVector; j < p[0]->order; j++) {
				ptr2 = ptr1;
				*data -= p[chan]->b1 * *(ptr1++);
				*data -= p[chan]->b2 * *ptr1;
				wn = *data;		/* Temp variable */
				*data *= p[chan]->a0;
				*data += p[chan]->a1 * *ptr2; /* Final Yn */
				*ptr1 = *ptr2;	/* Update Yn-1 to Yn-2 */
				*ptr2 = wn;
				ptr1++;
			}
	}
	free(stateVector);

}

/**************************** InitCoeffs...BandPass ***************************/

/*
 * This function initialises the band pass filter coefficents, and returns
 * a pointer the the allocated structure.
 */
 
BandPassCoeffs *
InitCoeffs_Filters_BandPass(double lowerCutOffFreq, double upperCutOffFreq,
  double dt)
{
	double	theta, tanTheta, frequencyDiff;
	BandPassCoeffs	*p;
	
	if ((frequencyDiff = upperCutOffFreq - lowerCutOffFreq) < 0.001) {
		NotifyError("InitCoeffs_Filters_BandPass: Illegal frequency band "\
		  "width = %g Hz.", frequencyDiff);
		exit(1);
	}
	if ((p = (BandPassCoeffs *) malloc(sizeof(BandPassCoeffs))) == NULL) {
		NotifyError("InitCoeffs_Filters_BandPass: Cannot allocate memory for "\
		  "band pass filter coefficients.");
		exit(1);
	}
	theta = PI * dt * (frequencyDiff);
	tanTheta = tan(theta);
	p->j = 1.0 / (1.0 + 1.0 / tanTheta);
	p->k = 2.0 * cos(PI * dt * (upperCutOffFreq + lowerCutOffFreq)) / ((1.0 +
	  tanTheta) * cos(theta));
	p->l = (tanTheta - 1.0) / (tanTheta + 1.0);
	p->gainLossFactor = 1.0 / p->j;
	return(p);

}

/**************************** BandPass *************************************************/

/*
 * This routine filters the signal, returning the result in the same signal
 * passed to the routine.
 * The signal is a band-pass filter.  Yi = J*X[i] - J*X[i-2] + K*Y[i-1] + L*Y[i-2]
 * - where X and Y are the input and output values respectively.  See Beauchamp p.257 -
 * 258.
 * The band-pass filter operates forwards then backwards to: (i) increase the attenuation
 * at the cut-off frequencies; (ii) to negate the phase change effects of the filter
 * (latency effects.).
 * No checks are made here as to whether or not the coefficients, passed as an argument
 * array, have been initialised.
 */

void
BandPass_Filters(SignalData *theSignal, BandPassCoeffs *p)
{
	int		i, chan;
	double		stateVector[B_PASS_STATE_VECTOR_LENGTH];
	register	double	*ptr;		/* Inner loop variables */
	register	ChanData *yi, xi;	/*  "		"	"		*/
	
	if (!CheckPars_SignalData(theSignal)) {
		NotifyError("BandPass_Filters: Signal not correctly initialised.");
		exit(1);
	}	
	for (chan = 0; chan < theSignal->numChannels; chan++) {
		ZeroArray_Filters(stateVector,B_PASS_STATE_VECTOR_LENGTH);
		/* Forward pass. */
		yi = theSignal->channel[chan];
		for (i = 0; i < theSignal->length; i++, yi++) {
			ptr = &stateVector[1];
			xi = *yi;
			/* Start of formula evaluation */
			*yi *= p->j;
			*yi -= p->j * *ptr;			/* -= J * X[i-2]	*/
			*ptr = *(ptr - 1);			/* X[i-2] = X[i-1]	*/
			*(ptr - 1) = xi;			/* X[i-1] = Xi		*/
			*yi += p->l * *(++ptr);		/* += L * Y[i-2]	*/
			*ptr = *(ptr + 1);			/* Y[i-2] = Y[i-1]	*/
			*yi += p->k * *(++ptr);		/* += K * Y[i-1]	*/
			*ptr = *yi;					/* Y[i-1] = Yi		*/
		}
		/* Backwards pass. */
		yi--;					/* Was set past the last data point. */		
		for (i = 0; i < theSignal->length; i++, yi--) {
			ptr = &stateVector[1];
			xi = *yi;
			/* Start of formula evaluation */
			*yi *= p->j;
			*yi -= p->j * *ptr;			/* -= J * X[i-2]	*/
			*ptr = *(ptr - 1);			/* X[i-2] = X[i-1]	*/
			*(ptr - 1) = xi;			/* X[i-1] = Xi		*/
			*yi += p->l * *(++ptr);		/* += L * Y[i-2]	*/
			*ptr = *(ptr + 1);			/* Y[i-2] = Y[i-1]	*/
			*yi += p->k * *(++ptr);		/* += K * Y[i-1]	*/
			*ptr = *yi;					/* Y[i-1] = Yi		*/
		}
	}

}

/*************************** InitCoeffs...IIR2Cont ******************************/ 

/*
 * Second order continuous Butterworth IIR Filter - see "Digital Methods for
 * Signal Analysis" by K. Beauchamp and C. Yuen, page 261 and previous.
 * The cut-off frequency is the frequency at which the signal is 3dB down.
 * The coefficients for both low- and high-pass filters can be returned.
 */

ContButtCoeffsPtr
InitCoeffs_Filters_IIR2Cont(double cutOffFrequency, double samplingInterval,
 int highOrLowPass)
{
	double	sqrt2, theta, cotOrTan, cotOrTanSquared, cC, dD, eE;
	ContButtCoeffsPtr	p;

	sqrt2 = sqrt(2.0);
	theta = Pi * cutOffFrequency * samplingInterval;
	
	cotOrTan = (highOrLowPass == HIGHPASS)? tan(theta): 1.0 / tan(theta);
	cotOrTanSquared = cotOrTan * cotOrTan;
	
	cC = 1.0 / (1.0 + sqrt2 * cotOrTan + cotOrTanSquared);
	dD = 2.0 * (1.0 - cotOrTanSquared);
	eE = 1.0 - sqrt2 * cotOrTan + cotOrTanSquared;
	
	if ((p = (ContButtCoeffsPtr) malloc(sizeof(ContButtCoeffs))) == NULL) {
		NotifyError("InitCoeffs_Filters_IIR2Cont: Out of memory.");
		return(NULL);
	}
	p->cC = cC;
	p->eE = eE;
	if (highOrLowPass == HIGHPASS) {
		p->twoC = -2.0 * cC;
		p->dD = -dD;
	} else {
		p->twoC = 2.0 * cC;
		p->dD = dD;
	}
	return(p);
	
}

/*************************** IIR2Cont...Single ***********************************/ 

/*
 * Second order continuous Butterworth IIR Filter - see "Digital Methods for
 * Signal Analysis" by K. Beauchamp and C. Yuen, page 261 and previous.
 *
 * This routine filters the signal, returning the result in the same signal
 * passed to the routine.
 * This is a infinite impulse response filters:
 *  Yi = C * X[i] + 2C * X[i-1] + C * X[i-2] - D * Y[i-1] - E * Y[i-2].
 * - where X and Y are the input and output values respectively.
 * No checks are made here as to whether or not the coefficients, passed as an
 * argument array, have been initialised.
 */

void
IIR2Cont_Filters_Single(SignalData *theSignal, ContButtCoeffsPtr p)
{
	int		chan;
	double		stateVector[IIRCONT_STATE_VECTOR_LENGTH];
	ChanLen		i;
	register	double	*ptr;		/* Inner loop variables */
	register	ChanData *yi, xi;	/*  "		"	"		*/
	
	if (!CheckPars_SignalData(theSignal)) {
		NotifyError("IIR2Cont_Filters_Single: Signal not correctly initialised.");
		exit(1);
	}	
	for (chan = 0; chan < theSignal->numChannels; chan++) {
		ZeroArray_Filters(stateVector, IIRCONT_STATE_VECTOR_LENGTH);
		yi = theSignal->channel[chan];
		for (i = 0; i < theSignal->length; i++, yi++) {
			ptr = &stateVector[0];
			xi = *yi;
			/* Start of formula evaluation */
			*yi *= p->cC;
			*yi += p->twoC * *ptr++;	/* += 2C * X[i-1]	*/
			*yi += p->cC * *ptr;		/* += C * X[i-2]	*/
			*ptr = *(ptr - 1);			/* X[i-2] = X[i-1]	*/
			*(ptr - 1) = xi;			/* X[i-1] = Xi		*/
			*yi -= p->eE * *(++ptr);	/* -= E * Y[i-2]	*/
			*ptr = *(ptr + 1);			/* Y[i-2] = Y[i-1]	*/
			*yi -= p->dD * *(++ptr);	/* -= D * Y[i-1]	*/
			*ptr = *yi;					/* Y[i-1] = Yi		*/
		}
	}

}

/*************************** InitCoeffs...IIR1Cont ******************************/ 

/*
 * First order continuous Butterworth IIR Filter - see "Digital Methods for
 * Signal Analysis" by K. Beauchamp and C. Yuen, page 261 and previous.
 * The cut-off frequency is the frequency at which the signal is 3dB down.
 * The coefficients for both low- and high-pass filters can be returned.
 */

ContButt1CoeffsPtr
InitCoeffs_Filters_IIR1Cont(double cutOffFrequency, double samplingInterval,
 int highOrLowPass)
{
	double	theta, cotOrTan;
	ContButt1CoeffsPtr	p;

	theta = Pi * cutOffFrequency * samplingInterval;
	
	cotOrTan = (highOrLowPass == HIGHPASS)? tan(theta): 1.0 / tan(theta);
	
	if ((p = (ContButt1CoeffsPtr) malloc(sizeof(ContButtCoeffs))) == NULL) {
		NotifyError("InitCoeffs_Filters_IIR2Cont: Out of memory.");
		return(NULL);
	}
	p->gG = 1.0 / (1.0 + cotOrTan);
	p->gGG = (highOrLowPass == HIGHPASS)? -p->gG: p->gG;
	p->hH = (1 - cotOrTan) / (1 + cotOrTan);
	p->gainLossFactor = 1.0 / p->gG;
	return(p);
	
}

/*************************** IIR1Cont...Single ***********************************/ 

/*
 * First order continuous Butterworth IIR Filter - see "Digital Methods for
 * Signal Analysis" by K. Beauchamp and C. Yuen, page 261 and previous.
 *
 * This routine filters the signal, returning the result in the same signal
 * passed to the routine.
 * This is a infinite impulse response filters:
 *  Yi = C * X[i] + 2C * X[i-1] + C * X[i-2] - D * Y[i-1] - E * Y[i-2].
 * - where X and Y are the input and output values respectively.
 * No checks are made here as to whether or not the coefficients, passed as an
 * argument array, have been initialised.
 */

void
IIR1Cont_Filters_Single(SignalData *theSignal, ContButt1CoeffsPtr p)
{
	int		i, chan;
	double		stateVector[IIRCONT_STATE_VECTOR_LENGTH];
	register	double	*ptr;		/* Inner loop variables */
	register	ChanData *yi, xi;	/*  "		"	"		*/
	
	if (!CheckPars_SignalData(theSignal)) {
		NotifyError("IIR2Cont_Filters_Single: Signal not correctly initialised.");
		exit(1);
	}	
	for (chan = 0; chan < theSignal->numChannels; chan++) {
		ZeroArray_Filters(stateVector, IIRCONT_STATE_VECTOR_LENGTH);
		yi = theSignal->channel[chan];
		for (i = 0; i < theSignal->length; i++, yi++) {
			ptr = &stateVector[0];
			xi = *yi;
			/* Start of formula evaluation */
			*yi *= p->gG;
			*yi += p->gGG * *ptr;		/* += G * X[i-1]	*/
			*ptr++ = xi;				/* X[i-1] = Xi		*/
			*yi -= p->hH * *ptr;		/* -= H * Y[i-1]	*/
			*ptr = *yi;					/* Y[i-1] = Yi		*/
		}
	}

}

