/*************
 *
 * UtFilters.h
 *
 * The filters.c module contains all the files that initialise and 
 * generate the 2-pole GammaTone filters, the 2-pole Butterworth low pass
 * filter, the Bessel low pass filter and the non-linear compresion filter.
 *
 * Header file for the IIR functions.
 *
 * IIR performs z plane filter for 2 pole coefficients.
 *
 **************/

#ifndef	_UTFILTERS_H
#define _UTFILTERS_H	1
 
/******************************************************************************/
/*************************** Constant Definitions *****************************/
/******************************************************************************/ 

#ifndef PI
# define PI 3.1415926538
#endif
#define Pi PI
#define Piby2 (PI/2.)
#define TwoPi (2.* Pi)
#define LOWPASS 1
#define HIGHPASS 0

#define DELAY_SF		(double) (NGAMMA - 1)	/* Delay scale factor to
												 * time align FB output */
#define AiBi			5	/* Max. No. of coefficients per LP filter stage
							 * (2 Pole). */
#define	B_PASS_STATE_VECTOR_LENGTH	4	/* State vector for band pass filter. */
#define	IIRCONT_STATE_VECTOR_LENGTH	4	/* State vector for IIR cont. filter. */

/******************************************************************************/
/*************************** Type definitions *********************************/
/******************************************************************************/ 

typedef struct {

	int		order;				/* The order of the filter */
	double	a0, a1;
	double	b1, b2;

} GammaToneCoeffs, *GammaToneCoeffsPtr;

typedef struct {

	double	b1, b2;
    double	a0, a1, a2;
    double	gainLossFactor;

} TwoPoleCoeffs, *TwoPoleCoeffsPtr;

typedef struct {
	double	j;
	double	k;
	double	l;
	double	gainLossFactor;

} BandPassCoeffs, *BandPassCoeffsPtr;

/*
 * Second order continuous Butterworth IIR Filter - see "Digital Methods for
 * Signal Analysis" by K. Beauchamp and C. Yuen, page 261 and previous.
 */

typedef struct {

	double	cC;
	double	twoC;
	double	dD;
	double	eE;
	double	gainLossFactor;

} ContButtCoeffs, *ContButtCoeffsPtr;

/*
 * First order continuous Butterworth IIR Filter - see "Digital Methods for
 * Signal Analysis" by K. Beauchamp and C. Yuen, page 261 and previous.
 */

typedef struct {

	double	gG;
	double	gGG;
	double	hH;
	double	gainLossFactor;

} ContButt1Coeffs, *ContButt1CoeffsPtr;

	
/******************************************************************************/
/*************************** External Variables *******************************/
/******************************************************************************/

extern	double	Filters_bess2Poly[];         	/* Bessel coefficents */

extern	double	Filters_butt2Poly[];			/* Butterworth */

/******************************************************************************/
/*************************** Function Prototypes ******************************/
/******************************************************************************/ 

void	BandPass_Filters(SignalDataPtr theSignal, BandPassCoeffsPtr p);

void	Compression_Filters(SignalDataPtr theSignal, double nrwthr,
		  double nrwcr);

void	GammaTone_Filters(SignalDataPtr theSignal, GammaToneCoeffsPtr p[]);

void	IIR2_Filters(SignalDataPtr theSignal, TwoPoleCoeffsPtr p[]);

void	II2R_Filters_Single(SignalDataPtr theSignal, TwoPoleCoeffsPtr p);

void	IIR2Cont_Filters_Single(SignalDataPtr theSignal, ContButtCoeffsPtr p);

void	IIR1Cont_Filters_Single(SignalDataPtr theSignal, ContButt1CoeffsPtr p);

BandPassCoeffsPtr	InitCoeffs_Filters_BandPass(double lowerCutOffFreq,
					  double upperCutOffFreq, double dt);

ContButtCoeffsPtr	InitCoeffs_Filters_IIR2Cont(double cutOffFrequency,
					  double samplingInterval, int highOrLowPass);

ContButt1CoeffsPtr	InitCoeffs_Filters_IIR1Cont(double cutOffFrequency,
					  double samplingInterval, int highOrLowPass);

GammaToneCoeffsPtr	InitCoeffs_Filters_GammaTone(double freq_at_E,
					  double sampleClk, int order);

GammaToneCoeffsPtr	InitCoeffs_Filters_GammaTVWidth(double freq_at_E,
					  double sampleClk, int order, double widthBroadening);

TwoPoleCoeffsPtr 	InitCoeffs_Filters_IIR2(double *splane, double f3dB,
					  double fs, int low_or_high);
					   
		/* latter generates z plane coefficients from 2 pole blocks, f3dB is
		 * the corner frequency, fs sampling freq, coeffs points to an array the corner
		 * the corner splane contains the coefficients of the denominator
		 * polynomial. */

void	ZeroArray_Filters(double *p, int length);

#endif
