/***********************
 *
 * UtRamp.c
 *
 * This is the signal ramp module.
 * The ramps gradually raise or lower the signal intensity to full strength or
 * from full strength to zero at the beginning or end of the signal respectively.
 *
 * Valid ramp functions are defined as a function which varies from 0 to 1 as a function
 * of the ramp interval.
 * The functions used must be defined in the format given below (see Sine_Ramp and
 * Ramp_linear) and are passed to the RampUpOutSignal_Ramp and RampDownOutSignal_Ramp
 * functions as arguments.
 * Impulse signals, such as clicks, should not be ramped: if a warning message is given,
 * in this case ignore it.
 *
 * Unless otherwise stated all code was written by L. P. O'Mard.
 *
 *	E-mail: L.P.OMard@uk.ac.lut
 *
 * COPYRIGHT 
 *
 *	Speech and Hearing Laboratory,
 *	Department of Human Sciences,
 *	Loughborough University of Technology,
 *	Loughborough,
 *	Leicestershire LE11 3TU.
 *	United Kingdom.
 *
 *	12th July 1993.
 *
 *	DISCLAIMER OF WARRANTY
 *
 * Although every effort has been made to ensure the integrity of this code,
 * the author, Loughborough University and its Laboratories disclaim all
 * liability for direct or consequential damages resulting from its use.
 *
 ***********************/

#include <stdio.h>
#include <stdlib.h> 
#include <math.h>

#include "LUTEar.h"
#include "SignalData.h"
#include "EarObject.h"
#include "UtRamp.h"

/****************************** CheckPars ***********************************************/

/*
 * This function checks that the signal to be ramped has been initialised, and that
 * the ramp interval is not greater than the signal length.
 * It also sets the signalData rampFlag to TRUE.
 */
 
BOOLEAN
CheckPars_Ramp(SignalData *theSignal, double timeInterval) 
{
	if (!CheckPars_SignalData(theSignal)) {
		NotifyError("CheckPars_Ramp: Signal not correctly initialised.");
		return(FALSE);
	}
	if ((ChanLen) (timeInterval / theSignal->dt) > theSignal->length) {
		NotifyError("CheckPars_Ramp: Ramp time interval is longer than signal!");
		return(FALSE);
	}
	theSignal->rampFlag = TRUE;
	return(TRUE);

}

/****************************** Sine ****************************************************/

/*
 * This is a standard ramp function, to be passed as an argument to the Ramp_UP and
 * Ramp_Down routines.  It varies from 0 to 1 as the step ranges from 0 to the interval
 * index length.
 */

double
Sine_Ramp(ChanLen step, ChanLen intervalIndex)
{
	return(sin(step * PI / 2.0 / intervalIndex));
}

/****************************** Linear **************************************************/

/*
 * This is a standard ramp function, to be passed as an argument to the Ramp_UP and
 * Ramp_Down routines.  It varies from 0 to 1 as the step ranges from 0 to the interval
 * index length.
 */

double
Linear_Ramp(ChanLen step, ChanLen intervalIndex)
{
	return((double) step / intervalIndex);
}

/****************************** RampUpOutSignal *****************************************/

/*
 * This function searches for the first non-zero signal intensity in each channel of a
 * signal data set.  It then starts the ramp from the preceding zero (or the start of
 * the signal if there are no preceding zero's).
 * The ramp function is passed as an argument.
 * The routine assumes that the EarObject has been initialised.
 */

void
RampUpOutSignal_Ramp(EarObjectPtr data, double (* RampFunction)(ChanLen, ChanLen),
  double timeInterval)
{
	int		chan;
	ChanLen	i, start, intervalIndex;
	
	if (data == NULL) {
		NotifyError("RampUpOutSignal_Ramp: EarObject not initialised.");
		exit(1);
	}	
	if (!CheckPars_Ramp(data->outSignal, timeInterval)) {
		NotifyError("RampUpOutSignal_Ramp: Function arguments not correctly "\
		  "initialised\nfor EarObject: %s.", data->processName);
		exit(1);
	}
	intervalIndex = (ChanLen) (timeInterval / data->outSignal->dt);
	for (chan = 0; chan < data->outSignal->numChannels; chan++) {
		for (start = 0; (start < data->outSignal->length) &&
		  (data->outSignal->channel[chan][start] == 0.0); start++);
		if (start > 0)
			start--;
		for (i = 0; (i < intervalIndex) && (start + i < data->outSignal->length); i++)
			data->outSignal->channel[chan][start + i] *= (*RampFunction)(i,
			  intervalIndex);
	}

}

/****************************** RampDownOutSignal ***************************************/

/*
 * This function searches for the last non-zero signal intensity from the end of each
 * channel of a signal data set.
 * It then starts the ramp from the following zero (or from the end of
 * the signal if there are no following zero's).
 */

void
RampDownOutSignal_Ramp(EarObjectPtr data, double (* RampFunction)(ChanLen, ChanLen),
  double timeInterval)
{
	int		chan;
	ChanLen	i, signalEnd, intervalIndex;
	
	if (data == NULL) {
		NotifyError("RampDownOutSignal_Ramp: EarObject not initialised.");
		exit(1);
	}	
	if (!CheckPars_Ramp(data->outSignal, timeInterval)) {
		NotifyError("RampDownOutSignal_Ramp: Function arguments not correctly "\
		  "initialised for\nEarObject: %s.", data->processName);
		exit(1);
	}
	intervalIndex = (ChanLen) (timeInterval / data->outSignal->dt);
	for (chan = 0; chan < data->outSignal->numChannels; chan++) {
		for (signalEnd = data->outSignal->length - 1; (signalEnd > 0) &&
		  (data->outSignal->channel[chan][signalEnd] == 0.0); signalEnd--);
		if (signalEnd < data->outSignal->length - 1)
			signalEnd++;
		for (i = 0; (i < intervalIndex) && ( signalEnd - i > 0); i++)
			data->outSignal->channel[chan][signalEnd - i] *= (*RampFunction)(i,
			  intervalIndex);
	}

}

