/***********************
 *
 * UtRandom.c
 *
 * This is the random number routines module.
 *
 * 29-3-93 L. P. O'Mard
 *
 ***********************/
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LUTEar.h"
#include "UtRandom.h"

/******************************************************************************/ 
/************************** Constant definitions ******************************/ 
/******************************************************************************/

#define DEBUG

/******************************************************************************/ 
/************************** Global Variables **********************************/ 
/******************************************************************************/

long	randomNumSeed = 3141; 

/******************************************************************************/ 
/************************** Subroutines and functions *************************/ 
/******************************************************************************/ 

/************************** Ran01_Random **************************************/ 

/*
 * This routine initialises the random number generator.
 * It has been revised from the "Numerical Recipes 2nd ed." Ran1 (p.280).
 * Set "seed" to any negative value to initialise or re-intialise the sequence.
 */

#define IA 16807
#define IM 2147483647
#define AM (1.0 / IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1 + (IM - 1) / NTAB)
#define EPS 1.2e-7
#define RNMX (1.0 - EPS)

double
Ran01_Random(long *idum)
{
	int j;
	long k;
	static long iy=0;
	static long iv[NTAB];
	double temp;

	if (*idum <= 0 || !iy) {
		*idum = (-(*idum) < 1)? 1: -(*idum);
		for (j = NTAB + 7; j >= 0; j--) {
			k = (*idum) / IQ;
			*idum = IA * (*idum - k * IQ) - IR * k;
			if (*idum < 0)
				*idum += IM;
			if (j < NTAB)
				iv[j] = *idum;
		}
		iy=iv[0];
	}
	k = (*idum) / IQ;
	*idum = IA * (*idum - k * IQ) - IR * k;
	if (*idum < 0)
		*idum += IM;
	j = iy / NDIV;
	iy = iv[j];
	iv[j] = *idum;
	return( ((temp = AM * iy) > RNMX)? RNMX: temp);

}
#undef IA
#undef IM
#undef AM
#undef IQ
#undef IR
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX

/************************** GeomDist_Random ***********************************/

/*
 * This routine calculates the number of times that an event occurs with 
 * a set probability and number of trials.
 * It uses the gemotric distribution method, but the probability must be
 * small.
 */

int
GeomDist_Random(double probability, int numTrials)
{
	int		i;
	double	sum;
	
	#ifdef DEBUG
		if (probability >= 1.0) {
			NotifyError("GeomDist_Random: probability = %g: greater than 1.",
			  probability);
			exit(1);
		}
	#endif
	for (i = 0, sum = 0; (sum += ceil(log(Ran01_Random(&randomNumSeed)) /
	  log(1.0 - probability))) < numTrials; i++);
	  
	return (i);
	
}
